# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mars/serialize/protos/operand.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mars.serialize.protos import value_pb2 as mars_dot_serialize_dot_protos_dot_value__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='mars/serialize/protos/operand.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n#mars/serialize/protos/operand.proto\x1a!mars/serialize/protos/value.proto\"\x83\'\n\nOperandDef\x12\x0c\n\x04type\x18\x01 \x01(\t\x12#\n\x04\x61ttr\x18\x02 \x03(\x0b\x32\x15.OperandDef.AttrEntry\x12\x1d\n\x08operands\x18\x03 \x03(\x0b\x32\x0b.OperandDef\x12\n\n\x02id\x18\x04 \x01(\t\x1a\x33\n\tAttrEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x15\n\x05value\x18\x02 \x01(\x0b\x32\x06.Value:\x02\x38\x01\"\xe1%\n\x0bOperandType\x12\x08\n\x04NULL\x10\x00\x12\n\n\x06SCALAR\x10\x01\x12\x16\n\x12TENSOR_DATA_SOURCE\x10\x02\x12\x0f\n\x0bTENSOR_ONES\x10\x03\x12\x14\n\x10TENSOR_ONES_LIKE\x10\x04\x12\x10\n\x0cTENSOR_ZEROS\x10\x05\x12\x15\n\x11TENSOR_ZEROS_LIKE\x10\x06\x12\x10\n\x0cTENSOR_EMPTY\x10\x07\x12\x15\n\x11TENSOR_EMPTY_LIKE\x10\x08\x12\x0f\n\x0bTENSOR_FULL\x10\t\x12\x11\n\rTENSOR_ARANGE\x10\n\x12\x12\n\x0eTENSOR_INDICES\x10\x0b\x12\x0f\n\x0bTENSOR_DIAG\x10\x0c\x12\x0e\n\nTENSOR_EYE\x10\r\x12\x13\n\x0fTENSOR_LINSPACE\x10\x0e\x12\x0f\n\x0bTENSOR_TRIU\x10\x0f\x12\x0f\n\x0bTENSOR_TRIL\x10\x10\x12\r\n\tRAND_RAND\x10)\x12\x0e\n\nRAND_RANDN\x10*\x12\x10\n\x0cRAND_RANDINT\x10+\x12\x18\n\x14RAND_RANDOM_INTEGERS\x10,\x12\x16\n\x12RAND_RANDOM_SAMPLE\x10-\x12\x0f\n\x0bRAND_RANDOM\x10.\x12\r\n\tRAND_RANF\x10/\x12\x0f\n\x0bRAND_SAMPLE\x10\x30\x12\x0e\n\nRAND_BYTES\x10\x31\x12\r\n\tRAND_BETA\x10\x32\x12\x11\n\rRAND_BINOMIAL\x10\x33\x12\x12\n\x0eRAND_CHISQUARE\x10\x34\x12\x0f\n\x0bRAND_CHOICE\x10\x35\x12\x12\n\x0eRAND_DIRICHLET\x10\x36\x12\x14\n\x10RAND_EXPONENTIAL\x10\x37\x12\n\n\x06RAND_F\x10\x38\x12\x0e\n\nRAND_GAMMA\x10\x39\x12\x12\n\x0eRAND_GEOMETRIC\x10:\x12\x0f\n\x0bRAND_GUMBEL\x10;\x12\x17\n\x13RAND_HYPERGEOMETRIC\x10<\x12\x10\n\x0cRAND_LAPLACE\x10=\x12\x11\n\rRAND_LOGISTIC\x10>\x12\x12\n\x0eRAND_LOGNORMAL\x10?\x12\x12\n\x0eRAND_LOGSERIES\x10@\x12\x14\n\x10RAND_MULTINOMIAL\x10\x41\x12\x1c\n\x18RAND_MULTIVARIATE_NORMAL\x10\x42\x12\x1a\n\x16RAND_NEGATIVE_BINOMIAL\x10\x43\x12\x1c\n\x18RAND_NONCENTRAL_CHISQURE\x10\x44\x12\x15\n\x11RAND_NONCENTRAL_F\x10\x45\x12\x0f\n\x0bRAND_NORMAL\x10\x46\x12\x0f\n\x0bRAND_PARETO\x10G\x12\x14\n\x10RAND_PERMUTATION\x10H\x12\x10\n\x0cRAND_POSSION\x10I\x12\x0e\n\nRAND_POWER\x10J\x12\x11\n\rRAND_RAYLEIGH\x10K\x12\x10\n\x0cRAND_SHUFFLE\x10L\x12\x18\n\x14RAND_STANDARD_CAUCHY\x10M\x12\x1d\n\x19RAND_STANDARD_EXPONENTIAL\x10N\x12\x18\n\x14RAND_STANDARD_GAMMMA\x10O\x12\x18\n\x14RAND_STANDARD_NORMAL\x10P\x12\x13\n\x0fRAND_STANDARD_T\x10Q\x12\x11\n\rRAND_TOMAXINT\x10R\x12\x13\n\x0fRAND_TRIANGULAR\x10S\x12\x10\n\x0cRAND_UNIFORM\x10T\x12\x11\n\rRAND_VONMISES\x10U\x12\r\n\tRAND_WALD\x10V\x12\x10\n\x0cRAND_WEIBULL\x10W\x12\r\n\tRAND_ZIPF\x10X\x12\x07\n\x03\x41\x44\x44\x10\x65\x12\x10\n\x0c\x41\x44\x44_CONSTANT\x10\x66\x12\x07\n\x03SUB\x10g\x12\x10\n\x0cSUB_CONSTANT\x10h\x12\x07\n\x03MUL\x10i\x12\x10\n\x0cMUL_CONSTANT\x10j\x12\x07\n\x03\x44IV\x10k\x12\x10\n\x0c\x44IV_CONSTANT\x10l\x12\x0b\n\x07TRUEDIV\x10m\x12\x11\n\rTDIV_CONSTANT\x10n\x12\x0c\n\x08\x46LOORDIV\x10o\x12\x11\n\rFDIV_CONSTANT\x10p\x12\x07\n\x03POW\x10q\x12\x10\n\x0cPOW_CONSTANT\x10r\x12\x07\n\x03MOD\x10s\x12\x10\n\x0cMOD_CONSTANT\x10t\x12\x08\n\x04\x46MOD\x10u\x12\x11\n\rFMOD_CONSTANT\x10v\x12\r\n\tLOGADDEXP\x10w\x12\x10\n\x0cLAE_CONSTANT\x10x\x12\x0e\n\nLOGADDEXP2\x10y\x12\x11\n\rLAE2_CONSTANT\x10z\x12\x0c\n\x08NEGATIVE\x10{\x12\x0c\n\x08POSITIVE\x10|\x12\x0c\n\x08\x41\x42SOLUTE\x10}\x12\x08\n\x04\x46\x41\x42S\x10~\x12\x07\n\x03\x41\x42S\x10\x7f\x12\t\n\x04RINT\x10\x80\x01\x12\t\n\x04SIGN\x10\x81\x01\x12\t\n\x04\x43ONJ\x10\x82\x01\x12\x08\n\x03\x45XP\x10\x83\x01\x12\t\n\x04\x45XP2\x10\x84\x01\x12\x08\n\x03LOG\x10\x85\x01\x12\t\n\x04LOG2\x10\x86\x01\x12\n\n\x05LOG10\x10\x87\x01\x12\n\n\x05\x45XPM1\x10\x88\x01\x12\n\n\x05LOG1P\x10\x89\x01\x12\t\n\x04SQRT\x10\x8a\x01\x12\x0b\n\x06SQUARE\x10\x8b\x01\x12\t\n\x04\x43\x42RT\x10\x8c\x01\x12\x0f\n\nRECIPROCAL\x10\x8d\x01\x12\x07\n\x02\x45Q\x10\x8e\x01\x12\x10\n\x0b\x45Q_CONSTANT\x10\x8f\x01\x12\x07\n\x02NE\x10\x90\x01\x12\x10\n\x0bNE_CONSTANT\x10\x91\x01\x12\x07\n\x02LT\x10\x92\x01\x12\x10\n\x0bLT_CONSTANT\x10\x93\x01\x12\x07\n\x02LE\x10\x94\x01\x12\x10\n\x0bLE_CONSTANT\x10\x95\x01\x12\x07\n\x02GT\x10\x96\x01\x12\x10\n\x0bGT_CONSTANT\x10\x97\x01\x12\x07\n\x02GE\x10\x98\x01\x12\x10\n\x0bGE_CONSTANT\x10\x99\x01\x12\x08\n\x03SIN\x10\x9a\x01\x12\x08\n\x03\x43OS\x10\x9b\x01\x12\x08\n\x03TAN\x10\x9c\x01\x12\x0b\n\x06\x41RCSIN\x10\x9d\x01\x12\x0b\n\x06\x41RCCOS\x10\x9e\x01\x12\x0b\n\x06\x41RCTAN\x10\x9f\x01\x12\x0c\n\x07\x41RCTAN2\x10\xa0\x01\x12\x13\n\x0e\x41RCT2_CONSTANT\x10\xa1\x01\x12\n\n\x05HYPOT\x10\xa2\x01\x12\x13\n\x0eHYPOT_CONSTANT\x10\xa3\x01\x12\t\n\x04SINH\x10\xa4\x01\x12\t\n\x04\x43OSH\x10\xa5\x01\x12\t\n\x04TANH\x10\xa6\x01\x12\x0c\n\x07\x41RCSINH\x10\xa7\x01\x12\x0c\n\x07\x41RCCOSH\x10\xa8\x01\x12\x0c\n\x07\x41RCTANH\x10\xa9\x01\x12\x0c\n\x07\x44\x45G2RAD\x10\xaa\x01\x12\x0c\n\x07RAD2DEG\x10\xab\x01\x12\x0b\n\x06\x42ITAND\x10\xac\x01\x12\x14\n\x0f\x42ITAND_CONSTANT\x10\xad\x01\x12\n\n\x05\x42ITOR\x10\xae\x01\x12\x13\n\x0e\x42ITOR_CONSTANT\x10\xaf\x01\x12\x0b\n\x06\x42ITXOR\x10\xb0\x01\x12\x14\n\x0f\x42ITXOR_CONSTANT\x10\xb1\x01\x12\x0b\n\x06INVERT\x10\xb2\x01\x12\x0b\n\x06LSHIFT\x10\xb3\x01\x12\x14\n\x0fLSHIFT_CONSTANT\x10\xb4\x01\x12\x0b\n\x06RSHIFT\x10\xb5\x01\x12\x14\n\x0fRSHIFT_CONSTANT\x10\xb6\x01\x12\x08\n\x03\x41ND\x10\xb7\x01\x12\x11\n\x0c\x41ND_CONSTANT\x10\xb8\x01\x12\x07\n\x02OR\x10\xb9\x01\x12\x10\n\x0bOR_CONSTANT\x10\xba\x01\x12\x08\n\x03XOR\x10\xbb\x01\x12\x11\n\x0cXOR_CONSTANT\x10\xbc\x01\x12\x08\n\x03NOT\x10\xbd\x01\x12\x0c\n\x07MAXIMUM\x10\xbe\x01\x12\x15\n\x10MAXIMUM_CONSTANT\x10\xbf\x01\x12\x0c\n\x07MINIMUM\x10\xc0\x01\x12\x15\n\x10MINIMUM_CONSTANT\x10\xc1\x01\x12\x0b\n\x06\x41ROUND\x10\xc2\x01\x12\x10\n\x0b\x46LOAT_POWER\x10\xc3\x01\x12\x19\n\x14\x46LOAT_POWER_CONSTANT\x10\xc4\x01\x12\t\n\x04\x46MAX\x10\xc5\x01\x12\x12\n\rFMAX_CONSTANT\x10\xc6\x01\x12\t\n\x04\x46MIN\x10\xc7\x01\x12\x12\n\rFMIN_CONSTANT\x10\xc8\x01\x12\r\n\x08ISFINITE\x10\xc9\x01\x12\n\n\x05ISINF\x10\xca\x01\x12\n\n\x05ISNAN\x10\xcb\x01\x12\x0c\n\x07SIGNBIT\x10\xcc\x01\x12\r\n\x08\x43OPYSIGN\x10\xcd\x01\x12\x16\n\x11\x43OPYSIGN_CONSTANT\x10\xce\x01\x12\x0e\n\tNEXTAFTER\x10\xcf\x01\x12\x17\n\x12NEXTAFTER_CONSTANT\x10\xd0\x01\x12\x0c\n\x07SPACING\x10\xd1\x01\x12\n\n\x05LDEXP\x10\xd2\x01\x12\x13\n\x0eLDEXP_CONSTANT\x10\xd3\x01\x12\n\n\x05\x46REXP\x10\xd4\x01\x12\t\n\x04MODF\x10\xd5\x01\x12\n\n\x05\x46LOOR\x10\xd6\x01\x12\t\n\x04\x43\x45IL\x10\xd7\x01\x12\n\n\x05TRUNC\x10\xd8\x01\x12\x0c\n\x07\x44\x45GREES\x10\xd9\x01\x12\x0c\n\x07RADIANS\x10\xda\x01\x12\t\n\x04\x43LIP\x10\xdb\x01\x12\x0b\n\x06ISREAL\x10\xdc\x01\x12\x0e\n\tISCOMPLEX\x10\xdd\x01\x12\t\n\x04REAL\x10\xde\x01\x12\t\n\x04IMAG\x10\xdf\x01\x12\x08\n\x03\x46IX\x10\xe0\x01\x12\x07\n\x02I0\x10\xe1\x01\x12\t\n\x04SINC\x10\xe2\x01\x12\x0f\n\nNAN_TO_NUM\x10\xe3\x01\x12\x0c\n\x07ISCLOSE\x10\xe4\x01\x12\x15\n\x10ISCLOSE_CONSTANT\x10\xe5\x01\x12\x0b\n\x06\x44IVMOD\x10\xe7\x01\x12\n\n\x05\x41NGLE\x10\xe8\x01\x12\r\n\x08SET_REAL\x10\xe9\x01\x12\x16\n\x11SET_REAL_CONSTANT\x10\xea\x01\x12\r\n\x08SET_IMAG\x10\xeb\x01\x12\x16\n\x11SET_IMAG_CONSTANT\x10\xec\x01\x12\r\n\x08TREE_ADD\x10\xfb\x01\x12\x12\n\rTREE_MULTIPLY\x10\xfc\x01\x12\x0b\n\x06\x43UMSUM\x10\xad\x02\x12\x0c\n\x07\x43UMPROD\x10\xae\x02\x12\t\n\x04PROD\x10\xaf\x02\x12\x08\n\x03SUM\x10\xb0\x02\x12\x08\n\x03MAX\x10\xb1\x02\x12\x08\n\x03MIN\x10\xb2\x02\x12\x08\n\x03\x41LL\x10\xb3\x02\x12\x08\n\x03\x41NY\x10\xb4\x02\x12\x0f\n\nMEAN_CHUNK\x10\xb5\x02\x12\x11\n\x0cMEAN_COMBINE\x10\xb6\x02\x12\t\n\x04MEAN\x10\xb7\x02\x12\x0b\n\x06\x41RGMAX\x10\xb8\x02\x12\x11\n\x0c\x41RGMAX_CHUNK\x10\xb9\x02\x12\x13\n\x0e\x41RGMAX_COMBINE\x10\xba\x02\x12\x0b\n\x06\x41RGMIN\x10\xbb\x02\x12\x11\n\x0c\x41RGMIN_CHUNK\x10\xbc\x02\x12\x13\n\x0e\x41RGMIN_COMBINE\x10\xbd\x02\x12\x0b\n\x06NANSUM\x10\xbe\x02\x12\x0b\n\x06NANMAX\x10\xbf\x02\x12\x0b\n\x06NANMIN\x10\xc0\x02\x12\x0c\n\x07NANPROD\x10\xc1\x02\x12\x0c\n\x07NANMEAN\x10\xc2\x02\x12\x12\n\rNANMEAN_CHUNK\x10\xc3\x02\x12\x0e\n\tNANARGMAX\x10\xc4\x02\x12\x14\n\x0fNANARGMAX_CHUNK\x10\xc5\x02\x12\x16\n\x11NANARGMAX_COMBINE\x10\xc6\x02\x12\x0e\n\tNANARGMIN\x10\xc7\x02\x12\x14\n\x0fNANARGMIN_CHUNK\x10\xc8\x02\x12\x16\n\x11NANARGMIN_COMBINE\x10\xc9\x02\x12\x12\n\rCOUNT_NONZERO\x10\xca\x02\x12\x11\n\x0cMOMENT_CHUNK\x10\xcb\x02\x12\x14\n\x0fNANMOMENT_CHUNK\x10\xcc\x02\x12\x13\n\x0eMOMENT_COMBINE\x10\xcd\x02\x12\x16\n\x11NANMOMENT_COMBINE\x10\xce\x02\x12\x0b\n\x06MOMENT\x10\xcf\x02\x12\x0e\n\tNANMOMENT\x10\xd0\x02\x12\x08\n\x03VAR\x10\xd1\x02\x12\x08\n\x03STD\x10\xd2\x02\x12\x0b\n\x06NANVAR\x10\xd3\x02\x12\x0b\n\x06NANSTD\x10\xd4\x02\x12\x0e\n\tNANCUMSUM\x10\xd5\x02\x12\x0f\n\nNANCUMPROD\x10\xd6\x02\x12\x0c\n\x07RESHAPE\x10\x91\x03\x12\n\n\x05SLICE\x10\x92\x03\x12\n\n\x05INDEX\x10\x93\x03\x12\x12\n\rINDEXSETVALUE\x10\x94\x03\x12\x10\n\x0b\x43ONCATENATE\x10\x95\x03\x12\x0c\n\x07RECHUNK\x10\x96\x03\x12\x0b\n\x06\x41STYPE\x10\x97\x03\x12\x0e\n\tTRANSPOSE\x10\x98\x03\x12\r\n\x08SWAPAXES\x10\x99\x03\x12\x11\n\x0c\x42ROADCAST_TO\x10\x9a\x03\x12\n\n\x05STACK\x10\x9b\x03\x12\n\n\x05WHERE\x10\x9c\x03\x12\x0b\n\x06\x43HOOSE\x10\x9d\x03\x12\x0c\n\x07NONZERO\x10\x9e\x03\x12\r\n\x08\x41RGWHERE\x10\x9f\x03\x12\x12\n\rUNRAVEL_INDEX\x10\xa0\x03\x12\x16\n\x11RAVEL_MULTI_INDEX\x10\xa1\x03\x12\x10\n\x0b\x41RRAY_SPLIT\x10\xa2\x03\x12\x0c\n\x07SQUEEZE\x10\xa3\x03\x12\r\n\x08\x44IGITIZE\x10\xa4\x03\x12\x0b\n\x06REPEAT\x10\xa5\x03\x12\x0b\n\x06\x43OPYTO\x10\xa6\x03\x12\t\n\x04ISIN\x10\xa7\x03\x12\x0e\n\tTENSORDOT\x10\xf5\x03\x12\x08\n\x03\x44OT\x10\xf6\x03\x12\x0b\n\x06MATMUL\x10\xf7\x03\x12\r\n\x08\x43HOLESKY\x10\xfe\x03\x12\x07\n\x02QR\x10\xff\x03\x12\x08\n\x03SVD\x10\x80\x04\x12\x07\n\x02LU\x10\x81\x04\x12\x15\n\x10SOLVE_TRIANGULAR\x10\x88\x04\x12\t\n\x04NORM\x10\x92\x04\x12\x08\n\x03\x46\x46T\x10\xd9\x04\x12\t\n\x04IFFT\x10\xda\x04\x12\t\n\x04\x46\x46T2\x10\xdb\x04\x12\n\n\x05IFFT2\x10\xdc\x04\x12\t\n\x04\x46\x46TN\x10\xdd\x04\x12\n\n\x05IFFTN\x10\xde\x04\x12\t\n\x04RFFT\x10\xdf\x04\x12\n\n\x05IRFFT\x10\xe0\x04\x12\n\n\x05RFFT2\x10\xe1\x04\x12\x0b\n\x06IRFFT2\x10\xe2\x04\x12\n\n\x05RFFTN\x10\xe3\x04\x12\x0b\n\x06IRFFTN\x10\xe4\x04\x12\t\n\x04HFFT\x10\xe5\x04\x12\n\n\x05IHFFT\x10\xe6\x04\x12\x0c\n\x07\x46\x46TFREQ\x10\xe7\x04\x12\x12\n\rFFTFREQ_CHUNK\x10\xe8\x04\x12\r\n\x08RFFTFREQ\x10\xe9\x04\x12\r\n\x08\x46\x46TSHIFT\x10\xea\x04\x12\x0e\n\tIFFTSHIFT\x10\xeb\x04\x12\x1e\n\x19SPARSE_MATRIX_DATA_SOURCE\x10\xbd\x05\x12\x14\n\x0f\x44\x45NSE_TO_SPARSE\x10\xbe\x05\x12\x14\n\x0fSPARSE_TO_DENSE\x10\xbf\x05\x12\t\n\x04\x46USE\x10\xa1\x06\x12\n\n\x05\x45NTER\x10\x85\x07\x12\n\n\x05LEAVE\x10\x86\x07\x12\x0f\n\nFIX_LATEST\x10\x87\x07\x12\x0c\n\x07IF_ELSE\x10\x88\x07\x12\x0e\n\tNEXT_ITER\x10\x89\x07\x12\x0e\n\tTABLE_COO\x10\xeb\x07\x12\x0e\n\tSTORE_COO\x10\xec\x07\x12\x0b\n\x05\x46\x45TCH\x10\xbf\x84=b\x06proto3')
  ,
  dependencies=[mars_dot_serialize_dot_protos_dot_value__pb2.DESCRIPTOR,])



_OPERANDDEF_OPERANDTYPE = _descriptor.EnumDescriptor(
  name='OperandType',
  full_name='OperandDef.OperandType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NULL', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SCALAR', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_DATA_SOURCE', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ONES', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ONES_LIKE', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ZEROS', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ZEROS_LIKE', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_EMPTY', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_EMPTY_LIKE', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_FULL', index=9, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ARANGE', index=10, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_INDICES', index=11, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_DIAG', index=12, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_EYE', index=13, number=13,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_LINSPACE', index=14, number=14,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_TRIU', index=15, number=15,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_TRIL', index=16, number=16,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RAND', index=17, number=41,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDN', index=18, number=42,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDINT', index=19, number=43,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDOM_INTEGERS', index=20, number=44,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDOM_SAMPLE', index=21, number=45,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDOM', index=22, number=46,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANF', index=23, number=47,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_SAMPLE', index=24, number=48,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_BYTES', index=25, number=49,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_BETA', index=26, number=50,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_BINOMIAL', index=27, number=51,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_CHISQUARE', index=28, number=52,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_CHOICE', index=29, number=53,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_DIRICHLET', index=30, number=54,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_EXPONENTIAL', index=31, number=55,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_F', index=32, number=56,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_GAMMA', index=33, number=57,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_GEOMETRIC', index=34, number=58,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_GUMBEL', index=35, number=59,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_HYPERGEOMETRIC', index=36, number=60,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LAPLACE', index=37, number=61,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LOGISTIC', index=38, number=62,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LOGNORMAL', index=39, number=63,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LOGSERIES', index=40, number=64,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_MULTINOMIAL', index=41, number=65,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_MULTIVARIATE_NORMAL', index=42, number=66,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NEGATIVE_BINOMIAL', index=43, number=67,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NONCENTRAL_CHISQURE', index=44, number=68,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NONCENTRAL_F', index=45, number=69,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NORMAL', index=46, number=70,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_PARETO', index=47, number=71,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_PERMUTATION', index=48, number=72,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_POSSION', index=49, number=73,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_POWER', index=50, number=74,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RAYLEIGH', index=51, number=75,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_SHUFFLE', index=52, number=76,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_CAUCHY', index=53, number=77,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_EXPONENTIAL', index=54, number=78,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_GAMMMA', index=55, number=79,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_NORMAL', index=56, number=80,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_T', index=57, number=81,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_TOMAXINT', index=58, number=82,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_TRIANGULAR', index=59, number=83,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_UNIFORM', index=60, number=84,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_VONMISES', index=61, number=85,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_WALD', index=62, number=86,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_WEIBULL', index=63, number=87,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_ZIPF', index=64, number=88,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ADD', index=65, number=101,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ADD_CONSTANT', index=66, number=102,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUB', index=67, number=103,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUB_CONSTANT', index=68, number=104,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MUL', index=69, number=105,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MUL_CONSTANT', index=70, number=106,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIV', index=71, number=107,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIV_CONSTANT', index=72, number=108,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRUEDIV', index=73, number=109,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TDIV_CONSTANT', index=74, number=110,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOORDIV', index=75, number=111,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FDIV_CONSTANT', index=76, number=112,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POW', index=77, number=113,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POW_CONSTANT', index=78, number=114,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOD', index=79, number=115,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOD_CONSTANT', index=80, number=116,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMOD', index=81, number=117,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMOD_CONSTANT', index=82, number=118,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOGADDEXP', index=83, number=119,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LAE_CONSTANT', index=84, number=120,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOGADDEXP2', index=85, number=121,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LAE2_CONSTANT', index=86, number=122,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEGATIVE', index=87, number=123,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POSITIVE', index=88, number=124,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ABSOLUTE', index=89, number=125,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FABS', index=90, number=126,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ABS', index=91, number=127,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RINT', index=92, number=128,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIGN', index=93, number=129,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONJ', index=94, number=130,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXP', index=95, number=131,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXP2', index=96, number=132,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG', index=97, number=133,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG2', index=98, number=134,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG10', index=99, number=135,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXPM1', index=100, number=136,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG1P', index=101, number=137,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SQRT', index=102, number=138,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SQUARE', index=103, number=139,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CBRT', index=104, number=140,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RECIPROCAL', index=105, number=141,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EQ', index=106, number=142,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EQ_CONSTANT', index=107, number=143,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NE', index=108, number=144,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NE_CONSTANT', index=109, number=145,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LT', index=110, number=146,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LT_CONSTANT', index=111, number=147,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LE', index=112, number=148,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LE_CONSTANT', index=113, number=149,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GT', index=114, number=150,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GT_CONSTANT', index=115, number=151,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GE', index=116, number=152,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GE_CONSTANT', index=117, number=153,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIN', index=118, number=154,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COS', index=119, number=155,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TAN', index=120, number=156,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCSIN', index=121, number=157,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCCOS', index=122, number=158,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCTAN', index=123, number=159,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCTAN2', index=124, number=160,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCT2_CONSTANT', index=125, number=161,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HYPOT', index=126, number=162,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HYPOT_CONSTANT', index=127, number=163,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SINH', index=128, number=164,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COSH', index=129, number=165,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TANH', index=130, number=166,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCSINH', index=131, number=167,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCCOSH', index=132, number=168,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCTANH', index=133, number=169,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DEG2RAD', index=134, number=170,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAD2DEG', index=135, number=171,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITAND', index=136, number=172,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITAND_CONSTANT', index=137, number=173,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITOR', index=138, number=174,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITOR_CONSTANT', index=139, number=175,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITXOR', index=140, number=176,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITXOR_CONSTANT', index=141, number=177,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INVERT', index=142, number=178,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LSHIFT', index=143, number=179,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LSHIFT_CONSTANT', index=144, number=180,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RSHIFT', index=145, number=181,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RSHIFT_CONSTANT', index=146, number=182,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AND', index=147, number=183,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AND_CONSTANT', index=148, number=184,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OR', index=149, number=185,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OR_CONSTANT', index=150, number=186,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='XOR', index=151, number=187,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='XOR_CONSTANT', index=152, number=188,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NOT', index=153, number=189,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MAXIMUM', index=154, number=190,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MAXIMUM_CONSTANT', index=155, number=191,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MINIMUM', index=156, number=192,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MINIMUM_CONSTANT', index=157, number=193,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AROUND', index=158, number=194,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT_POWER', index=159, number=195,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT_POWER_CONSTANT', index=160, number=196,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMAX', index=161, number=197,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMAX_CONSTANT', index=162, number=198,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMIN', index=163, number=199,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMIN_CONSTANT', index=164, number=200,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISFINITE', index=165, number=201,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISINF', index=166, number=202,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISNAN', index=167, number=203,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIGNBIT', index=168, number=204,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COPYSIGN', index=169, number=205,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COPYSIGN_CONSTANT', index=170, number=206,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEXTAFTER', index=171, number=207,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEXTAFTER_CONSTANT', index=172, number=208,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SPACING', index=173, number=209,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LDEXP', index=174, number=210,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LDEXP_CONSTANT', index=175, number=211,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FREXP', index=176, number=212,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MODF', index=177, number=213,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOOR', index=178, number=214,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CEIL', index=179, number=215,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRUNC', index=180, number=216,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DEGREES', index=181, number=217,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RADIANS', index=182, number=218,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIP', index=183, number=219,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISREAL', index=184, number=220,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISCOMPLEX', index=185, number=221,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REAL', index=186, number=222,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IMAG', index=187, number=223,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIX', index=188, number=224,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='I0', index=189, number=225,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SINC', index=190, number=226,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NAN_TO_NUM', index=191, number=227,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISCLOSE', index=192, number=228,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISCLOSE_CONSTANT', index=193, number=229,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIVMOD', index=194, number=231,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ANGLE', index=195, number=232,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET_REAL', index=196, number=233,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET_REAL_CONSTANT', index=197, number=234,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET_IMAG', index=198, number=235,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET_IMAG_CONSTANT', index=199, number=236,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TREE_ADD', index=200, number=251,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TREE_MULTIPLY', index=201, number=252,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMSUM', index=202, number=301,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMPROD', index=203, number=302,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PROD', index=204, number=303,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUM', index=205, number=304,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MAX', index=206, number=305,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MIN', index=207, number=306,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ALL', index=208, number=307,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ANY', index=209, number=308,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEAN_CHUNK', index=210, number=309,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEAN_COMBINE', index=211, number=310,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEAN', index=212, number=311,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMAX', index=213, number=312,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMAX_CHUNK', index=214, number=313,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMAX_COMBINE', index=215, number=314,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMIN', index=216, number=315,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMIN_CHUNK', index=217, number=316,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMIN_COMBINE', index=218, number=317,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANSUM', index=219, number=318,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMAX', index=220, number=319,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMIN', index=221, number=320,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANPROD', index=222, number=321,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMEAN', index=223, number=322,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMEAN_CHUNK', index=224, number=323,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMAX', index=225, number=324,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMAX_CHUNK', index=226, number=325,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMAX_COMBINE', index=227, number=326,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMIN', index=228, number=327,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMIN_CHUNK', index=229, number=328,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMIN_COMBINE', index=230, number=329,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COUNT_NONZERO', index=231, number=330,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOMENT_CHUNK', index=232, number=331,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMOMENT_CHUNK', index=233, number=332,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOMENT_COMBINE', index=234, number=333,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMOMENT_COMBINE', index=235, number=334,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOMENT', index=236, number=335,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMOMENT', index=237, number=336,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='VAR', index=238, number=337,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STD', index=239, number=338,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANVAR', index=240, number=339,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANSTD', index=241, number=340,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANCUMSUM', index=242, number=341,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANCUMPROD', index=243, number=342,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESHAPE', index=244, number=401,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SLICE', index=245, number=402,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INDEX', index=246, number=403,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INDEXSETVALUE', index=247, number=404,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONCATENATE', index=248, number=405,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RECHUNK', index=249, number=406,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ASTYPE', index=250, number=407,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRANSPOSE', index=251, number=408,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SWAPAXES', index=252, number=409,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BROADCAST_TO', index=253, number=410,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STACK', index=254, number=411,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WHERE', index=255, number=412,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CHOOSE', index=256, number=413,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NONZERO', index=257, number=414,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGWHERE', index=258, number=415,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNRAVEL_INDEX', index=259, number=416,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAVEL_MULTI_INDEX', index=260, number=417,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARRAY_SPLIT', index=261, number=418,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SQUEEZE', index=262, number=419,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIGITIZE', index=263, number=420,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REPEAT', index=264, number=421,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COPYTO', index=265, number=422,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISIN', index=266, number=423,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSORDOT', index=267, number=501,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DOT', index=268, number=502,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MATMUL', index=269, number=503,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CHOLESKY', index=270, number=510,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='QR', index=271, number=511,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVD', index=272, number=512,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LU', index=273, number=513,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SOLVE_TRIANGULAR', index=274, number=520,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NORM', index=275, number=530,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFT', index=276, number=601,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFT', index=277, number=602,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFT2', index=278, number=603,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFT2', index=279, number=604,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTN', index=280, number=605,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFTN', index=281, number=606,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFT', index=282, number=607,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IRFFT', index=283, number=608,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFT2', index=284, number=609,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IRFFT2', index=285, number=610,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFTN', index=286, number=611,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IRFFTN', index=287, number=612,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HFFT', index=288, number=613,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IHFFT', index=289, number=614,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTFREQ', index=290, number=615,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTFREQ_CHUNK', index=291, number=616,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFTFREQ', index=292, number=617,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTSHIFT', index=293, number=618,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFTSHIFT', index=294, number=619,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SPARSE_MATRIX_DATA_SOURCE', index=295, number=701,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DENSE_TO_SPARSE', index=296, number=702,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SPARSE_TO_DENSE', index=297, number=703,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FUSE', index=298, number=801,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENTER', index=299, number=901,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LEAVE', index=300, number=902,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIX_LATEST', index=301, number=903,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IF_ELSE', index=302, number=904,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEXT_ITER', index=303, number=905,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TABLE_COO', index=304, number=1003,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STORE_COO', index=305, number=1004,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FETCH', index=306, number=999999,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=237,
  serialized_end=5070,
)
_sym_db.RegisterEnumDescriptor(_OPERANDDEF_OPERANDTYPE)


_OPERANDDEF_ATTRENTRY = _descriptor.Descriptor(
  name='AttrEntry',
  full_name='OperandDef.AttrEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='OperandDef.AttrEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='OperandDef.AttrEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=183,
  serialized_end=234,
)

_OPERANDDEF = _descriptor.Descriptor(
  name='OperandDef',
  full_name='OperandDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='OperandDef.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='attr', full_name='OperandDef.attr', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='operands', full_name='OperandDef.operands', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='OperandDef.id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_OPERANDDEF_ATTRENTRY, ],
  enum_types=[
    _OPERANDDEF_OPERANDTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=5070,
)

_OPERANDDEF_ATTRENTRY.fields_by_name['value'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_OPERANDDEF_ATTRENTRY.containing_type = _OPERANDDEF
_OPERANDDEF.fields_by_name['attr'].message_type = _OPERANDDEF_ATTRENTRY
_OPERANDDEF.fields_by_name['operands'].message_type = _OPERANDDEF
_OPERANDDEF_OPERANDTYPE.containing_type = _OPERANDDEF
DESCRIPTOR.message_types_by_name['OperandDef'] = _OPERANDDEF
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

OperandDef = _reflection.GeneratedProtocolMessageType('OperandDef', (_message.Message,), dict(

  AttrEntry = _reflection.GeneratedProtocolMessageType('AttrEntry', (_message.Message,), dict(
    DESCRIPTOR = _OPERANDDEF_ATTRENTRY,
    __module__ = 'mars.serialize.protos.operand_pb2'
    # @@protoc_insertion_point(class_scope:OperandDef.AttrEntry)
    ))
  ,
  DESCRIPTOR = _OPERANDDEF,
  __module__ = 'mars.serialize.protos.operand_pb2'
  # @@protoc_insertion_point(class_scope:OperandDef)
  ))
_sym_db.RegisterMessage(OperandDef)
_sym_db.RegisterMessage(OperandDef.AttrEntry)


_OPERANDDEF_ATTRENTRY._options = None
# @@protoc_insertion_point(module_scope)
