/* Generated by Cython 0.29.2 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "C:\\Python36-x64\\lib\\site-packages\\numpy\\core\\include\\numpy\\arrayobject.h",
            "C:\\Python36-x64\\lib\\site-packages\\numpy\\core\\include\\numpy\\ufuncobject.h"
        ],
        "extra_compile_args": [
            "/Ot",
            "/IC:\\projects\\mars\\misc"
        ],
        "include_dirs": [
            "C:\\Python36-x64\\lib\\site-packages\\numpy\\core\\include"
        ],
        "name": "mars.serialize.jsonserializer",
        "sources": [
            "mars/serialize/jsonserializer.pyx"
        ]
    },
    "module_name": "mars.serialize.jsonserializer"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_2"
#define CYTHON_HEX_VERSION 0x001D02F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_USE_DICT_VERSIONS
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
        static PY_UINT64_T __pyx_dict_version = 0;\
        static PyObject *__pyx_dict_cached_value = NULL;\
        if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
            (VAR) = __pyx_dict_cached_value;\
        } else {\
            (VAR) = __pyx_dict_cached_value = (LOOKUP);\
            __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
        }\
    }
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0; // PyThread_create_key reports success always
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif // TSS (Thread Specific Storage) API
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__mars__serialize__jsonserializer
#define __PYX_HAVE_API__mars__serialize__jsonserializer
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "mars\\serialize\\jsonserializer.pyx",
  "stringsource",
  "__init__.pxd",
  "type.pxd",
  "mars\\serialize\\core.pxd",
};

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":776
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":778
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":779
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":784
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":785
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":786
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":791
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":800
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":801
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":804
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":805
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":806
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":808
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":809
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":811
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":812
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_4mars_9serialize_4core_Identity;
struct __pyx_obj_4mars_9serialize_4core_List;
struct __pyx_obj_4mars_9serialize_4core_Tuple;
struct __pyx_obj_4mars_9serialize_4core_Dict;
struct __pyx_obj_4mars_9serialize_4core_Reference;
struct __pyx_obj_4mars_9serialize_4core_OneOf;
struct __pyx_obj_4mars_9serialize_4core_ValueType;
struct __pyx_obj_4mars_9serialize_4core_Field;
struct __pyx_obj_4mars_9serialize_4core_AnyField;
struct __pyx_obj_4mars_9serialize_4core_IdentityField;
struct __pyx_obj_4mars_9serialize_4core_BoolField;
struct __pyx_obj_4mars_9serialize_4core_Int8Field;
struct __pyx_obj_4mars_9serialize_4core_Int16Field;
struct __pyx_obj_4mars_9serialize_4core_Int32Field;
struct __pyx_obj_4mars_9serialize_4core_Int64Field;
struct __pyx_obj_4mars_9serialize_4core_UInt8Field;
struct __pyx_obj_4mars_9serialize_4core_UInt16Field;
struct __pyx_obj_4mars_9serialize_4core_UInt32Field;
struct __pyx_obj_4mars_9serialize_4core_UInt64Field;
struct __pyx_obj_4mars_9serialize_4core_Float16Field;
struct __pyx_obj_4mars_9serialize_4core_Float32Field;
struct __pyx_obj_4mars_9serialize_4core_Float64Field;
struct __pyx_obj_4mars_9serialize_4core_StringField;
struct __pyx_obj_4mars_9serialize_4core_BytesField;
struct __pyx_obj_4mars_9serialize_4core_UnicodeField;
struct __pyx_obj_4mars_9serialize_4core_KeyField;
struct __pyx_obj_4mars_9serialize_4core_NDArrayField;
struct __pyx_obj_4mars_9serialize_4core_Datetime64Field;
struct __pyx_obj_4mars_9serialize_4core_Timedelta64Field;
struct __pyx_obj_4mars_9serialize_4core_DataTypeField;
struct __pyx_obj_4mars_9serialize_4core_ListField;
struct __pyx_obj_4mars_9serialize_4core_TupleField;
struct __pyx_obj_4mars_9serialize_4core_DictField;
struct __pyx_obj_4mars_9serialize_4core_ReferenceField;
struct __pyx_obj_4mars_9serialize_4core_OneOfField;
struct __pyx_obj_4mars_9serialize_4core_KeyPlaceholder;
struct __pyx_obj_4mars_9serialize_4core_AttrWrapper;
struct __pyx_obj_4mars_9serialize_4core_Provider;
struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider;
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key;
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr;
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list;
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb;
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple;
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr;
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict;
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb;
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value;
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":815
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":816
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":817
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":819
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_4mars_9serialize_4core_8Provider_serialize_model;
struct __pyx_opt_args_4mars_9serialize_4core_8Provider_serialize_attribute_as_dict;

/* "core.pxd":18
 * 
 * 
 * cpdef enum PrimitiveType:             # <<<<<<<<<<<<<<
 *     bool = 1
 *     int8 = 2
 */
enum __pyx_t_4mars_9serialize_4core_PrimitiveType {
  __pyx_e_4mars_9serialize_4core_bool = 1,
  __pyx_e_4mars_9serialize_4core_int8 = 2,
  __pyx_e_4mars_9serialize_4core_int16 = 3,
  __pyx_e_4mars_9serialize_4core_int32 = 4,
  __pyx_e_4mars_9serialize_4core_int64 = 5,
  __pyx_e_4mars_9serialize_4core_uint8 = 6,
  __pyx_e_4mars_9serialize_4core_uint16 = 7,
  __pyx_e_4mars_9serialize_4core_uint32 = 8,
  __pyx_e_4mars_9serialize_4core_uint64 = 9,
  __pyx_e_4mars_9serialize_4core_float16 = 10,
  __pyx_e_4mars_9serialize_4core_float32 = 11,
  __pyx_e_4mars_9serialize_4core_float64 = 12,
  __pyx_e_4mars_9serialize_4core_bytes = 13,
  __pyx_e_4mars_9serialize_4core_unicode = 14
};

/* "core.pxd":35
 * 
 * 
 * cpdef enum ExtendType:             # <<<<<<<<<<<<<<
 *     slice = 15
 *     arr = 16
 */
enum __pyx_t_4mars_9serialize_4core_ExtendType {
  __pyx_e_4mars_9serialize_4core_slice = 15,
  __pyx_e_4mars_9serialize_4core_arr = 16,
  __pyx_e_4mars_9serialize_4core_dtype = 17,
  __pyx_e_4mars_9serialize_4core_key = 18,
  __pyx_e_4mars_9serialize_4core_datetime64 = 19,
  __pyx_e_4mars_9serialize_4core_timedelta64 = 20
};

/* "core.pxd":214
 * 
 * 
 * cpdef enum ProviderType:             # <<<<<<<<<<<<<<
 *     protobuf = 1
 *     json = 2
 */
enum __pyx_t_4mars_9serialize_4core_ProviderType {
  __pyx_e_4mars_9serialize_4core_protobuf = 1,
  __pyx_e_4mars_9serialize_4core_json = 2
};

/* "core.pxd":223
 * 
 *     cpdef serialize_field(self, Field field, model_instance, obj)
 *     cpdef serialize_model(self, model_instance, obj=?)             # <<<<<<<<<<<<<<
 *     cpdef serialize_attribute_as_dict(self, model_instance, obj=?)
 *     cpdef deserialize_model(self, model_cls, obj, list callbacks, dict key_to_instance)
 */
struct __pyx_opt_args_4mars_9serialize_4core_8Provider_serialize_model {
  int __pyx_n;
  PyObject *obj;
};

/* "core.pxd":224
 *     cpdef serialize_field(self, Field field, model_instance, obj)
 *     cpdef serialize_model(self, model_instance, obj=?)
 *     cpdef serialize_attribute_as_dict(self, model_instance, obj=?)             # <<<<<<<<<<<<<<
 *     cpdef deserialize_model(self, model_cls, obj, list callbacks, dict key_to_instance)
 */
struct __pyx_opt_args_4mars_9serialize_4core_8Provider_serialize_attribute_as_dict {
  int __pyx_n;
  PyObject *obj;
};
struct __pyx_opt_args_4mars_7utils_c_to_str;
struct __pyx_opt_args_4mars_7utils_c_to_binary;
struct __pyx_opt_args_4mars_7utils_c_to_text;

/* "utils_c.pxd":15
 * # limitations under the License.
 * 
 * cpdef str to_str(s, encoding=*)             # <<<<<<<<<<<<<<
 * cpdef bytes to_binary(s, encoding=*)
 * cpdef unicode to_text(s, encoding=*)
 */
struct __pyx_opt_args_4mars_7utils_c_to_str {
  int __pyx_n;
  PyObject *encoding;
};

/* "utils_c.pxd":16
 * 
 * cpdef str to_str(s, encoding=*)
 * cpdef bytes to_binary(s, encoding=*)             # <<<<<<<<<<<<<<
 * cpdef unicode to_text(s, encoding=*)
 */
struct __pyx_opt_args_4mars_7utils_c_to_binary {
  int __pyx_n;
  PyObject *encoding;
};

/* "utils_c.pxd":17
 * cpdef str to_str(s, encoding=*)
 * cpdef bytes to_binary(s, encoding=*)
 * cpdef unicode to_text(s, encoding=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_4mars_7utils_c_to_text {
  int __pyx_n;
  PyObject *encoding;
};
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list;
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple;
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict;
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value;
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value;
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value;

/* "mars/serialize/jsonserializer.pyx":156
 *             return KeyPlaceholder(*(to_str(v) for v in obj['value']))
 * 
 *     cdef inline dict _serialize_list(self, list value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.list),
 */
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list {
  int __pyx_n;
  PyObject *tp;
  int weak_ref;
};

/* "mars/serialize/jsonserializer.pyx":185
 *         return res
 * 
 *     cdef inline dict _serialize_tuple(self, tuple value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if tp is None or not isinstance(tp.type, tuple):
 *             return {
 */
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple {
  int __pyx_n;
  PyObject *tp;
  int weak_ref;
};

/* "mars/serialize/jsonserializer.pyx":205
 *                      for it_val in obj['value'])
 * 
 *     cdef inline dict _serialize_dict(self, object value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.dict),
 */
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict {
  int __pyx_n;
  PyObject *tp;
  int weak_ref;
};

/* "mars/serialize/jsonserializer.pyx":263
 *         return self._deserialize_datetime64_timedelta64(obj, callbacks)
 * 
 *     cdef inline object _serialize_typed_value(self, value, tp, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if type(tp) not in (List, Tuple, Dict) and weak_ref:
 *             # not iterable, and is weak ref
 */
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value {
  int __pyx_n;
  int weak_ref;
};

/* "mars/serialize/jsonserializer.pyx":302
 *             raise TypeError('Unknown type to serialize: {0}'.format(tp))
 * 
 *     cdef inline object _serialize_untyped_value(self, value, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (list, tuple, dict)) and weak_ref:
 *             # not iterable, and is weak ref
 */
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value {
  int __pyx_n;
  int weak_ref;
};

/* "mars/serialize/jsonserializer.pyx":343
 *             raise TypeError('Unknown type to serialize: {0}'.format(type(value)))
 * 
 *     cdef inline object _serialize_value(self, value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if tp is None:
 *             return self._serialize_untyped_value(value, weak_ref=weak_ref)
 */
struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value {
  int __pyx_n;
  PyObject *tp;
  int weak_ref;
};

/* "core.pxd":44
 * 
 * 
 * cdef class Identity:             # <<<<<<<<<<<<<<
 *     cdef public object type
 *     cdef public str name
 */
struct __pyx_obj_4mars_9serialize_4core_Identity {
  PyObject_HEAD
  PyObject *type;
  PyObject *name;
};


/* "core.pxd":49
 * 
 * 
 * cdef class List:             # <<<<<<<<<<<<<<
 *     cdef public object type
 *     cdef public str name
 */
struct __pyx_obj_4mars_9serialize_4core_List {
  PyObject_HEAD
  PyObject *type;
  PyObject *name;
};


/* "core.pxd":54
 * 
 * 
 * cdef class Tuple:             # <<<<<<<<<<<<<<
 *     cdef public object type
 *     cdef public str name
 */
struct __pyx_obj_4mars_9serialize_4core_Tuple {
  PyObject_HEAD
  PyObject *type;
  PyObject *name;
};


/* "core.pxd":59
 * 
 * 
 * cdef class Dict:             # <<<<<<<<<<<<<<
 *     cdef public object key_type
 *     cdef public object value_type
 */
struct __pyx_obj_4mars_9serialize_4core_Dict {
  PyObject_HEAD
  PyObject *key_type;
  PyObject *value_type;
  PyObject *name;
};


/* "core.pxd":65
 * 
 * 
 * cdef class Reference:             # <<<<<<<<<<<<<<
 *     cdef public object model
 *     cdef public str name
 */
struct __pyx_obj_4mars_9serialize_4core_Reference {
  PyObject_HEAD
  PyObject *model;
  PyObject *name;
};


/* "core.pxd":70
 * 
 * 
 * cdef class OneOf:             # <<<<<<<<<<<<<<
 *     cdef public tuple references
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_OneOf {
  PyObject_HEAD
  PyObject *references;
};


/* "core.pxd":74
 * 
 * 
 * cdef class ValueType:             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_ValueType {
  PyObject_HEAD
};


/* "core.pxd":78
 * 
 * 
 * cdef class Field:             # <<<<<<<<<<<<<<
 *     cdef object tag
 *     cdef object default_val
 */
struct __pyx_obj_4mars_9serialize_4core_Field {
  PyObject_HEAD
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field *__pyx_vtab;
  PyObject *tag;
  PyObject *default_val;
  PyObject *_tag_name;
  PyObject *_type;
  int weak_ref;
  PyObject *model;
  PyObject *attr;
  PyObject *on_serialize;
  PyObject *on_deserialize;
};


/* "core.pxd":95
 * 
 * 
 * cdef class AnyField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_AnyField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":99
 * 
 * 
 * cdef class IdentityField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_IdentityField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":103
 * 
 * 
 * cdef class BoolField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_BoolField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":107
 * 
 * 
 * cdef class Int8Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_Int8Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":111
 * 
 * 
 * cdef class Int16Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_Int16Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":115
 * 
 * 
 * cdef class Int32Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_Int32Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":119
 * 
 * 
 * cdef class Int64Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_Int64Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":123
 * 
 * 
 * cdef class UInt8Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_UInt8Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":127
 * 
 * 
 * cdef class UInt16Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_UInt16Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":131
 * 
 * 
 * cdef class UInt32Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_UInt32Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":135
 * 
 * 
 * cdef class UInt64Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_UInt64Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":139
 * 
 * 
 * cdef class Float16Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_Float16Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":143
 * 
 * 
 * cdef class Float32Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_Float32Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":147
 * 
 * 
 * cdef class Float64Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_Float64Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":151
 * 
 * 
 * cdef class StringField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_StringField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":155
 * 
 * 
 * cdef class BytesField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_BytesField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":159
 * 
 * 
 * cdef class UnicodeField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_UnicodeField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":163
 * 
 * 
 * cdef class KeyField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_KeyField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":167
 * 
 * 
 * cdef class NDArrayField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_NDArrayField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":171
 * 
 * 
 * cdef class Datetime64Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_Datetime64Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":175
 * 
 * 
 * cdef class Timedelta64Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_Timedelta64Field {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":179
 * 
 * 
 * cdef class DataTypeField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_DataTypeField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":183
 * 
 * 
 * cdef class ListField(Field):             # <<<<<<<<<<<<<<
 *     cdef object _nest_ref
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_ListField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
  PyObject *_nest_ref;
};


/* "core.pxd":187
 * 
 * 
 * cdef class TupleField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_TupleField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":191
 * 
 * 
 * cdef class DictField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_DictField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
};


/* "core.pxd":195
 * 
 * 
 * cdef class ReferenceField(Field):             # <<<<<<<<<<<<<<
 *     cdef object _model
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_ReferenceField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
  PyObject *_model;
};


/* "core.pxd":199
 * 
 * 
 * cdef class OneOfField(Field):             # <<<<<<<<<<<<<<
 *     cdef public list fields
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_OneOfField {
  struct __pyx_obj_4mars_9serialize_4core_Field __pyx_base;
  PyObject *fields;
};


/* "core.pxd":203
 * 
 * 
 * cdef class KeyPlaceholder:             # <<<<<<<<<<<<<<
 *     cdef public str key
 *     cdef public str id
 */
struct __pyx_obj_4mars_9serialize_4core_KeyPlaceholder {
  PyObject_HEAD
  PyObject *key;
  PyObject *id;
};


/* "core.pxd":208
 * 
 * 
 * cdef class AttrWrapper:             # <<<<<<<<<<<<<<
 *     cdef object _obj
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_AttrWrapper {
  PyObject_HEAD
  struct __pyx_vtabstruct_4mars_9serialize_4core_AttrWrapper *__pyx_vtab;
  PyObject *_obj;
};


/* "core.pxd":219
 * 
 * 
 * cdef class Provider:             # <<<<<<<<<<<<<<
 *     cdef public object type
 * 
 */
struct __pyx_obj_4mars_9serialize_4core_Provider {
  PyObject_HEAD
  struct __pyx_vtabstruct_4mars_9serialize_4core_Provider *__pyx_vtab;
  PyObject *type;
};


/* "mars/serialize/jsonserializer.pyx":72
 * 
 * 
 * cdef class JsonSerializeProvider(Provider):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         self.type = ProviderType.json
 */
struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider {
  struct __pyx_obj_4mars_9serialize_4core_Provider __pyx_base;
};


/* "mars/serialize/jsonserializer.pyx":150
 *         }
 * 
 *     cdef inline KeyPlaceholder _deserialize_key(self, object obj, list callbacks):             # <<<<<<<<<<<<<<
 *         if PY_MAJOR_VERSION >= 3:
 *             return KeyPlaceholder(*obj['value'])
 */
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key {
  PyObject_HEAD
  PyObject *__pyx_v_obj;
};


/* "mars/serialize/jsonserializer.pyx":154
 *             return KeyPlaceholder(*obj['value'])
 *         else:
 *             return KeyPlaceholder(*(to_str(v) for v in obj['value']))             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_list(self, list value, tp=None, bint weak_ref=False):
 */
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *__pyx_outer_scope;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "mars/serialize/jsonserializer.pyx":164
 *         }
 * 
 *     cdef inline list _deserialize_list(self, object obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         cdef list res
 *         cdef int i
 */
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list {
  PyObject_HEAD
  PyObject *__pyx_v_res;
  int __pyx_v_weak_ref;
};


/* "mars/serialize/jsonserializer.pyx":174
 *             res.append(val)
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(j, v):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[v.key, v.id]
 */
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb {
  PyObject_HEAD
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *__pyx_outer_scope;
  PyObject *__pyx_v_j;
  PyObject *__pyx_v_v;
};


/* "mars/serialize/jsonserializer.pyx":201
 *             }
 * 
 *     cdef inline tuple _deserialize_tuple(self, object obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         return tuple(self._deserialize_value(it_val, callbacks, weak_ref)
 *                      for it_val in obj['value'])
 */
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple {
  PyObject_HEAD
  PyObject *__pyx_v_callbacks;
  PyObject *__pyx_v_obj;
  struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self;
  int __pyx_v_weak_ref;
};


/* "mars/serialize/jsonserializer.pyx":202
 * 
 *     cdef inline tuple _deserialize_tuple(self, object obj, list callbacks, bint weak_ref):
 *         return tuple(self._deserialize_value(it_val, callbacks, weak_ref)             # <<<<<<<<<<<<<<
 *                      for it_val in obj['value'])
 * 
 */
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr {
  PyObject_HEAD
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *__pyx_outer_scope;
  PyObject *__pyx_v_it_val;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "mars/serialize/jsonserializer.pyx":215
 *         }
 * 
 *     cdef inline object _deserialize_dict(self, object obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         res = OrderedDict()
 *         for k, v in obj['value']:
 */
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict {
  PyObject_HEAD
  PyObject *__pyx_v_res;
  int __pyx_v_weak_ref;
};


/* "mars/serialize/jsonserializer.pyx":222
 *             res[key] = value
 *             if isinstance(value, KeyPlaceholder):
 *                 def cb(ko, vo):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[vo.key, vo.id]
 */
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb {
  PyObject_HEAD
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *__pyx_outer_scope;
  PyObject *__pyx_v_ko;
  PyObject *__pyx_v_vo;
};


/* "mars/serialize/jsonserializer.pyx":407
 *             obj[tag] = self._serialize_value(val, field.type, weak_ref=field.weak_ref)
 * 
 *     cdef inline _deserialize_value(self, obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, dict):
 *             return obj
 */
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value {
  PyObject_HEAD
  int __pyx_v_weak_ref;
};


/* "mars/serialize/jsonserializer.pyx":445
 *         return field.on_deserialize(x) if field.on_deserialize is not None else x
 * 
 *     def deserialize_field(self, Field field, model_instance, obj, list callbacks, dict key_to_instance):             # <<<<<<<<<<<<<<
 *         cdef str tag
 *         cdef object val
 */
struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field {
  PyObject_HEAD
  struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field;
  PyObject *__pyx_v_model_instance;
  struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self;
  PyObject *__pyx_v_val;
};



/* "core.pxd":78
 * 
 * 
 * cdef class Field:             # <<<<<<<<<<<<<<
 *     cdef object tag
 *     cdef object default_val
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_Field {
  PyObject *(*tag_name)(struct __pyx_obj_4mars_9serialize_4core_Field *, struct __pyx_obj_4mars_9serialize_4core_Provider *, int __pyx_skip_dispatch);
  PyObject *(*serialize)(struct __pyx_obj_4mars_9serialize_4core_Field *, struct __pyx_obj_4mars_9serialize_4core_Provider *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*deserialize)(struct __pyx_obj_4mars_9serialize_4core_Field *, struct __pyx_obj_4mars_9serialize_4core_Provider *, PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_Field *__pyx_vtabptr_4mars_9serialize_4core_Field;


/* "core.pxd":95
 * 
 * 
 * cdef class AnyField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_AnyField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_AnyField *__pyx_vtabptr_4mars_9serialize_4core_AnyField;


/* "core.pxd":99
 * 
 * 
 * cdef class IdentityField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_IdentityField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_IdentityField *__pyx_vtabptr_4mars_9serialize_4core_IdentityField;


/* "core.pxd":103
 * 
 * 
 * cdef class BoolField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_BoolField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_BoolField *__pyx_vtabptr_4mars_9serialize_4core_BoolField;


/* "core.pxd":107
 * 
 * 
 * cdef class Int8Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_Int8Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_Int8Field *__pyx_vtabptr_4mars_9serialize_4core_Int8Field;


/* "core.pxd":111
 * 
 * 
 * cdef class Int16Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_Int16Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_Int16Field *__pyx_vtabptr_4mars_9serialize_4core_Int16Field;


/* "core.pxd":115
 * 
 * 
 * cdef class Int32Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_Int32Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_Int32Field *__pyx_vtabptr_4mars_9serialize_4core_Int32Field;


/* "core.pxd":119
 * 
 * 
 * cdef class Int64Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_Int64Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_Int64Field *__pyx_vtabptr_4mars_9serialize_4core_Int64Field;


/* "core.pxd":123
 * 
 * 
 * cdef class UInt8Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_UInt8Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_UInt8Field *__pyx_vtabptr_4mars_9serialize_4core_UInt8Field;


/* "core.pxd":127
 * 
 * 
 * cdef class UInt16Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_UInt16Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_UInt16Field *__pyx_vtabptr_4mars_9serialize_4core_UInt16Field;


/* "core.pxd":131
 * 
 * 
 * cdef class UInt32Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_UInt32Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_UInt32Field *__pyx_vtabptr_4mars_9serialize_4core_UInt32Field;


/* "core.pxd":135
 * 
 * 
 * cdef class UInt64Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_UInt64Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_UInt64Field *__pyx_vtabptr_4mars_9serialize_4core_UInt64Field;


/* "core.pxd":139
 * 
 * 
 * cdef class Float16Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_Float16Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_Float16Field *__pyx_vtabptr_4mars_9serialize_4core_Float16Field;


/* "core.pxd":143
 * 
 * 
 * cdef class Float32Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_Float32Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_Float32Field *__pyx_vtabptr_4mars_9serialize_4core_Float32Field;


/* "core.pxd":147
 * 
 * 
 * cdef class Float64Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_Float64Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_Float64Field *__pyx_vtabptr_4mars_9serialize_4core_Float64Field;


/* "core.pxd":151
 * 
 * 
 * cdef class StringField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_StringField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_StringField *__pyx_vtabptr_4mars_9serialize_4core_StringField;


/* "core.pxd":155
 * 
 * 
 * cdef class BytesField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_BytesField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_BytesField *__pyx_vtabptr_4mars_9serialize_4core_BytesField;


/* "core.pxd":159
 * 
 * 
 * cdef class UnicodeField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_UnicodeField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_UnicodeField *__pyx_vtabptr_4mars_9serialize_4core_UnicodeField;


/* "core.pxd":163
 * 
 * 
 * cdef class KeyField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_KeyField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_KeyField *__pyx_vtabptr_4mars_9serialize_4core_KeyField;


/* "core.pxd":167
 * 
 * 
 * cdef class NDArrayField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_NDArrayField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_NDArrayField *__pyx_vtabptr_4mars_9serialize_4core_NDArrayField;


/* "core.pxd":171
 * 
 * 
 * cdef class Datetime64Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_Datetime64Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_Datetime64Field *__pyx_vtabptr_4mars_9serialize_4core_Datetime64Field;


/* "core.pxd":175
 * 
 * 
 * cdef class Timedelta64Field(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_Timedelta64Field {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_Timedelta64Field *__pyx_vtabptr_4mars_9serialize_4core_Timedelta64Field;


/* "core.pxd":179
 * 
 * 
 * cdef class DataTypeField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_DataTypeField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_DataTypeField *__pyx_vtabptr_4mars_9serialize_4core_DataTypeField;


/* "core.pxd":183
 * 
 * 
 * cdef class ListField(Field):             # <<<<<<<<<<<<<<
 *     cdef object _nest_ref
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_ListField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_ListField *__pyx_vtabptr_4mars_9serialize_4core_ListField;


/* "core.pxd":187
 * 
 * 
 * cdef class TupleField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_TupleField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_TupleField *__pyx_vtabptr_4mars_9serialize_4core_TupleField;


/* "core.pxd":191
 * 
 * 
 * cdef class DictField(Field):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_DictField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_DictField *__pyx_vtabptr_4mars_9serialize_4core_DictField;


/* "core.pxd":195
 * 
 * 
 * cdef class ReferenceField(Field):             # <<<<<<<<<<<<<<
 *     cdef object _model
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_ReferenceField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_ReferenceField *__pyx_vtabptr_4mars_9serialize_4core_ReferenceField;


/* "core.pxd":199
 * 
 * 
 * cdef class OneOfField(Field):             # <<<<<<<<<<<<<<
 *     cdef public list fields
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_OneOfField {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Field __pyx_base;
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_OneOfField *__pyx_vtabptr_4mars_9serialize_4core_OneOfField;


/* "core.pxd":208
 * 
 * 
 * cdef class AttrWrapper:             # <<<<<<<<<<<<<<
 *     cdef object _obj
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_AttrWrapper {
  PyObject *(*asdict)(struct __pyx_obj_4mars_9serialize_4core_AttrWrapper *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_AttrWrapper *__pyx_vtabptr_4mars_9serialize_4core_AttrWrapper;


/* "core.pxd":219
 * 
 * 
 * cdef class Provider:             # <<<<<<<<<<<<<<
 *     cdef public object type
 * 
 */

struct __pyx_vtabstruct_4mars_9serialize_4core_Provider {
  PyObject *(*serialize_field)(struct __pyx_obj_4mars_9serialize_4core_Provider *, struct __pyx_obj_4mars_9serialize_4core_Field *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*serialize_model)(struct __pyx_obj_4mars_9serialize_4core_Provider *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_9serialize_4core_8Provider_serialize_model *__pyx_optional_args);
  PyObject *(*serialize_attribute_as_dict)(struct __pyx_obj_4mars_9serialize_4core_Provider *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_9serialize_4core_8Provider_serialize_attribute_as_dict *__pyx_optional_args);
  PyObject *(*deserialize_model)(struct __pyx_obj_4mars_9serialize_4core_Provider *, PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4mars_9serialize_4core_Provider *__pyx_vtabptr_4mars_9serialize_4core_Provider;


/* "mars/serialize/jsonserializer.pyx":72
 * 
 * 
 * cdef class JsonSerializeProvider(Provider):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         self.type = ProviderType.json
 */

struct __pyx_vtabstruct_4mars_9serialize_14jsonserializer_JsonSerializeProvider {
  struct __pyx_vtabstruct_4mars_9serialize_4core_Provider __pyx_base;
  PyObject *(*_to_str)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
  PyObject *(*_to_bytes)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
  PyObject *(*_serialize_slice)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
  PyObject *(*_deserialize_slice)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
  PyObject *(*_serialize_arr)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyArrayObject *);
  PyArrayObject *(*_deserialize_arr)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
  PyObject *(*_serialize_dtype)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyArray_Descr *);
  PyArray_Descr *(*_deserialize_dtype)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
  PyObject *(*_serialize_key)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
  struct __pyx_obj_4mars_9serialize_4core_KeyPlaceholder *(*_deserialize_key)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
  PyObject *(*_serialize_list)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list *__pyx_optional_args);
  PyObject *(*_deserialize_list)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int);
  PyObject *(*_serialize_tuple)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple *__pyx_optional_args);
  PyObject *(*_deserialize_tuple)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int);
  PyObject *(*_serialize_dict)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict *__pyx_optional_args);
  PyObject *(*_deserialize_dict)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int);
  PyObject *(*_serialize_datetime64_timedelta64)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
  PyObject *(*_deserialize_datetime64_timedelta64)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
  PyObject *(*_serialize_datetime64)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
  PyObject *(*_deserialize_datetime64)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
  PyObject *(*_serialize_timedelta64)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
  PyObject *(*_deserialize_timedelta64)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
  PyObject *(*_serialize_typed_value)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value *__pyx_optional_args);
  PyObject *(*_serialize_untyped_value)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value *__pyx_optional_args);
  PyObject *(*_serialize_value)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value *__pyx_optional_args);
  PyObject *(*_on_serial)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, struct __pyx_obj_4mars_9serialize_4core_Field *, PyObject *);
  PyObject *(*_deserialize_value)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int);
  PyObject *(*_on_deserial)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, struct __pyx_obj_4mars_9serialize_4core_Field *, PyObject *);
};
static struct __pyx_vtabstruct_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_vtabptr_4mars_9serialize_14jsonserializer_JsonSerializeProvider;
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_str(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_bytes(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_slice(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_slice(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_arr(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyArrayObject *);
static CYTHON_INLINE PyArrayObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_arr(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dtype(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyArray_Descr *);
static CYTHON_INLINE PyArray_Descr *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_dtype(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_key(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
static CYTHON_INLINE struct __pyx_obj_4mars_9serialize_4core_KeyPlaceholder *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_key(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list *__pyx_optional_args);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_list(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple *__pyx_optional_args);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_tuple(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict *__pyx_optional_args);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_dict(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64_timedelta64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_datetime64_timedelta64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_datetime64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_timedelta64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_timedelta64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value *__pyx_optional_args);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value *__pyx_optional_args);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value *__pyx_optional_args);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_serial(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, struct __pyx_obj_4mars_9serialize_4core_Field *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int);
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_deserial(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, struct __pyx_obj_4mars_9serialize_4core_Field *, PyObject *);

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* IncludeStringH.proto */
#include <string.h>

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_4mars_9serialize_4core_ProviderType(enum __pyx_t_4mars_9serialize_4core_ProviderType value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_str(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_val); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_bytes(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_val); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_slice(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_slice(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, CYTHON_UNUSED PyObject *__pyx_v_callbacks); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_arr(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyArrayObject *__pyx_v_value); /* proto*/
static CYTHON_INLINE PyArrayObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_arr(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, CYTHON_UNUSED PyObject *__pyx_v_callbacks); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dtype(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyArray_Descr *__pyx_v_value); /* proto*/
static CYTHON_INLINE PyArray_Descr *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_dtype(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED PyObject *__pyx_v_callbacks); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_key(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static CYTHON_INLINE struct __pyx_obj_4mars_9serialize_4core_KeyPlaceholder *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_key(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, CYTHON_UNUSED PyObject *__pyx_v_callbacks); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list *__pyx_optional_args); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_list(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks, int __pyx_v_weak_ref); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple *__pyx_optional_args); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_tuple(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks, int __pyx_v_weak_ref); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict *__pyx_optional_args); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_dict(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks, int __pyx_v_weak_ref); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64_timedelta64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_tp); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_datetime64_timedelta64(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, CYTHON_UNUSED PyObject *__pyx_v_callbacks); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_datetime64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_timedelta64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_timedelta64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_tp, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value *__pyx_optional_args); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value *__pyx_optional_args); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value *__pyx_optional_args); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_serial(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field, PyObject *__pyx_v_x); /* proto*/
static PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_serialize_field(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field, PyObject *__pyx_v_model_instance, PyObject *__pyx_v_obj, int __pyx_skip_dispatch); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks, int __pyx_v_weak_ref); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_deserial(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field, PyObject *__pyx_v_x); /* proto*/

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'cpython.version' */

/* Module declarations from 'mars.serialize.core' */
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Identity = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_List = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Tuple = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Dict = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Reference = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_OneOf = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_ValueType = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_AnyField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_IdentityField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_BoolField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Int8Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Int16Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Int32Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Int64Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_UInt8Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_UInt16Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_UInt32Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_UInt64Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Float16Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Float32Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Float64Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_StringField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_BytesField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_UnicodeField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_KeyField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_NDArrayField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Datetime64Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Timedelta64Field = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_DataTypeField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_ListField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_TupleField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_DictField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_ReferenceField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_OneOfField = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_KeyPlaceholder = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_AttrWrapper = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_4core_Provider = 0;

/* Module declarations from 'mars.utils_c' */
static PyObject *(*__pyx_f_4mars_7utils_c_to_str)(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_7utils_c_to_str *__pyx_optional_args); /*proto*/

/* Module declarations from 'mars.serialize.jsonserializer' */
static PyTypeObject *__pyx_ptype_4mars_9serialize_14jsonserializer_JsonSerializeProvider = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value = 0;
static PyTypeObject *__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field = 0;
static PyObject *__pyx_v_4mars_9serialize_14jsonserializer_PRIMITIVE_TYPE_TO_NAME = 0;
static PyObject *__pyx_v_4mars_9serialize_14jsonserializer_EXTEND_TYPE_TO_NAME = 0;
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer__get_name(PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer__get_type(PyObject *); /*proto*/
static PyObject *__pyx_f_4mars_9serialize_14jsonserializer___pyx_unpickle_JsonSerializeProvider__set_state(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "mars.serialize.jsonserializer"
extern int __pyx_module_is_main_mars__serialize__jsonserializer;
int __pyx_module_is_main_mars__serialize__jsonserializer = 0;

/* Implementation of 'mars.serialize.jsonserializer' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_V[] = "V";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_o[] = "o";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_cb[] = "cb";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_ko[] = "ko";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_vo[] = "vo";
static const char __pyx_k_arr[] = "arr";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_ref[] = "ref";
static const char __pyx_k_six[] = "six";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_attr[] = "attr";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_core[] = "core";
static const char __pyx_k_dict[] = "dict";
static const char __pyx_k_int8[] = "int8";
static const char __pyx_k_izip[] = "izip";
static const char __pyx_k_list[] = "list";
static const char __pyx_k_load[] = "load";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_save[] = "save";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_subs[] = "subs";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_attrs[] = "attrs";
static const char __pyx_k_bytes[] = "bytes";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_dumps[] = "dumps";
static const char __pyx_k_field[] = "field";
static const char __pyx_k_inner[] = "inner";
static const char __pyx_k_int16[] = "int16";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_int64[] = "int64";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_loads[] = "loads";
static const char __pyx_k_model[] = "model";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_slice[] = "slice";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_tuple[] = "tuple";
static const char __pyx_k_uint8[] = "uint8";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_base64[] = "base64";
static const char __pyx_k_compat[] = "compat";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_dict_2[] = "__dict__";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_uint16[] = "uint16";
static const char __pyx_k_uint32[] = "uint32";
static const char __pyx_k_uint64[] = "uint64";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_BytesIO[] = "BytesIO";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_float16[] = "float16";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_unicode[] = "unicode";
static const char __pyx_k_weakref[] = "weakref";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_getvalue[] = "getvalue";
static const char __pyx_k_key_type[] = "key_type";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_tag_name[] = "tag_name";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_b64decode[] = "b64decode";
static const char __pyx_k_b64encode[] = "b64encode";
static const char __pyx_k_callbacks[] = "callbacks";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_serialize[] = "serialize";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_datetime64[] = "datetime64";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_value_type[] = "value_type";
static const char __pyx_k_BaseWithKey[] = "BaseWithKey";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_OrderedDict[] = "OrderedDict";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_deserialize[] = "deserialize";
static const char __pyx_k_timedelta64[] = "timedelta64";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_model_instance[] = "model_instance";
static const char __pyx_k_key_to_instance[] = "key_to_instance";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_serialize_field[] = "serialize_field";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_JsonSerializeProvider[] = "JsonSerializeProvider";
static const char __pyx_k_Unknown_type_to_serialize_0[] = "Unknown type to serialize: {0}";
static const char __pyx_k_deserialize_field_locals_cb[] = "deserialize_field.<locals>.cb";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_Unknown_type_to_deserialize_0[] = "Unknown type to deserialize {0}";
static const char __pyx_k_mars_serialize_jsonserializer[] = "mars.serialize.jsonserializer";
static const char __pyx_k_pyx_unpickle_JsonSerializeProv[] = "__pyx_unpickle_JsonSerializeProvider";
static const char __pyx_k_Only_one_of_attributes_0_can_be[] = "Only one of attributes({0}) can be specified";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Incompatible_checksums_s_vs_0x59[] = "Incompatible checksums (%s vs 0x599dcce = (type))";
static const char __pyx_k_JsonSerializeProvider__deseriali[] = "JsonSerializeProvider._deserialize_key.<locals>.genexpr";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_mars_serialize_jsonserializer_py[] = "mars\\serialize\\jsonserializer.pyx";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static const char __pyx_k_JsonSerializeProvider__deseriali_2[] = "JsonSerializeProvider._deserialize_list.<locals>.cb.<locals>.inner";
static const char __pyx_k_JsonSerializeProvider__deseriali_3[] = "JsonSerializeProvider._deserialize_list.<locals>.cb";
static const char __pyx_k_JsonSerializeProvider__deseriali_4[] = "JsonSerializeProvider._deserialize_tuple.<locals>.genexpr";
static const char __pyx_k_JsonSerializeProvider__deseriali_5[] = "JsonSerializeProvider._deserialize_dict.<locals>.cb.<locals>.inner";
static const char __pyx_k_JsonSerializeProvider__deseriali_6[] = "JsonSerializeProvider._deserialize_dict.<locals>.cb";
static const char __pyx_k_JsonSerializeProvider__deseriali_7[] = "JsonSerializeProvider._deserialize_value.<locals>.<lambda>";
static PyObject *__pyx_n_s_BaseWithKey;
static PyObject *__pyx_n_s_BytesIO;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x59;
static PyObject *__pyx_n_s_JsonSerializeProvider;
static PyObject *__pyx_n_s_JsonSerializeProvider__deseriali;
static PyObject *__pyx_n_s_JsonSerializeProvider__deseriali_2;
static PyObject *__pyx_n_s_JsonSerializeProvider__deseriali_3;
static PyObject *__pyx_n_s_JsonSerializeProvider__deseriali_4;
static PyObject *__pyx_n_s_JsonSerializeProvider__deseriali_5;
static PyObject *__pyx_n_s_JsonSerializeProvider__deseriali_6;
static PyObject *__pyx_n_s_JsonSerializeProvider__deseriali_7;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_kp_s_Only_one_of_attributes_0_can_be;
static PyObject *__pyx_n_s_OrderedDict;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unknown_type_to_deserialize_0;
static PyObject *__pyx_kp_s_Unknown_type_to_serialize_0;
static PyObject *__pyx_n_s_V;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_arr;
static PyObject *__pyx_n_s_attr;
static PyObject *__pyx_n_s_attrs;
static PyObject *__pyx_n_s_b64decode;
static PyObject *__pyx_n_s_b64encode;
static PyObject *__pyx_n_s_base64;
static PyObject *__pyx_n_s_bool;
static PyObject *__pyx_n_s_bytes;
static PyObject *__pyx_n_s_callbacks;
static PyObject *__pyx_n_s_cb;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_compat;
static PyObject *__pyx_n_s_core;
static PyObject *__pyx_n_s_datetime64;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_default;
static PyObject *__pyx_n_s_deserialize;
static PyObject *__pyx_n_s_deserialize_field_locals_cb;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dict_2;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dumps;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_field;
static PyObject *__pyx_n_s_float16;
static PyObject *__pyx_n_s_float32;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_getvalue;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inner;
static PyObject *__pyx_n_s_int16;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_int64;
static PyObject *__pyx_n_s_int8;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_izip;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_key_to_instance;
static PyObject *__pyx_n_s_key_type;
static PyObject *__pyx_n_s_ko;
static PyObject *__pyx_n_s_list;
static PyObject *__pyx_n_s_load;
static PyObject *__pyx_n_s_loads;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mars_serialize_jsonserializer;
static PyObject *__pyx_kp_s_mars_serialize_jsonserializer_py;
static PyObject *__pyx_n_s_model;
static PyObject *__pyx_n_s_model_instance;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_o;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_JsonSerializeProv;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_ref;
static PyObject *__pyx_n_s_save;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_serialize;
static PyObject *__pyx_n_s_serialize_field;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_six;
static PyObject *__pyx_n_s_slice;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_subs;
static PyObject *__pyx_n_s_tag_name;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_timedelta64;
static PyObject *__pyx_n_s_tuple;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_uint16;
static PyObject *__pyx_n_s_uint32;
static PyObject *__pyx_n_s_uint64;
static PyObject *__pyx_n_s_uint8;
static PyObject *__pyx_n_s_unicode;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_kp_s_utf_8;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_value_type;
static PyObject *__pyx_n_s_vo;
static PyObject *__pyx_n_s_weakref;
static int __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider___init__(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_16_deserialize_key_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_2cb_inner(PyObject *__pyx_self, PyObject *__pyx_v_subs); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_cb(PyObject *__pyx_self, PyObject *__pyx_v_j, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_18_deserialize_tuple_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_2cb_inner(PyObject *__pyx_self, PyObject *__pyx_v_subs); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_cb(PyObject *__pyx_self, PyObject *__pyx_v_ko, PyObject *__pyx_v_vo); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_2serialize_field(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field, PyObject *__pyx_v_model_instance, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda2(PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17deserialize_field_cb(PyObject *__pyx_self, PyObject *__pyx_v_subs); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_4deserialize_field(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field, PyObject *__pyx_v_model_instance, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks, PyObject *__pyx_v_key_to_instance); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_6__reduce_cython__(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_8__setstate_cython__(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer___pyx_unpickle_JsonSerializeProvider(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer_JsonSerializeProvider(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_93969614;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_codeobj__2;
static PyObject *__pyx_codeobj__4;
static PyObject *__pyx_codeobj__6;
static PyObject *__pyx_codeobj__8;
static PyObject *__pyx_codeobj__10;
static PyObject *__pyx_codeobj__19;
/* Late includes */

/* "mars/serialize/jsonserializer.pyx":60
 * 
 * 
 * cdef inline str _get_name(value_type):             # <<<<<<<<<<<<<<
 *     if value_type in PRIMITIVE_TYPE_TO_NAME:
 *         return PRIMITIVE_TYPE_TO_NAME[value_type]
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer__get_name(PyObject *__pyx_v_value_type) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_get_name", 0);

  /* "mars/serialize/jsonserializer.pyx":61
 * 
 * cdef inline str _get_name(value_type):
 *     if value_type in PRIMITIVE_TYPE_TO_NAME:             # <<<<<<<<<<<<<<
 *         return PRIMITIVE_TYPE_TO_NAME[value_type]
 *     if value_type in EXTEND_TYPE_TO_NAME:
 */
  if (unlikely(__pyx_v_4mars_9serialize_14jsonserializer_PRIMITIVE_TYPE_TO_NAME == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 61, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_value_type, __pyx_v_4mars_9serialize_14jsonserializer_PRIMITIVE_TYPE_TO_NAME, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "mars/serialize/jsonserializer.pyx":62
 * cdef inline str _get_name(value_type):
 *     if value_type in PRIMITIVE_TYPE_TO_NAME:
 *         return PRIMITIVE_TYPE_TO_NAME[value_type]             # <<<<<<<<<<<<<<
 *     if value_type in EXTEND_TYPE_TO_NAME:
 *         return EXTEND_TYPE_TO_NAME[value_type]
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(__pyx_v_4mars_9serialize_14jsonserializer_PRIMITIVE_TYPE_TO_NAME == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 62, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_4mars_9serialize_14jsonserializer_PRIMITIVE_TYPE_TO_NAME, __pyx_v_value_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyString_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 62, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":61
 * 
 * cdef inline str _get_name(value_type):
 *     if value_type in PRIMITIVE_TYPE_TO_NAME:             # <<<<<<<<<<<<<<
 *         return PRIMITIVE_TYPE_TO_NAME[value_type]
 *     if value_type in EXTEND_TYPE_TO_NAME:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":63
 *     if value_type in PRIMITIVE_TYPE_TO_NAME:
 *         return PRIMITIVE_TYPE_TO_NAME[value_type]
 *     if value_type in EXTEND_TYPE_TO_NAME:             # <<<<<<<<<<<<<<
 *         return EXTEND_TYPE_TO_NAME[value_type]
 *     return value_type.name
 */
  if (unlikely(__pyx_v_4mars_9serialize_14jsonserializer_EXTEND_TYPE_TO_NAME == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 63, __pyx_L1_error)
  }
  __pyx_t_2 = (__Pyx_PyDict_ContainsTF(__pyx_v_value_type, __pyx_v_4mars_9serialize_14jsonserializer_EXTEND_TYPE_TO_NAME, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 63, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":64
 *         return PRIMITIVE_TYPE_TO_NAME[value_type]
 *     if value_type in EXTEND_TYPE_TO_NAME:
 *         return EXTEND_TYPE_TO_NAME[value_type]             # <<<<<<<<<<<<<<
 *     return value_type.name
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(__pyx_v_4mars_9serialize_14jsonserializer_EXTEND_TYPE_TO_NAME == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 64, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_4mars_9serialize_14jsonserializer_EXTEND_TYPE_TO_NAME, __pyx_v_value_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyString_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 64, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":63
 *     if value_type in PRIMITIVE_TYPE_TO_NAME:
 *         return PRIMITIVE_TYPE_TO_NAME[value_type]
 *     if value_type in EXTEND_TYPE_TO_NAME:             # <<<<<<<<<<<<<<
 *         return EXTEND_TYPE_TO_NAME[value_type]
 *     return value_type.name
 */
  }

  /* "mars/serialize/jsonserializer.pyx":65
 *     if value_type in EXTEND_TYPE_TO_NAME:
 *         return EXTEND_TYPE_TO_NAME[value_type]
 *     return value_type.name             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value_type, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyString_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 65, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":60
 * 
 * 
 * cdef inline str _get_name(value_type):             # <<<<<<<<<<<<<<
 *     if value_type in PRIMITIVE_TYPE_TO_NAME:
 *         return PRIMITIVE_TYPE_TO_NAME[value_type]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mars.serialize.jsonserializer._get_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":68
 * 
 * 
 * cdef inline object _get_type(str value_type_name):             # <<<<<<<<<<<<<<
 *     return getattr(ValueType, value_type_name)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer__get_type(PyObject *__pyx_v_value_type_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_get_type", 0);

  /* "mars/serialize/jsonserializer.pyx":69
 * 
 * cdef inline object _get_type(str value_type_name):
 *     return getattr(ValueType, value_type_name)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetAttr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_v_value_type_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":68
 * 
 * 
 * cdef inline object _get_type(str value_type_name):             # <<<<<<<<<<<<<<
 *     return getattr(ValueType, value_type_name)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.serialize.jsonserializer._get_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":73
 * 
 * cdef class JsonSerializeProvider(Provider):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.type = ProviderType.json
 * 
 */

/* Python wrapper */
static int __pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider___init__(((struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider___init__(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "mars/serialize/jsonserializer.pyx":74
 * cdef class JsonSerializeProvider(Provider):
 *     def __init__(self):
 *         self.type = ProviderType.json             # <<<<<<<<<<<<<<
 * 
 *     cdef inline str _to_str(self, val):
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_4mars_9serialize_4core_ProviderType(__pyx_e_4mars_9serialize_4core_json); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.type);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.type);
  __pyx_v_self->__pyx_base.type = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":73
 * 
 * cdef class JsonSerializeProvider(Provider):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.type = ProviderType.json
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":76
 *         self.type = ProviderType.json
 * 
 *     cdef inline str _to_str(self, val):             # <<<<<<<<<<<<<<
 *         assert isinstance(val, (bytes, unicode))
 *         if unicode is str and isinstance(val, bytes):
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_str(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("_to_str", 0);

  /* "mars/serialize/jsonserializer.pyx":77
 * 
 *     cdef inline str _to_str(self, val):
 *         assert isinstance(val, (bytes, unicode))             # <<<<<<<<<<<<<<
 *         if unicode is str and isinstance(val, bytes):
 *             return val.decode('utf-8')
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_2 = PyBytes_Check(__pyx_v_val); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L3_bool_binop_done;
    }
    __pyx_t_3 = PyUnicode_Check(__pyx_v_val); 
    __pyx_t_2 = (__pyx_t_3 != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L3_bool_binop_done:;
    if (unlikely(!(__pyx_t_1 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 77, __pyx_L1_error)
    }
  }
  #endif

  /* "mars/serialize/jsonserializer.pyx":78
 *     cdef inline str _to_str(self, val):
 *         assert isinstance(val, (bytes, unicode))
 *         if unicode is str and isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *             return val.decode('utf-8')
 *         return val
 */
  __pyx_t_2 = ((&PyUnicode_Type) == (&PyString_Type));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = PyBytes_Check(__pyx_v_val); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":79
 *         assert isinstance(val, (bytes, unicode))
 *         if unicode is str and isinstance(val, bytes):
 *             return val.decode('utf-8')             # <<<<<<<<<<<<<<
 *         return val
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_decode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_s_utf_8);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(PyString_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 79, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":78
 *     cdef inline str _to_str(self, val):
 *         assert isinstance(val, (bytes, unicode))
 *         if unicode is str and isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *             return val.decode('utf-8')
 *         return val
 */
  }

  /* "mars/serialize/jsonserializer.pyx":80
 *         if unicode is str and isinstance(val, bytes):
 *             return val.decode('utf-8')
 *         return val             # <<<<<<<<<<<<<<
 * 
 *     cdef inline bytes _to_bytes(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyString_CheckExact(__pyx_v_val))||((__pyx_v_val) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_val)->tp_name), 0))) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_INCREF(__pyx_v_val);
  __pyx_r = ((PyObject*)__pyx_v_val);
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":76
 *         self.type = ProviderType.json
 * 
 *     cdef inline str _to_str(self, val):             # <<<<<<<<<<<<<<
 *         assert isinstance(val, (bytes, unicode))
 *         if unicode is str and isinstance(val, bytes):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":82
 *         return val
 * 
 *     cdef inline bytes _to_bytes(self, val):             # <<<<<<<<<<<<<<
 *         assert isinstance(val, (bytes, unicode))
 *         if isinstance(val, unicode):
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_bytes(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("_to_bytes", 0);

  /* "mars/serialize/jsonserializer.pyx":83
 * 
 *     cdef inline bytes _to_bytes(self, val):
 *         assert isinstance(val, (bytes, unicode))             # <<<<<<<<<<<<<<
 *         if isinstance(val, unicode):
 *             return val.encode('utf-8')
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_2 = PyBytes_Check(__pyx_v_val); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L3_bool_binop_done;
    }
    __pyx_t_3 = PyUnicode_Check(__pyx_v_val); 
    __pyx_t_2 = (__pyx_t_3 != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L3_bool_binop_done:;
    if (unlikely(!(__pyx_t_1 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 83, __pyx_L1_error)
    }
  }
  #endif

  /* "mars/serialize/jsonserializer.pyx":84
 *     cdef inline bytes _to_bytes(self, val):
 *         assert isinstance(val, (bytes, unicode))
 *         if isinstance(val, unicode):             # <<<<<<<<<<<<<<
 *             return val.encode('utf-8')
 *         return val
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_val); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "mars/serialize/jsonserializer.pyx":85
 *         assert isinstance(val, (bytes, unicode))
 *         if isinstance(val, unicode):
 *             return val.encode('utf-8')             # <<<<<<<<<<<<<<
 *         return val
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_s_utf_8);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 85, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":84
 *     cdef inline bytes _to_bytes(self, val):
 *         assert isinstance(val, (bytes, unicode))
 *         if isinstance(val, unicode):             # <<<<<<<<<<<<<<
 *             return val.encode('utf-8')
 *         return val
 */
  }

  /* "mars/serialize/jsonserializer.pyx":86
 *         if isinstance(val, unicode):
 *             return val.encode('utf-8')
 *         return val             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_slice(self, slice value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyBytes_CheckExact(__pyx_v_val))||((__pyx_v_val) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_val)->tp_name), 0))) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_INCREF(__pyx_v_val);
  __pyx_r = ((PyObject*)__pyx_v_val);
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":82
 *         return val
 * 
 *     cdef inline bytes _to_bytes(self, val):             # <<<<<<<<<<<<<<
 *         assert isinstance(val, (bytes, unicode))
 *         if isinstance(val, unicode):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._to_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":88
 *         return val
 * 
 *     cdef inline dict _serialize_slice(self, slice value):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.slice),
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_slice(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_serialize_slice", 0);

  /* "mars/serialize/jsonserializer.pyx":89
 * 
 *     cdef inline dict _serialize_slice(self, slice value):
 *         return {             # <<<<<<<<<<<<<<
 *             'type': _get_name(ValueType.slice),
 *             'value': {
 */
  __Pyx_XDECREF(__pyx_r);

  /* "mars/serialize/jsonserializer.pyx":90
 *     cdef inline dict _serialize_slice(self, slice value):
 *         return {
 *             'type': _get_name(ValueType.slice),             # <<<<<<<<<<<<<<
 *             'value': {
 *                 'start': value.start,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_slice); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_4mars_9serialize_14jsonserializer__get_name(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_t_3) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mars/serialize/jsonserializer.pyx":92
 *             'type': _get_name(ValueType.slice),
 *             'value': {
 *                 'start': value.start,             # <<<<<<<<<<<<<<
 *                 'stop': value.stop,
 *                 'step': value.step
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_start, ((PySliceObject*)__pyx_v_value)->start) < 0) __PYX_ERR(0, 92, __pyx_L1_error)

  /* "mars/serialize/jsonserializer.pyx":93
 *             'value': {
 *                 'start': value.start,
 *                 'stop': value.stop,             # <<<<<<<<<<<<<<
 *                 'step': value.step
 *             }
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stop, ((PySliceObject*)__pyx_v_value)->stop) < 0) __PYX_ERR(0, 92, __pyx_L1_error)

  /* "mars/serialize/jsonserializer.pyx":94
 *                 'start': value.start,
 *                 'stop': value.stop,
 *                 'step': value.step             # <<<<<<<<<<<<<<
 *             }
 *         }
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_step, ((PySliceObject*)__pyx_v_value)->step) < 0) __PYX_ERR(0, 92, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_value, __pyx_t_3) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":88
 *         return val
 * 
 *     cdef inline dict _serialize_slice(self, slice value):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.slice),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":98
 *         }
 * 
 *     cdef inline slice _deserialize_slice(self, object obj, list callbacks):             # <<<<<<<<<<<<<<
 *         value = obj['value']
 *         return slice(value['start'], value['stop'], value['step'])
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_slice(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, CYTHON_UNUSED PyObject *__pyx_v_callbacks) {
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_deserialize_slice", 0);

  /* "mars/serialize/jsonserializer.pyx":99
 * 
 *     cdef inline slice _deserialize_slice(self, object obj, list callbacks):
 *         value = obj['value']             # <<<<<<<<<<<<<<
 *         return slice(value['start'], value['stop'], value['step'])
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":100
 *     cdef inline slice _deserialize_slice(self, object obj, list callbacks):
 *         value = obj['value']
 *         return slice(value['start'], value['stop'], value['step'])             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_arr(self, np.ndarray value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_value, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_value, __pyx_n_s_stop); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_value, __pyx_n_s_step); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PySlice_New(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":98
 *         }
 * 
 *     cdef inline slice _deserialize_slice(self, object obj, list callbacks):             # <<<<<<<<<<<<<<
 *         value = obj['value']
 *         return slice(value['start'], value['stop'], value['step'])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":102
 *         return slice(value['start'], value['stop'], value['step'])
 * 
 *     cdef inline dict _serialize_arr(self, np.ndarray value):             # <<<<<<<<<<<<<<
 *         cdef bytes bt
 *         cdef str res
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_arr(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyArrayObject *__pyx_v_value) {
  PyObject *__pyx_v_bio = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_serialize_arr", 0);

  /* "mars/serialize/jsonserializer.pyx":106
 *         cdef str res
 * 
 *         bio = six.BytesIO()             # <<<<<<<<<<<<<<
 *         np.save(bio, value)
 *         return {
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_six); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BytesIO); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_bio = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":107
 * 
 *         bio = six.BytesIO()
 *         np.save(bio, value)             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.arr),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_save); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_bio, ((PyObject *)__pyx_v_value)};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_bio, ((PyObject *)__pyx_v_value)};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_bio);
    __Pyx_GIVEREF(__pyx_v_bio);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_bio);
    __Pyx_INCREF(((PyObject *)__pyx_v_value));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_value));
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, ((PyObject *)__pyx_v_value));
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":108
 *         bio = six.BytesIO()
 *         np.save(bio, value)
 *         return {             # <<<<<<<<<<<<<<
 *             'type': _get_name(ValueType.arr),
 *             'value': self._to_str(base64.b64encode(bio.getvalue()))
 */
  __Pyx_XDECREF(__pyx_r);

  /* "mars/serialize/jsonserializer.pyx":109
 *         np.save(bio, value)
 *         return {
 *             'type': _get_name(ValueType.arr),             # <<<<<<<<<<<<<<
 *             'value': self._to_str(base64.b64encode(bio.getvalue()))
 *         }
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_arr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer__get_name(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_t_5) < 0) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "mars/serialize/jsonserializer.pyx":110
 *         return {
 *             'type': _get_name(ValueType.arr),
 *             'value': self._to_str(base64.b64encode(bio.getvalue()))             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_base64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_b64encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_bio, __pyx_n_s_getvalue); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_str(__pyx_v_self, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_value, __pyx_t_3) < 0) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":102
 *         return slice(value['start'], value['stop'], value['step'])
 * 
 *     cdef inline dict _serialize_arr(self, np.ndarray value):             # <<<<<<<<<<<<<<
 *         cdef bytes bt
 *         cdef str res
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_arr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_bio);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":113
 *         }
 * 
 *     cdef inline np.ndarray _deserialize_arr(self, object obj, list callbacks):             # <<<<<<<<<<<<<<
 *         cdef bytes bt
 * 
 */

static CYTHON_INLINE PyArrayObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_arr(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, CYTHON_UNUSED PyObject *__pyx_v_callbacks) {
  PyObject *__pyx_v_bt = 0;
  PyObject *__pyx_v_value = NULL;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_deserialize_arr", 0);

  /* "mars/serialize/jsonserializer.pyx":116
 *         cdef bytes bt
 * 
 *         value = obj['value']             # <<<<<<<<<<<<<<
 * 
 *         bt = self._to_bytes(base64.b64decode(value))
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":118
 *         value = obj['value']
 * 
 *         bt = self._to_bytes(base64.b64decode(value))             # <<<<<<<<<<<<<<
 * 
 *         if bt is not None:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_base64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_b64decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_value);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_bytes(__pyx_v_self, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_bt = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "mars/serialize/jsonserializer.pyx":120
 *         bt = self._to_bytes(base64.b64decode(value))
 * 
 *         if bt is not None:             # <<<<<<<<<<<<<<
 *             return np.load(six.BytesIO(bt))
 * 
 */
  __pyx_t_4 = (__pyx_v_bt != ((PyObject*)Py_None));
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "mars/serialize/jsonserializer.pyx":121
 * 
 *         if bt is not None:
 *             return np.load(six.BytesIO(bt))             # <<<<<<<<<<<<<<
 * 
 *         return None
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_load); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_six); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_BytesIO); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_v_bt) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_bt);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 121, __pyx_L1_error)
    __pyx_r = ((PyArrayObject *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":120
 *         bt = self._to_bytes(base64.b64decode(value))
 * 
 *         if bt is not None:             # <<<<<<<<<<<<<<
 *             return np.load(six.BytesIO(bt))
 * 
 */
  }

  /* "mars/serialize/jsonserializer.pyx":123
 *             return np.load(six.BytesIO(bt))
 * 
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_dtype(self, np.dtype value):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_r = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":113
 *         }
 * 
 *     cdef inline np.ndarray _deserialize_arr(self, object obj, list callbacks):             # <<<<<<<<<<<<<<
 *         cdef bytes bt
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_arr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_bt);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":125
 *         return None
 * 
 *     cdef inline dict _serialize_dtype(self, np.dtype value):             # <<<<<<<<<<<<<<
 *         cdef str v
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dtype(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyArray_Descr *__pyx_v_value) {
  PyObject *__pyx_v_v = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_serialize_dtype", 0);

  /* "mars/serialize/jsonserializer.pyx":128
 *         cdef str v
 * 
 *         if 'V' not in value.str:             # <<<<<<<<<<<<<<
 *             v = value.str
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_value), __pyx_n_s_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_n_s_V, __pyx_t_1, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "mars/serialize/jsonserializer.pyx":129
 * 
 *         if 'V' not in value.str:
 *             v = value.str             # <<<<<<<<<<<<<<
 *         else:
 *             v = self._to_str(base64.b64encode(pickle.dumps(value)))
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_value), __pyx_n_s_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 129, __pyx_L1_error)
    __pyx_v_v = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":128
 *         cdef str v
 * 
 *         if 'V' not in value.str:             # <<<<<<<<<<<<<<
 *             v = value.str
 *         else:
 */
    goto __pyx_L3;
  }

  /* "mars/serialize/jsonserializer.pyx":131
 *             v = value.str
 *         else:
 *             v = self._to_str(base64.b64encode(pickle.dumps(value)))             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.dtype),
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_base64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_b64encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pickle); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_dumps); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, ((PyObject *)__pyx_v_value)) : __Pyx_PyObject_CallOneArg(__pyx_t_7, ((PyObject *)__pyx_v_value));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_str(__pyx_v_self, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_v = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
  }
  __pyx_L3:;

  /* "mars/serialize/jsonserializer.pyx":132
 *         else:
 *             v = self._to_str(base64.b64encode(pickle.dumps(value)))
 *         return {             # <<<<<<<<<<<<<<
 *             'type': _get_name(ValueType.dtype),
 *             'value': v
 */
  __Pyx_XDECREF(__pyx_r);

  /* "mars/serialize/jsonserializer.pyx":133
 *             v = self._to_str(base64.b64encode(pickle.dumps(value)))
 *         return {
 *             'type': _get_name(ValueType.dtype),             # <<<<<<<<<<<<<<
 *             'value': v
 *         }
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_f_4mars_9serialize_14jsonserializer__get_name(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_type, __pyx_t_4) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "mars/serialize/jsonserializer.pyx":134
 *         return {
 *             'type': _get_name(ValueType.dtype),
 *             'value': v             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_value, __pyx_v_v) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":125
 *         return None
 * 
 *     cdef inline dict _serialize_dtype(self, np.dtype value):             # <<<<<<<<<<<<<<
 *         cdef str v
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_dtype", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":137
 *         }
 * 
 *     cdef inline np.dtype _deserialize_dtype(self, object value, list callbacks):             # <<<<<<<<<<<<<<
 *         try:
 *             return np.dtype(value['value'])
 */

static CYTHON_INLINE PyArray_Descr *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_dtype(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED PyObject *__pyx_v_callbacks) {
  PyObject *__pyx_v_val = NULL;
  PyArray_Descr *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("_deserialize_dtype", 0);

  /* "mars/serialize/jsonserializer.pyx":138
 * 
 *     cdef inline np.dtype _deserialize_dtype(self, object value, list callbacks):
 *         try:             # <<<<<<<<<<<<<<
 *             return np.dtype(value['value'])
 *         except TypeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "mars/serialize/jsonserializer.pyx":139
 *     cdef inline np.dtype _deserialize_dtype(self, object value, list callbacks):
 *         try:
 *             return np.dtype(value['value'])             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             val = value['value']
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_value, __pyx_n_s_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 139, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 139, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = ((PyArray_Descr *)__pyx_t_5);
      __pyx_t_5 = 0;
      goto __pyx_L7_try_return;

      /* "mars/serialize/jsonserializer.pyx":138
 * 
 *     cdef inline np.dtype _deserialize_dtype(self, object value, list callbacks):
 *         try:             # <<<<<<<<<<<<<<
 *             return np.dtype(value['value'])
 *         except TypeError:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "mars/serialize/jsonserializer.pyx":140
 *         try:
 *             return np.dtype(value['value'])
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             val = value['value']
 *             return np.dtype(pickle.loads(base64.b64decode(val)))
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_dtype", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 140, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_7);

      /* "mars/serialize/jsonserializer.pyx":141
 *             return np.dtype(value['value'])
 *         except TypeError:
 *             val = value['value']             # <<<<<<<<<<<<<<
 *             return np.dtype(pickle.loads(base64.b64decode(val)))
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_value, __pyx_n_s_value); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 141, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_v_val = __pyx_t_8;
      __pyx_t_8 = 0;

      /* "mars/serialize/jsonserializer.pyx":142
 *         except TypeError:
 *             val = value['value']
 *             return np.dtype(pickle.loads(base64.b64decode(val)))             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_key(self, value):
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_pickle); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 142, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_loads); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 142, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_base64); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 142, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_b64decode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 142, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_9 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_11, __pyx_v_val) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_v_val);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 142, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_8 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_12, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_9);
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 142, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 142, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_r = ((PyArray_Descr *)__pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "mars/serialize/jsonserializer.pyx":138
 * 
 *     cdef inline np.dtype _deserialize_dtype(self, object value, list callbacks):
 *         try:             # <<<<<<<<<<<<<<
 *             return np.dtype(value['value'])
 *         except TypeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "mars/serialize/jsonserializer.pyx":137
 *         }
 * 
 *     cdef inline np.dtype _deserialize_dtype(self, object value, list callbacks):             # <<<<<<<<<<<<<<
 *         try:
 *             return np.dtype(value['value'])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_dtype", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":144
 *             return np.dtype(pickle.loads(base64.b64decode(val)))
 * 
 *     cdef inline dict _serialize_key(self, value):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.key),
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_key(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_serialize_key", 0);

  /* "mars/serialize/jsonserializer.pyx":145
 * 
 *     cdef inline dict _serialize_key(self, value):
 *         return {             # <<<<<<<<<<<<<<
 *             'type': _get_name(ValueType.key),
 *             'value': (value.key, value.id)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "mars/serialize/jsonserializer.pyx":146
 *     cdef inline dict _serialize_key(self, value):
 *         return {
 *             'type': _get_name(ValueType.key),             # <<<<<<<<<<<<<<
 *             'value': (value.key, value.id)
 *         }
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_4mars_9serialize_14jsonserializer__get_name(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_t_3) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mars/serialize/jsonserializer.pyx":147
 *         return {
 *             'type': _get_name(ValueType.key),
 *             'value': (value.key, value.id)             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_value, __pyx_t_4) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":144
 *             return np.dtype(pickle.loads(base64.b64decode(val)))
 * 
 *     cdef inline dict _serialize_key(self, value):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.key),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_16_deserialize_key_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "mars/serialize/jsonserializer.pyx":154
 *             return KeyPlaceholder(*obj['value'])
 *         else:
 *             return KeyPlaceholder(*(to_str(v) for v in obj['value']))             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_list(self, list value, tp=None, bint weak_ref=False):
 */

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_16_deserialize_key_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr *)__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr(__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 154, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_16_deserialize_key_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_JsonSerializeProvider__deseriali, __pyx_n_s_mars_serialize_jsonserializer); if (unlikely(!gen)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_key.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_16_deserialize_key_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 154, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_obj)) { __Pyx_RaiseClosureNameError("obj"); __PYX_ERR(0, 154, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_obj, __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 154, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 154, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 154, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 154, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __pyx_f_4mars_7utils_c_to_str(__pyx_cur_scope->__pyx_v_v, 0, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 154, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":150
 *         }
 * 
 *     cdef inline KeyPlaceholder _deserialize_key(self, object obj, list callbacks):             # <<<<<<<<<<<<<<
 *         if PY_MAJOR_VERSION >= 3:
 *             return KeyPlaceholder(*obj['value'])
 */

static CYTHON_INLINE struct __pyx_obj_4mars_9serialize_4core_KeyPlaceholder *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_key(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, CYTHON_UNUSED PyObject *__pyx_v_callbacks) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *__pyx_cur_scope;
  struct __pyx_obj_4mars_9serialize_4core_KeyPlaceholder *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_deserialize_key", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *)__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key(__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 150, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_obj = __pyx_v_obj;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_obj);

  /* "mars/serialize/jsonserializer.pyx":151
 * 
 *     cdef inline KeyPlaceholder _deserialize_key(self, object obj, list callbacks):
 *         if PY_MAJOR_VERSION >= 3:             # <<<<<<<<<<<<<<
 *             return KeyPlaceholder(*obj['value'])
 *         else:
 */
  __pyx_t_1 = ((PY_MAJOR_VERSION >= 3) != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":152
 *     cdef inline KeyPlaceholder _deserialize_key(self, object obj, list callbacks):
 *         if PY_MAJOR_VERSION >= 3:
 *             return KeyPlaceholder(*obj['value'])             # <<<<<<<<<<<<<<
 *         else:
 *             return KeyPlaceholder(*(to_str(v) for v in obj['value']))
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_cur_scope->__pyx_v_obj, __pyx_n_s_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4mars_9serialize_4core_KeyPlaceholder), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = ((struct __pyx_obj_4mars_9serialize_4core_KeyPlaceholder *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":151
 * 
 *     cdef inline KeyPlaceholder _deserialize_key(self, object obj, list callbacks):
 *         if PY_MAJOR_VERSION >= 3:             # <<<<<<<<<<<<<<
 *             return KeyPlaceholder(*obj['value'])
 *         else:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":154
 *             return KeyPlaceholder(*obj['value'])
 *         else:
 *             return KeyPlaceholder(*(to_str(v) for v in obj['value']))             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_list(self, list value, tp=None, bint weak_ref=False):
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_t_2 = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_16_deserialize_key_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4mars_9serialize_4core_KeyPlaceholder), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = ((struct __pyx_obj_4mars_9serialize_4core_KeyPlaceholder *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "mars/serialize/jsonserializer.pyx":150
 *         }
 * 
 *     cdef inline KeyPlaceholder _deserialize_key(self, object obj, list callbacks):             # <<<<<<<<<<<<<<
 *         if PY_MAJOR_VERSION >= 3:
 *             return KeyPlaceholder(*obj['value'])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":156
 *             return KeyPlaceholder(*(to_str(v) for v in obj['value']))
 * 
 *     cdef inline dict _serialize_list(self, list value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.list),
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list *__pyx_optional_args) {
  PyObject *__pyx_v_tp = ((PyObject *)Py_None);
  int __pyx_v_weak_ref = ((int)0);
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value __pyx_t_10;
  __Pyx_RefNannySetupContext("_serialize_list", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_tp = __pyx_optional_args->tp;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_weak_ref = __pyx_optional_args->weak_ref;
      }
    }
  }

  /* "mars/serialize/jsonserializer.pyx":157
 * 
 *     cdef inline dict _serialize_list(self, list value, tp=None, bint weak_ref=False):
 *         return {             # <<<<<<<<<<<<<<
 *             'type': _get_name(ValueType.list),
 *             'value': [self._serialize_value(val if not weak_ref else val(),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "mars/serialize/jsonserializer.pyx":158
 *     cdef inline dict _serialize_list(self, list value, tp=None, bint weak_ref=False):
 *         return {
 *             'type': _get_name(ValueType.list),             # <<<<<<<<<<<<<<
 *             'value': [self._serialize_value(val if not weak_ref else val(),
 *                                             tp.type if tp is not None else None)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_4mars_9serialize_14jsonserializer__get_name(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_t_3) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mars/serialize/jsonserializer.pyx":159
 *         return {
 *             'type': _get_name(ValueType.list),
 *             'value': [self._serialize_value(val if not weak_ref else val(),             # <<<<<<<<<<<<<<
 *                                             tp.type if tp is not None else None)
 *                       for val in value]
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "mars/serialize/jsonserializer.pyx":161
 *             'value': [self._serialize_value(val if not weak_ref else val(),
 *                                             tp.type if tp is not None else None)
 *                       for val in value]             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  if (unlikely(__pyx_v_value == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 161, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_value; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
  for (;;) {
    if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 161, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "mars/serialize/jsonserializer.pyx":159
 *         return {
 *             'type': _get_name(ValueType.list),
 *             'value': [self._serialize_value(val if not weak_ref else val(),             # <<<<<<<<<<<<<<
 *                                             tp.type if tp is not None else None)
 *                       for val in value]
 */
    if (((!(__pyx_v_weak_ref != 0)) != 0)) {
      __Pyx_INCREF(__pyx_v_val);
      __pyx_t_5 = __pyx_v_val;
    } else {
      __Pyx_INCREF(__pyx_v_val);
      __pyx_t_7 = __pyx_v_val; __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 159, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_5 = __pyx_t_6;
      __pyx_t_6 = 0;
    }

    /* "mars/serialize/jsonserializer.pyx":160
 *             'type': _get_name(ValueType.list),
 *             'value': [self._serialize_value(val if not weak_ref else val(),
 *                                             tp.type if tp is not None else None)             # <<<<<<<<<<<<<<
 *                       for val in value]
 *         }
 */
    __pyx_t_9 = (__pyx_v_tp != Py_None);
    if ((__pyx_t_9 != 0)) {
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_tp, __pyx_n_s_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __pyx_t_7;
      __pyx_t_7 = 0;
    } else {
      __Pyx_INCREF(Py_None);
      __pyx_t_6 = Py_None;
    }

    /* "mars/serialize/jsonserializer.pyx":159
 *         return {
 *             'type': _get_name(ValueType.list),
 *             'value': [self._serialize_value(val if not weak_ref else val(),             # <<<<<<<<<<<<<<
 *                                             tp.type if tp is not None else None)
 *                       for val in value]
 */
    __pyx_t_10.__pyx_n = 1;
    __pyx_t_10.tp = __pyx_t_6;
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value(__pyx_v_self, __pyx_t_5, &__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "mars/serialize/jsonserializer.pyx":161
 *             'value': [self._serialize_value(val if not weak_ref else val(),
 *                                             tp.type if tp is not None else None)
 *                       for val in value]             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_value, __pyx_t_3) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":156
 *             return KeyPlaceholder(*(to_str(v) for v in obj['value']))
 * 
 *     cdef inline dict _serialize_list(self, list value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.list),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":174
 *             res.append(val)
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(j, v):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[v.key, v.id]
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_1cb(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_1cb = {"cb", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_1cb, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_1cb(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_j = 0;
  PyObject *__pyx_v_v = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cb (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_j,&__pyx_n_s_v,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_j)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cb", 1, 2, 2, 1); __PYX_ERR(0, 174, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cb") < 0)) __PYX_ERR(0, 174, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_j = values[0];
    __pyx_v_v = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cb", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 174, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_list.cb", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_cb(__pyx_self, __pyx_v_j, __pyx_v_v);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":175
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(j, v):
 *                     def inner(subs):             # <<<<<<<<<<<<<<
 *                         o = subs[v.key, v.id]
 *                         if weak_ref:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_2cb_1inner(PyObject *__pyx_self, PyObject *__pyx_v_subs); /*proto*/
static PyMethodDef __pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_2cb_1inner = {"inner", (PyCFunction)__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_2cb_1inner, METH_O, 0};
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_2cb_1inner(PyObject *__pyx_self, PyObject *__pyx_v_subs) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("inner (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_2cb_inner(__pyx_self, ((PyObject *)__pyx_v_subs));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_2cb_inner(PyObject *__pyx_self, PyObject *__pyx_v_subs) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *__pyx_cur_scope;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *__pyx_outer_scope;
  PyObject *__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("inner", 0);
  __pyx_outer_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "mars/serialize/jsonserializer.pyx":176
 *                 def cb(j, v):
 *                     def inner(subs):
 *                         o = subs[v.key, v.id]             # <<<<<<<<<<<<<<
 *                         if weak_ref:
 *                             o = weakref.ref(o)
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_v)) { __Pyx_RaiseClosureNameError("v"); __PYX_ERR(0, 176, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_v, __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_cur_scope->__pyx_v_v)) { __Pyx_RaiseClosureNameError("v"); __PYX_ERR(0, 176, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_v, __pyx_n_s_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_subs, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_o = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":177
 *                     def inner(subs):
 *                         o = subs[v.key, v.id]
 *                         if weak_ref:             # <<<<<<<<<<<<<<
 *                             o = weakref.ref(o)
 *                         res[j] = o
 */
  __pyx_t_4 = (__pyx_cur_scope->__pyx_outer_scope->__pyx_v_weak_ref != 0);
  if (__pyx_t_4) {

    /* "mars/serialize/jsonserializer.pyx":178
 *                         o = subs[v.key, v.id]
 *                         if weak_ref:
 *                             o = weakref.ref(o)             # <<<<<<<<<<<<<<
 *                         res[j] = o
 *                     return inner
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_weakref); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ref); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_o) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_o);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_o, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "mars/serialize/jsonserializer.pyx":177
 *                     def inner(subs):
 *                         o = subs[v.key, v.id]
 *                         if weak_ref:             # <<<<<<<<<<<<<<
 *                             o = weakref.ref(o)
 *                         res[j] = o
 */
  }

  /* "mars/serialize/jsonserializer.pyx":179
 *                         if weak_ref:
 *                             o = weakref.ref(o)
 *                         res[j] = o             # <<<<<<<<<<<<<<
 *                     return inner
 *                 callbacks.append(cb(i, val))
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_res)) { __Pyx_RaiseClosureNameError("res"); __PYX_ERR(0, 179, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_res == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 179, __pyx_L1_error)
  }
  if (unlikely(!__pyx_cur_scope->__pyx_v_j)) { __Pyx_RaiseClosureNameError("j"); __PYX_ERR(0, 179, __pyx_L1_error) }
  if (unlikely(PyObject_SetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_res, __pyx_cur_scope->__pyx_v_j, __pyx_v_o) < 0)) __PYX_ERR(0, 179, __pyx_L1_error)

  /* "mars/serialize/jsonserializer.pyx":175
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(j, v):
 *                     def inner(subs):             # <<<<<<<<<<<<<<
 *                         o = subs[v.key, v.id]
 *                         if weak_ref:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_list.cb.inner", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":174
 *             res.append(val)
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(j, v):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[v.key, v.id]
 */

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_cb(PyObject *__pyx_self, PyObject *__pyx_v_j, PyObject *__pyx_v_v) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *__pyx_cur_scope;
  PyObject *__pyx_v_inner = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cb", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *)__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb(__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 174, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_v_j = __pyx_v_j;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_j);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_j);
  __pyx_cur_scope->__pyx_v_v = __pyx_v_v;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_v);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_v);

  /* "mars/serialize/jsonserializer.pyx":175
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(j, v):
 *                     def inner(subs):             # <<<<<<<<<<<<<<
 *                         o = subs[v.key, v.id]
 *                         if weak_ref:
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_2cb_1inner, 0, __pyx_n_s_JsonSerializeProvider__deseriali_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_mars_serialize_jsonserializer, __pyx_d, ((PyObject *)__pyx_codeobj__2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_inner = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":180
 *                             o = weakref.ref(o)
 *                         res[j] = o
 *                     return inner             # <<<<<<<<<<<<<<
 *                 callbacks.append(cb(i, val))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_inner);
  __pyx_r = __pyx_v_inner;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":174
 *             res.append(val)
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(j, v):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[v.key, v.id]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_list.cb", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_inner);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":164
 *         }
 * 
 *     cdef inline list _deserialize_list(self, object obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         cdef list res
 *         cdef int i
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_list(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks, int __pyx_v_weak_ref) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *__pyx_cur_scope;
  int __pyx_v_i;
  PyObject *__pyx_v_it_val = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_cb = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("_deserialize_list", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *)__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list(__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 164, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_weak_ref = __pyx_v_weak_ref;

  /* "mars/serialize/jsonserializer.pyx":169
 *         cdef int j
 * 
 *         res = []             # <<<<<<<<<<<<<<
 *         for i, it_val in enumerate(obj['value']):
 *             val = self._deserialize_value(it_val, callbacks, weak_ref)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_res = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":170
 * 
 *         res = []
 *         for i, it_val in enumerate(obj['value']):             # <<<<<<<<<<<<<<
 *             val = self._deserialize_value(it_val, callbacks, weak_ref)
 *             res.append(val)
 */
  __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 170, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 170, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 170, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 170, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_it_val, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_i = __pyx_t_2;
    __pyx_t_2 = (__pyx_t_2 + 1);

    /* "mars/serialize/jsonserializer.pyx":171
 *         res = []
 *         for i, it_val in enumerate(obj['value']):
 *             val = self._deserialize_value(it_val, callbacks, weak_ref)             # <<<<<<<<<<<<<<
 *             res.append(val)
 *             if isinstance(val, KeyPlaceholder):
 */
    __pyx_t_1 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_value(__pyx_v_self, __pyx_v_it_val, __pyx_v_callbacks, __pyx_cur_scope->__pyx_v_weak_ref); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":172
 *         for i, it_val in enumerate(obj['value']):
 *             val = self._deserialize_value(it_val, callbacks, weak_ref)
 *             res.append(val)             # <<<<<<<<<<<<<<
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(j, v):
 */
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_res, __pyx_v_val); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 172, __pyx_L1_error)

    /* "mars/serialize/jsonserializer.pyx":173
 *             val = self._deserialize_value(it_val, callbacks, weak_ref)
 *             res.append(val)
 *             if isinstance(val, KeyPlaceholder):             # <<<<<<<<<<<<<<
 *                 def cb(j, v):
 *                     def inner(subs):
 */
    __pyx_t_7 = __Pyx_TypeCheck(__pyx_v_val, __pyx_ptype_4mars_9serialize_4core_KeyPlaceholder); 
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (__pyx_t_8) {

      /* "mars/serialize/jsonserializer.pyx":174
 *             res.append(val)
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(j, v):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[v.key, v.id]
 */
      __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_1cb, 0, __pyx_n_s_JsonSerializeProvider__deseriali_3, ((PyObject*)__pyx_cur_scope), __pyx_n_s_mars_serialize_jsonserializer, __pyx_d, ((PyObject *)__pyx_codeobj__4)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_cb, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "mars/serialize/jsonserializer.pyx":181
 *                         res[j] = o
 *                     return inner
 *                 callbacks.append(cb(i, val))             # <<<<<<<<<<<<<<
 * 
 *         return res
 */
      if (unlikely(__pyx_v_callbacks == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 181, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_list_cb(__pyx_v_cb, __pyx_t_1, __pyx_v_val); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_callbacks, __pyx_t_9); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "mars/serialize/jsonserializer.pyx":173
 *             val = self._deserialize_value(it_val, callbacks, weak_ref)
 *             res.append(val)
 *             if isinstance(val, KeyPlaceholder):             # <<<<<<<<<<<<<<
 *                 def cb(j, v):
 *                     def inner(subs):
 */
    }

    /* "mars/serialize/jsonserializer.pyx":170
 * 
 *         res = []
 *         for i, it_val in enumerate(obj['value']):             # <<<<<<<<<<<<<<
 *             val = self._deserialize_value(it_val, callbacks, weak_ref)
 *             res.append(val)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mars/serialize/jsonserializer.pyx":183
 *                 callbacks.append(cb(i, val))
 * 
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_tuple(self, tuple value, tp=None, bint weak_ref=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_res);
  __pyx_r = __pyx_cur_scope->__pyx_v_res;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":164
 *         }
 * 
 *     cdef inline list _deserialize_list(self, object obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         cdef list res
 *         cdef int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_it_val);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_cb);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":185
 *         return res
 * 
 *     cdef inline dict _serialize_tuple(self, tuple value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if tp is None or not isinstance(tp.type, tuple):
 *             return {
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple *__pyx_optional_args) {
  PyObject *__pyx_v_tp = ((PyObject *)Py_None);
  int __pyx_v_weak_ref = ((int)0);
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_it_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value __pyx_t_12;
  int __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *(*__pyx_t_15)(PyObject *);
  __Pyx_RefNannySetupContext("_serialize_tuple", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_tp = __pyx_optional_args->tp;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_weak_ref = __pyx_optional_args->weak_ref;
      }
    }
  }

  /* "mars/serialize/jsonserializer.pyx":186
 * 
 *     cdef inline dict _serialize_tuple(self, tuple value, tp=None, bint weak_ref=False):
 *         if tp is None or not isinstance(tp.type, tuple):             # <<<<<<<<<<<<<<
 *             return {
 *                 'type': _get_name(ValueType.tuple),
 */
  __pyx_t_2 = (__pyx_v_tp == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_tp, __pyx_n_s_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_Check(__pyx_t_4); 
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = ((!(__pyx_t_3 != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":187
 *     cdef inline dict _serialize_tuple(self, tuple value, tp=None, bint weak_ref=False):
 *         if tp is None or not isinstance(tp.type, tuple):
 *             return {             # <<<<<<<<<<<<<<
 *                 'type': _get_name(ValueType.tuple),
 *                 'value': [self._serialize_value(val if not weak_ref else val(),
 */
    __Pyx_XDECREF(__pyx_r);

    /* "mars/serialize/jsonserializer.pyx":188
 *         if tp is None or not isinstance(tp.type, tuple):
 *             return {
 *                 'type': _get_name(ValueType.tuple),             # <<<<<<<<<<<<<<
 *                 'value': [self._serialize_value(val if not weak_ref else val(),
 *                                                 tp.type if tp is not None else None)
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_tuple); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_f_4mars_9serialize_14jsonserializer__get_name(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_type, __pyx_t_6) < 0) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "mars/serialize/jsonserializer.pyx":189
 *             return {
 *                 'type': _get_name(ValueType.tuple),
 *                 'value': [self._serialize_value(val if not weak_ref else val(),             # <<<<<<<<<<<<<<
 *                                                 tp.type if tp is not None else None)
 *                           for val in value]
 */
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "mars/serialize/jsonserializer.pyx":191
 *                 'value': [self._serialize_value(val if not weak_ref else val(),
 *                                                 tp.type if tp is not None else None)
 *                           for val in value]             # <<<<<<<<<<<<<<
 *             }
 *         else:
 */
    if (unlikely(__pyx_v_value == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 191, __pyx_L1_error)
    }
    __pyx_t_5 = __pyx_v_value; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
    for (;;) {
      if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_8); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 191, __pyx_L1_error)
      #else
      __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "mars/serialize/jsonserializer.pyx":189
 *             return {
 *                 'type': _get_name(ValueType.tuple),
 *                 'value': [self._serialize_value(val if not weak_ref else val(),             # <<<<<<<<<<<<<<
 *                                                 tp.type if tp is not None else None)
 *                           for val in value]
 */
      if (((!(__pyx_v_weak_ref != 0)) != 0)) {
        __Pyx_INCREF(__pyx_v_val);
        __pyx_t_8 = __pyx_v_val;
      } else {
        __Pyx_INCREF(__pyx_v_val);
        __pyx_t_10 = __pyx_v_val; __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
          }
        }
        __pyx_t_9 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_8 = __pyx_t_9;
        __pyx_t_9 = 0;
      }

      /* "mars/serialize/jsonserializer.pyx":190
 *                 'type': _get_name(ValueType.tuple),
 *                 'value': [self._serialize_value(val if not weak_ref else val(),
 *                                                 tp.type if tp is not None else None)             # <<<<<<<<<<<<<<
 *                           for val in value]
 *             }
 */
      __pyx_t_1 = (__pyx_v_tp != Py_None);
      if ((__pyx_t_1 != 0)) {
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_tp, __pyx_n_s_type); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 190, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_9 = __pyx_t_10;
        __pyx_t_10 = 0;
      } else {
        __Pyx_INCREF(Py_None);
        __pyx_t_9 = Py_None;
      }

      /* "mars/serialize/jsonserializer.pyx":189
 *             return {
 *                 'type': _get_name(ValueType.tuple),
 *                 'value': [self._serialize_value(val if not weak_ref else val(),             # <<<<<<<<<<<<<<
 *                                                 tp.type if tp is not None else None)
 *                           for val in value]
 */
      __pyx_t_12.__pyx_n = 1;
      __pyx_t_12.tp = __pyx_t_9;
      __pyx_t_10 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value(__pyx_v_self, __pyx_t_8, &__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 189, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "mars/serialize/jsonserializer.pyx":191
 *                 'value': [self._serialize_value(val if not weak_ref else val(),
 *                                                 tp.type if tp is not None else None)
 *                           for val in value]             # <<<<<<<<<<<<<<
 *             }
 *         else:
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_value, __pyx_t_6) < 0) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":186
 * 
 *     cdef inline dict _serialize_tuple(self, tuple value, tp=None, bint weak_ref=False):
 *         if tp is None or not isinstance(tp.type, tuple):             # <<<<<<<<<<<<<<
 *             return {
 *                 'type': _get_name(ValueType.tuple),
 */
  }

  /* "mars/serialize/jsonserializer.pyx":194
 *             }
 *         else:
 *             return {             # <<<<<<<<<<<<<<
 *                 'type': _get_name(ValueType.tuple),
 *                 'value': [self._serialize_value(val if not weak_ref else val(),
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);

    /* "mars/serialize/jsonserializer.pyx":195
 *         else:
 *             return {
 *                 'type': _get_name(ValueType.tuple),             # <<<<<<<<<<<<<<
 *                 'value': [self._serialize_value(val if not weak_ref else val(),
 *                                                 it_type)
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_tuple); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer__get_name(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_type, __pyx_t_5) < 0) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "mars/serialize/jsonserializer.pyx":196
 *             return {
 *                 'type': _get_name(ValueType.tuple),
 *                 'value': [self._serialize_value(val if not weak_ref else val(),             # <<<<<<<<<<<<<<
 *                                                 it_type)
 *                           for val, it_type in izip(value, tp.type)]
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "mars/serialize/jsonserializer.pyx":198
 *                 'value': [self._serialize_value(val if not weak_ref else val(),
 *                                                 it_type)
 *                           for val, it_type in izip(value, tp.type)]             # <<<<<<<<<<<<<<
 *             }
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_izip); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_tp, __pyx_n_s_type); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = NULL;
    __pyx_t_13 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_13 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_value, __pyx_t_9};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 198, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_value, __pyx_t_9};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 198, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    {
      __pyx_t_11 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 198, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_13, __pyx_v_value);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_13, __pyx_t_9);
      __pyx_t_9 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_11, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 198, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_10 = __pyx_t_6; __Pyx_INCREF(__pyx_t_10); __pyx_t_7 = 0;
      __pyx_t_14 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 198, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_14 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 198, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_14)) {
        if (likely(PyList_CheckExact(__pyx_t_10))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_10)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 198, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 198, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 198, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 198, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_14(__pyx_t_10);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 198, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
        PyObject* sequence = __pyx_t_6;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 198, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_11 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_11 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_9);
        #else
        __pyx_t_11 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 198, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 198, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_8 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 198, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_15 = Py_TYPE(__pyx_t_8)->tp_iternext;
        index = 0; __pyx_t_11 = __pyx_t_15(__pyx_t_8); if (unlikely(!__pyx_t_11)) goto __pyx_L10_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_11);
        index = 1; __pyx_t_9 = __pyx_t_15(__pyx_t_8); if (unlikely(!__pyx_t_9)) goto __pyx_L10_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_9);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_8), 2) < 0) __PYX_ERR(0, 198, __pyx_L1_error)
        __pyx_t_15 = NULL;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L11_unpacking_done;
        __pyx_L10_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_15 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 198, __pyx_L1_error)
        __pyx_L11_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_XDECREF_SET(__pyx_v_it_type, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "mars/serialize/jsonserializer.pyx":196
 *             return {
 *                 'type': _get_name(ValueType.tuple),
 *                 'value': [self._serialize_value(val if not weak_ref else val(),             # <<<<<<<<<<<<<<
 *                                                 it_type)
 *                           for val, it_type in izip(value, tp.type)]
 */
      if (((!(__pyx_v_weak_ref != 0)) != 0)) {
        __Pyx_INCREF(__pyx_v_val);
        __pyx_t_6 = __pyx_v_val;
      } else {
        __Pyx_INCREF(__pyx_v_val);
        __pyx_t_11 = __pyx_v_val; __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
          }
        }
        __pyx_t_9 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 196, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_6 = __pyx_t_9;
        __pyx_t_9 = 0;
      }

      /* "mars/serialize/jsonserializer.pyx":197
 *                 'type': _get_name(ValueType.tuple),
 *                 'value': [self._serialize_value(val if not weak_ref else val(),
 *                                                 it_type)             # <<<<<<<<<<<<<<
 *                           for val, it_type in izip(value, tp.type)]
 *             }
 */
      __pyx_t_12.__pyx_n = 1;
      __pyx_t_12.tp = __pyx_v_it_type;
      __pyx_t_9 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value(__pyx_v_self, __pyx_t_6, &__pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 196, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "mars/serialize/jsonserializer.pyx":198
 *                 'value': [self._serialize_value(val if not weak_ref else val(),
 *                                                 it_type)
 *                           for val, it_type in izip(value, tp.type)]             # <<<<<<<<<<<<<<
 *             }
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_value, __pyx_t_5) < 0) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "mars/serialize/jsonserializer.pyx":185
 *         return res
 * 
 *     cdef inline dict _serialize_tuple(self, tuple value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if tp is None or not isinstance(tp.type, tuple):
 *             return {
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_tuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_it_type);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_18_deserialize_tuple_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "mars/serialize/jsonserializer.pyx":202
 * 
 *     cdef inline tuple _deserialize_tuple(self, object obj, list callbacks, bint weak_ref):
 *         return tuple(self._deserialize_value(it_val, callbacks, weak_ref)             # <<<<<<<<<<<<<<
 *                      for it_val in obj['value'])
 * 
 */

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_18_deserialize_tuple_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr *)__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr(__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 202, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_18_deserialize_tuple_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_JsonSerializeProvider__deseriali_4, __pyx_n_s_mars_serialize_jsonserializer); if (unlikely(!gen)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_tuple.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_18_deserialize_tuple_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr *__pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 202, __pyx_L1_error)

  /* "mars/serialize/jsonserializer.pyx":203
 *     cdef inline tuple _deserialize_tuple(self, object obj, list callbacks, bint weak_ref):
 *         return tuple(self._deserialize_value(it_val, callbacks, weak_ref)
 *                      for it_val in obj['value'])             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_dict(self, object value, tp=None, bint weak_ref=False):
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_obj)) { __Pyx_RaiseClosureNameError("obj"); __PYX_ERR(0, 203, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_obj, __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 203, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 203, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 203, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 203, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_it_val);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_it_val, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":202
 * 
 *     cdef inline tuple _deserialize_tuple(self, object obj, list callbacks, bint weak_ref):
 *         return tuple(self._deserialize_value(it_val, callbacks, weak_ref)             # <<<<<<<<<<<<<<
 *                      for it_val in obj['value'])
 * 
 */
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 202, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_callbacks)) { __Pyx_RaiseClosureNameError("callbacks"); __PYX_ERR(0, 202, __pyx_L1_error) }
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_callbacks;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_value(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_it_val, ((PyObject*)__pyx_t_1), __pyx_cur_scope->__pyx_outer_scope->__pyx_v_weak_ref); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 202, __pyx_L1_error)

    /* "mars/serialize/jsonserializer.pyx":203
 *     cdef inline tuple _deserialize_tuple(self, object obj, list callbacks, bint weak_ref):
 *         return tuple(self._deserialize_value(it_val, callbacks, weak_ref)
 *                      for it_val in obj['value'])             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_dict(self, object value, tp=None, bint weak_ref=False):
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "mars/serialize/jsonserializer.pyx":202
 * 
 *     cdef inline tuple _deserialize_tuple(self, object obj, list callbacks, bint weak_ref):
 *         return tuple(self._deserialize_value(it_val, callbacks, weak_ref)             # <<<<<<<<<<<<<<
 *                      for it_val in obj['value'])
 * 
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":201
 *             }
 * 
 *     cdef inline tuple _deserialize_tuple(self, object obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         return tuple(self._deserialize_value(it_val, callbacks, weak_ref)
 *                      for it_val in obj['value'])
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_tuple(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks, int __pyx_v_weak_ref) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_deserialize_tuple", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *)__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple(__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 201, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_obj = __pyx_v_obj;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_obj);
  __pyx_cur_scope->__pyx_v_callbacks = __pyx_v_callbacks;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_callbacks);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_callbacks);
  __pyx_cur_scope->__pyx_v_weak_ref = __pyx_v_weak_ref;

  /* "mars/serialize/jsonserializer.pyx":202
 * 
 *     cdef inline tuple _deserialize_tuple(self, object obj, list callbacks, bint weak_ref):
 *         return tuple(self._deserialize_value(it_val, callbacks, weak_ref)             # <<<<<<<<<<<<<<
 *                      for it_val in obj['value'])
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_18_deserialize_tuple_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":201
 *             }
 * 
 *     cdef inline tuple _deserialize_tuple(self, object obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         return tuple(self._deserialize_value(it_val, callbacks, weak_ref)
 *                      for it_val in obj['value'])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_tuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":205
 *                      for it_val in obj['value'])
 * 
 *     cdef inline dict _serialize_dict(self, object value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.dict),
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict *__pyx_optional_args) {
  PyObject *__pyx_v_tp = ((PyObject *)Py_None);
  int __pyx_v_weak_ref = ((int)0);
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  __Pyx_RefNannySetupContext("_serialize_dict", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_tp = __pyx_optional_args->tp;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_weak_ref = __pyx_optional_args->weak_ref;
      }
    }
  }

  /* "mars/serialize/jsonserializer.pyx":206
 * 
 *     cdef inline dict _serialize_dict(self, object value, tp=None, bint weak_ref=False):
 *         return {             # <<<<<<<<<<<<<<
 *             'type': _get_name(ValueType.dict),
 *             'value': [(self._serialize_value(k if not weak_ref else k(),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "mars/serialize/jsonserializer.pyx":207
 *     cdef inline dict _serialize_dict(self, object value, tp=None, bint weak_ref=False):
 *         return {
 *             'type': _get_name(ValueType.dict),             # <<<<<<<<<<<<<<
 *             'value': [(self._serialize_value(k if not weak_ref else k(),
 *                                              tp.key_type if tp is not None else tp),
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_4mars_9serialize_14jsonserializer__get_name(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_t_3) < 0) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mars/serialize/jsonserializer.pyx":208
 *         return {
 *             'type': _get_name(ValueType.dict),
 *             'value': [(self._serialize_value(k if not weak_ref else k(),             # <<<<<<<<<<<<<<
 *                                              tp.key_type if tp is not None else tp),
 *                        self._serialize_value(v if not weak_ref else v(),
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "mars/serialize/jsonserializer.pyx":212
 *                        self._serialize_value(v if not weak_ref else v(),
 *                                              tp.value_type if tp is not None else tp))
 *                       for k, v in value.items()]             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 212, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 212, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 212, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_4);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 212, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 212, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 212, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 212, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "mars/serialize/jsonserializer.pyx":208
 *         return {
 *             'type': _get_name(ValueType.dict),
 *             'value': [(self._serialize_value(k if not weak_ref else k(),             # <<<<<<<<<<<<<<
 *                                              tp.key_type if tp is not None else tp),
 *                        self._serialize_value(v if not weak_ref else v(),
 */
    if (((!(__pyx_v_weak_ref != 0)) != 0)) {
      __Pyx_INCREF(__pyx_v_k);
      __pyx_t_2 = __pyx_v_k;
    } else {
      __Pyx_INCREF(__pyx_v_k);
      __pyx_t_5 = __pyx_v_k; __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_8 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 208, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_2 = __pyx_t_8;
      __pyx_t_8 = 0;
    }

    /* "mars/serialize/jsonserializer.pyx":209
 *             'type': _get_name(ValueType.dict),
 *             'value': [(self._serialize_value(k if not weak_ref else k(),
 *                                              tp.key_type if tp is not None else tp),             # <<<<<<<<<<<<<<
 *                        self._serialize_value(v if not weak_ref else v(),
 *                                              tp.value_type if tp is not None else tp))
 */
    __pyx_t_11 = (__pyx_v_tp != Py_None);
    if ((__pyx_t_11 != 0)) {
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_tp, __pyx_n_s_key_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __pyx_t_5;
      __pyx_t_5 = 0;
    } else {
      __Pyx_INCREF(__pyx_v_tp);
      __pyx_t_8 = __pyx_v_tp;
    }

    /* "mars/serialize/jsonserializer.pyx":208
 *         return {
 *             'type': _get_name(ValueType.dict),
 *             'value': [(self._serialize_value(k if not weak_ref else k(),             # <<<<<<<<<<<<<<
 *                                              tp.key_type if tp is not None else tp),
 *                        self._serialize_value(v if not weak_ref else v(),
 */
    __pyx_t_12.__pyx_n = 1;
    __pyx_t_12.tp = __pyx_t_8;
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value(__pyx_v_self, __pyx_t_2, &__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "mars/serialize/jsonserializer.pyx":210
 *             'value': [(self._serialize_value(k if not weak_ref else k(),
 *                                              tp.key_type if tp is not None else tp),
 *                        self._serialize_value(v if not weak_ref else v(),             # <<<<<<<<<<<<<<
 *                                              tp.value_type if tp is not None else tp))
 *                       for k, v in value.items()]
 */
    if (((!(__pyx_v_weak_ref != 0)) != 0)) {
      __Pyx_INCREF(__pyx_v_v);
      __pyx_t_8 = __pyx_v_v;
    } else {
      __Pyx_INCREF(__pyx_v_v);
      __pyx_t_9 = __pyx_v_v; __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_2 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_8 = __pyx_t_2;
      __pyx_t_2 = 0;
    }

    /* "mars/serialize/jsonserializer.pyx":211
 *                                              tp.key_type if tp is not None else tp),
 *                        self._serialize_value(v if not weak_ref else v(),
 *                                              tp.value_type if tp is not None else tp))             # <<<<<<<<<<<<<<
 *                       for k, v in value.items()]
 *         }
 */
    __pyx_t_11 = (__pyx_v_tp != Py_None);
    if ((__pyx_t_11 != 0)) {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_tp, __pyx_n_s_value_type); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 211, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_2 = __pyx_t_9;
      __pyx_t_9 = 0;
    } else {
      __Pyx_INCREF(__pyx_v_tp);
      __pyx_t_2 = __pyx_v_tp;
    }

    /* "mars/serialize/jsonserializer.pyx":210
 *             'value': [(self._serialize_value(k if not weak_ref else k(),
 *                                              tp.key_type if tp is not None else tp),
 *                        self._serialize_value(v if not weak_ref else v(),             # <<<<<<<<<<<<<<
 *                                              tp.value_type if tp is not None else tp))
 *                       for k, v in value.items()]
 */
    __pyx_t_12.__pyx_n = 1;
    __pyx_t_12.tp = __pyx_t_2;
    __pyx_t_9 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value(__pyx_v_self, __pyx_t_8, &__pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "mars/serialize/jsonserializer.pyx":208
 *         return {
 *             'type': _get_name(ValueType.dict),
 *             'value': [(self._serialize_value(k if not weak_ref else k(),             # <<<<<<<<<<<<<<
 *                                              tp.key_type if tp is not None else tp),
 *                        self._serialize_value(v if not weak_ref else v(),
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_9);
    __pyx_t_5 = 0;
    __pyx_t_9 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "mars/serialize/jsonserializer.pyx":212
 *                        self._serialize_value(v if not weak_ref else v(),
 *                                              tp.value_type if tp is not None else tp))
 *                       for k, v in value.items()]             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_value, __pyx_t_3) < 0) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":205
 *                      for it_val in obj['value'])
 * 
 *     cdef inline dict _serialize_dict(self, object value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(ValueType.dict),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":222
 *             res[key] = value
 *             if isinstance(value, KeyPlaceholder):
 *                 def cb(ko, vo):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[vo.key, vo.id]
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_1cb(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_1cb = {"cb", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_1cb, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_1cb(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ko = 0;
  PyObject *__pyx_v_vo = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cb (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ko,&__pyx_n_s_vo,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ko)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_vo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cb", 1, 2, 2, 1); __PYX_ERR(0, 222, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cb") < 0)) __PYX_ERR(0, 222, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_ko = values[0];
    __pyx_v_vo = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cb", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 222, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_dict.cb", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_cb(__pyx_self, __pyx_v_ko, __pyx_v_vo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":223
 *             if isinstance(value, KeyPlaceholder):
 *                 def cb(ko, vo):
 *                     def inner(subs):             # <<<<<<<<<<<<<<
 *                         o = subs[vo.key, vo.id]
 *                         if weak_ref:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_2cb_1inner(PyObject *__pyx_self, PyObject *__pyx_v_subs); /*proto*/
static PyMethodDef __pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_2cb_1inner = {"inner", (PyCFunction)__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_2cb_1inner, METH_O, 0};
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_2cb_1inner(PyObject *__pyx_self, PyObject *__pyx_v_subs) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("inner (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_2cb_inner(__pyx_self, ((PyObject *)__pyx_v_subs));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_2cb_inner(PyObject *__pyx_self, PyObject *__pyx_v_subs) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *__pyx_cur_scope;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *__pyx_outer_scope;
  PyObject *__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("inner", 0);
  __pyx_outer_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "mars/serialize/jsonserializer.pyx":224
 *                 def cb(ko, vo):
 *                     def inner(subs):
 *                         o = subs[vo.key, vo.id]             # <<<<<<<<<<<<<<
 *                         if weak_ref:
 *                             o = weakref.ref(o)
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_vo)) { __Pyx_RaiseClosureNameError("vo"); __PYX_ERR(0, 224, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_vo, __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_cur_scope->__pyx_v_vo)) { __Pyx_RaiseClosureNameError("vo"); __PYX_ERR(0, 224, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_vo, __pyx_n_s_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_subs, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_o = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":225
 *                     def inner(subs):
 *                         o = subs[vo.key, vo.id]
 *                         if weak_ref:             # <<<<<<<<<<<<<<
 *                             o = weakref.ref(o)
 *                         res[ko] = o
 */
  __pyx_t_4 = (__pyx_cur_scope->__pyx_outer_scope->__pyx_v_weak_ref != 0);
  if (__pyx_t_4) {

    /* "mars/serialize/jsonserializer.pyx":226
 *                         o = subs[vo.key, vo.id]
 *                         if weak_ref:
 *                             o = weakref.ref(o)             # <<<<<<<<<<<<<<
 *                         res[ko] = o
 *                     return inner
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_weakref); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ref); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_o) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_o);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_o, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "mars/serialize/jsonserializer.pyx":225
 *                     def inner(subs):
 *                         o = subs[vo.key, vo.id]
 *                         if weak_ref:             # <<<<<<<<<<<<<<
 *                             o = weakref.ref(o)
 *                         res[ko] = o
 */
  }

  /* "mars/serialize/jsonserializer.pyx":227
 *                         if weak_ref:
 *                             o = weakref.ref(o)
 *                         res[ko] = o             # <<<<<<<<<<<<<<
 *                     return inner
 *                 callbacks.append(cb(key, value))
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_res)) { __Pyx_RaiseClosureNameError("res"); __PYX_ERR(0, 227, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_v_ko)) { __Pyx_RaiseClosureNameError("ko"); __PYX_ERR(0, 227, __pyx_L1_error) }
  if (unlikely(PyObject_SetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_res, __pyx_cur_scope->__pyx_v_ko, __pyx_v_o) < 0)) __PYX_ERR(0, 227, __pyx_L1_error)

  /* "mars/serialize/jsonserializer.pyx":223
 *             if isinstance(value, KeyPlaceholder):
 *                 def cb(ko, vo):
 *                     def inner(subs):             # <<<<<<<<<<<<<<
 *                         o = subs[vo.key, vo.id]
 *                         if weak_ref:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_dict.cb.inner", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":222
 *             res[key] = value
 *             if isinstance(value, KeyPlaceholder):
 *                 def cb(ko, vo):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[vo.key, vo.id]
 */

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_cb(PyObject *__pyx_self, PyObject *__pyx_v_ko, PyObject *__pyx_v_vo) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *__pyx_cur_scope;
  PyObject *__pyx_v_inner = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cb", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *)__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb(__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 222, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_v_ko = __pyx_v_ko;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ko);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ko);
  __pyx_cur_scope->__pyx_v_vo = __pyx_v_vo;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_vo);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_vo);

  /* "mars/serialize/jsonserializer.pyx":223
 *             if isinstance(value, KeyPlaceholder):
 *                 def cb(ko, vo):
 *                     def inner(subs):             # <<<<<<<<<<<<<<
 *                         o = subs[vo.key, vo.id]
 *                         if weak_ref:
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_2cb_1inner, 0, __pyx_n_s_JsonSerializeProvider__deseriali_5, ((PyObject*)__pyx_cur_scope), __pyx_n_s_mars_serialize_jsonserializer, __pyx_d, ((PyObject *)__pyx_codeobj__6)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_inner = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":228
 *                             o = weakref.ref(o)
 *                         res[ko] = o
 *                     return inner             # <<<<<<<<<<<<<<
 *                 callbacks.append(cb(key, value))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_inner);
  __pyx_r = __pyx_v_inner;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":222
 *             res[key] = value
 *             if isinstance(value, KeyPlaceholder):
 *                 def cb(ko, vo):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[vo.key, vo.id]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_dict.cb", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_inner);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":215
 *         }
 * 
 *     cdef inline object _deserialize_dict(self, object obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         res = OrderedDict()
 *         for k, v in obj['value']:
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_dict(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks, int __pyx_v_weak_ref) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *__pyx_cur_scope;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_cb = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("_deserialize_dict", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *)__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict(__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 215, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_weak_ref = __pyx_v_weak_ref;

  /* "mars/serialize/jsonserializer.pyx":216
 * 
 *     cdef inline object _deserialize_dict(self, object obj, list callbacks, bint weak_ref):
 *         res = OrderedDict()             # <<<<<<<<<<<<<<
 *         for k, v in obj['value']:
 *             key = self._deserialize_value(k, callbacks, weak_ref)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_res = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":217
 *     cdef inline object _deserialize_dict(self, object obj, list callbacks, bint weak_ref):
 *         res = OrderedDict()
 *         for k, v in obj['value']:             # <<<<<<<<<<<<<<
 *             key = self._deserialize_value(k, callbacks, weak_ref)
 *             value = self._deserialize_value(v, callbacks, weak_ref)
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 217, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 217, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 217, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 217, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 217, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 217, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 217, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "mars/serialize/jsonserializer.pyx":218
 *         res = OrderedDict()
 *         for k, v in obj['value']:
 *             key = self._deserialize_value(k, callbacks, weak_ref)             # <<<<<<<<<<<<<<
 *             value = self._deserialize_value(v, callbacks, weak_ref)
 *             res[key] = value
 */
    __pyx_t_1 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_value(__pyx_v_self, __pyx_v_k, __pyx_v_callbacks, __pyx_cur_scope->__pyx_v_weak_ref); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":219
 *         for k, v in obj['value']:
 *             key = self._deserialize_value(k, callbacks, weak_ref)
 *             value = self._deserialize_value(v, callbacks, weak_ref)             # <<<<<<<<<<<<<<
 *             res[key] = value
 *             if isinstance(value, KeyPlaceholder):
 */
    __pyx_t_1 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_value(__pyx_v_self, __pyx_v_v, __pyx_v_callbacks, __pyx_cur_scope->__pyx_v_weak_ref); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":220
 *             key = self._deserialize_value(k, callbacks, weak_ref)
 *             value = self._deserialize_value(v, callbacks, weak_ref)
 *             res[key] = value             # <<<<<<<<<<<<<<
 *             if isinstance(value, KeyPlaceholder):
 *                 def cb(ko, vo):
 */
    if (unlikely(PyObject_SetItem(__pyx_cur_scope->__pyx_v_res, __pyx_v_key, __pyx_v_value) < 0)) __PYX_ERR(0, 220, __pyx_L1_error)

    /* "mars/serialize/jsonserializer.pyx":221
 *             value = self._deserialize_value(v, callbacks, weak_ref)
 *             res[key] = value
 *             if isinstance(value, KeyPlaceholder):             # <<<<<<<<<<<<<<
 *                 def cb(ko, vo):
 *                     def inner(subs):
 */
    __pyx_t_9 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_4mars_9serialize_4core_KeyPlaceholder); 
    __pyx_t_10 = (__pyx_t_9 != 0);
    if (__pyx_t_10) {

      /* "mars/serialize/jsonserializer.pyx":222
 *             res[key] = value
 *             if isinstance(value, KeyPlaceholder):
 *                 def cb(ko, vo):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[vo.key, vo.id]
 */
      __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_1cb, 0, __pyx_n_s_JsonSerializeProvider__deseriali_6, ((PyObject*)__pyx_cur_scope), __pyx_n_s_mars_serialize_jsonserializer, __pyx_d, ((PyObject *)__pyx_codeobj__8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_cb, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "mars/serialize/jsonserializer.pyx":229
 *                         res[ko] = o
 *                     return inner
 *                 callbacks.append(cb(key, value))             # <<<<<<<<<<<<<<
 * 
 *         return res
 */
      if (unlikely(__pyx_v_callbacks == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 229, __pyx_L1_error)
      }
      __pyx_t_1 = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17_deserialize_dict_cb(__pyx_v_cb, __pyx_v_key, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 229, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_callbacks, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 229, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "mars/serialize/jsonserializer.pyx":221
 *             value = self._deserialize_value(v, callbacks, weak_ref)
 *             res[key] = value
 *             if isinstance(value, KeyPlaceholder):             # <<<<<<<<<<<<<<
 *                 def cb(ko, vo):
 *                     def inner(subs):
 */
    }

    /* "mars/serialize/jsonserializer.pyx":217
 *     cdef inline object _deserialize_dict(self, object obj, list callbacks, bint weak_ref):
 *         res = OrderedDict()
 *         for k, v in obj['value']:             # <<<<<<<<<<<<<<
 *             key = self._deserialize_value(k, callbacks, weak_ref)
 *             value = self._deserialize_value(v, callbacks, weak_ref)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":231
 *                 callbacks.append(cb(key, value))
 * 
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_datetime64_timedelta64(self, value, tp):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_res);
  __pyx_r = __pyx_cur_scope->__pyx_v_res;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":215
 *         }
 * 
 *     cdef inline object _deserialize_dict(self, object obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         res = OrderedDict()
 *         for k, v in obj['value']:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_cb);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":233
 *         return res
 * 
 *     cdef inline dict _serialize_datetime64_timedelta64(self, value, tp):             # <<<<<<<<<<<<<<
 *         bio = six.BytesIO()
 *         np.save(bio, value)
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64_timedelta64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_tp) {
  PyObject *__pyx_v_bio = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_serialize_datetime64_timedelta64", 0);

  /* "mars/serialize/jsonserializer.pyx":234
 * 
 *     cdef inline dict _serialize_datetime64_timedelta64(self, value, tp):
 *         bio = six.BytesIO()             # <<<<<<<<<<<<<<
 *         np.save(bio, value)
 *         return {
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_six); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BytesIO); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_bio = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":235
 *     cdef inline dict _serialize_datetime64_timedelta64(self, value, tp):
 *         bio = six.BytesIO()
 *         np.save(bio, value)             # <<<<<<<<<<<<<<
 *         return {
 *             'type': _get_name(tp),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_save); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_bio, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_bio, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_bio);
    __Pyx_GIVEREF(__pyx_v_bio);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_bio);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_value);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":236
 *         bio = six.BytesIO()
 *         np.save(bio, value)
 *         return {             # <<<<<<<<<<<<<<
 *             'type': _get_name(tp),
 *             'value': self._to_str(base64.b64encode(bio.getvalue()))
 */
  __Pyx_XDECREF(__pyx_r);

  /* "mars/serialize/jsonserializer.pyx":237
 *         np.save(bio, value)
 *         return {
 *             'type': _get_name(tp),             # <<<<<<<<<<<<<<
 *             'value': self._to_str(base64.b64encode(bio.getvalue()))
 *         }
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_4mars_9serialize_14jsonserializer__get_name(__pyx_v_tp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_t_2) < 0) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":238
 *         return {
 *             'type': _get_name(tp),
 *             'value': self._to_str(base64.b64encode(bio.getvalue()))             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_base64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_b64encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_bio, __pyx_n_s_getvalue); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_str(__pyx_v_self, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_value, __pyx_t_3) < 0) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":233
 *         return res
 * 
 *     cdef inline dict _serialize_datetime64_timedelta64(self, value, tp):             # <<<<<<<<<<<<<<
 *         bio = six.BytesIO()
 *         np.save(bio, value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_datetime64_timedelta64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_bio);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":241
 *         }
 * 
 *     cdef inline _deserialize_datetime64_timedelta64(self, obj, list callbacks):             # <<<<<<<<<<<<<<
 *         cdef bytes v
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_datetime64_timedelta64(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, CYTHON_UNUSED PyObject *__pyx_v_callbacks) {
  PyObject *__pyx_v_v = 0;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_deserialize_datetime64_timedelta64", 0);

  /* "mars/serialize/jsonserializer.pyx":244
 *         cdef bytes v
 * 
 *         value = obj['value']             # <<<<<<<<<<<<<<
 *         v = base64.b64decode(value)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":245
 * 
 *         value = obj['value']
 *         v = base64.b64decode(value)             # <<<<<<<<<<<<<<
 * 
 *         if v is not None:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_base64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_b64decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_value);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 245, __pyx_L1_error)
  __pyx_v_v = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":247
 *         v = base64.b64decode(value)
 * 
 *         if v is not None:             # <<<<<<<<<<<<<<
 *             return np.load(six.BytesIO(v))
 *         return None
 */
  __pyx_t_4 = (__pyx_v_v != ((PyObject*)Py_None));
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "mars/serialize/jsonserializer.pyx":248
 * 
 *         if v is not None:
 *             return np.load(six.BytesIO(v))             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_load); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_six); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_BytesIO); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_v_v) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_v);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":247
 *         v = base64.b64decode(value)
 * 
 *         if v is not None:             # <<<<<<<<<<<<<<
 *             return np.load(six.BytesIO(v))
 *         return None
 */
  }

  /* "mars/serialize/jsonserializer.pyx":249
 *         if v is not None:
 *             return np.load(six.BytesIO(v))
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_datetime64(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":241
 *         }
 * 
 *     cdef inline _deserialize_datetime64_timedelta64(self, obj, list callbacks):             # <<<<<<<<<<<<<<
 *         cdef bytes v
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_datetime64_timedelta64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":251
 *         return None
 * 
 *     cdef inline dict _serialize_datetime64(self, value):             # <<<<<<<<<<<<<<
 *         return self._serialize_datetime64_timedelta64(value, ValueType.datetime64)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_serialize_datetime64", 0);

  /* "mars/serialize/jsonserializer.pyx":252
 * 
 *     cdef inline dict _serialize_datetime64(self, value):
 *         return self._serialize_datetime64_timedelta64(value, ValueType.datetime64)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _deserialize_datetime64(self, obj, list callbacks):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_datetime64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64_timedelta64(__pyx_v_self, __pyx_v_value, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":251
 *         return None
 * 
 *     cdef inline dict _serialize_datetime64(self, value):             # <<<<<<<<<<<<<<
 *         return self._serialize_datetime64_timedelta64(value, ValueType.datetime64)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_datetime64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":254
 *         return self._serialize_datetime64_timedelta64(value, ValueType.datetime64)
 * 
 *     cdef inline _deserialize_datetime64(self, obj, list callbacks):             # <<<<<<<<<<<<<<
 *         return self._deserialize_datetime64_timedelta64(obj, callbacks)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_datetime64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_deserialize_datetime64", 0);

  /* "mars/serialize/jsonserializer.pyx":255
 * 
 *     cdef inline _deserialize_datetime64(self, obj, list callbacks):
 *         return self._deserialize_datetime64_timedelta64(obj, callbacks)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline dict _serialize_timedelta64(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_datetime64_timedelta64(__pyx_v_self, __pyx_v_obj, __pyx_v_callbacks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":254
 *         return self._serialize_datetime64_timedelta64(value, ValueType.datetime64)
 * 
 *     cdef inline _deserialize_datetime64(self, obj, list callbacks):             # <<<<<<<<<<<<<<
 *         return self._deserialize_datetime64_timedelta64(obj, callbacks)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_datetime64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":257
 *         return self._deserialize_datetime64_timedelta64(obj, callbacks)
 * 
 *     cdef inline dict _serialize_timedelta64(self, value):             # <<<<<<<<<<<<<<
 *         return self._serialize_datetime64_timedelta64(value, ValueType.timedelta64)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_timedelta64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_serialize_timedelta64", 0);

  /* "mars/serialize/jsonserializer.pyx":258
 * 
 *     cdef inline dict _serialize_timedelta64(self, value):
 *         return self._serialize_datetime64_timedelta64(value, ValueType.timedelta64)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _deserialize_timedelta64(self, obj, list callbacks):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_timedelta64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64_timedelta64(__pyx_v_self, __pyx_v_value, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":257
 *         return self._deserialize_datetime64_timedelta64(obj, callbacks)
 * 
 *     cdef inline dict _serialize_timedelta64(self, value):             # <<<<<<<<<<<<<<
 *         return self._serialize_datetime64_timedelta64(value, ValueType.timedelta64)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_timedelta64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":260
 *         return self._serialize_datetime64_timedelta64(value, ValueType.timedelta64)
 * 
 *     cdef inline _deserialize_timedelta64(self, obj, list callbacks):             # <<<<<<<<<<<<<<
 *         return self._deserialize_datetime64_timedelta64(obj, callbacks)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_timedelta64(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_deserialize_timedelta64", 0);

  /* "mars/serialize/jsonserializer.pyx":261
 * 
 *     cdef inline _deserialize_timedelta64(self, obj, list callbacks):
 *         return self._deserialize_datetime64_timedelta64(obj, callbacks)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline object _serialize_typed_value(self, value, tp, bint weak_ref=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_datetime64_timedelta64(__pyx_v_self, __pyx_v_obj, __pyx_v_callbacks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":260
 *         return self._serialize_datetime64_timedelta64(value, ValueType.timedelta64)
 * 
 *     cdef inline _deserialize_timedelta64(self, obj, list callbacks):             # <<<<<<<<<<<<<<
 *         return self._deserialize_datetime64_timedelta64(obj, callbacks)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_timedelta64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":263
 *         return self._deserialize_datetime64_timedelta64(obj, callbacks)
 * 
 *     cdef inline object _serialize_typed_value(self, value, tp, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if type(tp) not in (List, Tuple, Dict) and weak_ref:
 *             # not iterable, and is weak ref
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_tp, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value *__pyx_optional_args) {
  int __pyx_v_weak_ref = ((int)0);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value __pyx_t_8;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list __pyx_t_9;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple __pyx_t_10;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict __pyx_t_11;
  __Pyx_RefNannySetupContext("_serialize_typed_value", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_weak_ref = __pyx_optional_args->weak_ref;
    }
  }
  __Pyx_INCREF(__pyx_v_value);

  /* "mars/serialize/jsonserializer.pyx":264
 * 
 *     cdef inline object _serialize_typed_value(self, value, tp, bint weak_ref=False):
 *         if type(tp) not in (List, Tuple, Dict) and weak_ref:             # <<<<<<<<<<<<<<
 *             # not iterable, and is weak ref
 *             value = value()
 */
  __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_tp)));
  __pyx_t_2 = ((PyObject *)Py_TYPE(__pyx_v_tp));
  __pyx_t_4 = PyObject_RichCompare(((PyObject *)__pyx_t_2), ((PyObject *)__pyx_ptype_4mars_9serialize_4core_List), Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 264, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = PyObject_RichCompare(((PyObject *)__pyx_t_2), ((PyObject *)__pyx_ptype_4mars_9serialize_4core_Tuple), Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 264, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = PyObject_RichCompare(((PyObject *)__pyx_t_2), ((PyObject *)__pyx_ptype_4mars_9serialize_4core_Dict), Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 264, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __pyx_t_5;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_3 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = (__pyx_v_weak_ref != 0);
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":266
 *         if type(tp) not in (List, Tuple, Dict) and weak_ref:
 *             # not iterable, and is weak ref
 *             value = value()             # <<<<<<<<<<<<<<
 *         if value is None:
 *             return
 */
    __Pyx_INCREF(__pyx_v_value);
    __pyx_t_4 = __pyx_v_value; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "mars/serialize/jsonserializer.pyx":264
 * 
 *     cdef inline object _serialize_typed_value(self, value, tp, bint weak_ref=False):
 *         if type(tp) not in (List, Tuple, Dict) and weak_ref:             # <<<<<<<<<<<<<<
 *             # not iterable, and is weak ref
 *             value = value()
 */
  }

  /* "mars/serialize/jsonserializer.pyx":267
 *             # not iterable, and is weak ref
 *             value = value()
 *         if value is None:             # <<<<<<<<<<<<<<
 *             return
 *         if tp == ValueType.bytes:
 */
  __pyx_t_1 = (__pyx_v_value == Py_None);
  __pyx_t_5 = (__pyx_t_1 != 0);
  if (__pyx_t_5) {

    /* "mars/serialize/jsonserializer.pyx":268
 *             value = value()
 *         if value is None:
 *             return             # <<<<<<<<<<<<<<
 *         if tp == ValueType.bytes:
 *             return {
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":267
 *             # not iterable, and is weak ref
 *             value = value()
 *         if value is None:             # <<<<<<<<<<<<<<
 *             return
 *         if tp == ValueType.bytes:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":269
 *         if value is None:
 *             return
 *         if tp == ValueType.bytes:             # <<<<<<<<<<<<<<
 *             return {
 *                 'type': _get_name(tp),
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_bytes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_tp, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {

    /* "mars/serialize/jsonserializer.pyx":270
 *             return
 *         if tp == ValueType.bytes:
 *             return {             # <<<<<<<<<<<<<<
 *                 'type': _get_name(tp),
 *                 'value': self._to_str(base64.b64encode(value))
 */
    __Pyx_XDECREF(__pyx_r);

    /* "mars/serialize/jsonserializer.pyx":271
 *         if tp == ValueType.bytes:
 *             return {
 *                 'type': _get_name(tp),             # <<<<<<<<<<<<<<
 *                 'value': self._to_str(base64.b64encode(value))
 *             }
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __pyx_f_4mars_9serialize_14jsonserializer__get_name(__pyx_v_tp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_type, __pyx_t_2) < 0) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "mars/serialize/jsonserializer.pyx":272
 *             return {
 *                 'type': _get_name(tp),
 *                 'value': self._to_str(base64.b64encode(value))             # <<<<<<<<<<<<<<
 *             }
 *         elif tp in PRIMITIVE_TYPE_TO_NAME:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_base64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_b64encode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_value);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_str(__pyx_v_self, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_value, __pyx_t_7) < 0) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":269
 *         if value is None:
 *             return
 *         if tp == ValueType.bytes:             # <<<<<<<<<<<<<<
 *             return {
 *                 'type': _get_name(tp),
 */
  }

  /* "mars/serialize/jsonserializer.pyx":274
 *                 'value': self._to_str(base64.b64encode(value))
 *             }
 *         elif tp in PRIMITIVE_TYPE_TO_NAME:             # <<<<<<<<<<<<<<
 *             # primitive type, we do not do any type check here
 *             return value
 */
  if (unlikely(__pyx_v_4mars_9serialize_14jsonserializer_PRIMITIVE_TYPE_TO_NAME == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 274, __pyx_L1_error)
  }
  __pyx_t_5 = (__Pyx_PyDict_ContainsTF(__pyx_v_tp, __pyx_v_4mars_9serialize_14jsonserializer_PRIMITIVE_TYPE_TO_NAME, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 274, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_5 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":276
 *         elif tp in PRIMITIVE_TYPE_TO_NAME:
 *             # primitive type, we do not do any type check here
 *             return value             # <<<<<<<<<<<<<<
 *         elif type(tp) is Identity:
 *             return self._serialize_typed_value(value, tp.type, weak_ref=weak_ref)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":274
 *                 'value': self._to_str(base64.b64encode(value))
 *             }
 *         elif tp in PRIMITIVE_TYPE_TO_NAME:             # <<<<<<<<<<<<<<
 *             # primitive type, we do not do any type check here
 *             return value
 */
  }

  /* "mars/serialize/jsonserializer.pyx":277
 *             # primitive type, we do not do any type check here
 *             return value
 *         elif type(tp) is Identity:             # <<<<<<<<<<<<<<
 *             return self._serialize_typed_value(value, tp.type, weak_ref=weak_ref)
 *         elif tp is ValueType.slice:
 */
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_tp)) == ((PyObject *)__pyx_ptype_4mars_9serialize_4core_Identity));
  __pyx_t_5 = (__pyx_t_1 != 0);
  if (__pyx_t_5) {

    /* "mars/serialize/jsonserializer.pyx":278
 *             return value
 *         elif type(tp) is Identity:
 *             return self._serialize_typed_value(value, tp.type, weak_ref=weak_ref)             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.slice:
 *             return self._serialize_slice(value)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_tp, __pyx_n_s_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8.__pyx_n = 1;
    __pyx_t_8.weak_ref = __pyx_v_weak_ref;
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value(__pyx_v_self, __pyx_v_value, __pyx_t_4, &__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":277
 *             # primitive type, we do not do any type check here
 *             return value
 *         elif type(tp) is Identity:             # <<<<<<<<<<<<<<
 *             return self._serialize_typed_value(value, tp.type, weak_ref=weak_ref)
 *         elif tp is ValueType.slice:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":279
 *         elif type(tp) is Identity:
 *             return self._serialize_typed_value(value, tp.type, weak_ref=weak_ref)
 *         elif tp is ValueType.slice:             # <<<<<<<<<<<<<<
 *             return self._serialize_slice(value)
 *         elif tp is ValueType.arr:
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_slice); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = (__pyx_v_tp == __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_1 = (__pyx_t_5 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":280
 *             return self._serialize_typed_value(value, tp.type, weak_ref=weak_ref)
 *         elif tp is ValueType.slice:
 *             return self._serialize_slice(value)             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.arr:
 *             return self._serialize_arr(value)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PySlice_Check(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "slice", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 280, __pyx_L1_error)
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_slice(__pyx_v_self, ((PyObject*)__pyx_v_value)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":279
 *         elif type(tp) is Identity:
 *             return self._serialize_typed_value(value, tp.type, weak_ref=weak_ref)
 *         elif tp is ValueType.slice:             # <<<<<<<<<<<<<<
 *             return self._serialize_slice(value)
 *         elif tp is ValueType.arr:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":281
 *         elif tp is ValueType.slice:
 *             return self._serialize_slice(value)
 *         elif tp is ValueType.arr:             # <<<<<<<<<<<<<<
 *             return self._serialize_arr(value)
 *         elif tp is ValueType.dtype:
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_arr); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = (__pyx_v_tp == __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_5 = (__pyx_t_1 != 0);
  if (__pyx_t_5) {

    /* "mars/serialize/jsonserializer.pyx":282
 *             return self._serialize_slice(value)
 *         elif tp is ValueType.arr:
 *             return self._serialize_arr(value)             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.dtype:
 *             return self._serialize_dtype(value)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 282, __pyx_L1_error)
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_arr(__pyx_v_self, ((PyArrayObject *)__pyx_v_value)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":281
 *         elif tp is ValueType.slice:
 *             return self._serialize_slice(value)
 *         elif tp is ValueType.arr:             # <<<<<<<<<<<<<<
 *             return self._serialize_arr(value)
 *         elif tp is ValueType.dtype:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":283
 *         elif tp is ValueType.arr:
 *             return self._serialize_arr(value)
 *         elif tp is ValueType.dtype:             # <<<<<<<<<<<<<<
 *             return self._serialize_dtype(value)
 *         elif tp is ValueType.key:
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_dtype); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = (__pyx_v_tp == __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_1 = (__pyx_t_5 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":284
 *             return self._serialize_arr(value)
 *         elif tp is ValueType.dtype:
 *             return self._serialize_dtype(value)             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.key:
 *             return self._serialize_key(value)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(0, 284, __pyx_L1_error)
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dtype(__pyx_v_self, ((PyArray_Descr *)__pyx_v_value)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":283
 *         elif tp is ValueType.arr:
 *             return self._serialize_arr(value)
 *         elif tp is ValueType.dtype:             # <<<<<<<<<<<<<<
 *             return self._serialize_dtype(value)
 *         elif tp is ValueType.key:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":285
 *         elif tp is ValueType.dtype:
 *             return self._serialize_dtype(value)
 *         elif tp is ValueType.key:             # <<<<<<<<<<<<<<
 *             return self._serialize_key(value)
 *         elif tp == ValueType.datetime64:
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_key); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = (__pyx_v_tp == __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_5 = (__pyx_t_1 != 0);
  if (__pyx_t_5) {

    /* "mars/serialize/jsonserializer.pyx":286
 *             return self._serialize_dtype(value)
 *         elif tp is ValueType.key:
 *             return self._serialize_key(value)             # <<<<<<<<<<<<<<
 *         elif tp == ValueType.datetime64:
 *             return self._serialize_datetime64(value)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_key(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":285
 *         elif tp is ValueType.dtype:
 *             return self._serialize_dtype(value)
 *         elif tp is ValueType.key:             # <<<<<<<<<<<<<<
 *             return self._serialize_key(value)
 *         elif tp == ValueType.datetime64:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":287
 *         elif tp is ValueType.key:
 *             return self._serialize_key(value)
 *         elif tp == ValueType.datetime64:             # <<<<<<<<<<<<<<
 *             return self._serialize_datetime64(value)
 *         elif tp == ValueType.timedelta64:
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_datetime64); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_tp, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {

    /* "mars/serialize/jsonserializer.pyx":288
 *             return self._serialize_key(value)
 *         elif tp == ValueType.datetime64:
 *             return self._serialize_datetime64(value)             # <<<<<<<<<<<<<<
 *         elif tp == ValueType.timedelta64:
 *             return self._serialize_timedelta64(value)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":287
 *         elif tp is ValueType.key:
 *             return self._serialize_key(value)
 *         elif tp == ValueType.datetime64:             # <<<<<<<<<<<<<<
 *             return self._serialize_datetime64(value)
 *         elif tp == ValueType.timedelta64:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":289
 *         elif tp == ValueType.datetime64:
 *             return self._serialize_datetime64(value)
 *         elif tp == ValueType.timedelta64:             # <<<<<<<<<<<<<<
 *             return self._serialize_timedelta64(value)
 *         elif isinstance(tp, List):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_timedelta64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyObject_RichCompare(__pyx_v_tp, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_5) {

    /* "mars/serialize/jsonserializer.pyx":290
 *             return self._serialize_datetime64(value)
 *         elif tp == ValueType.timedelta64:
 *             return self._serialize_timedelta64(value)             # <<<<<<<<<<<<<<
 *         elif isinstance(tp, List):
 *             if not isinstance(value, list):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_timedelta64(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":289
 *         elif tp == ValueType.datetime64:
 *             return self._serialize_datetime64(value)
 *         elif tp == ValueType.timedelta64:             # <<<<<<<<<<<<<<
 *             return self._serialize_timedelta64(value)
 *         elif isinstance(tp, List):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":291
 *         elif tp == ValueType.timedelta64:
 *             return self._serialize_timedelta64(value)
 *         elif isinstance(tp, List):             # <<<<<<<<<<<<<<
 *             if not isinstance(value, list):
 *                 value = list(value)
 */
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_tp, __pyx_ptype_4mars_9serialize_4core_List); 
  __pyx_t_1 = (__pyx_t_5 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":292
 *             return self._serialize_timedelta64(value)
 *         elif isinstance(tp, List):
 *             if not isinstance(value, list):             # <<<<<<<<<<<<<<
 *                 value = list(value)
 *             return self._serialize_list(value, tp, weak_ref=weak_ref)
 */
    __pyx_t_1 = PyList_Check(__pyx_v_value); 
    __pyx_t_5 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_5) {

      /* "mars/serialize/jsonserializer.pyx":293
 *         elif isinstance(tp, List):
 *             if not isinstance(value, list):
 *                 value = list(value)             # <<<<<<<<<<<<<<
 *             return self._serialize_list(value, tp, weak_ref=weak_ref)
 *         elif isinstance(tp, Tuple):
 */
      __pyx_t_7 = PySequence_List(__pyx_v_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "mars/serialize/jsonserializer.pyx":292
 *             return self._serialize_timedelta64(value)
 *         elif isinstance(tp, List):
 *             if not isinstance(value, list):             # <<<<<<<<<<<<<<
 *                 value = list(value)
 *             return self._serialize_list(value, tp, weak_ref=weak_ref)
 */
    }

    /* "mars/serialize/jsonserializer.pyx":294
 *             if not isinstance(value, list):
 *                 value = list(value)
 *             return self._serialize_list(value, tp, weak_ref=weak_ref)             # <<<<<<<<<<<<<<
 *         elif isinstance(tp, Tuple):
 *             return self._serialize_tuple(value, tp, weak_ref=weak_ref)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 294, __pyx_L1_error)
    __pyx_t_9.__pyx_n = 2;
    __pyx_t_9.tp = __pyx_v_tp;
    __pyx_t_9.weak_ref = __pyx_v_weak_ref;
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list(__pyx_v_self, ((PyObject*)__pyx_v_value), &__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":291
 *         elif tp == ValueType.timedelta64:
 *             return self._serialize_timedelta64(value)
 *         elif isinstance(tp, List):             # <<<<<<<<<<<<<<
 *             if not isinstance(value, list):
 *                 value = list(value)
 */
  }

  /* "mars/serialize/jsonserializer.pyx":295
 *                 value = list(value)
 *             return self._serialize_list(value, tp, weak_ref=weak_ref)
 *         elif isinstance(tp, Tuple):             # <<<<<<<<<<<<<<
 *             return self._serialize_tuple(value, tp, weak_ref=weak_ref)
 *         elif isinstance(tp, Dict):
 */
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_tp, __pyx_ptype_4mars_9serialize_4core_Tuple); 
  __pyx_t_1 = (__pyx_t_5 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":296
 *             return self._serialize_list(value, tp, weak_ref=weak_ref)
 *         elif isinstance(tp, Tuple):
 *             return self._serialize_tuple(value, tp, weak_ref=weak_ref)             # <<<<<<<<<<<<<<
 *         elif isinstance(tp, Dict):
 *             return self._serialize_dict(value, tp, weak_ref=weak_ref)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyTuple_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 296, __pyx_L1_error)
    __pyx_t_10.__pyx_n = 2;
    __pyx_t_10.tp = __pyx_v_tp;
    __pyx_t_10.weak_ref = __pyx_v_weak_ref;
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple(__pyx_v_self, ((PyObject*)__pyx_v_value), &__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":295
 *                 value = list(value)
 *             return self._serialize_list(value, tp, weak_ref=weak_ref)
 *         elif isinstance(tp, Tuple):             # <<<<<<<<<<<<<<
 *             return self._serialize_tuple(value, tp, weak_ref=weak_ref)
 *         elif isinstance(tp, Dict):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":297
 *         elif isinstance(tp, Tuple):
 *             return self._serialize_tuple(value, tp, weak_ref=weak_ref)
 *         elif isinstance(tp, Dict):             # <<<<<<<<<<<<<<
 *             return self._serialize_dict(value, tp, weak_ref=weak_ref)
 *         else:
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_tp, __pyx_ptype_4mars_9serialize_4core_Dict); 
  __pyx_t_5 = (__pyx_t_1 != 0);
  if (likely(__pyx_t_5)) {

    /* "mars/serialize/jsonserializer.pyx":298
 *             return self._serialize_tuple(value, tp, weak_ref=weak_ref)
 *         elif isinstance(tp, Dict):
 *             return self._serialize_dict(value, tp, weak_ref=weak_ref)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Unknown type to serialize: {0}'.format(tp))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_11.__pyx_n = 2;
    __pyx_t_11.tp = __pyx_v_tp;
    __pyx_t_11.weak_ref = __pyx_v_weak_ref;
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict(__pyx_v_self, __pyx_v_value, &__pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":297
 *         elif isinstance(tp, Tuple):
 *             return self._serialize_tuple(value, tp, weak_ref=weak_ref)
 *         elif isinstance(tp, Dict):             # <<<<<<<<<<<<<<
 *             return self._serialize_dict(value, tp, weak_ref=weak_ref)
 *         else:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":300
 *             return self._serialize_dict(value, tp, weak_ref=weak_ref)
 *         else:
 *             raise TypeError('Unknown type to serialize: {0}'.format(tp))             # <<<<<<<<<<<<<<
 * 
 *     cdef inline object _serialize_untyped_value(self, value, bint weak_ref=False):
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Unknown_type_to_serialize_0, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_7 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_v_tp) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_tp);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 300, __pyx_L1_error)
  }

  /* "mars/serialize/jsonserializer.pyx":263
 *         return self._deserialize_datetime64_timedelta64(obj, callbacks)
 * 
 *     cdef inline object _serialize_typed_value(self, value, tp, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if type(tp) not in (List, Tuple, Dict) and weak_ref:
 *             # not iterable, and is weak ref
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_typed_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":302
 *             raise TypeError('Unknown type to serialize: {0}'.format(tp))
 * 
 *     cdef inline object _serialize_untyped_value(self, value, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (list, tuple, dict)) and weak_ref:
 *             # not iterable, and is weak ref
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value *__pyx_optional_args) {
  int __pyx_v_weak_ref = ((int)0);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list __pyx_t_9;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple __pyx_t_10;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict __pyx_t_11;
  __Pyx_RefNannySetupContext("_serialize_untyped_value", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_weak_ref = __pyx_optional_args->weak_ref;
    }
  }
  __Pyx_INCREF(__pyx_v_value);

  /* "mars/serialize/jsonserializer.pyx":303
 * 
 *     cdef inline object _serialize_untyped_value(self, value, bint weak_ref=False):
 *         if not isinstance(value, (list, tuple, dict)) and weak_ref:             # <<<<<<<<<<<<<<
 *             # not iterable, and is weak ref
 *             value = value()
 */
  __pyx_t_3 = PyList_Check(__pyx_v_value); 
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = PyDict_Check(__pyx_v_value); 
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = (__pyx_v_weak_ref != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":305
 *         if not isinstance(value, (list, tuple, dict)) and weak_ref:
 *             # not iterable, and is weak ref
 *             value = value()             # <<<<<<<<<<<<<<
 *         if value is None:
 *             return
 */
    __Pyx_INCREF(__pyx_v_value);
    __pyx_t_6 = __pyx_v_value; __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "mars/serialize/jsonserializer.pyx":303
 * 
 *     cdef inline object _serialize_untyped_value(self, value, bint weak_ref=False):
 *         if not isinstance(value, (list, tuple, dict)) and weak_ref:             # <<<<<<<<<<<<<<
 *             # not iterable, and is weak ref
 *             value = value()
 */
  }

  /* "mars/serialize/jsonserializer.pyx":306
 *             # not iterable, and is weak ref
 *             value = value()
 *         if value is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_value == Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "mars/serialize/jsonserializer.pyx":307
 *             value = value()
 *         if value is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(value, bool):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":306
 *             # not iterable, and is weak ref
 *             value = value()
 *         if value is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "mars/serialize/jsonserializer.pyx":309
 *             return
 * 
 *         if isinstance(value, bool):             # <<<<<<<<<<<<<<
 *             return value
 *         elif isinstance(value, bytes):
 */
  __pyx_t_5 = ((PyObject*)&PyBool_Type);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_5); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":310
 * 
 *         if isinstance(value, bool):
 *             return value             # <<<<<<<<<<<<<<
 *         elif isinstance(value, bytes):
 *             return {
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":309
 *             return
 * 
 *         if isinstance(value, bool):             # <<<<<<<<<<<<<<
 *             return value
 *         elif isinstance(value, bytes):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":311
 *         if isinstance(value, bool):
 *             return value
 *         elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *             return {
 *                 'type': _get_name(ValueType.bytes),
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_value); 
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "mars/serialize/jsonserializer.pyx":312
 *             return value
 *         elif isinstance(value, bytes):
 *             return {             # <<<<<<<<<<<<<<
 *                 'type': _get_name(ValueType.bytes),
 *                 'value': self._to_str(base64.b64encode(value)),
 */
    __Pyx_XDECREF(__pyx_r);

    /* "mars/serialize/jsonserializer.pyx":313
 *         elif isinstance(value, bytes):
 *             return {
 *                 'type': _get_name(ValueType.bytes),             # <<<<<<<<<<<<<<
 *                 'value': self._to_str(base64.b64encode(value)),
 *             }
 */
    __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_bytes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer__get_name(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_type, __pyx_t_7) < 0) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "mars/serialize/jsonserializer.pyx":314
 *             return {
 *                 'type': _get_name(ValueType.bytes),
 *                 'value': self._to_str(base64.b64encode(value)),             # <<<<<<<<<<<<<<
 *             }
 *         elif isinstance(value, unicode):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_base64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_b64encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_7 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_6, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_value);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_str(__pyx_v_self, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_value, __pyx_t_8) < 0) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":311
 *         if isinstance(value, bool):
 *             return value
 *         elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *             return {
 *                 'type': _get_name(ValueType.bytes),
 */
  }

  /* "mars/serialize/jsonserializer.pyx":316
 *                 'value': self._to_str(base64.b64encode(value)),
 *             }
 *         elif isinstance(value, unicode):             # <<<<<<<<<<<<<<
 *             return value
 *         elif isinstance(value, (int, long)):
 */
  __pyx_t_4 = PyUnicode_Check(__pyx_v_value); 
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":317
 *             }
 *         elif isinstance(value, unicode):
 *             return value             # <<<<<<<<<<<<<<
 *         elif isinstance(value, (int, long)):
 *             return value
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":316
 *                 'value': self._to_str(base64.b64encode(value)),
 *             }
 *         elif isinstance(value, unicode):             # <<<<<<<<<<<<<<
 *             return value
 *         elif isinstance(value, (int, long)):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":318
 *         elif isinstance(value, unicode):
 *             return value
 *         elif isinstance(value, (int, long)):             # <<<<<<<<<<<<<<
 *             return value
 *         elif isinstance(value, float):
 */
  __pyx_t_4 = PyInt_Check(__pyx_v_value); 
  __pyx_t_2 = (__pyx_t_4 != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_2 = PyLong_Check(__pyx_v_value); 
  __pyx_t_4 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L11_bool_binop_done:;
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "mars/serialize/jsonserializer.pyx":319
 *             return value
 *         elif isinstance(value, (int, long)):
 *             return value             # <<<<<<<<<<<<<<
 *         elif isinstance(value, float):
 *             return value
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":318
 *         elif isinstance(value, unicode):
 *             return value
 *         elif isinstance(value, (int, long)):             # <<<<<<<<<<<<<<
 *             return value
 *         elif isinstance(value, float):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":320
 *         elif isinstance(value, (int, long)):
 *             return value
 *         elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *             return value
 *         elif isinstance(value, slice):
 */
  __pyx_t_4 = PyFloat_Check(__pyx_v_value); 
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":321
 *             return value
 *         elif isinstance(value, float):
 *             return value             # <<<<<<<<<<<<<<
 *         elif isinstance(value, slice):
 *             return self._serialize_slice(value)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":320
 *         elif isinstance(value, (int, long)):
 *             return value
 *         elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *             return value
 *         elif isinstance(value, slice):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":322
 *         elif isinstance(value, float):
 *             return value
 *         elif isinstance(value, slice):             # <<<<<<<<<<<<<<
 *             return self._serialize_slice(value)
 *         elif isinstance(value, np.ndarray):
 */
  __pyx_t_1 = PySlice_Check(__pyx_v_value); 
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "mars/serialize/jsonserializer.pyx":323
 *             return value
 *         elif isinstance(value, slice):
 *             return self._serialize_slice(value)             # <<<<<<<<<<<<<<
 *         elif isinstance(value, np.ndarray):
 *             return self._serialize_arr(value)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PySlice_Check(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "slice", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 323, __pyx_L1_error)
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_slice(__pyx_v_self, ((PyObject*)__pyx_v_value)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":322
 *         elif isinstance(value, float):
 *             return value
 *         elif isinstance(value, slice):             # <<<<<<<<<<<<<<
 *             return self._serialize_slice(value)
 *         elif isinstance(value, np.ndarray):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":324
 *         elif isinstance(value, slice):
 *             return self._serialize_slice(value)
 *         elif isinstance(value, np.ndarray):             # <<<<<<<<<<<<<<
 *             return self._serialize_arr(value)
 *         elif isinstance(value, np.dtype):
 */
  __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5numpy_ndarray); 
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":325
 *             return self._serialize_slice(value)
 *         elif isinstance(value, np.ndarray):
 *             return self._serialize_arr(value)             # <<<<<<<<<<<<<<
 *         elif isinstance(value, np.dtype):
 *             return self._serialize_dtype(value)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 325, __pyx_L1_error)
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_arr(__pyx_v_self, ((PyArrayObject *)__pyx_v_value)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":324
 *         elif isinstance(value, slice):
 *             return self._serialize_slice(value)
 *         elif isinstance(value, np.ndarray):             # <<<<<<<<<<<<<<
 *             return self._serialize_arr(value)
 *         elif isinstance(value, np.dtype):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":326
 *         elif isinstance(value, np.ndarray):
 *             return self._serialize_arr(value)
 *         elif isinstance(value, np.dtype):             # <<<<<<<<<<<<<<
 *             return self._serialize_dtype(value)
 *         elif isinstance(value, BaseWithKey):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5numpy_dtype); 
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "mars/serialize/jsonserializer.pyx":327
 *             return self._serialize_arr(value)
 *         elif isinstance(value, np.dtype):
 *             return self._serialize_dtype(value)             # <<<<<<<<<<<<<<
 *         elif isinstance(value, BaseWithKey):
 *             return self._serialize_key(value)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(0, 327, __pyx_L1_error)
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dtype(__pyx_v_self, ((PyArray_Descr *)__pyx_v_value)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":326
 *         elif isinstance(value, np.ndarray):
 *             return self._serialize_arr(value)
 *         elif isinstance(value, np.dtype):             # <<<<<<<<<<<<<<
 *             return self._serialize_dtype(value)
 *         elif isinstance(value, BaseWithKey):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":328
 *         elif isinstance(value, np.dtype):
 *             return self._serialize_dtype(value)
 *         elif isinstance(value, BaseWithKey):             # <<<<<<<<<<<<<<
 *             return self._serialize_key(value)
 *         elif isinstance(value, list):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_BaseWithKey); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_5); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":329
 *             return self._serialize_dtype(value)
 *         elif isinstance(value, BaseWithKey):
 *             return self._serialize_key(value)             # <<<<<<<<<<<<<<
 *         elif isinstance(value, list):
 *             return self._serialize_list(value, tp=None, weak_ref=weak_ref)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_key(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":328
 *         elif isinstance(value, np.dtype):
 *             return self._serialize_dtype(value)
 *         elif isinstance(value, BaseWithKey):             # <<<<<<<<<<<<<<
 *             return self._serialize_key(value)
 *         elif isinstance(value, list):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":330
 *         elif isinstance(value, BaseWithKey):
 *             return self._serialize_key(value)
 *         elif isinstance(value, list):             # <<<<<<<<<<<<<<
 *             return self._serialize_list(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, tuple):
 */
  __pyx_t_1 = PyList_Check(__pyx_v_value); 
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "mars/serialize/jsonserializer.pyx":331
 *             return self._serialize_key(value)
 *         elif isinstance(value, list):
 *             return self._serialize_list(value, tp=None, weak_ref=weak_ref)             # <<<<<<<<<<<<<<
 *         elif isinstance(value, tuple):
 *             return self._serialize_tuple(value, tp=None, weak_ref=weak_ref)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 331, __pyx_L1_error)
    __pyx_t_9.__pyx_n = 2;
    __pyx_t_9.tp = Py_None;
    __pyx_t_9.weak_ref = __pyx_v_weak_ref;
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list(__pyx_v_self, ((PyObject*)__pyx_v_value), &__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":330
 *         elif isinstance(value, BaseWithKey):
 *             return self._serialize_key(value)
 *         elif isinstance(value, list):             # <<<<<<<<<<<<<<
 *             return self._serialize_list(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, tuple):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":332
 *         elif isinstance(value, list):
 *             return self._serialize_list(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             return self._serialize_tuple(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, dict):
 */
  __pyx_t_4 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":333
 *             return self._serialize_list(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, tuple):
 *             return self._serialize_tuple(value, tp=None, weak_ref=weak_ref)             # <<<<<<<<<<<<<<
 *         elif isinstance(value, dict):
 *             return self._serialize_dict(value, tp=None, weak_ref=weak_ref)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyTuple_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 333, __pyx_L1_error)
    __pyx_t_10.__pyx_n = 2;
    __pyx_t_10.tp = Py_None;
    __pyx_t_10.weak_ref = __pyx_v_weak_ref;
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple(__pyx_v_self, ((PyObject*)__pyx_v_value), &__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":332
 *         elif isinstance(value, list):
 *             return self._serialize_list(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             return self._serialize_tuple(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, dict):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":334
 *         elif isinstance(value, tuple):
 *             return self._serialize_tuple(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, dict):             # <<<<<<<<<<<<<<
 *             return self._serialize_dict(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, np.datetime64):
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_value); 
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "mars/serialize/jsonserializer.pyx":335
 *             return self._serialize_tuple(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, dict):
 *             return self._serialize_dict(value, tp=None, weak_ref=weak_ref)             # <<<<<<<<<<<<<<
 *         elif isinstance(value, np.datetime64):
 *             return self._serialize_datetime64(value)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_11.__pyx_n = 2;
    __pyx_t_11.tp = Py_None;
    __pyx_t_11.weak_ref = __pyx_v_weak_ref;
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict(__pyx_v_self, __pyx_v_value, &__pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":334
 *         elif isinstance(value, tuple):
 *             return self._serialize_tuple(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, dict):             # <<<<<<<<<<<<<<
 *             return self._serialize_dict(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, np.datetime64):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":336
 *         elif isinstance(value, dict):
 *             return self._serialize_dict(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, np.datetime64):             # <<<<<<<<<<<<<<
 *             return self._serialize_datetime64(value)
 *         elif isinstance(value, np.timedelta64):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_datetime64); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_8); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":337
 *             return self._serialize_dict(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, np.datetime64):
 *             return self._serialize_datetime64(value)             # <<<<<<<<<<<<<<
 *         elif isinstance(value, np.timedelta64):
 *             return self._serialize_timedelta64(value)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":336
 *         elif isinstance(value, dict):
 *             return self._serialize_dict(value, tp=None, weak_ref=weak_ref)
 *         elif isinstance(value, np.datetime64):             # <<<<<<<<<<<<<<
 *             return self._serialize_datetime64(value)
 *         elif isinstance(value, np.timedelta64):
 */
  }

  /* "mars/serialize/jsonserializer.pyx":338
 *         elif isinstance(value, np.datetime64):
 *             return self._serialize_datetime64(value)
 *         elif isinstance(value, np.timedelta64):             # <<<<<<<<<<<<<<
 *             return self._serialize_timedelta64(value)
 *         else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_timedelta64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_value, __pyx_t_5); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (likely(__pyx_t_4)) {

    /* "mars/serialize/jsonserializer.pyx":339
 *             return self._serialize_datetime64(value)
 *         elif isinstance(value, np.timedelta64):
 *             return self._serialize_timedelta64(value)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Unknown type to serialize: {0}'.format(type(value)))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_timedelta64(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":338
 *         elif isinstance(value, np.datetime64):
 *             return self._serialize_datetime64(value)
 *         elif isinstance(value, np.timedelta64):             # <<<<<<<<<<<<<<
 *             return self._serialize_timedelta64(value)
 *         else:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":341
 *             return self._serialize_timedelta64(value)
 *         else:
 *             raise TypeError('Unknown type to serialize: {0}'.format(type(value)))             # <<<<<<<<<<<<<<
 * 
 *     cdef inline object _serialize_value(self, value, tp=None, bint weak_ref=False):
 */
  /*else*/ {
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Unknown_type_to_serialize_0, __pyx_n_s_format); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, ((PyObject *)Py_TYPE(__pyx_v_value))) : __Pyx_PyObject_CallOneArg(__pyx_t_8, ((PyObject *)Py_TYPE(__pyx_v_value)));
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 341, __pyx_L1_error)
  }

  /* "mars/serialize/jsonserializer.pyx":302
 *             raise TypeError('Unknown type to serialize: {0}'.format(tp))
 * 
 *     cdef inline object _serialize_untyped_value(self, value, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (list, tuple, dict)) and weak_ref:
 *             # not iterable, and is weak ref
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_untyped_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":343
 *             raise TypeError('Unknown type to serialize: {0}'.format(type(value)))
 * 
 *     cdef inline object _serialize_value(self, value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if tp is None:
 *             return self._serialize_untyped_value(value, weak_ref=weak_ref)
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_value, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value *__pyx_optional_args) {
  PyObject *__pyx_v_tp = ((PyObject *)Py_None);
  int __pyx_v_weak_ref = ((int)0);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value __pyx_t_4;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value __pyx_t_5;
  __Pyx_RefNannySetupContext("_serialize_value", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_tp = __pyx_optional_args->tp;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_weak_ref = __pyx_optional_args->weak_ref;
      }
    }
  }

  /* "mars/serialize/jsonserializer.pyx":344
 * 
 *     cdef inline object _serialize_value(self, value, tp=None, bint weak_ref=False):
 *         if tp is None:             # <<<<<<<<<<<<<<
 *             return self._serialize_untyped_value(value, weak_ref=weak_ref)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_tp == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "mars/serialize/jsonserializer.pyx":345
 *     cdef inline object _serialize_value(self, value, tp=None, bint weak_ref=False):
 *         if tp is None:
 *             return self._serialize_untyped_value(value, weak_ref=weak_ref)             # <<<<<<<<<<<<<<
 *         else:
 *             return self._serialize_typed_value(value, tp, weak_ref=weak_ref)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4.__pyx_n = 1;
    __pyx_t_4.weak_ref = __pyx_v_weak_ref;
    __pyx_t_3 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value(__pyx_v_self, __pyx_v_value, &__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":344
 * 
 *     cdef inline object _serialize_value(self, value, tp=None, bint weak_ref=False):
 *         if tp is None:             # <<<<<<<<<<<<<<
 *             return self._serialize_untyped_value(value, weak_ref=weak_ref)
 *         else:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":347
 *             return self._serialize_untyped_value(value, weak_ref=weak_ref)
 *         else:
 *             return self._serialize_typed_value(value, tp, weak_ref=weak_ref)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _on_serial(self, Field field, x):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5.__pyx_n = 1;
    __pyx_t_5.weak_ref = __pyx_v_weak_ref;
    __pyx_t_3 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value(__pyx_v_self, __pyx_v_value, __pyx_v_tp, &__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "mars/serialize/jsonserializer.pyx":343
 *             raise TypeError('Unknown type to serialize: {0}'.format(type(value)))
 * 
 *     cdef inline object _serialize_value(self, value, tp=None, bint weak_ref=False):             # <<<<<<<<<<<<<<
 *         if tp is None:
 *             return self._serialize_untyped_value(value, weak_ref=weak_ref)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._serialize_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":349
 *             return self._serialize_typed_value(value, tp, weak_ref=weak_ref)
 * 
 *     cdef inline _on_serial(self, Field field, x):             # <<<<<<<<<<<<<<
 *         x = x if x is not None else field.default
 *         return field.on_serialize(x) if field.on_serialize is not None else x
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_serial(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_on_serial", 0);
  __Pyx_INCREF(__pyx_v_x);

  /* "mars/serialize/jsonserializer.pyx":350
 * 
 *     cdef inline _on_serial(self, Field field, x):
 *         x = x if x is not None else field.default             # <<<<<<<<<<<<<<
 *         return field.on_serialize(x) if field.on_serialize is not None else x
 * 
 */
  __pyx_t_2 = (__pyx_v_x != Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_INCREF(__pyx_v_x);
    __pyx_t_1 = __pyx_v_x;
  } else {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_field), __pyx_n_s_default); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __Pyx_DECREF_SET(__pyx_v_x, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":351
 *     cdef inline _on_serial(self, Field field, x):
 *         x = x if x is not None else field.default
 *         return field.on_serialize(x) if field.on_serialize is not None else x             # <<<<<<<<<<<<<<
 * 
 *     cpdef serialize_field(self, Field field, model_instance, obj):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = (__pyx_v_field->on_serialize != Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_INCREF(__pyx_v_field->on_serialize);
    __pyx_t_4 = __pyx_v_field->on_serialize; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_x) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_x);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_v_x);
    __pyx_t_1 = __pyx_v_x;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":349
 *             return self._serialize_typed_value(value, tp, weak_ref=weak_ref)
 * 
 *     cdef inline _on_serial(self, Field field, x):             # <<<<<<<<<<<<<<
 *         x = x if x is not None else field.default
 *         return field.on_serialize(x) if field.on_serialize is not None else x
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._on_serial", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":353
 *         return field.on_serialize(x) if field.on_serialize is not None else x
 * 
 *     cpdef serialize_field(self, Field field, model_instance, obj):             # <<<<<<<<<<<<<<
 *         cdef str tag
 *         cdef object new_obj
 */

static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_3serialize_field(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_serialize_field(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field, PyObject *__pyx_v_model_instance, PyObject *__pyx_v_obj, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_tag = 0;
  PyObject *__pyx_v_new_obj = 0;
  int __pyx_v_has_val;
  PyObject *__pyx_v_value = 0;
  struct __pyx_obj_4mars_9serialize_4core_OneOfField *__pyx_v_oneoffield = 0;
  PyObject *__pyx_v_field_val = NULL;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value __pyx_t_14;
  __Pyx_RefNannySetupContext("serialize_field", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T tp_dict_version = 0, obj_dict_version = 0;
    if (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict && tp_dict_version == __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) && (!Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset || obj_dict_version == __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))))));
    else {
      PY_UINT64_T type_dict_guard = (likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict)) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_serialize_field); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_3serialize_field)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_field), __pyx_v_model_instance, __pyx_v_obj};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 353, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_field), __pyx_v_model_instance, __pyx_v_obj};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 353, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 353, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_field));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_field));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_field));
          __Pyx_INCREF(__pyx_v_model_instance);
          __Pyx_GIVEREF(__pyx_v_model_instance);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_model_instance);
          __Pyx_INCREF(__pyx_v_obj);
          __Pyx_GIVEREF(__pyx_v_obj);
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_obj);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 353, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      tp_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) ? __PYX_GET_DICT_VERSION(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dict) : 0;
      obj_dict_version = likely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset) ? __PYX_GET_DICT_VERSION(_PyObject_GetDictPtr(((PyObject *)__pyx_v_self))) : 0;
      if (unlikely(type_dict_guard != tp_dict_version)) {
        tp_dict_version = obj_dict_version = 0;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "mars/serialize/jsonserializer.pyx":360
 *         cdef OneOfField oneoffield
 * 
 *         if isinstance(field, ReferenceField):             # <<<<<<<<<<<<<<
 *             tag = field.tag_name(self)
 *             new_obj = obj[tag] = dict()
 */
  __pyx_t_7 = __Pyx_TypeCheck(((PyObject *)__pyx_v_field), __pyx_ptype_4mars_9serialize_4core_ReferenceField); 
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "mars/serialize/jsonserializer.pyx":361
 * 
 *         if isinstance(field, ReferenceField):
 *             tag = field.tag_name(self)             # <<<<<<<<<<<<<<
 *             new_obj = obj[tag] = dict()
 *             if hasattr(model_instance, field.attr):
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_4mars_9serialize_4core_Field *)__pyx_v_field->__pyx_vtab)->tag_name(__pyx_v_field, ((struct __pyx_obj_4mars_9serialize_4core_Provider *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_tag = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":362
 *         if isinstance(field, ReferenceField):
 *             tag = field.tag_name(self)
 *             new_obj = obj[tag] = dict()             # <<<<<<<<<<<<<<
 *             if hasattr(model_instance, field.attr):
 *                 field_val = getattr(model_instance, field.attr)
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_new_obj = __pyx_t_1;
    if (unlikely(PyObject_SetItem(__pyx_v_obj, __pyx_v_tag, __pyx_t_1) < 0)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":363
 *             tag = field.tag_name(self)
 *             new_obj = obj[tag] = dict()
 *             if hasattr(model_instance, field.attr):             # <<<<<<<<<<<<<<
 *                 field_val = getattr(model_instance, field.attr)
 *                 if field.weak_ref:
 */
    __pyx_t_1 = __pyx_v_field->attr;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_HasAttr(__pyx_v_model_instance, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 363, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = (__pyx_t_8 != 0);
    if (__pyx_t_7) {

      /* "mars/serialize/jsonserializer.pyx":364
 *             new_obj = obj[tag] = dict()
 *             if hasattr(model_instance, field.attr):
 *                 field_val = getattr(model_instance, field.attr)             # <<<<<<<<<<<<<<
 *                 if field.weak_ref:
 *                     field_val = field_val()
 */
      __pyx_t_1 = __pyx_v_field->attr;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_GetAttr(__pyx_v_model_instance, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_field_val = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "mars/serialize/jsonserializer.pyx":365
 *             if hasattr(model_instance, field.attr):
 *                 field_val = getattr(model_instance, field.attr)
 *                 if field.weak_ref:             # <<<<<<<<<<<<<<
 *                     field_val = field_val()
 *                 value = self._on_serial(field, field_val)
 */
      __pyx_t_7 = (__pyx_v_field->weak_ref != 0);
      if (__pyx_t_7) {

        /* "mars/serialize/jsonserializer.pyx":366
 *                 field_val = getattr(model_instance, field.attr)
 *                 if field.weak_ref:
 *                     field_val = field_val()             # <<<<<<<<<<<<<<
 *                 value = self._on_serial(field, field_val)
 *                 value.serialize(self, new_obj)
 */
        __Pyx_INCREF(__pyx_v_field_val);
        __pyx_t_1 = __pyx_v_field_val; __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_field_val, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "mars/serialize/jsonserializer.pyx":365
 *             if hasattr(model_instance, field.attr):
 *                 field_val = getattr(model_instance, field.attr)
 *                 if field.weak_ref:             # <<<<<<<<<<<<<<
 *                     field_val = field_val()
 *                 value = self._on_serial(field, field_val)
 */
      }

      /* "mars/serialize/jsonserializer.pyx":367
 *                 if field.weak_ref:
 *                     field_val = field_val()
 *                 value = self._on_serial(field, field_val)             # <<<<<<<<<<<<<<
 *                 value.serialize(self, new_obj)
 *         elif isinstance(field, OneOfField):
 */
      __pyx_t_2 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_serial(__pyx_v_self, __pyx_v_field, __pyx_v_field_val); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 367, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_value = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "mars/serialize/jsonserializer.pyx":368
 *                     field_val = field_val()
 *                 value = self._on_serial(field, field_val)
 *                 value.serialize(self, new_obj)             # <<<<<<<<<<<<<<
 *         elif isinstance(field, OneOfField):
 *             has_val = False
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_serialize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_5 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_new_obj};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_new_obj};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(((PyObject *)__pyx_v_self));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
        PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_self));
        __Pyx_INCREF(__pyx_v_new_obj);
        __Pyx_GIVEREF(__pyx_v_new_obj);
        PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_new_obj);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "mars/serialize/jsonserializer.pyx":363
 *             tag = field.tag_name(self)
 *             new_obj = obj[tag] = dict()
 *             if hasattr(model_instance, field.attr):             # <<<<<<<<<<<<<<
 *                 field_val = getattr(model_instance, field.attr)
 *                 if field.weak_ref:
 */
    }

    /* "mars/serialize/jsonserializer.pyx":360
 *         cdef OneOfField oneoffield
 * 
 *         if isinstance(field, ReferenceField):             # <<<<<<<<<<<<<<
 *             tag = field.tag_name(self)
 *             new_obj = obj[tag] = dict()
 */
    goto __pyx_L3;
  }

  /* "mars/serialize/jsonserializer.pyx":369
 *                 value = self._on_serial(field, field_val)
 *                 value.serialize(self, new_obj)
 *         elif isinstance(field, OneOfField):             # <<<<<<<<<<<<<<
 *             has_val = False
 *             field_val = getattr(model_instance, field.attr, None)
 */
  __pyx_t_7 = __Pyx_TypeCheck(((PyObject *)__pyx_v_field), __pyx_ptype_4mars_9serialize_4core_OneOfField); 
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "mars/serialize/jsonserializer.pyx":370
 *                 value.serialize(self, new_obj)
 *         elif isinstance(field, OneOfField):
 *             has_val = False             # <<<<<<<<<<<<<<
 *             field_val = getattr(model_instance, field.attr, None)
 *             if field.weak_ref and field_val is not None:
 */
    __pyx_v_has_val = 0;

    /* "mars/serialize/jsonserializer.pyx":371
 *         elif isinstance(field, OneOfField):
 *             has_val = False
 *             field_val = getattr(model_instance, field.attr, None)             # <<<<<<<<<<<<<<
 *             if field.weak_ref and field_val is not None:
 *                 field_val = field_val()
 */
    __pyx_t_2 = __pyx_v_field->attr;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_model_instance, __pyx_t_2, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_field_val = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":372
 *             has_val = False
 *             field_val = getattr(model_instance, field.attr, None)
 *             if field.weak_ref and field_val is not None:             # <<<<<<<<<<<<<<
 *                 field_val = field_val()
 *             value = self._on_serial(field, field_val)
 */
    __pyx_t_7 = (__pyx_v_field->weak_ref != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_8 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_field_val != Py_None);
    __pyx_t_9 = (__pyx_t_7 != 0);
    __pyx_t_8 = __pyx_t_9;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_8) {

      /* "mars/serialize/jsonserializer.pyx":373
 *             field_val = getattr(model_instance, field.attr, None)
 *             if field.weak_ref and field_val is not None:
 *                 field_val = field_val()             # <<<<<<<<<<<<<<
 *             value = self._on_serial(field, field_val)
 *             oneoffield = <OneOfField>field
 */
      __Pyx_INCREF(__pyx_v_field_val);
      __pyx_t_2 = __pyx_v_field_val; __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 373, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_field_val, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "mars/serialize/jsonserializer.pyx":372
 *             has_val = False
 *             field_val = getattr(model_instance, field.attr, None)
 *             if field.weak_ref and field_val is not None:             # <<<<<<<<<<<<<<
 *                 field_val = field_val()
 *             value = self._on_serial(field, field_val)
 */
    }

    /* "mars/serialize/jsonserializer.pyx":374
 *             if field.weak_ref and field_val is not None:
 *                 field_val = field_val()
 *             value = self._on_serial(field, field_val)             # <<<<<<<<<<<<<<
 *             oneoffield = <OneOfField>field
 *             for f in oneoffield.fields:
 */
    __pyx_t_1 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_serial(__pyx_v_self, __pyx_v_field, __pyx_v_field_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_value = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":375
 *                 field_val = field_val()
 *             value = self._on_serial(field, field_val)
 *             oneoffield = <OneOfField>field             # <<<<<<<<<<<<<<
 *             for f in oneoffield.fields:
 *                 tag = f.tag_name(self)
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_field);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_oneoffield = ((struct __pyx_obj_4mars_9serialize_4core_OneOfField *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":376
 *             value = self._on_serial(field, field_val)
 *             oneoffield = <OneOfField>field
 *             for f in oneoffield.fields:             # <<<<<<<<<<<<<<
 *                 tag = f.tag_name(self)
 *                 if isinstance(value, f.type.model):
 */
    if (unlikely(__pyx_v_oneoffield->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 376, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_oneoffield->fields; __Pyx_INCREF(__pyx_t_1); __pyx_t_10 = 0;
    for (;;) {
      if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 376, __pyx_L1_error)
      #else
      __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 376, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_f, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "mars/serialize/jsonserializer.pyx":377
 *             oneoffield = <OneOfField>field
 *             for f in oneoffield.fields:
 *                 tag = f.tag_name(self)             # <<<<<<<<<<<<<<
 *                 if isinstance(value, f.type.model):
 *                     if not has_val:
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_tag_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 377, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, ((PyObject *)__pyx_v_self)) : __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_self));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 377, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 377, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_tag, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "mars/serialize/jsonserializer.pyx":378
 *             for f in oneoffield.fields:
 *                 tag = f.tag_name(self)
 *                 if isinstance(value, f.type.model):             # <<<<<<<<<<<<<<
 *                     if not has_val:
 *                         has_val = True
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 378, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_model); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 378, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = PyObject_IsInstance(__pyx_v_value, __pyx_t_6); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 378, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_9 = (__pyx_t_8 != 0);
      if (__pyx_t_9) {

        /* "mars/serialize/jsonserializer.pyx":379
 *                 tag = f.tag_name(self)
 *                 if isinstance(value, f.type.model):
 *                     if not has_val:             # <<<<<<<<<<<<<<
 *                         has_val = True
 *                     else:
 */
        __pyx_t_9 = ((!(__pyx_v_has_val != 0)) != 0);
        if (likely(__pyx_t_9)) {

          /* "mars/serialize/jsonserializer.pyx":380
 *                 if isinstance(value, f.type.model):
 *                     if not has_val:
 *                         has_val = True             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise ValueError('Only one of attributes({0}) can be specified'.format(field.attrs))
 */
          __pyx_v_has_val = 1;

          /* "mars/serialize/jsonserializer.pyx":379
 *                 tag = f.tag_name(self)
 *                 if isinstance(value, f.type.model):
 *                     if not has_val:             # <<<<<<<<<<<<<<
 *                         has_val = True
 *                     else:
 */
          goto __pyx_L12;
        }

        /* "mars/serialize/jsonserializer.pyx":382
 *                         has_val = True
 *                     else:
 *                         raise ValueError('Only one of attributes({0}) can be specified'.format(field.attrs))             # <<<<<<<<<<<<<<
 * 
 *                     new_obj = obj[tag] = dict()
 */
        /*else*/ {
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Only_one_of_attributes_0_can_be, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 382, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_field), __pyx_n_s_attrs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 382, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 382, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 382, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_Raise(__pyx_t_2, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __PYX_ERR(0, 382, __pyx_L1_error)
        }
        __pyx_L12:;

        /* "mars/serialize/jsonserializer.pyx":384
 *                         raise ValueError('Only one of attributes({0}) can be specified'.format(field.attrs))
 * 
 *                     new_obj = obj[tag] = dict()             # <<<<<<<<<<<<<<
 *                     value.serialize(self, new_obj)
 *                     return
 */
        __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 384, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_new_obj = __pyx_t_2;
        if (unlikely(PyObject_SetItem(__pyx_v_obj, __pyx_v_tag, __pyx_t_2) < 0)) __PYX_ERR(0, 384, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "mars/serialize/jsonserializer.pyx":385
 * 
 *                     new_obj = obj[tag] = dict()
 *                     value.serialize(self, new_obj)             # <<<<<<<<<<<<<<
 *                     return
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_serialize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 385, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_3 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_new_obj};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_new_obj};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 385, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_self));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
          PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, ((PyObject *)__pyx_v_self));
          __Pyx_INCREF(__pyx_v_new_obj);
          __Pyx_GIVEREF(__pyx_v_new_obj);
          PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_v_new_obj);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "mars/serialize/jsonserializer.pyx":386
 *                     new_obj = obj[tag] = dict()
 *                     value.serialize(self, new_obj)
 *                     return             # <<<<<<<<<<<<<<
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:
 *             tag = field.tag_name(self)
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;

        /* "mars/serialize/jsonserializer.pyx":378
 *             for f in oneoffield.fields:
 *                 tag = f.tag_name(self)
 *                 if isinstance(value, f.type.model):             # <<<<<<<<<<<<<<
 *                     if not has_val:
 *                         has_val = True
 */
      }

      /* "mars/serialize/jsonserializer.pyx":376
 *             value = self._on_serial(field, field_val)
 *             oneoffield = <OneOfField>field
 *             for f in oneoffield.fields:             # <<<<<<<<<<<<<<
 *                 tag = f.tag_name(self)
 *                 if isinstance(value, f.type.model):
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":369
 *                 value = self._on_serial(field, field_val)
 *                 value.serialize(self, new_obj)
 *         elif isinstance(field, OneOfField):             # <<<<<<<<<<<<<<
 *             has_val = False
 *             field_val = getattr(model_instance, field.attr, None)
 */
    goto __pyx_L3;
  }

  /* "mars/serialize/jsonserializer.pyx":387
 *                     value.serialize(self, new_obj)
 *                     return
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:             # <<<<<<<<<<<<<<
 *             tag = field.tag_name(self)
 *             value = self._on_serial(field, getattr(model_instance, field.attr, None))
 */
  __pyx_t_8 = __Pyx_TypeCheck(((PyObject *)__pyx_v_field), __pyx_ptype_4mars_9serialize_4core_ListField); 
  __pyx_t_7 = (__pyx_t_8 != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_9 = __pyx_t_7;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_field), __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_t_2)), ((PyObject *)__pyx_ptype_4mars_9serialize_4core_Reference), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = __pyx_t_7;
  __pyx_L13_bool_binop_done:;
  if (__pyx_t_9) {

    /* "mars/serialize/jsonserializer.pyx":388
 *                     return
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:
 *             tag = field.tag_name(self)             # <<<<<<<<<<<<<<
 *             value = self._on_serial(field, getattr(model_instance, field.attr, None))
 *             if value is None:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_4mars_9serialize_4core_Field *)__pyx_v_field->__pyx_vtab)->tag_name(__pyx_v_field, ((struct __pyx_obj_4mars_9serialize_4core_Provider *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_tag = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":389
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:
 *             tag = field.tag_name(self)
 *             value = self._on_serial(field, getattr(model_instance, field.attr, None))             # <<<<<<<<<<<<<<
 *             if value is None:
 *                 return
 */
    __pyx_t_1 = __pyx_v_field->attr;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_GetAttr3(__pyx_v_model_instance, __pyx_t_1, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_serial(__pyx_v_self, __pyx_v_field, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_value = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":390
 *             tag = field.tag_name(self)
 *             value = self._on_serial(field, getattr(model_instance, field.attr, None))
 *             if value is None:             # <<<<<<<<<<<<<<
 *                 return
 *             new_obj = obj[tag] = list()
 */
    __pyx_t_9 = (__pyx_v_value == Py_None);
    __pyx_t_7 = (__pyx_t_9 != 0);
    if (__pyx_t_7) {

      /* "mars/serialize/jsonserializer.pyx":391
 *             value = self._on_serial(field, getattr(model_instance, field.attr, None))
 *             if value is None:
 *                 return             # <<<<<<<<<<<<<<
 *             new_obj = obj[tag] = list()
 *             for val in value:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "mars/serialize/jsonserializer.pyx":390
 *             tag = field.tag_name(self)
 *             value = self._on_serial(field, getattr(model_instance, field.attr, None))
 *             if value is None:             # <<<<<<<<<<<<<<
 *                 return
 *             new_obj = obj[tag] = list()
 */
    }

    /* "mars/serialize/jsonserializer.pyx":392
 *             if value is None:
 *                 return
 *             new_obj = obj[tag] = list()             # <<<<<<<<<<<<<<
 *             for val in value:
 *                 if field.weak_ref:
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_new_obj = __pyx_t_1;
    if (unlikely(PyObject_SetItem(__pyx_v_obj, __pyx_v_tag, __pyx_t_1) < 0)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":393
 *                 return
 *             new_obj = obj[tag] = list()
 *             for val in value:             # <<<<<<<<<<<<<<
 *                 if field.weak_ref:
 *                     val = val()
 */
    if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
      __pyx_t_1 = __pyx_v_value; __Pyx_INCREF(__pyx_t_1); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 393, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 393, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 393, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_11(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 393, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "mars/serialize/jsonserializer.pyx":394
 *             new_obj = obj[tag] = list()
 *             for val in value:
 *                 if field.weak_ref:             # <<<<<<<<<<<<<<
 *                     val = val()
 *                 if val is not None:
 */
      __pyx_t_7 = (__pyx_v_field->weak_ref != 0);
      if (__pyx_t_7) {

        /* "mars/serialize/jsonserializer.pyx":395
 *             for val in value:
 *                 if field.weak_ref:
 *                     val = val()             # <<<<<<<<<<<<<<
 *                 if val is not None:
 *                     new_obj.append(val.serialize(self, dict()))
 */
        __Pyx_INCREF(__pyx_v_val);
        __pyx_t_6 = __pyx_v_val; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF_SET(__pyx_v_val, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "mars/serialize/jsonserializer.pyx":394
 *             new_obj = obj[tag] = list()
 *             for val in value:
 *                 if field.weak_ref:             # <<<<<<<<<<<<<<
 *                     val = val()
 *                 if val is not None:
 */
      }

      /* "mars/serialize/jsonserializer.pyx":396
 *                 if field.weak_ref:
 *                     val = val()
 *                 if val is not None:             # <<<<<<<<<<<<<<
 *                     new_obj.append(val.serialize(self, dict()))
 *                 else:
 */
      __pyx_t_7 = (__pyx_v_val != Py_None);
      __pyx_t_9 = (__pyx_t_7 != 0);
      if (__pyx_t_9) {

        /* "mars/serialize/jsonserializer.pyx":397
 *                     val = val()
 *                 if val is not None:
 *                     new_obj.append(val.serialize(self, dict()))             # <<<<<<<<<<<<<<
 *                 else:
 *                     new_obj.append(None)
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_serialize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 397, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 397, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_12 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 397, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_self));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
          PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_5, ((PyObject *)__pyx_v_self));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_5, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 397, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_new_obj, __pyx_t_2); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 397, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "mars/serialize/jsonserializer.pyx":396
 *                 if field.weak_ref:
 *                     val = val()
 *                 if val is not None:             # <<<<<<<<<<<<<<
 *                     new_obj.append(val.serialize(self, dict()))
 *                 else:
 */
        goto __pyx_L19;
      }

      /* "mars/serialize/jsonserializer.pyx":399
 *                     new_obj.append(val.serialize(self, dict()))
 *                 else:
 *                     new_obj.append(None)             # <<<<<<<<<<<<<<
 *         else:
 *             tag = field.tag_name(self)
 */
      /*else*/ {
        __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_new_obj, Py_None); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 399, __pyx_L1_error)
      }
      __pyx_L19:;

      /* "mars/serialize/jsonserializer.pyx":393
 *                 return
 *             new_obj = obj[tag] = list()
 *             for val in value:             # <<<<<<<<<<<<<<
 *                 if field.weak_ref:
 *                     val = val()
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":387
 *                     value.serialize(self, new_obj)
 *                     return
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:             # <<<<<<<<<<<<<<
 *             tag = field.tag_name(self)
 *             value = self._on_serial(field, getattr(model_instance, field.attr, None))
 */
    goto __pyx_L3;
  }

  /* "mars/serialize/jsonserializer.pyx":401
 *                     new_obj.append(None)
 *         else:
 *             tag = field.tag_name(self)             # <<<<<<<<<<<<<<
 *             val = self._on_serial(field, getattr(model_instance, field.attr, None))
 *             if val is None:
 */
  /*else*/ {
    __pyx_t_1 = ((struct __pyx_vtabstruct_4mars_9serialize_4core_Field *)__pyx_v_field->__pyx_vtab)->tag_name(__pyx_v_field, ((struct __pyx_obj_4mars_9serialize_4core_Provider *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_tag = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":402
 *         else:
 *             tag = field.tag_name(self)
 *             val = self._on_serial(field, getattr(model_instance, field.attr, None))             # <<<<<<<<<<<<<<
 *             if val is None:
 *                 return
 */
    __pyx_t_1 = __pyx_v_field->attr;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_GetAttr3(__pyx_v_model_instance, __pyx_t_1, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_serial(__pyx_v_self, __pyx_v_field, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_val = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":403
 *             tag = field.tag_name(self)
 *             val = self._on_serial(field, getattr(model_instance, field.attr, None))
 *             if val is None:             # <<<<<<<<<<<<<<
 *                 return
 *             obj[tag] = self._serialize_value(val, field.type, weak_ref=field.weak_ref)
 */
    __pyx_t_9 = (__pyx_v_val == Py_None);
    __pyx_t_7 = (__pyx_t_9 != 0);
    if (__pyx_t_7) {

      /* "mars/serialize/jsonserializer.pyx":404
 *             val = self._on_serial(field, getattr(model_instance, field.attr, None))
 *             if val is None:
 *                 return             # <<<<<<<<<<<<<<
 *             obj[tag] = self._serialize_value(val, field.type, weak_ref=field.weak_ref)
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "mars/serialize/jsonserializer.pyx":403
 *             tag = field.tag_name(self)
 *             val = self._on_serial(field, getattr(model_instance, field.attr, None))
 *             if val is None:             # <<<<<<<<<<<<<<
 *                 return
 *             obj[tag] = self._serialize_value(val, field.type, weak_ref=field.weak_ref)
 */
    }

    /* "mars/serialize/jsonserializer.pyx":405
 *             if val is None:
 *                 return
 *             obj[tag] = self._serialize_value(val, field.type, weak_ref=field.weak_ref)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _deserialize_value(self, obj, list callbacks, bint weak_ref):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_field), __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_14.__pyx_n = 2;
    __pyx_t_14.tp = __pyx_t_1;
    __pyx_t_14.weak_ref = __pyx_v_field->weak_ref;
    __pyx_t_2 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value(__pyx_v_self, __pyx_v_val, &__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_obj, __pyx_v_tag, __pyx_t_2) < 0)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "mars/serialize/jsonserializer.pyx":353
 *         return field.on_serialize(x) if field.on_serialize is not None else x
 * 
 *     cpdef serialize_field(self, Field field, model_instance, obj):             # <<<<<<<<<<<<<<
 *         cdef str tag
 *         cdef object new_obj
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider.serialize_field", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tag);
  __Pyx_XDECREF(__pyx_v_new_obj);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF((PyObject *)__pyx_v_oneoffield);
  __Pyx_XDECREF(__pyx_v_field_val);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_3serialize_field(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_3serialize_field(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field = 0;
  PyObject *__pyx_v_model_instance = 0;
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("serialize_field (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_field,&__pyx_n_s_model_instance,&__pyx_n_s_obj,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_field)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_model_instance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("serialize_field", 1, 3, 3, 1); __PYX_ERR(0, 353, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("serialize_field", 1, 3, 3, 2); __PYX_ERR(0, 353, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "serialize_field") < 0)) __PYX_ERR(0, 353, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_field = ((struct __pyx_obj_4mars_9serialize_4core_Field *)values[0]);
    __pyx_v_model_instance = values[1];
    __pyx_v_obj = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("serialize_field", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 353, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider.serialize_field", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_field), __pyx_ptype_4mars_9serialize_4core_Field, 1, "field", 0))) __PYX_ERR(0, 353, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_2serialize_field(((struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *)__pyx_v_self), __pyx_v_field, __pyx_v_model_instance, __pyx_v_obj);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_2serialize_field(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field, PyObject *__pyx_v_model_instance, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("serialize_field", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_serialize_field(__pyx_v_self, __pyx_v_field, __pyx_v_model_instance, __pyx_v_obj, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider.serialize_field", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":411
 *             return obj
 * 
 *         ref = lambda x: weakref.ref(x) if weak_ref else x             # <<<<<<<<<<<<<<
 * 
 *         if PY_MAJOR_VERSION >= 3:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_18_deserialize_value_lambda2(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static PyMethodDef __pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_18_deserialize_value_lambda2 = {"lambda2", (PyCFunction)__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_18_deserialize_value_lambda2, METH_O, 0};
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_18_deserialize_value_lambda2(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda2 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda2(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda2(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value *__pyx_cur_scope;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("lambda2", 0);
  __pyx_outer_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  if ((__pyx_cur_scope->__pyx_v_weak_ref != 0)) {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_weakref); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ref); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_x) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_x);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_v_x);
    __pyx_t_1 = __pyx_v_x;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_value.lambda2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":407
 *             obj[tag] = self._serialize_value(val, field.type, weak_ref=field.weak_ref)
 * 
 *     cdef inline _deserialize_value(self, obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, dict):
 *             return obj
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_value(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks, int __pyx_v_weak_ref) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value *__pyx_cur_scope;
  PyObject *__pyx_v_ref = NULL;
  PyObject *__pyx_v_tp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("_deserialize_value", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value *)__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value(__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 407, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_weak_ref = __pyx_v_weak_ref;

  /* "mars/serialize/jsonserializer.pyx":408
 * 
 *     cdef inline _deserialize_value(self, obj, list callbacks, bint weak_ref):
 *         if not isinstance(obj, dict):             # <<<<<<<<<<<<<<
 *             return obj
 * 
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_obj); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "mars/serialize/jsonserializer.pyx":409
 *     cdef inline _deserialize_value(self, obj, list callbacks, bint weak_ref):
 *         if not isinstance(obj, dict):
 *             return obj             # <<<<<<<<<<<<<<
 * 
 *         ref = lambda x: weakref.ref(x) if weak_ref else x
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_obj);
    __pyx_r = __pyx_v_obj;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":408
 * 
 *     cdef inline _deserialize_value(self, obj, list callbacks, bint weak_ref):
 *         if not isinstance(obj, dict):             # <<<<<<<<<<<<<<
 *             return obj
 * 
 */
  }

  /* "mars/serialize/jsonserializer.pyx":411
 *             return obj
 * 
 *         ref = lambda x: weakref.ref(x) if weak_ref else x             # <<<<<<<<<<<<<<
 * 
 *         if PY_MAJOR_VERSION >= 3:
 */
  __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_18_deserialize_value_lambda2, 0, __pyx_n_s_JsonSerializeProvider__deseriali_7, ((PyObject*)__pyx_cur_scope), __pyx_n_s_mars_serialize_jsonserializer, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_ref = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "mars/serialize/jsonserializer.pyx":413
 *         ref = lambda x: weakref.ref(x) if weak_ref else x
 * 
 *         if PY_MAJOR_VERSION >= 3:             # <<<<<<<<<<<<<<
 *             tp = _get_type(obj['type'])
 *         else:
 */
  __pyx_t_2 = ((PY_MAJOR_VERSION >= 3) != 0);
  if (__pyx_t_2) {

    /* "mars/serialize/jsonserializer.pyx":414
 * 
 *         if PY_MAJOR_VERSION >= 3:
 *             tp = _get_type(obj['type'])             # <<<<<<<<<<<<<<
 *         else:
 *             tp = _get_type(to_str(obj['type']))
 */
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyString_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 414, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_4mars_9serialize_14jsonserializer__get_type(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tp = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "mars/serialize/jsonserializer.pyx":413
 *         ref = lambda x: weakref.ref(x) if weak_ref else x
 * 
 *         if PY_MAJOR_VERSION >= 3:             # <<<<<<<<<<<<<<
 *             tp = _get_type(obj['type'])
 *         else:
 */
    goto __pyx_L4;
  }

  /* "mars/serialize/jsonserializer.pyx":416
 *             tp = _get_type(obj['type'])
 *         else:
 *             tp = _get_type(to_str(obj['type']))             # <<<<<<<<<<<<<<
 * 
 *         if tp is ValueType.bytes:
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_f_4mars_7utils_c_to_str(__pyx_t_4, 0, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_f_4mars_9serialize_14jsonserializer__get_type(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tp = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_L4:;

  /* "mars/serialize/jsonserializer.pyx":418
 *             tp = _get_type(to_str(obj['type']))
 * 
 *         if tp is ValueType.bytes:             # <<<<<<<<<<<<<<
 *             return ref(base64.b64decode(obj['value']))
 *         elif tp is ValueType.slice:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_bytes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = (__pyx_v_tp == __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":419
 * 
 *         if tp is ValueType.bytes:
 *             return ref(base64.b64decode(obj['value']))             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.slice:
 *             return ref(self._deserialize_slice(obj, callbacks))
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_base64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_b64decode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __pyx_lambda_funcdef_lambda2(__pyx_v_ref, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":418
 *             tp = _get_type(to_str(obj['type']))
 * 
 *         if tp is ValueType.bytes:             # <<<<<<<<<<<<<<
 *             return ref(base64.b64decode(obj['value']))
 *         elif tp is ValueType.slice:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":420
 *         if tp is ValueType.bytes:
 *             return ref(base64.b64decode(obj['value']))
 *         elif tp is ValueType.slice:             # <<<<<<<<<<<<<<
 *             return ref(self._deserialize_slice(obj, callbacks))
 *         elif tp is ValueType.arr:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_slice); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = (__pyx_v_tp == __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "mars/serialize/jsonserializer.pyx":421
 *             return ref(base64.b64decode(obj['value']))
 *         elif tp is ValueType.slice:
 *             return ref(self._deserialize_slice(obj, callbacks))             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.arr:
 *             return ref(self._deserialize_arr(obj, callbacks))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_slice(__pyx_v_self, __pyx_v_obj, __pyx_v_callbacks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __pyx_lambda_funcdef_lambda2(__pyx_v_ref, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":420
 *         if tp is ValueType.bytes:
 *             return ref(base64.b64decode(obj['value']))
 *         elif tp is ValueType.slice:             # <<<<<<<<<<<<<<
 *             return ref(self._deserialize_slice(obj, callbacks))
 *         elif tp is ValueType.arr:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":422
 *         elif tp is ValueType.slice:
 *             return ref(self._deserialize_slice(obj, callbacks))
 *         elif tp is ValueType.arr:             # <<<<<<<<<<<<<<
 *             return ref(self._deserialize_arr(obj, callbacks))
 *         elif tp is ValueType.dtype:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_arr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = (__pyx_v_tp == __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":423
 *             return ref(self._deserialize_slice(obj, callbacks))
 *         elif tp is ValueType.arr:
 *             return ref(self._deserialize_arr(obj, callbacks))             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.dtype:
 *             return ref(self._deserialize_dtype(obj, callbacks))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = ((PyObject *)__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_arr(__pyx_v_self, __pyx_v_obj, __pyx_v_callbacks)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __pyx_lambda_funcdef_lambda2(__pyx_v_ref, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":422
 *         elif tp is ValueType.slice:
 *             return ref(self._deserialize_slice(obj, callbacks))
 *         elif tp is ValueType.arr:             # <<<<<<<<<<<<<<
 *             return ref(self._deserialize_arr(obj, callbacks))
 *         elif tp is ValueType.dtype:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":424
 *         elif tp is ValueType.arr:
 *             return ref(self._deserialize_arr(obj, callbacks))
 *         elif tp is ValueType.dtype:             # <<<<<<<<<<<<<<
 *             return ref(self._deserialize_dtype(obj, callbacks))
 *         elif tp is ValueType.key:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = (__pyx_v_tp == __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "mars/serialize/jsonserializer.pyx":425
 *             return ref(self._deserialize_arr(obj, callbacks))
 *         elif tp is ValueType.dtype:
 *             return ref(self._deserialize_dtype(obj, callbacks))             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.key:
 *             return self._deserialize_key(obj, callbacks)  # the weakref will do in the callback, so skip
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = ((PyObject *)__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_dtype(__pyx_v_self, __pyx_v_obj, __pyx_v_callbacks)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __pyx_lambda_funcdef_lambda2(__pyx_v_ref, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":424
 *         elif tp is ValueType.arr:
 *             return ref(self._deserialize_arr(obj, callbacks))
 *         elif tp is ValueType.dtype:             # <<<<<<<<<<<<<<
 *             return ref(self._deserialize_dtype(obj, callbacks))
 *         elif tp is ValueType.key:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":426
 *         elif tp is ValueType.dtype:
 *             return ref(self._deserialize_dtype(obj, callbacks))
 *         elif tp is ValueType.key:             # <<<<<<<<<<<<<<
 *             return self._deserialize_key(obj, callbacks)  # the weakref will do in the callback, so skip
 *         elif tp is ValueType.datetime64:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_key); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = (__pyx_v_tp == __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":427
 *             return ref(self._deserialize_dtype(obj, callbacks))
 *         elif tp is ValueType.key:
 *             return self._deserialize_key(obj, callbacks)  # the weakref will do in the callback, so skip             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.datetime64:
 *             return ref(self._deserialize_datetime64(obj, callbacks))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = ((PyObject *)__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_key(__pyx_v_self, __pyx_v_obj, __pyx_v_callbacks)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":426
 *         elif tp is ValueType.dtype:
 *             return ref(self._deserialize_dtype(obj, callbacks))
 *         elif tp is ValueType.key:             # <<<<<<<<<<<<<<
 *             return self._deserialize_key(obj, callbacks)  # the weakref will do in the callback, so skip
 *         elif tp is ValueType.datetime64:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":428
 *         elif tp is ValueType.key:
 *             return self._deserialize_key(obj, callbacks)  # the weakref will do in the callback, so skip
 *         elif tp is ValueType.datetime64:             # <<<<<<<<<<<<<<
 *             return ref(self._deserialize_datetime64(obj, callbacks))
 *         elif tp is ValueType.timedelta64:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_datetime64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = (__pyx_v_tp == __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "mars/serialize/jsonserializer.pyx":429
 *             return self._deserialize_key(obj, callbacks)  # the weakref will do in the callback, so skip
 *         elif tp is ValueType.datetime64:
 *             return ref(self._deserialize_datetime64(obj, callbacks))             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.timedelta64:
 *             return ref(self._deserialize_timedelta64(obj, callbacks))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_datetime64(__pyx_v_self, __pyx_v_obj, __pyx_v_callbacks); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __pyx_lambda_funcdef_lambda2(__pyx_v_ref, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":428
 *         elif tp is ValueType.key:
 *             return self._deserialize_key(obj, callbacks)  # the weakref will do in the callback, so skip
 *         elif tp is ValueType.datetime64:             # <<<<<<<<<<<<<<
 *             return ref(self._deserialize_datetime64(obj, callbacks))
 *         elif tp is ValueType.timedelta64:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":430
 *         elif tp is ValueType.datetime64:
 *             return ref(self._deserialize_datetime64(obj, callbacks))
 *         elif tp is ValueType.timedelta64:             # <<<<<<<<<<<<<<
 *             return ref(self._deserialize_timedelta64(obj, callbacks))
 *         elif tp is ValueType.list:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_timedelta64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = (__pyx_v_tp == __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":431
 *             return ref(self._deserialize_datetime64(obj, callbacks))
 *         elif tp is ValueType.timedelta64:
 *             return ref(self._deserialize_timedelta64(obj, callbacks))             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.list:
 *             return self._deserialize_list(obj, callbacks, weak_ref)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_timedelta64(__pyx_v_self, __pyx_v_obj, __pyx_v_callbacks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __pyx_lambda_funcdef_lambda2(__pyx_v_ref, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":430
 *         elif tp is ValueType.datetime64:
 *             return ref(self._deserialize_datetime64(obj, callbacks))
 *         elif tp is ValueType.timedelta64:             # <<<<<<<<<<<<<<
 *             return ref(self._deserialize_timedelta64(obj, callbacks))
 *         elif tp is ValueType.list:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":432
 *         elif tp is ValueType.timedelta64:
 *             return ref(self._deserialize_timedelta64(obj, callbacks))
 *         elif tp is ValueType.list:             # <<<<<<<<<<<<<<
 *             return self._deserialize_list(obj, callbacks, weak_ref)
 *         elif tp is ValueType.tuple:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = (__pyx_v_tp == __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "mars/serialize/jsonserializer.pyx":433
 *             return ref(self._deserialize_timedelta64(obj, callbacks))
 *         elif tp is ValueType.list:
 *             return self._deserialize_list(obj, callbacks, weak_ref)             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.tuple:
 *             return self._deserialize_tuple(obj, callbacks, weak_ref)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_list(__pyx_v_self, __pyx_v_obj, __pyx_v_callbacks, __pyx_cur_scope->__pyx_v_weak_ref); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":432
 *         elif tp is ValueType.timedelta64:
 *             return ref(self._deserialize_timedelta64(obj, callbacks))
 *         elif tp is ValueType.list:             # <<<<<<<<<<<<<<
 *             return self._deserialize_list(obj, callbacks, weak_ref)
 *         elif tp is ValueType.tuple:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":434
 *         elif tp is ValueType.list:
 *             return self._deserialize_list(obj, callbacks, weak_ref)
 *         elif tp is ValueType.tuple:             # <<<<<<<<<<<<<<
 *             return self._deserialize_tuple(obj, callbacks, weak_ref)
 *         elif tp is ValueType.dict:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_tuple); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = (__pyx_v_tp == __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "mars/serialize/jsonserializer.pyx":435
 *             return self._deserialize_list(obj, callbacks, weak_ref)
 *         elif tp is ValueType.tuple:
 *             return self._deserialize_tuple(obj, callbacks, weak_ref)             # <<<<<<<<<<<<<<
 *         elif tp is ValueType.dict:
 *             return self._deserialize_dict(obj, callbacks, weak_ref)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_tuple(__pyx_v_self, __pyx_v_obj, __pyx_v_callbacks, __pyx_cur_scope->__pyx_v_weak_ref); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":434
 *         elif tp is ValueType.list:
 *             return self._deserialize_list(obj, callbacks, weak_ref)
 *         elif tp is ValueType.tuple:             # <<<<<<<<<<<<<<
 *             return self._deserialize_tuple(obj, callbacks, weak_ref)
 *         elif tp is ValueType.dict:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":436
 *         elif tp is ValueType.tuple:
 *             return self._deserialize_tuple(obj, callbacks, weak_ref)
 *         elif tp is ValueType.dict:             # <<<<<<<<<<<<<<
 *             return self._deserialize_dict(obj, callbacks, weak_ref)
 *         else:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_dict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = (__pyx_v_tp == __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (likely(__pyx_t_2)) {

    /* "mars/serialize/jsonserializer.pyx":437
 *             return self._deserialize_tuple(obj, callbacks, weak_ref)
 *         elif tp is ValueType.dict:
 *             return self._deserialize_dict(obj, callbacks, weak_ref)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Unknown type to deserialize {0}'.format(obj['type']))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_dict(__pyx_v_self, __pyx_v_obj, __pyx_v_callbacks, __pyx_cur_scope->__pyx_v_weak_ref); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/serialize/jsonserializer.pyx":436
 *         elif tp is ValueType.tuple:
 *             return self._deserialize_tuple(obj, callbacks, weak_ref)
 *         elif tp is ValueType.dict:             # <<<<<<<<<<<<<<
 *             return self._deserialize_dict(obj, callbacks, weak_ref)
 *         else:
 */
  }

  /* "mars/serialize/jsonserializer.pyx":439
 *             return self._deserialize_dict(obj, callbacks, weak_ref)
 *         else:
 *             raise TypeError('Unknown type to deserialize {0}'.format(obj['type']))             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _on_deserial(self, Field field, x):
 */
  /*else*/ {
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Unknown_type_to_deserialize_0, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 439, __pyx_L1_error)
  }

  /* "mars/serialize/jsonserializer.pyx":407
 *             obj[tag] = self._serialize_value(val, field.type, weak_ref=field.weak_ref)
 * 
 *     cdef inline _deserialize_value(self, obj, list callbacks, bint weak_ref):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, dict):
 *             return obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._deserialize_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ref);
  __Pyx_XDECREF(__pyx_v_tp);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":441
 *             raise TypeError('Unknown type to deserialize {0}'.format(obj['type']))
 * 
 *     cdef inline _on_deserial(self, Field field, x):             # <<<<<<<<<<<<<<
 *         x = x if x is not None else field.default
 *         return field.on_deserialize(x) if field.on_deserialize is not None else x
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_deserial(CYTHON_UNUSED struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_on_deserial", 0);
  __Pyx_INCREF(__pyx_v_x);

  /* "mars/serialize/jsonserializer.pyx":442
 * 
 *     cdef inline _on_deserial(self, Field field, x):
 *         x = x if x is not None else field.default             # <<<<<<<<<<<<<<
 *         return field.on_deserialize(x) if field.on_deserialize is not None else x
 * 
 */
  __pyx_t_2 = (__pyx_v_x != Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_INCREF(__pyx_v_x);
    __pyx_t_1 = __pyx_v_x;
  } else {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_field), __pyx_n_s_default); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __Pyx_DECREF_SET(__pyx_v_x, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":443
 *     cdef inline _on_deserial(self, Field field, x):
 *         x = x if x is not None else field.default
 *         return field.on_deserialize(x) if field.on_deserialize is not None else x             # <<<<<<<<<<<<<<
 * 
 *     def deserialize_field(self, Field field, model_instance, obj, list callbacks, dict key_to_instance):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = (__pyx_v_field->on_deserialize != Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_INCREF(__pyx_v_field->on_deserialize);
    __pyx_t_4 = __pyx_v_field->on_deserialize; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_x) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_x);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_v_x);
    __pyx_t_1 = __pyx_v_x;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/serialize/jsonserializer.pyx":441
 *             raise TypeError('Unknown type to deserialize {0}'.format(obj['type']))
 * 
 *     cdef inline _on_deserial(self, Field field, x):             # <<<<<<<<<<<<<<
 *         x = x if x is not None else field.default
 *         return field.on_deserialize(x) if field.on_deserialize is not None else x
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider._on_deserial", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":445
 *         return field.on_deserialize(x) if field.on_deserialize is not None else x
 * 
 *     def deserialize_field(self, Field field, model_instance, obj, list callbacks, dict key_to_instance):             # <<<<<<<<<<<<<<
 *         cdef str tag
 *         cdef object val
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_5deserialize_field(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_5deserialize_field(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field = 0;
  PyObject *__pyx_v_model_instance = 0;
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_callbacks = 0;
  PyObject *__pyx_v_key_to_instance = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("deserialize_field (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_field,&__pyx_n_s_model_instance,&__pyx_n_s_obj,&__pyx_n_s_callbacks,&__pyx_n_s_key_to_instance,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_field)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_model_instance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("deserialize_field", 1, 5, 5, 1); __PYX_ERR(0, 445, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("deserialize_field", 1, 5, 5, 2); __PYX_ERR(0, 445, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_callbacks)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("deserialize_field", 1, 5, 5, 3); __PYX_ERR(0, 445, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key_to_instance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("deserialize_field", 1, 5, 5, 4); __PYX_ERR(0, 445, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "deserialize_field") < 0)) __PYX_ERR(0, 445, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_field = ((struct __pyx_obj_4mars_9serialize_4core_Field *)values[0]);
    __pyx_v_model_instance = values[1];
    __pyx_v_obj = values[2];
    __pyx_v_callbacks = ((PyObject*)values[3]);
    __pyx_v_key_to_instance = ((PyObject*)values[4]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("deserialize_field", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 445, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider.deserialize_field", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_field), __pyx_ptype_4mars_9serialize_4core_Field, 1, "field", 0))) __PYX_ERR(0, 445, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_callbacks), (&PyList_Type), 1, "callbacks", 1))) __PYX_ERR(0, 445, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key_to_instance), (&PyDict_Type), 1, "key_to_instance", 1))) __PYX_ERR(0, 445, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_4deserialize_field(((struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *)__pyx_v_self), __pyx_v_field, __pyx_v_model_instance, __pyx_v_obj, __pyx_v_callbacks, __pyx_v_key_to_instance);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":489
 *             setattr(model_instance, field.attr, self._on_deserial(field, val))
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(subs):             # <<<<<<<<<<<<<<
 *                     o = subs[val.key, val.id]
 *                     if field.weak_ref:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17deserialize_field_1cb(PyObject *__pyx_self, PyObject *__pyx_v_subs); /*proto*/
static PyMethodDef __pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17deserialize_field_1cb = {"cb", (PyCFunction)__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17deserialize_field_1cb, METH_O, 0};
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17deserialize_field_1cb(PyObject *__pyx_self, PyObject *__pyx_v_subs) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cb (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17deserialize_field_cb(__pyx_self, ((PyObject *)__pyx_v_subs));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17deserialize_field_cb(PyObject *__pyx_self, PyObject *__pyx_v_subs) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *__pyx_cur_scope;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *__pyx_outer_scope;
  PyObject *__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("cb", 0);
  __pyx_outer_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "mars/serialize/jsonserializer.pyx":490
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(subs):
 *                     o = subs[val.key, val.id]             # <<<<<<<<<<<<<<
 *                     if field.weak_ref:
 *                         o = weakref.ref(o)
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_val)) { __Pyx_RaiseClosureNameError("val"); __PYX_ERR(0, 490, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_val, __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_cur_scope->__pyx_v_val)) { __Pyx_RaiseClosureNameError("val"); __PYX_ERR(0, 490, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_val, __pyx_n_s_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_subs, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_o = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":491
 *                 def cb(subs):
 *                     o = subs[val.key, val.id]
 *                     if field.weak_ref:             # <<<<<<<<<<<<<<
 *                         o = weakref.ref(o)
 *                     setattr(model_instance, field.attr, self._on_deserial(field, o))
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_field)) { __Pyx_RaiseClosureNameError("field"); __PYX_ERR(0, 491, __pyx_L1_error) }
  __pyx_t_4 = (__pyx_cur_scope->__pyx_v_field->weak_ref != 0);
  if (__pyx_t_4) {

    /* "mars/serialize/jsonserializer.pyx":492
 *                     o = subs[val.key, val.id]
 *                     if field.weak_ref:
 *                         o = weakref.ref(o)             # <<<<<<<<<<<<<<
 *                     setattr(model_instance, field.attr, self._on_deserial(field, o))
 *                 callbacks.append(cb)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_weakref); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ref); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_o) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_o);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_o, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "mars/serialize/jsonserializer.pyx":491
 *                 def cb(subs):
 *                     o = subs[val.key, val.id]
 *                     if field.weak_ref:             # <<<<<<<<<<<<<<
 *                         o = weakref.ref(o)
 *                     setattr(model_instance, field.attr, self._on_deserial(field, o))
 */
  }

  /* "mars/serialize/jsonserializer.pyx":493
 *                     if field.weak_ref:
 *                         o = weakref.ref(o)
 *                     setattr(model_instance, field.attr, self._on_deserial(field, o))             # <<<<<<<<<<<<<<
 *                 callbacks.append(cb)
 * 
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_model_instance)) { __Pyx_RaiseClosureNameError("model_instance"); __PYX_ERR(0, 493, __pyx_L1_error) }
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_model_instance;
  __Pyx_INCREF(__pyx_t_2);
  if (unlikely(!__pyx_cur_scope->__pyx_v_field)) { __Pyx_RaiseClosureNameError("field"); __PYX_ERR(0, 493, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_field->attr;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 493, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_v_field)) { __Pyx_RaiseClosureNameError("field"); __PYX_ERR(0, 493, __pyx_L1_error) }
  __pyx_t_3 = ((PyObject *)__pyx_cur_scope->__pyx_v_field);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_deserial(__pyx_cur_scope->__pyx_v_self, ((struct __pyx_obj_4mars_9serialize_4core_Field *)__pyx_t_3), __pyx_v_o); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = PyObject_SetAttr(__pyx_t_2, __pyx_t_1, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "mars/serialize/jsonserializer.pyx":489
 *             setattr(model_instance, field.attr, self._on_deserial(field, val))
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(subs):             # <<<<<<<<<<<<<<
 *                     o = subs[val.key, val.id]
 *                     if field.weak_ref:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider.deserialize_field.cb", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/serialize/jsonserializer.pyx":445
 *         return field.on_deserialize(x) if field.on_deserialize is not None else x
 * 
 *     def deserialize_field(self, Field field, model_instance, obj, list callbacks, dict key_to_instance):             # <<<<<<<<<<<<<<
 *         cdef str tag
 *         cdef object val
 */

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_4deserialize_field(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, struct __pyx_obj_4mars_9serialize_4core_Field *__pyx_v_field, PyObject *__pyx_v_model_instance, PyObject *__pyx_v_obj, PyObject *__pyx_v_callbacks, PyObject *__pyx_v_key_to_instance) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *__pyx_cur_scope;
  PyObject *__pyx_v_tag = 0;
  int __pyx_v_has_val;
  struct __pyx_obj_4mars_9serialize_4core_OneOfField *__pyx_v_oneoffield = 0;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_cb = 0;
  PyObject *__pyx_v_it_obj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  PyObject *__pyx_t_17 = NULL;
  __Pyx_RefNannySetupContext("deserialize_field", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *)__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field(__pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 445, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_field = __pyx_v_field;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_field);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_field);
  __pyx_cur_scope->__pyx_v_model_instance = __pyx_v_model_instance;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_model_instance);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_model_instance);

  /* "mars/serialize/jsonserializer.pyx":451
 *         cdef OneOfField oneoffield
 * 
 *         if isinstance(field, ReferenceField):             # <<<<<<<<<<<<<<
 *             tag = field.tag_name(self)
 *             val = obj.get(tag)
 */
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_field);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_4mars_9serialize_4core_ReferenceField); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "mars/serialize/jsonserializer.pyx":452
 * 
 *         if isinstance(field, ReferenceField):
 *             tag = field.tag_name(self)             # <<<<<<<<<<<<<<
 *             val = obj.get(tag)
 *             if val is None:
 */
    __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_self);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = ((struct __pyx_vtabstruct_4mars_9serialize_4core_Field *)__pyx_cur_scope->__pyx_v_field->__pyx_vtab)->tag_name(__pyx_cur_scope->__pyx_v_field, ((struct __pyx_obj_4mars_9serialize_4core_Provider *)__pyx_t_1), 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_tag = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "mars/serialize/jsonserializer.pyx":453
 *         if isinstance(field, ReferenceField):
 *             tag = field.tag_name(self)
 *             val = obj.get(tag)             # <<<<<<<<<<<<<<
 *             if val is None:
 *                 return
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_tag) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_tag);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_cur_scope->__pyx_v_val = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "mars/serialize/jsonserializer.pyx":454
 *             tag = field.tag_name(self)
 *             val = obj.get(tag)
 *             if val is None:             # <<<<<<<<<<<<<<
 *                 return
 *             setattr(model_instance, field.attr,
 */
    __pyx_t_3 = (__pyx_cur_scope->__pyx_v_val == Py_None);
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {

      /* "mars/serialize/jsonserializer.pyx":455
 *             val = obj.get(tag)
 *             if val is None:
 *                 return             # <<<<<<<<<<<<<<
 *             setattr(model_instance, field.attr,
 *                     self._on_deserial(field, field.type.model.deserialize(self, val, callbacks, key_to_instance)))
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "mars/serialize/jsonserializer.pyx":454
 *             tag = field.tag_name(self)
 *             val = obj.get(tag)
 *             if val is None:             # <<<<<<<<<<<<<<
 *                 return
 *             setattr(model_instance, field.attr,
 */
    }

    /* "mars/serialize/jsonserializer.pyx":456
 *             if val is None:
 *                 return
 *             setattr(model_instance, field.attr,             # <<<<<<<<<<<<<<
 *                     self._on_deserial(field, field.type.model.deserialize(self, val, callbacks, key_to_instance)))
 *         elif isinstance(field, OneOfField):
 */
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_model_instance;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_field->attr;
    __Pyx_INCREF(__pyx_t_1);

    /* "mars/serialize/jsonserializer.pyx":457
 *                 return
 *             setattr(model_instance, field.attr,
 *                     self._on_deserial(field, field.type.model.deserialize(self, val, callbacks, key_to_instance)))             # <<<<<<<<<<<<<<
 *         elif isinstance(field, OneOfField):
 *             oneoffield = <OneOfField>field
 */
    __pyx_t_5 = ((PyObject *)__pyx_cur_scope->__pyx_v_field);
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_field), __pyx_n_s_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_model); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_deserialize); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, ((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_val, __pyx_v_callbacks, __pyx_v_key_to_instance};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, ((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_val, __pyx_v_callbacks, __pyx_v_key_to_instance};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_cur_scope->__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, ((PyObject *)__pyx_cur_scope->__pyx_v_self));
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_val);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_val);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_cur_scope->__pyx_v_val);
      __Pyx_INCREF(__pyx_v_callbacks);
      __Pyx_GIVEREF(__pyx_v_callbacks);
      PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_v_callbacks);
      __Pyx_INCREF(__pyx_v_key_to_instance);
      __Pyx_GIVEREF(__pyx_v_key_to_instance);
      PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, __pyx_v_key_to_instance);
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_deserial(__pyx_cur_scope->__pyx_v_self, ((struct __pyx_obj_4mars_9serialize_4core_Field *)__pyx_t_5), __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "mars/serialize/jsonserializer.pyx":456
 *             if val is None:
 *                 return
 *             setattr(model_instance, field.attr,             # <<<<<<<<<<<<<<
 *                     self._on_deserial(field, field.type.model.deserialize(self, val, callbacks, key_to_instance)))
 *         elif isinstance(field, OneOfField):
 */
    __pyx_t_11 = PyObject_SetAttr(__pyx_t_4, __pyx_t_1, __pyx_t_7); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "mars/serialize/jsonserializer.pyx":451
 *         cdef OneOfField oneoffield
 * 
 *         if isinstance(field, ReferenceField):             # <<<<<<<<<<<<<<
 *             tag = field.tag_name(self)
 *             val = obj.get(tag)
 */
    goto __pyx_L3;
  }

  /* "mars/serialize/jsonserializer.pyx":458
 *             setattr(model_instance, field.attr,
 *                     self._on_deserial(field, field.type.model.deserialize(self, val, callbacks, key_to_instance)))
 *         elif isinstance(field, OneOfField):             # <<<<<<<<<<<<<<
 *             oneoffield = <OneOfField>field
 *             has_val = False
 */
  __pyx_t_7 = ((PyObject *)__pyx_cur_scope->__pyx_v_field);
  __Pyx_INCREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_7, __pyx_ptype_4mars_9serialize_4core_OneOfField); 
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "mars/serialize/jsonserializer.pyx":459
 *                     self._on_deserial(field, field.type.model.deserialize(self, val, callbacks, key_to_instance)))
 *         elif isinstance(field, OneOfField):
 *             oneoffield = <OneOfField>field             # <<<<<<<<<<<<<<
 *             has_val = False
 *             for f in oneoffield.fields:
 */
    __pyx_t_7 = ((PyObject *)__pyx_cur_scope->__pyx_v_field);
    __Pyx_INCREF(__pyx_t_7);
    __pyx_v_oneoffield = ((struct __pyx_obj_4mars_9serialize_4core_OneOfField *)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "mars/serialize/jsonserializer.pyx":460
 *         elif isinstance(field, OneOfField):
 *             oneoffield = <OneOfField>field
 *             has_val = False             # <<<<<<<<<<<<<<
 *             for f in oneoffield.fields:
 *                 tag = f.tag_name(self)
 */
    __pyx_v_has_val = 0;

    /* "mars/serialize/jsonserializer.pyx":461
 *             oneoffield = <OneOfField>field
 *             has_val = False
 *             for f in oneoffield.fields:             # <<<<<<<<<<<<<<
 *                 tag = f.tag_name(self)
 *                 val = obj.get(tag)
 */
    if (unlikely(__pyx_v_oneoffield->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 461, __pyx_L1_error)
    }
    __pyx_t_7 = __pyx_v_oneoffield->fields; __Pyx_INCREF(__pyx_t_7); __pyx_t_12 = 0;
    for (;;) {
      if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_7)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 461, __pyx_L1_error)
      #else
      __pyx_t_1 = PySequence_ITEM(__pyx_t_7, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_f, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "mars/serialize/jsonserializer.pyx":462
 *             has_val = False
 *             for f in oneoffield.fields:
 *                 tag = f.tag_name(self)             # <<<<<<<<<<<<<<
 *                 val = obj.get(tag)
 *                 if val:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_tag_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, ((PyObject *)__pyx_cur_scope->__pyx_v_self)) : __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_cur_scope->__pyx_v_self));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 462, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_tag, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "mars/serialize/jsonserializer.pyx":463
 *             for f in oneoffield.fields:
 *                 tag = f.tag_name(self)
 *                 val = obj.get(tag)             # <<<<<<<<<<<<<<
 *                 if val:
 *                     if not has_val:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 463, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_tag) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_tag);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_val);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_val, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_t_1 = 0;

      /* "mars/serialize/jsonserializer.pyx":464
 *                 tag = f.tag_name(self)
 *                 val = obj.get(tag)
 *                 if val:             # <<<<<<<<<<<<<<
 *                     if not has_val:
 *                         has_val = True
 */
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_val); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 464, __pyx_L1_error)
      if (__pyx_t_3) {

        /* "mars/serialize/jsonserializer.pyx":465
 *                 val = obj.get(tag)
 *                 if val:
 *                     if not has_val:             # <<<<<<<<<<<<<<
 *                         has_val = True
 *                     else:
 */
        __pyx_t_3 = ((!(__pyx_v_has_val != 0)) != 0);
        if (likely(__pyx_t_3)) {

          /* "mars/serialize/jsonserializer.pyx":466
 *                 if val:
 *                     if not has_val:
 *                         has_val = True             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise ValueError(
 */
          __pyx_v_has_val = 1;

          /* "mars/serialize/jsonserializer.pyx":465
 *                 val = obj.get(tag)
 *                 if val:
 *                     if not has_val:             # <<<<<<<<<<<<<<
 *                         has_val = True
 *                     else:
 */
          goto __pyx_L8;
        }

        /* "mars/serialize/jsonserializer.pyx":468
 *                         has_val = True
 *                     else:
 *                         raise ValueError(             # <<<<<<<<<<<<<<
 *                             'Only one of attributes({0}) can be specified'.format(field.attrs))
 * 
 */
        /*else*/ {

          /* "mars/serialize/jsonserializer.pyx":469
 *                     else:
 *                         raise ValueError(
 *                             'Only one of attributes({0}) can be specified'.format(field.attrs))             # <<<<<<<<<<<<<<
 * 
 *                     setattr(model_instance, f.attr,
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Only_one_of_attributes_0_can_be, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 469, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_field), __pyx_n_s_attrs); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 469, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "mars/serialize/jsonserializer.pyx":468
 *                         has_val = True
 *                     else:
 *                         raise ValueError(             # <<<<<<<<<<<<<<
 *                             'Only one of attributes({0}) can be specified'.format(field.attrs))
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 468, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_Raise(__pyx_t_4, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __PYX_ERR(0, 468, __pyx_L1_error)
        }
        __pyx_L8:;

        /* "mars/serialize/jsonserializer.pyx":471
 *                             'Only one of attributes({0}) can be specified'.format(field.attrs))
 * 
 *                     setattr(model_instance, f.attr,             # <<<<<<<<<<<<<<
 *                             self._on_deserial(field, f.type.model.deserialize(self, obj[tag], callbacks, key_to_instance)))
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:
 */
        __pyx_t_4 = __pyx_cur_scope->__pyx_v_model_instance;
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_attr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "mars/serialize/jsonserializer.pyx":472
 * 
 *                     setattr(model_instance, f.attr,
 *                             self._on_deserial(field, f.type.model.deserialize(self, obj[tag], callbacks, key_to_instance)))             # <<<<<<<<<<<<<<
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:
 *             tag = field.tag_name(self)
 */
        __pyx_t_6 = ((PyObject *)__pyx_cur_scope->__pyx_v_field);
        __Pyx_INCREF(__pyx_t_6);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_type); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_model); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_deserialize); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_v_tag); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_13 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[5] = {__pyx_t_13, ((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_t_8, __pyx_v_callbacks, __pyx_v_key_to_instance};
          __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 472, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[5] = {__pyx_t_13, ((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_t_8, __pyx_v_callbacks, __pyx_v_key_to_instance};
          __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 472, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        } else
        #endif
        {
          __pyx_t_14 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 472, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (__pyx_t_13) {
            __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13); __pyx_t_13 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_self));
          __Pyx_GIVEREF(((PyObject *)__pyx_cur_scope->__pyx_v_self));
          PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_9, ((PyObject *)__pyx_cur_scope->__pyx_v_self));
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_9, __pyx_t_8);
          __Pyx_INCREF(__pyx_v_callbacks);
          __Pyx_GIVEREF(__pyx_v_callbacks);
          PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_9, __pyx_v_callbacks);
          __Pyx_INCREF(__pyx_v_key_to_instance);
          __Pyx_GIVEREF(__pyx_v_key_to_instance);
          PyTuple_SET_ITEM(__pyx_t_14, 3+__pyx_t_9, __pyx_v_key_to_instance);
          __pyx_t_8 = 0;
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_14, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 472, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_deserial(__pyx_cur_scope->__pyx_v_self, ((struct __pyx_obj_4mars_9serialize_4core_Field *)__pyx_t_6), __pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "mars/serialize/jsonserializer.pyx":471
 *                             'Only one of attributes({0}) can be specified'.format(field.attrs))
 * 
 *                     setattr(model_instance, f.attr,             # <<<<<<<<<<<<<<
 *                             self._on_deserial(field, f.type.model.deserialize(self, obj[tag], callbacks, key_to_instance)))
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:
 */
        __pyx_t_11 = PyObject_SetAttr(__pyx_t_4, __pyx_t_1, __pyx_t_10); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 471, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "mars/serialize/jsonserializer.pyx":464
 *                 tag = f.tag_name(self)
 *                 val = obj.get(tag)
 *                 if val:             # <<<<<<<<<<<<<<
 *                     if not has_val:
 *                         has_val = True
 */
      }

      /* "mars/serialize/jsonserializer.pyx":461
 *             oneoffield = <OneOfField>field
 *             has_val = False
 *             for f in oneoffield.fields:             # <<<<<<<<<<<<<<
 *                 tag = f.tag_name(self)
 *                 val = obj.get(tag)
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "mars/serialize/jsonserializer.pyx":458
 *             setattr(model_instance, field.attr,
 *                     self._on_deserial(field, field.type.model.deserialize(self, val, callbacks, key_to_instance)))
 *         elif isinstance(field, OneOfField):             # <<<<<<<<<<<<<<
 *             oneoffield = <OneOfField>field
 *             has_val = False
 */
    goto __pyx_L3;
  }

  /* "mars/serialize/jsonserializer.pyx":473
 *                     setattr(model_instance, f.attr,
 *                             self._on_deserial(field, f.type.model.deserialize(self, obj[tag], callbacks, key_to_instance)))
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:             # <<<<<<<<<<<<<<
 *             tag = field.tag_name(self)
 *             if tag not in obj:
 */
  __pyx_t_7 = ((PyObject *)__pyx_cur_scope->__pyx_v_field);
  __Pyx_INCREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_7, __pyx_ptype_4mars_9serialize_4core_ListField); 
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_15 = (__pyx_t_2 != 0);
  if (__pyx_t_15) {
  } else {
    __pyx_t_3 = __pyx_t_15;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_field), __pyx_n_s_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_type); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_t_10)), ((PyObject *)__pyx_ptype_4mars_9serialize_4core_Reference), Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_3 = __pyx_t_15;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_3) {

    /* "mars/serialize/jsonserializer.pyx":474
 *                             self._on_deserial(field, f.type.model.deserialize(self, obj[tag], callbacks, key_to_instance)))
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:
 *             tag = field.tag_name(self)             # <<<<<<<<<<<<<<
 *             if tag not in obj:
 *                 return
 */
    __pyx_t_7 = ((PyObject *)__pyx_cur_scope->__pyx_v_self);
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_10 = ((struct __pyx_vtabstruct_4mars_9serialize_4core_Field *)__pyx_cur_scope->__pyx_v_field->__pyx_vtab)->tag_name(__pyx_cur_scope->__pyx_v_field, ((struct __pyx_obj_4mars_9serialize_4core_Provider *)__pyx_t_7), 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_tag = ((PyObject*)__pyx_t_10);
    __pyx_t_10 = 0;

    /* "mars/serialize/jsonserializer.pyx":475
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:
 *             tag = field.tag_name(self)
 *             if tag not in obj:             # <<<<<<<<<<<<<<
 *                 return
 *             setattr(model_instance, field.attr,
 */
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_tag, __pyx_v_obj, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 475, __pyx_L1_error)
    __pyx_t_15 = (__pyx_t_3 != 0);
    if (__pyx_t_15) {

      /* "mars/serialize/jsonserializer.pyx":476
 *             tag = field.tag_name(self)
 *             if tag not in obj:
 *                 return             # <<<<<<<<<<<<<<
 *             setattr(model_instance, field.attr,
 *                     self._on_deserial(
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "mars/serialize/jsonserializer.pyx":475
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:
 *             tag = field.tag_name(self)
 *             if tag not in obj:             # <<<<<<<<<<<<<<
 *                 return
 *             setattr(model_instance, field.attr,
 */
    }

    /* "mars/serialize/jsonserializer.pyx":477
 *             if tag not in obj:
 *                 return
 *             setattr(model_instance, field.attr,             # <<<<<<<<<<<<<<
 *                     self._on_deserial(
 *                         field, [field.type.type.model.deserialize(self, it_obj, callbacks, key_to_instance)
 */
    __pyx_t_10 = __pyx_cur_scope->__pyx_v_model_instance;
    __Pyx_INCREF(__pyx_t_10);
    __pyx_t_7 = __pyx_cur_scope->__pyx_v_field->attr;
    __Pyx_INCREF(__pyx_t_7);

    /* "mars/serialize/jsonserializer.pyx":479
 *             setattr(model_instance, field.attr,
 *                     self._on_deserial(
 *                         field, [field.type.type.model.deserialize(self, it_obj, callbacks, key_to_instance)             # <<<<<<<<<<<<<<
 *                                 if it_obj is not None else None
 *                                 for it_obj in obj[tag]]))
 */
    __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_field);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "mars/serialize/jsonserializer.pyx":481
 *                         field, [field.type.type.model.deserialize(self, it_obj, callbacks, key_to_instance)
 *                                 if it_obj is not None else None
 *                                 for it_obj in obj[tag]]))             # <<<<<<<<<<<<<<
 *         else:
 *             tag = field.tag_name(self)
 */
    __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_v_tag); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_6 = __pyx_t_5; __Pyx_INCREF(__pyx_t_6); __pyx_t_12 = 0;
      __pyx_t_16 = NULL;
    } else {
      __pyx_t_12 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_16 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 481, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_16)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_12); __Pyx_INCREF(__pyx_t_5); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 481, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_6, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_12); __Pyx_INCREF(__pyx_t_5); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 481, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_6, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_16(__pyx_t_6);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 481, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_v_it_obj, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "mars/serialize/jsonserializer.pyx":480
 *                     self._on_deserial(
 *                         field, [field.type.type.model.deserialize(self, it_obj, callbacks, key_to_instance)
 *                                 if it_obj is not None else None             # <<<<<<<<<<<<<<
 *                                 for it_obj in obj[tag]]))
 *         else:
 */
      __pyx_t_15 = (__pyx_v_it_obj != Py_None);
      if ((__pyx_t_15 != 0)) {

        /* "mars/serialize/jsonserializer.pyx":479
 *             setattr(model_instance, field.attr,
 *                     self._on_deserial(
 *                         field, [field.type.type.model.deserialize(self, it_obj, callbacks, key_to_instance)             # <<<<<<<<<<<<<<
 *                                 if it_obj is not None else None
 *                                 for it_obj in obj[tag]]))
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_field), __pyx_n_s_type); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 479, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_type); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 479, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_model); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 479, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_deserialize); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 479, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[5] = {__pyx_t_8, ((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_v_it_obj, __pyx_v_callbacks, __pyx_v_key_to_instance};
          __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 479, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_14);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
          PyObject *__pyx_temp[5] = {__pyx_t_8, ((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_v_it_obj, __pyx_v_callbacks, __pyx_v_key_to_instance};
          __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 479, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_14);
        } else
        #endif
        {
          __pyx_t_17 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 479, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_self));
          __Pyx_GIVEREF(((PyObject *)__pyx_cur_scope->__pyx_v_self));
          PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_9, ((PyObject *)__pyx_cur_scope->__pyx_v_self));
          __Pyx_INCREF(__pyx_v_it_obj);
          __Pyx_GIVEREF(__pyx_v_it_obj);
          PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_9, __pyx_v_it_obj);
          __Pyx_INCREF(__pyx_v_callbacks);
          __Pyx_GIVEREF(__pyx_v_callbacks);
          PyTuple_SET_ITEM(__pyx_t_17, 2+__pyx_t_9, __pyx_v_callbacks);
          __Pyx_INCREF(__pyx_v_key_to_instance);
          __Pyx_GIVEREF(__pyx_v_key_to_instance);
          PyTuple_SET_ITEM(__pyx_t_17, 3+__pyx_t_9, __pyx_v_key_to_instance);
          __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_17, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 479, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_5 = __pyx_t_14;
        __pyx_t_14 = 0;
      } else {

        /* "mars/serialize/jsonserializer.pyx":480
 *                     self._on_deserial(
 *                         field, [field.type.type.model.deserialize(self, it_obj, callbacks, key_to_instance)
 *                                 if it_obj is not None else None             # <<<<<<<<<<<<<<
 *                                 for it_obj in obj[tag]]))
 *         else:
 */
        __Pyx_INCREF(Py_None);
        __pyx_t_5 = Py_None;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 479, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "mars/serialize/jsonserializer.pyx":481
 *                         field, [field.type.type.model.deserialize(self, it_obj, callbacks, key_to_instance)
 *                                 if it_obj is not None else None
 *                                 for it_obj in obj[tag]]))             # <<<<<<<<<<<<<<
 *         else:
 *             tag = field.tag_name(self)
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "mars/serialize/jsonserializer.pyx":478
 *                 return
 *             setattr(model_instance, field.attr,
 *                     self._on_deserial(             # <<<<<<<<<<<<<<
 *                         field, [field.type.type.model.deserialize(self, it_obj, callbacks, key_to_instance)
 *                                 if it_obj is not None else None
 */
    __pyx_t_6 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_deserial(__pyx_cur_scope->__pyx_v_self, ((struct __pyx_obj_4mars_9serialize_4core_Field *)__pyx_t_1), __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 478, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "mars/serialize/jsonserializer.pyx":477
 *             if tag not in obj:
 *                 return
 *             setattr(model_instance, field.attr,             # <<<<<<<<<<<<<<
 *                     self._on_deserial(
 *                         field, [field.type.type.model.deserialize(self, it_obj, callbacks, key_to_instance)
 */
    __pyx_t_11 = PyObject_SetAttr(__pyx_t_10, __pyx_t_7, __pyx_t_6); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "mars/serialize/jsonserializer.pyx":473
 *                     setattr(model_instance, f.attr,
 *                             self._on_deserial(field, f.type.model.deserialize(self, obj[tag], callbacks, key_to_instance)))
 *         elif isinstance(field, ListField) and type(field.type.type) == Reference:             # <<<<<<<<<<<<<<
 *             tag = field.tag_name(self)
 *             if tag not in obj:
 */
    goto __pyx_L3;
  }

  /* "mars/serialize/jsonserializer.pyx":483
 *                                 for it_obj in obj[tag]]))
 *         else:
 *             tag = field.tag_name(self)             # <<<<<<<<<<<<<<
 *             val = self._deserialize_value(obj.get(tag), callbacks, field.weak_ref)
 *             if val is None:
 */
  /*else*/ {
    __pyx_t_6 = ((PyObject *)__pyx_cur_scope->__pyx_v_self);
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_7 = ((struct __pyx_vtabstruct_4mars_9serialize_4core_Field *)__pyx_cur_scope->__pyx_v_field->__pyx_vtab)->tag_name(__pyx_cur_scope->__pyx_v_field, ((struct __pyx_obj_4mars_9serialize_4core_Provider *)__pyx_t_6), 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_tag = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "mars/serialize/jsonserializer.pyx":484
 *         else:
 *             tag = field.tag_name(self)
 *             val = self._deserialize_value(obj.get(tag), callbacks, field.weak_ref)             # <<<<<<<<<<<<<<
 *             if val is None:
 *                 return
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_get); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_7 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_10, __pyx_v_tag) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_tag);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_value(__pyx_cur_scope->__pyx_v_self, __pyx_t_7, __pyx_v_callbacks, __pyx_cur_scope->__pyx_v_field->weak_ref); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_cur_scope->__pyx_v_val = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "mars/serialize/jsonserializer.pyx":485
 *             tag = field.tag_name(self)
 *             val = self._deserialize_value(obj.get(tag), callbacks, field.weak_ref)
 *             if val is None:             # <<<<<<<<<<<<<<
 *                 return
 *             setattr(model_instance, field.attr, self._on_deserial(field, val))
 */
    __pyx_t_15 = (__pyx_cur_scope->__pyx_v_val == Py_None);
    __pyx_t_3 = (__pyx_t_15 != 0);
    if (__pyx_t_3) {

      /* "mars/serialize/jsonserializer.pyx":486
 *             val = self._deserialize_value(obj.get(tag), callbacks, field.weak_ref)
 *             if val is None:
 *                 return             # <<<<<<<<<<<<<<
 *             setattr(model_instance, field.attr, self._on_deserial(field, val))
 *             if isinstance(val, KeyPlaceholder):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "mars/serialize/jsonserializer.pyx":485
 *             tag = field.tag_name(self)
 *             val = self._deserialize_value(obj.get(tag), callbacks, field.weak_ref)
 *             if val is None:             # <<<<<<<<<<<<<<
 *                 return
 *             setattr(model_instance, field.attr, self._on_deserial(field, val))
 */
    }

    /* "mars/serialize/jsonserializer.pyx":487
 *             if val is None:
 *                 return
 *             setattr(model_instance, field.attr, self._on_deserial(field, val))             # <<<<<<<<<<<<<<
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(subs):
 */
    __pyx_t_6 = __pyx_cur_scope->__pyx_v_model_instance;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_7 = __pyx_cur_scope->__pyx_v_field->attr;
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_10 = ((PyObject *)__pyx_cur_scope->__pyx_v_field);
    __Pyx_INCREF(__pyx_t_10);
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_val;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = __pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_deserial(__pyx_cur_scope->__pyx_v_self, ((struct __pyx_obj_4mars_9serialize_4core_Field *)__pyx_t_10), __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_11 = PyObject_SetAttr(__pyx_t_6, __pyx_t_7, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mars/serialize/jsonserializer.pyx":488
 *                 return
 *             setattr(model_instance, field.attr, self._on_deserial(field, val))
 *             if isinstance(val, KeyPlaceholder):             # <<<<<<<<<<<<<<
 *                 def cb(subs):
 *                     o = subs[val.key, val.id]
 */
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_val;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_4mars_9serialize_4core_KeyPlaceholder); 
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_15 = (__pyx_t_3 != 0);
    if (__pyx_t_15) {

      /* "mars/serialize/jsonserializer.pyx":489
 *             setattr(model_instance, field.attr, self._on_deserial(field, val))
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(subs):             # <<<<<<<<<<<<<<
 *                     o = subs[val.key, val.id]
 *                     if field.weak_ref:
 */
      __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_17deserialize_field_1cb, 0, __pyx_n_s_deserialize_field_locals_cb, ((PyObject*)__pyx_cur_scope), __pyx_n_s_mars_serialize_jsonserializer, __pyx_d, ((PyObject *)__pyx_codeobj__10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_cb = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "mars/serialize/jsonserializer.pyx":494
 *                         o = weakref.ref(o)
 *                     setattr(model_instance, field.attr, self._on_deserial(field, o))
 *                 callbacks.append(cb)             # <<<<<<<<<<<<<<
 * 
 */
      if (unlikely(__pyx_v_callbacks == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 494, __pyx_L1_error)
      }
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_callbacks, __pyx_v_cb); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 494, __pyx_L1_error)

      /* "mars/serialize/jsonserializer.pyx":488
 *                 return
 *             setattr(model_instance, field.attr, self._on_deserial(field, val))
 *             if isinstance(val, KeyPlaceholder):             # <<<<<<<<<<<<<<
 *                 def cb(subs):
 *                     o = subs[val.key, val.id]
 */
    }
  }
  __pyx_L3:;

  /* "mars/serialize/jsonserializer.pyx":445
 *         return field.on_deserialize(x) if field.on_deserialize is not None else x
 * 
 *     def deserialize_field(self, Field field, model_instance, obj, list callbacks, dict key_to_instance):             # <<<<<<<<<<<<<<
 *         cdef str tag
 *         cdef object val
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider.deserialize_field", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tag);
  __Pyx_XDECREF((PyObject *)__pyx_v_oneoffield);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_cb);
  __Pyx_XDECREF(__pyx_v_it_obj);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_6__reduce_cython__(((struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_6__reduce_cython__(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.type,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.type);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.type);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.type);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.type,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.type,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.type is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.type,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.type is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_JsonSerializeProvider, (type(self), 0x599dcce, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->__pyx_base.type != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_JsonSerializeProvider, (type(self), 0x599dcce, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.type is not None
 *     if use_setstate:
 *         return __pyx_unpickle_JsonSerializeProvider, (type(self), 0x599dcce, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_JsonSerializeProvider, (type(self), 0x599dcce, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_JsonSerializeProv); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_93969614);
    __Pyx_GIVEREF(__pyx_int_93969614);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_93969614);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_JsonSerializeProvider, (type(self), 0x599dcce, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_JsonSerializeProvider, (type(self), 0x599dcce, None), state
 *     else:
 *         return __pyx_unpickle_JsonSerializeProvider, (type(self), 0x599dcce, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_JsonSerializeProvider__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_JsonSerializeProv); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_93969614);
    __Pyx_GIVEREF(__pyx_int_93969614);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_93969614);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_JsonSerializeProvider, (type(self), 0x599dcce, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_JsonSerializeProvider__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_8__setstate_cython__(((struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_8__setstate_cython__(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_JsonSerializeProvider, (type(self), 0x599dcce, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_JsonSerializeProvider__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_4mars_9serialize_14jsonserializer___pyx_unpickle_JsonSerializeProvider__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_JsonSerializeProvider, (type(self), 0x599dcce, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_JsonSerializeProvider__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.JsonSerializeProvider.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_JsonSerializeProvider(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_1__pyx_unpickle_JsonSerializeProvider(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4mars_9serialize_14jsonserializer_1__pyx_unpickle_JsonSerializeProvider = {"__pyx_unpickle_JsonSerializeProvider", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_9serialize_14jsonserializer_1__pyx_unpickle_JsonSerializeProvider, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4mars_9serialize_14jsonserializer_1__pyx_unpickle_JsonSerializeProvider(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_JsonSerializeProvider (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_JsonSerializeProvider", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_JsonSerializeProvider", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_JsonSerializeProvider") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_JsonSerializeProvider", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.serialize.jsonserializer.__pyx_unpickle_JsonSerializeProvider", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4mars_9serialize_14jsonserializer___pyx_unpickle_JsonSerializeProvider(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_9serialize_14jsonserializer___pyx_unpickle_JsonSerializeProvider(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_JsonSerializeProvider", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x599dcce:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x599dcce = (type))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x599dcce) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x599dcce:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x599dcce = (type))" % __pyx_checksum)
 *     __pyx_result = JsonSerializeProvider.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x599dcce:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x599dcce = (type))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = JsonSerializeProvider.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x59, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x599dcce:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x599dcce = (type))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x599dcce = (type))" % __pyx_checksum)
 *     __pyx_result = JsonSerializeProvider.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_JsonSerializeProvider__set_state(<JsonSerializeProvider> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_14jsonserializer_JsonSerializeProvider), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x599dcce = (type))" % __pyx_checksum)
 *     __pyx_result = JsonSerializeProvider.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_JsonSerializeProvider__set_state(<JsonSerializeProvider> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = JsonSerializeProvider.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_JsonSerializeProvider__set_state(<JsonSerializeProvider> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_JsonSerializeProvider__set_state(JsonSerializeProvider __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_4mars_9serialize_14jsonserializer___pyx_unpickle_JsonSerializeProvider__set_state(((struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x599dcce = (type))" % __pyx_checksum)
 *     __pyx_result = JsonSerializeProvider.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_JsonSerializeProvider__set_state(<JsonSerializeProvider> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_JsonSerializeProvider__set_state(<JsonSerializeProvider> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_JsonSerializeProvider__set_state(JsonSerializeProvider __pyx_result, tuple __pyx_state):
 *     __pyx_result.type = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_JsonSerializeProvider(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.__pyx_unpickle_JsonSerializeProvider", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_JsonSerializeProvider__set_state(<JsonSerializeProvider> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_JsonSerializeProvider__set_state(JsonSerializeProvider __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.type = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_4mars_9serialize_14jsonserializer___pyx_unpickle_JsonSerializeProvider__set_state(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_JsonSerializeProvider__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_JsonSerializeProvider__set_state(JsonSerializeProvider __pyx_result, tuple __pyx_state):
 *     __pyx_result.type = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.type);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.type);
  __pyx_v___pyx_result->__pyx_base.type = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_JsonSerializeProvider__set_state(JsonSerializeProvider __pyx_result, tuple __pyx_state):
 *     __pyx_result.type = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.type = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_JsonSerializeProvider__set_state(JsonSerializeProvider __pyx_result, tuple __pyx_state):
 *     __pyx_result.type = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_JsonSerializeProvider__set_state(<JsonSerializeProvider> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_JsonSerializeProvider__set_state(JsonSerializeProvider __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.type = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("mars.serialize.jsonserializer.__pyx_unpickle_JsonSerializeProvider__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyArray_Descr *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":265
 * 
 *             cdef int i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *             cdef int i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":271
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 272, __pyx_L1_error)

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":275
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 276, __pyx_L1_error)

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":279
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * 2) * ((size_t)__pyx_v_ndim))));

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":286
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":287
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L9;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":289
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L9:;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":291
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":296
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":297
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_7 = PyArray_DESCR(__pyx_v_self);
  __pyx_t_3 = ((PyObject *)__pyx_t_7);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":300
 *             cdef int offset
 * 
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not PyDataType_HASFIELDS(descr):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(PyDataType_HASFIELDS(__pyx_v_descr) != 0)) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":303
 * 
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L15_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_L15_next_or:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":305
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_1)) {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 306, __pyx_L1_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":307
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;
      case NPY_UBYTE:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":308
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      __pyx_v_f = ((char *)"B");
      break;
      case NPY_SHORT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":309
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      __pyx_v_f = ((char *)"h");
      break;
      case NPY_USHORT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":310
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      __pyx_v_f = ((char *)"H");
      break;
      case NPY_INT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":311
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      __pyx_v_f = ((char *)"i");
      break;
      case NPY_UINT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":312
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      __pyx_v_f = ((char *)"I");
      break;
      case NPY_LONG:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":313
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      __pyx_v_f = ((char *)"l");
      break;
      case NPY_ULONG:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":314
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      __pyx_v_f = ((char *)"L");
      break;
      case NPY_LONGLONG:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":315
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      __pyx_v_f = ((char *)"q");
      break;
      case NPY_ULONGLONG:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":316
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      __pyx_v_f = ((char *)"Q");
      break;
      case NPY_FLOAT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":317
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      __pyx_v_f = ((char *)"f");
      break;
      case NPY_DOUBLE:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":318
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      __pyx_v_f = ((char *)"d");
      break;
      case NPY_LONGDOUBLE:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":319
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      __pyx_v_f = ((char *)"g");
      break;
      case NPY_CFLOAT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":320
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      __pyx_v_f = ((char *)"Zf");
      break;
      case NPY_CDOUBLE:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":321
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      __pyx_v_f = ((char *)"Zd");
      break;
      case NPY_CLONGDOUBLE:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":322
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      __pyx_v_f = ((char *)"Zg");
      break;
      case NPY_OBJECT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":323
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":325
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 325, __pyx_L1_error)
      break;
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":326
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":327
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":329
 *                 return
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)PyObject_Malloc(0xFF));

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":330
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":331
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":332
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(2, 332, __pyx_L1_error)
    __pyx_v_f = __pyx_t_9;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":335
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":339
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)
 */
    PyObject_Free(__pyx_v_info->format);

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":341
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    PyObject_Free(__pyx_v_info->strides);

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":822
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":825
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":828
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":831
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":834
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":838
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":847
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":848
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(2, 851, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 851, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":852
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 852, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(2, 852, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":853
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(2, 853, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 853, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(2, 853, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 856, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 856, __pyx_L1_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":859
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_6)) {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":860
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 860, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 860, __pyx_L1_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":870
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":871
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":872
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":873
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":875
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":878
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (unlikely(__pyx_t_6)) {

        /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 880, __pyx_L1_error)

        /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":883
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":884
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":885
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":886
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":887
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":888
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":889
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":890
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":891
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":892
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":893
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":894
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":895
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":896
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":897
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":898
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":899
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(__pyx_t_6)) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":901
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 901, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":902
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":906
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(2, 906, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":907
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1023
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1024
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1029
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1030
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1036
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1036, __pyx_L3_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1037, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1038, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1038, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1042
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1042, __pyx_L3_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1044, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1044, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1048
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1048, __pyx_L3_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1049
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1050
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1050, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1050, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_4mars_9serialize_14jsonserializer_JsonSerializeProvider __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider;

static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer_JsonSerializeProvider(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *p;
  PyObject *o = __pyx_ptype_4mars_9serialize_4core_Provider->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_4mars_9serialize_4core_Provider*)__pyx_vtabptr_4mars_9serialize_14jsonserializer_JsonSerializeProvider;
  return o;
}

static void __pyx_tp_dealloc_4mars_9serialize_14jsonserializer_JsonSerializeProvider(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_4mars_9serialize_4core_Provider)) __pyx_ptype_4mars_9serialize_4core_Provider->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_4mars_9serialize_14jsonserializer_JsonSerializeProvider);
}

static int __pyx_tp_traverse_4mars_9serialize_14jsonserializer_JsonSerializeProvider(PyObject *o, visitproc v, void *a) {
  int e;
  e = ((likely(__pyx_ptype_4mars_9serialize_4core_Provider)) ? ((__pyx_ptype_4mars_9serialize_4core_Provider->tp_traverse) ? __pyx_ptype_4mars_9serialize_4core_Provider->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_4mars_9serialize_14jsonserializer_JsonSerializeProvider)); if (e) return e;
  return 0;
}

static int __pyx_tp_clear_4mars_9serialize_14jsonserializer_JsonSerializeProvider(PyObject *o) {
  if (likely(__pyx_ptype_4mars_9serialize_4core_Provider)) { if (__pyx_ptype_4mars_9serialize_4core_Provider->tp_clear) __pyx_ptype_4mars_9serialize_4core_Provider->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_4mars_9serialize_14jsonserializer_JsonSerializeProvider);
  return 0;
}

static PyMethodDef __pyx_methods_4mars_9serialize_14jsonserializer_JsonSerializeProvider[] = {
  {"serialize_field", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_3serialize_field, METH_VARARGS|METH_KEYWORDS, 0},
  {"deserialize_field", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_5deserialize_field, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4mars_9serialize_14jsonserializer_JsonSerializeProvider = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.serialize.jsonserializer.JsonSerializeProvider", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_9serialize_14jsonserializer_JsonSerializeProvider, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_9serialize_14jsonserializer_JsonSerializeProvider, /*tp_traverse*/
  __pyx_tp_clear_4mars_9serialize_14jsonserializer_JsonSerializeProvider, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4mars_9serialize_14jsonserializer_JsonSerializeProvider, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_9serialize_14jsonserializer_JsonSerializeProvider, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key[8];
static int __pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key = 0;

static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key)))) {
    o = (PyObject*)__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key[--__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key(PyObject *o) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_obj);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key)))) {
    __pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key[__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key++] = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *)o;
  if (p->__pyx_v_obj) {
    e = (*v)(p->__pyx_v_obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key *)o;
  tmp = ((PyObject*)p->__pyx_v_obj);
  p->__pyx_v_obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.serialize.jsonserializer.__pyx_scope_struct___deserialize_key", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key, /*tp_traverse*/
  __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr *__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr[--__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr[__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.serialize.jsonserializer.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list[8];
static int __pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list = 0;

static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list)))) {
    o = (PyObject*)__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list[--__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list(PyObject *o) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_res);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list)))) {
    __pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list[__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list++] = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *)o;
  if (p->__pyx_v_res) {
    e = (*v)(p->__pyx_v_res, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *)o;
  tmp = ((PyObject*)p->__pyx_v_res);
  p->__pyx_v_res = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.serialize.jsonserializer.__pyx_scope_struct_2__deserialize_list", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list, /*tp_traverse*/
  __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb[8];
static int __pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb = 0;

static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb)))) {
    o = (PyObject*)__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb[--__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb(PyObject *o) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_j);
  Py_CLEAR(p->__pyx_v_v);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb)))) {
    __pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb[__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb++] = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_j) {
    e = (*v)(p->__pyx_v_j, a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_j);
  p->__pyx_v_j = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_v);
  p->__pyx_v_v = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.serialize.jsonserializer.__pyx_scope_struct_3_cb", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb, /*tp_traverse*/
  __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple[8];
static int __pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple = 0;

static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple)))) {
    o = (PyObject*)__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple[--__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple(PyObject *o) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_callbacks);
  Py_CLEAR(p->__pyx_v_obj);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple)))) {
    __pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple[__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple++] = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *)o;
  if (p->__pyx_v_callbacks) {
    e = (*v)(p->__pyx_v_callbacks, a); if (e) return e;
  }
  if (p->__pyx_v_obj) {
    e = (*v)(p->__pyx_v_obj, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple *)o;
  tmp = ((PyObject*)p->__pyx_v_callbacks);
  p->__pyx_v_callbacks = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_obj);
  p->__pyx_v_obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.serialize.jsonserializer.__pyx_scope_struct_4__deserialize_tuple", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple, /*tp_traverse*/
  __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr *__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr[8];
static int __pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr = 0;

static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr)))) {
    o = (PyObject*)__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr[--__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr(PyObject *o) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_it_val);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr)))) {
    __pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr[__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr++] = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_it_val) {
    e = (*v)(p->__pyx_v_it_val, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.serialize.jsonserializer.__pyx_scope_struct_5_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict[8];
static int __pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict = 0;

static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict)))) {
    o = (PyObject*)__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict[--__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict(PyObject *o) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_res);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict)))) {
    __pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict[__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict++] = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *)o;
  if (p->__pyx_v_res) {
    e = (*v)(p->__pyx_v_res, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *)o;
  tmp = ((PyObject*)p->__pyx_v_res);
  p->__pyx_v_res = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.serialize.jsonserializer.__pyx_scope_struct_6__deserialize_dict", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict, /*tp_traverse*/
  __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb[8];
static int __pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb = 0;

static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb)))) {
    o = (PyObject*)__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb[--__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb(PyObject *o) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_ko);
  Py_CLEAR(p->__pyx_v_vo);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb)))) {
    __pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb[__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb++] = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_ko) {
    e = (*v)(p->__pyx_v_ko, a); if (e) return e;
  }
  if (p->__pyx_v_vo) {
    e = (*v)(p->__pyx_v_vo, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_ko);
  p->__pyx_v_ko = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_vo);
  p->__pyx_v_vo = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.serialize.jsonserializer.__pyx_scope_struct_7_cb", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb, /*tp_traverse*/
  __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value *__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value[8];
static int __pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value = 0;

static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value)))) {
    o = (PyObject*)__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value[--__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value)))) {
    __pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value[__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value++] = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.serialize.jsonserializer.__pyx_scope_struct_8__deserialize_value", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field[8];
static int __pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field = 0;

static PyObject *__pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field)))) {
    o = (PyObject*)__pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field[--__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field(PyObject *o) {
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_field);
  Py_CLEAR(p->__pyx_v_model_instance);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_val);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field)))) {
    __pyx_freelist_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field[__pyx_freecount_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field++] = ((struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *)o;
  if (p->__pyx_v_field) {
    e = (*v)(((PyObject *)p->__pyx_v_field), a); if (e) return e;
  }
  if (p->__pyx_v_model_instance) {
    e = (*v)(p->__pyx_v_model_instance, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_val) {
    e = (*v)(p->__pyx_v_val, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *p = (struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field *)o;
  tmp = ((PyObject*)p->__pyx_v_field);
  p->__pyx_v_field = ((struct __pyx_obj_4mars_9serialize_4core_Field *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_model_instance);
  p->__pyx_v_model_instance = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_val);
  p->__pyx_v_val = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.serialize.jsonserializer.__pyx_scope_struct_9_deserialize_field", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field, /*tp_traverse*/
  __pyx_tp_clear_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_jsonserializer(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_jsonserializer},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "jsonserializer",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_BaseWithKey, __pyx_k_BaseWithKey, sizeof(__pyx_k_BaseWithKey), 0, 0, 1, 1},
  {&__pyx_n_s_BytesIO, __pyx_k_BytesIO, sizeof(__pyx_k_BytesIO), 0, 0, 1, 1},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x59, __pyx_k_Incompatible_checksums_s_vs_0x59, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x59), 0, 0, 1, 0},
  {&__pyx_n_s_JsonSerializeProvider, __pyx_k_JsonSerializeProvider, sizeof(__pyx_k_JsonSerializeProvider), 0, 0, 1, 1},
  {&__pyx_n_s_JsonSerializeProvider__deseriali, __pyx_k_JsonSerializeProvider__deseriali, sizeof(__pyx_k_JsonSerializeProvider__deseriali), 0, 0, 1, 1},
  {&__pyx_n_s_JsonSerializeProvider__deseriali_2, __pyx_k_JsonSerializeProvider__deseriali_2, sizeof(__pyx_k_JsonSerializeProvider__deseriali_2), 0, 0, 1, 1},
  {&__pyx_n_s_JsonSerializeProvider__deseriali_3, __pyx_k_JsonSerializeProvider__deseriali_3, sizeof(__pyx_k_JsonSerializeProvider__deseriali_3), 0, 0, 1, 1},
  {&__pyx_n_s_JsonSerializeProvider__deseriali_4, __pyx_k_JsonSerializeProvider__deseriali_4, sizeof(__pyx_k_JsonSerializeProvider__deseriali_4), 0, 0, 1, 1},
  {&__pyx_n_s_JsonSerializeProvider__deseriali_5, __pyx_k_JsonSerializeProvider__deseriali_5, sizeof(__pyx_k_JsonSerializeProvider__deseriali_5), 0, 0, 1, 1},
  {&__pyx_n_s_JsonSerializeProvider__deseriali_6, __pyx_k_JsonSerializeProvider__deseriali_6, sizeof(__pyx_k_JsonSerializeProvider__deseriali_6), 0, 0, 1, 1},
  {&__pyx_n_s_JsonSerializeProvider__deseriali_7, __pyx_k_JsonSerializeProvider__deseriali_7, sizeof(__pyx_k_JsonSerializeProvider__deseriali_7), 0, 0, 1, 1},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_kp_s_Only_one_of_attributes_0_can_be, __pyx_k_Only_one_of_attributes_0_can_be, sizeof(__pyx_k_Only_one_of_attributes_0_can_be), 0, 0, 1, 0},
  {&__pyx_n_s_OrderedDict, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unknown_type_to_deserialize_0, __pyx_k_Unknown_type_to_deserialize_0, sizeof(__pyx_k_Unknown_type_to_deserialize_0), 0, 0, 1, 0},
  {&__pyx_kp_s_Unknown_type_to_serialize_0, __pyx_k_Unknown_type_to_serialize_0, sizeof(__pyx_k_Unknown_type_to_serialize_0), 0, 0, 1, 0},
  {&__pyx_n_s_V, __pyx_k_V, sizeof(__pyx_k_V), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_arr, __pyx_k_arr, sizeof(__pyx_k_arr), 0, 0, 1, 1},
  {&__pyx_n_s_attr, __pyx_k_attr, sizeof(__pyx_k_attr), 0, 0, 1, 1},
  {&__pyx_n_s_attrs, __pyx_k_attrs, sizeof(__pyx_k_attrs), 0, 0, 1, 1},
  {&__pyx_n_s_b64decode, __pyx_k_b64decode, sizeof(__pyx_k_b64decode), 0, 0, 1, 1},
  {&__pyx_n_s_b64encode, __pyx_k_b64encode, sizeof(__pyx_k_b64encode), 0, 0, 1, 1},
  {&__pyx_n_s_base64, __pyx_k_base64, sizeof(__pyx_k_base64), 0, 0, 1, 1},
  {&__pyx_n_s_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
  {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_callbacks, __pyx_k_callbacks, sizeof(__pyx_k_callbacks), 0, 0, 1, 1},
  {&__pyx_n_s_cb, __pyx_k_cb, sizeof(__pyx_k_cb), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_compat, __pyx_k_compat, sizeof(__pyx_k_compat), 0, 0, 1, 1},
  {&__pyx_n_s_core, __pyx_k_core, sizeof(__pyx_k_core), 0, 0, 1, 1},
  {&__pyx_n_s_datetime64, __pyx_k_datetime64, sizeof(__pyx_k_datetime64), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
  {&__pyx_n_s_deserialize, __pyx_k_deserialize, sizeof(__pyx_k_deserialize), 0, 0, 1, 1},
  {&__pyx_n_s_deserialize_field_locals_cb, __pyx_k_deserialize_field_locals_cb, sizeof(__pyx_k_deserialize_field_locals_cb), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dumps, __pyx_k_dumps, sizeof(__pyx_k_dumps), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_field, __pyx_k_field, sizeof(__pyx_k_field), 0, 0, 1, 1},
  {&__pyx_n_s_float16, __pyx_k_float16, sizeof(__pyx_k_float16), 0, 0, 1, 1},
  {&__pyx_n_s_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_getvalue, __pyx_k_getvalue, sizeof(__pyx_k_getvalue), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inner, __pyx_k_inner, sizeof(__pyx_k_inner), 0, 0, 1, 1},
  {&__pyx_n_s_int16, __pyx_k_int16, sizeof(__pyx_k_int16), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_int64, __pyx_k_int64, sizeof(__pyx_k_int64), 0, 0, 1, 1},
  {&__pyx_n_s_int8, __pyx_k_int8, sizeof(__pyx_k_int8), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_izip, __pyx_k_izip, sizeof(__pyx_k_izip), 0, 0, 1, 1},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_key_to_instance, __pyx_k_key_to_instance, sizeof(__pyx_k_key_to_instance), 0, 0, 1, 1},
  {&__pyx_n_s_key_type, __pyx_k_key_type, sizeof(__pyx_k_key_type), 0, 0, 1, 1},
  {&__pyx_n_s_ko, __pyx_k_ko, sizeof(__pyx_k_ko), 0, 0, 1, 1},
  {&__pyx_n_s_list, __pyx_k_list, sizeof(__pyx_k_list), 0, 0, 1, 1},
  {&__pyx_n_s_load, __pyx_k_load, sizeof(__pyx_k_load), 0, 0, 1, 1},
  {&__pyx_n_s_loads, __pyx_k_loads, sizeof(__pyx_k_loads), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mars_serialize_jsonserializer, __pyx_k_mars_serialize_jsonserializer, sizeof(__pyx_k_mars_serialize_jsonserializer), 0, 0, 1, 1},
  {&__pyx_kp_s_mars_serialize_jsonserializer_py, __pyx_k_mars_serialize_jsonserializer_py, sizeof(__pyx_k_mars_serialize_jsonserializer_py), 0, 0, 1, 0},
  {&__pyx_n_s_model, __pyx_k_model, sizeof(__pyx_k_model), 0, 0, 1, 1},
  {&__pyx_n_s_model_instance, __pyx_k_model_instance, sizeof(__pyx_k_model_instance), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_o, __pyx_k_o, sizeof(__pyx_k_o), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_JsonSerializeProv, __pyx_k_pyx_unpickle_JsonSerializeProv, sizeof(__pyx_k_pyx_unpickle_JsonSerializeProv), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_ref, __pyx_k_ref, sizeof(__pyx_k_ref), 0, 0, 1, 1},
  {&__pyx_n_s_save, __pyx_k_save, sizeof(__pyx_k_save), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_serialize, __pyx_k_serialize, sizeof(__pyx_k_serialize), 0, 0, 1, 1},
  {&__pyx_n_s_serialize_field, __pyx_k_serialize_field, sizeof(__pyx_k_serialize_field), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_six, __pyx_k_six, sizeof(__pyx_k_six), 0, 0, 1, 1},
  {&__pyx_n_s_slice, __pyx_k_slice, sizeof(__pyx_k_slice), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_subs, __pyx_k_subs, sizeof(__pyx_k_subs), 0, 0, 1, 1},
  {&__pyx_n_s_tag_name, __pyx_k_tag_name, sizeof(__pyx_k_tag_name), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_timedelta64, __pyx_k_timedelta64, sizeof(__pyx_k_timedelta64), 0, 0, 1, 1},
  {&__pyx_n_s_tuple, __pyx_k_tuple, sizeof(__pyx_k_tuple), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_uint16, __pyx_k_uint16, sizeof(__pyx_k_uint16), 0, 0, 1, 1},
  {&__pyx_n_s_uint32, __pyx_k_uint32, sizeof(__pyx_k_uint32), 0, 0, 1, 1},
  {&__pyx_n_s_uint64, __pyx_k_uint64, sizeof(__pyx_k_uint64), 0, 0, 1, 1},
  {&__pyx_n_s_uint8, __pyx_k_uint8, sizeof(__pyx_k_uint8), 0, 0, 1, 1},
  {&__pyx_n_s_unicode, __pyx_k_unicode, sizeof(__pyx_k_unicode), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_kp_s_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 0, 1, 0},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_value_type, __pyx_k_value_type, sizeof(__pyx_k_value_type), 0, 0, 1, 1},
  {&__pyx_n_s_vo, __pyx_k_vo, sizeof(__pyx_k_vo), 0, 0, 1, 1},
  {&__pyx_n_s_weakref, __pyx_k_weakref, sizeof(__pyx_k_weakref), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 140, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 170, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 382, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(2, 285, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(2, 856, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 1038, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "mars/serialize/jsonserializer.pyx":175
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(j, v):
 *                     def inner(subs):             # <<<<<<<<<<<<<<
 *                         o = subs[v.key, v.id]
 *                         if weak_ref:
 */
  __pyx_tuple_ = PyTuple_Pack(2, __pyx_n_s_subs, __pyx_n_s_o); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);
  __pyx_codeobj__2 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple_, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_mars_serialize_jsonserializer_py, __pyx_n_s_inner, 175, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2)) __PYX_ERR(0, 175, __pyx_L1_error)

  /* "mars/serialize/jsonserializer.pyx":174
 *             res.append(val)
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(j, v):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[v.key, v.id]
 */
  __pyx_tuple__3 = PyTuple_Pack(4, __pyx_n_s_j, __pyx_n_s_v, __pyx_n_s_inner, __pyx_n_s_inner); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);
  __pyx_codeobj__4 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__3, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_mars_serialize_jsonserializer_py, __pyx_n_s_cb, 174, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__4)) __PYX_ERR(0, 174, __pyx_L1_error)

  /* "mars/serialize/jsonserializer.pyx":223
 *             if isinstance(value, KeyPlaceholder):
 *                 def cb(ko, vo):
 *                     def inner(subs):             # <<<<<<<<<<<<<<
 *                         o = subs[vo.key, vo.id]
 *                         if weak_ref:
 */
  __pyx_tuple__5 = PyTuple_Pack(2, __pyx_n_s_subs, __pyx_n_s_o); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);
  __pyx_codeobj__6 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__5, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_mars_serialize_jsonserializer_py, __pyx_n_s_inner, 223, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__6)) __PYX_ERR(0, 223, __pyx_L1_error)

  /* "mars/serialize/jsonserializer.pyx":222
 *             res[key] = value
 *             if isinstance(value, KeyPlaceholder):
 *                 def cb(ko, vo):             # <<<<<<<<<<<<<<
 *                     def inner(subs):
 *                         o = subs[vo.key, vo.id]
 */
  __pyx_tuple__7 = PyTuple_Pack(4, __pyx_n_s_ko, __pyx_n_s_vo, __pyx_n_s_inner, __pyx_n_s_inner); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);
  __pyx_codeobj__8 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__7, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_mars_serialize_jsonserializer_py, __pyx_n_s_cb, 222, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__8)) __PYX_ERR(0, 222, __pyx_L1_error)

  /* "mars/serialize/jsonserializer.pyx":489
 *             setattr(model_instance, field.attr, self._on_deserial(field, val))
 *             if isinstance(val, KeyPlaceholder):
 *                 def cb(subs):             # <<<<<<<<<<<<<<
 *                     o = subs[val.key, val.id]
 *                     if field.weak_ref:
 */
  __pyx_tuple__9 = PyTuple_Pack(2, __pyx_n_s_subs, __pyx_n_s_o); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_codeobj__10 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__9, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_mars_serialize_jsonserializer_py, __pyx_n_s_cb, 489, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__10)) __PYX_ERR(0, 489, __pyx_L1_error)

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(2, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(2, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(2, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(2, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(2, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(2, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(2, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "(tree fragment)":1
 * def __pyx_unpickle_JsonSerializeProvider(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__18 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_JsonSerializeProv, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_93969614 = PyInt_FromLong(93969614L); if (unlikely(!__pyx_int_93969614)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_4mars_9serialize_14jsonserializer_PRIMITIVE_TYPE_TO_NAME = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_4mars_9serialize_14jsonserializer_EXTEND_TYPE_TO_NAME = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_t_1 = PyImport_ImportModule("mars.serialize.core"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4mars_9serialize_4core_Provider = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Provider", sizeof(struct __pyx_obj_4mars_9serialize_4core_Provider), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Provider) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_Provider = (struct __pyx_vtabstruct_4mars_9serialize_4core_Provider*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_Provider->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_Provider)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_14jsonserializer_JsonSerializeProvider = &__pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider.__pyx_base = *__pyx_vtabptr_4mars_9serialize_4core_Provider;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider.__pyx_base.serialize_field = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_4core_Provider *, struct __pyx_obj_4mars_9serialize_4core_Field *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider_serialize_field;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._to_str = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_str;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._to_bytes = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__to_bytes;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_slice = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_slice;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._deserialize_slice = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_slice;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_arr = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyArrayObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_arr;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._deserialize_arr = (PyArrayObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_arr;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_dtype = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyArray_Descr *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dtype;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._deserialize_dtype = (PyArray_Descr *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_dtype;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_key = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_key;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._deserialize_key = (struct __pyx_obj_4mars_9serialize_4core_KeyPlaceholder *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_key;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_list = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list *__pyx_optional_args))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_list;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._deserialize_list = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_list;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_tuple = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple *__pyx_optional_args))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_tuple;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._deserialize_tuple = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_tuple;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_dict = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict *__pyx_optional_args))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_dict;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._deserialize_dict = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_dict;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_datetime64_timedelta64 = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64_timedelta64;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._deserialize_datetime64_timedelta64 = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_datetime64_timedelta64;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_datetime64 = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_datetime64;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._deserialize_datetime64 = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_datetime64;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_timedelta64 = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_timedelta64;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._deserialize_timedelta64 = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_timedelta64;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_typed_value = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value *__pyx_optional_args))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_typed_value;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_untyped_value = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value *__pyx_optional_args))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_untyped_value;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._serialize_value = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, struct __pyx_opt_args_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value *__pyx_optional_args))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__serialize_value;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._on_serial = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, struct __pyx_obj_4mars_9serialize_4core_Field *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_serial;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._deserialize_value = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, PyObject *, PyObject *, int))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__deserialize_value;
  __pyx_vtable_4mars_9serialize_14jsonserializer_JsonSerializeProvider._on_deserial = (PyObject *(*)(struct __pyx_obj_4mars_9serialize_14jsonserializer_JsonSerializeProvider *, struct __pyx_obj_4mars_9serialize_4core_Field *, PyObject *))__pyx_f_4mars_9serialize_14jsonserializer_21JsonSerializeProvider__on_deserial;
  __pyx_type_4mars_9serialize_14jsonserializer_JsonSerializeProvider.tp_base = __pyx_ptype_4mars_9serialize_4core_Provider;
  if (PyType_Ready(&__pyx_type_4mars_9serialize_14jsonserializer_JsonSerializeProvider) < 0) __PYX_ERR(0, 72, __pyx_L1_error)
  __pyx_type_4mars_9serialize_14jsonserializer_JsonSerializeProvider.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_9serialize_14jsonserializer_JsonSerializeProvider.tp_dictoffset && __pyx_type_4mars_9serialize_14jsonserializer_JsonSerializeProvider.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_9serialize_14jsonserializer_JsonSerializeProvider.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_4mars_9serialize_14jsonserializer_JsonSerializeProvider.tp_dict, __pyx_vtabptr_4mars_9serialize_14jsonserializer_JsonSerializeProvider) < 0) __PYX_ERR(0, 72, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_JsonSerializeProvider, (PyObject *)&__pyx_type_4mars_9serialize_14jsonserializer_JsonSerializeProvider) < 0) __PYX_ERR(0, 72, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_4mars_9serialize_14jsonserializer_JsonSerializeProvider) < 0) __PYX_ERR(0, 72, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_14jsonserializer_JsonSerializeProvider = &__pyx_type_4mars_9serialize_14jsonserializer_JsonSerializeProvider;
  if (PyType_Ready(&__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
  __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key.tp_dictoffset && __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key = &__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct___deserialize_key;
  if (PyType_Ready(&__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 154, __pyx_L1_error)
  __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr = &__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list) < 0) __PYX_ERR(0, 164, __pyx_L1_error)
  __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list.tp_dictoffset && __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list = &__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_2__deserialize_list;
  if (PyType_Ready(&__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb.tp_dictoffset && __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb = &__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_3_cb;
  if (PyType_Ready(&__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple) < 0) __PYX_ERR(0, 201, __pyx_L1_error)
  __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple.tp_dictoffset && __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple = &__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_4__deserialize_tuple;
  if (PyType_Ready(&__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr) < 0) __PYX_ERR(0, 202, __pyx_L1_error)
  __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr.tp_dictoffset && __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr = &__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_5_genexpr;
  if (PyType_Ready(&__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict) < 0) __PYX_ERR(0, 215, __pyx_L1_error)
  __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict.tp_dictoffset && __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict = &__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_6__deserialize_dict;
  if (PyType_Ready(&__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb) < 0) __PYX_ERR(0, 222, __pyx_L1_error)
  __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb.tp_dictoffset && __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb = &__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_7_cb;
  if (PyType_Ready(&__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value) < 0) __PYX_ERR(0, 407, __pyx_L1_error)
  __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value.tp_dictoffset && __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value = &__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_8__deserialize_value;
  if (PyType_Ready(&__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field) < 0) __PYX_ERR(0, 445, __pyx_L1_error)
  __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field.tp_dictoffset && __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field = &__pyx_type_4mars_9serialize_14jsonserializer___pyx_scope_struct_9_deserialize_field;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 206, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 233, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 242, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("mars.serialize.core"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4mars_9serialize_4core_Identity = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Identity", sizeof(struct __pyx_obj_4mars_9serialize_4core_Identity), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Identity) __PYX_ERR(4, 44, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_List = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "List", sizeof(struct __pyx_obj_4mars_9serialize_4core_List), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_List) __PYX_ERR(4, 49, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Tuple = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Tuple", sizeof(struct __pyx_obj_4mars_9serialize_4core_Tuple), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Tuple) __PYX_ERR(4, 54, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Dict = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Dict", sizeof(struct __pyx_obj_4mars_9serialize_4core_Dict), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Dict) __PYX_ERR(4, 59, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Reference = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Reference", sizeof(struct __pyx_obj_4mars_9serialize_4core_Reference), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Reference) __PYX_ERR(4, 65, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_OneOf = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "OneOf", sizeof(struct __pyx_obj_4mars_9serialize_4core_OneOf), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_OneOf) __PYX_ERR(4, 70, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_ValueType = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "ValueType", sizeof(struct __pyx_obj_4mars_9serialize_4core_ValueType), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_ValueType) __PYX_ERR(4, 74, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Field) __PYX_ERR(4, 78, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_Field)) __PYX_ERR(4, 78, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_AnyField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "AnyField", sizeof(struct __pyx_obj_4mars_9serialize_4core_AnyField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_AnyField) __PYX_ERR(4, 95, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_AnyField = (struct __pyx_vtabstruct_4mars_9serialize_4core_AnyField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_AnyField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_AnyField)) __PYX_ERR(4, 95, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_IdentityField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "IdentityField", sizeof(struct __pyx_obj_4mars_9serialize_4core_IdentityField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_IdentityField) __PYX_ERR(4, 99, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_IdentityField = (struct __pyx_vtabstruct_4mars_9serialize_4core_IdentityField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_IdentityField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_IdentityField)) __PYX_ERR(4, 99, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_BoolField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "BoolField", sizeof(struct __pyx_obj_4mars_9serialize_4core_BoolField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_BoolField) __PYX_ERR(4, 103, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_BoolField = (struct __pyx_vtabstruct_4mars_9serialize_4core_BoolField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_BoolField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_BoolField)) __PYX_ERR(4, 103, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Int8Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Int8Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_Int8Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Int8Field) __PYX_ERR(4, 107, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_Int8Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_Int8Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_Int8Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_Int8Field)) __PYX_ERR(4, 107, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Int16Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Int16Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_Int16Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Int16Field) __PYX_ERR(4, 111, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_Int16Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_Int16Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_Int16Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_Int16Field)) __PYX_ERR(4, 111, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Int32Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Int32Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_Int32Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Int32Field) __PYX_ERR(4, 115, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_Int32Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_Int32Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_Int32Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_Int32Field)) __PYX_ERR(4, 115, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Int64Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Int64Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_Int64Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Int64Field) __PYX_ERR(4, 119, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_Int64Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_Int64Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_Int64Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_Int64Field)) __PYX_ERR(4, 119, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_UInt8Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "UInt8Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_UInt8Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_UInt8Field) __PYX_ERR(4, 123, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_UInt8Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_UInt8Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_UInt8Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_UInt8Field)) __PYX_ERR(4, 123, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_UInt16Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "UInt16Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_UInt16Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_UInt16Field) __PYX_ERR(4, 127, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_UInt16Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_UInt16Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_UInt16Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_UInt16Field)) __PYX_ERR(4, 127, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_UInt32Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "UInt32Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_UInt32Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_UInt32Field) __PYX_ERR(4, 131, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_UInt32Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_UInt32Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_UInt32Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_UInt32Field)) __PYX_ERR(4, 131, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_UInt64Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "UInt64Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_UInt64Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_UInt64Field) __PYX_ERR(4, 135, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_UInt64Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_UInt64Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_UInt64Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_UInt64Field)) __PYX_ERR(4, 135, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Float16Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Float16Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_Float16Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Float16Field) __PYX_ERR(4, 139, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_Float16Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_Float16Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_Float16Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_Float16Field)) __PYX_ERR(4, 139, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Float32Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Float32Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_Float32Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Float32Field) __PYX_ERR(4, 143, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_Float32Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_Float32Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_Float32Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_Float32Field)) __PYX_ERR(4, 143, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Float64Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Float64Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_Float64Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Float64Field) __PYX_ERR(4, 147, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_Float64Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_Float64Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_Float64Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_Float64Field)) __PYX_ERR(4, 147, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_StringField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "StringField", sizeof(struct __pyx_obj_4mars_9serialize_4core_StringField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_StringField) __PYX_ERR(4, 151, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_StringField = (struct __pyx_vtabstruct_4mars_9serialize_4core_StringField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_StringField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_StringField)) __PYX_ERR(4, 151, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_BytesField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "BytesField", sizeof(struct __pyx_obj_4mars_9serialize_4core_BytesField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_BytesField) __PYX_ERR(4, 155, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_BytesField = (struct __pyx_vtabstruct_4mars_9serialize_4core_BytesField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_BytesField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_BytesField)) __PYX_ERR(4, 155, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_UnicodeField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "UnicodeField", sizeof(struct __pyx_obj_4mars_9serialize_4core_UnicodeField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_UnicodeField) __PYX_ERR(4, 159, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_UnicodeField = (struct __pyx_vtabstruct_4mars_9serialize_4core_UnicodeField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_UnicodeField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_UnicodeField)) __PYX_ERR(4, 159, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_KeyField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "KeyField", sizeof(struct __pyx_obj_4mars_9serialize_4core_KeyField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_KeyField) __PYX_ERR(4, 163, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_KeyField = (struct __pyx_vtabstruct_4mars_9serialize_4core_KeyField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_KeyField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_KeyField)) __PYX_ERR(4, 163, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_NDArrayField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "NDArrayField", sizeof(struct __pyx_obj_4mars_9serialize_4core_NDArrayField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_NDArrayField) __PYX_ERR(4, 167, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_NDArrayField = (struct __pyx_vtabstruct_4mars_9serialize_4core_NDArrayField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_NDArrayField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_NDArrayField)) __PYX_ERR(4, 167, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Datetime64Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Datetime64Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_Datetime64Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Datetime64Field) __PYX_ERR(4, 171, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_Datetime64Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_Datetime64Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_Datetime64Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_Datetime64Field)) __PYX_ERR(4, 171, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_Timedelta64Field = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "Timedelta64Field", sizeof(struct __pyx_obj_4mars_9serialize_4core_Timedelta64Field), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_Timedelta64Field) __PYX_ERR(4, 175, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_Timedelta64Field = (struct __pyx_vtabstruct_4mars_9serialize_4core_Timedelta64Field*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_Timedelta64Field->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_Timedelta64Field)) __PYX_ERR(4, 175, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_DataTypeField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "DataTypeField", sizeof(struct __pyx_obj_4mars_9serialize_4core_DataTypeField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_DataTypeField) __PYX_ERR(4, 179, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_DataTypeField = (struct __pyx_vtabstruct_4mars_9serialize_4core_DataTypeField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_DataTypeField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_DataTypeField)) __PYX_ERR(4, 179, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_ListField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "ListField", sizeof(struct __pyx_obj_4mars_9serialize_4core_ListField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_ListField) __PYX_ERR(4, 183, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_ListField = (struct __pyx_vtabstruct_4mars_9serialize_4core_ListField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_ListField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_ListField)) __PYX_ERR(4, 183, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_TupleField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "TupleField", sizeof(struct __pyx_obj_4mars_9serialize_4core_TupleField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_TupleField) __PYX_ERR(4, 187, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_TupleField = (struct __pyx_vtabstruct_4mars_9serialize_4core_TupleField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_TupleField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_TupleField)) __PYX_ERR(4, 187, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_DictField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "DictField", sizeof(struct __pyx_obj_4mars_9serialize_4core_DictField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_DictField) __PYX_ERR(4, 191, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_DictField = (struct __pyx_vtabstruct_4mars_9serialize_4core_DictField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_DictField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_DictField)) __PYX_ERR(4, 191, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_ReferenceField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "ReferenceField", sizeof(struct __pyx_obj_4mars_9serialize_4core_ReferenceField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_ReferenceField) __PYX_ERR(4, 195, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_ReferenceField = (struct __pyx_vtabstruct_4mars_9serialize_4core_ReferenceField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_ReferenceField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_ReferenceField)) __PYX_ERR(4, 195, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_OneOfField = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "OneOfField", sizeof(struct __pyx_obj_4mars_9serialize_4core_OneOfField), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_OneOfField) __PYX_ERR(4, 199, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_OneOfField = (struct __pyx_vtabstruct_4mars_9serialize_4core_OneOfField*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_OneOfField->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_OneOfField)) __PYX_ERR(4, 199, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_KeyPlaceholder = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "KeyPlaceholder", sizeof(struct __pyx_obj_4mars_9serialize_4core_KeyPlaceholder), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_KeyPlaceholder) __PYX_ERR(4, 203, __pyx_L1_error)
  __pyx_ptype_4mars_9serialize_4core_AttrWrapper = __Pyx_ImportType(__pyx_t_1, "mars.serialize.core", "AttrWrapper", sizeof(struct __pyx_obj_4mars_9serialize_4core_AttrWrapper), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_9serialize_4core_AttrWrapper) __PYX_ERR(4, 208, __pyx_L1_error)
  __pyx_vtabptr_4mars_9serialize_4core_AttrWrapper = (struct __pyx_vtabstruct_4mars_9serialize_4core_AttrWrapper*)__Pyx_GetVtable(__pyx_ptype_4mars_9serialize_4core_AttrWrapper->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_9serialize_4core_AttrWrapper)) __PYX_ERR(4, 208, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("mars.utils_c"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "to_str", (void (**)(void))&__pyx_f_4mars_7utils_c_to_str, "PyObject *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_7utils_c_to_str *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initjsonserializer(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initjsonserializer(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_jsonserializer(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_jsonserializer(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_jsonserializer(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'jsonserializer' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_jsonserializer(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("jsonserializer", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_mars__serialize__jsonserializer) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "mars.serialize.jsonserializer")) {
      if (unlikely(PyDict_SetItemString(modules, "mars.serialize.jsonserializer", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() != 0)) goto __pyx_L1_error;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "mars/serialize/jsonserializer.pyx":17
 * # limitations under the License.
 * 
 * import base64             # <<<<<<<<<<<<<<
 * import pickle
 * import weakref
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_base64, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_base64, __pyx_t_1) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":18
 * 
 * import base64
 * import pickle             # <<<<<<<<<<<<<<
 * import weakref
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pickle, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pickle, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":19
 * import base64
 * import pickle
 * import weakref             # <<<<<<<<<<<<<<
 * 
 * import numpy as np
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_weakref, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_weakref, __pyx_t_1) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":21
 * import weakref
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * from cpython.version cimport PY_MAJOR_VERSION
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":25
 * from cpython.version cimport PY_MAJOR_VERSION
 * 
 * from ..compat import six, OrderedDict, izip             # <<<<<<<<<<<<<<
 * from ..core import BaseWithKey
 * from .core cimport Provider, ValueType, ProviderType, \
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_six);
  __Pyx_GIVEREF(__pyx_n_s_six);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_six);
  __Pyx_INCREF(__pyx_n_s_OrderedDict);
  __Pyx_GIVEREF(__pyx_n_s_OrderedDict);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_OrderedDict);
  __Pyx_INCREF(__pyx_n_s_izip);
  __Pyx_GIVEREF(__pyx_n_s_izip);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_izip);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_compat, __pyx_t_1, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_six); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_six, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderedDict, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_izip); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_izip, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":26
 * 
 * from ..compat import six, OrderedDict, izip
 * from ..core import BaseWithKey             # <<<<<<<<<<<<<<
 * from .core cimport Provider, ValueType, ProviderType, \
 *     Field, List, Tuple, Dict, Identity, Reference, OneOf, KeyPlaceholder, \
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_BaseWithKey);
  __Pyx_GIVEREF(__pyx_n_s_BaseWithKey);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_BaseWithKey);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_core, __pyx_t_2, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_BaseWithKey); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BaseWithKey, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":34
 * 
 * cdef dict PRIMITIVE_TYPE_TO_NAME = {
 *     ValueType.bool: 'bool',             # <<<<<<<<<<<<<<
 *     ValueType.int8: 'int8',
 *     ValueType.int16: 'int16',
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_bool); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_bool) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":35
 * cdef dict PRIMITIVE_TYPE_TO_NAME = {
 *     ValueType.bool: 'bool',
 *     ValueType.int8: 'int8',             # <<<<<<<<<<<<<<
 *     ValueType.int16: 'int16',
 *     ValueType.int32: 'int32',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_int8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_int8) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":36
 *     ValueType.bool: 'bool',
 *     ValueType.int8: 'int8',
 *     ValueType.int16: 'int16',             # <<<<<<<<<<<<<<
 *     ValueType.int32: 'int32',
 *     ValueType.int64: 'int64',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_int16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_int16) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":37
 *     ValueType.int8: 'int8',
 *     ValueType.int16: 'int16',
 *     ValueType.int32: 'int32',             # <<<<<<<<<<<<<<
 *     ValueType.int64: 'int64',
 *     ValueType.uint8: 'uint8',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_int32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_int32) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":38
 *     ValueType.int16: 'int16',
 *     ValueType.int32: 'int32',
 *     ValueType.int64: 'int64',             # <<<<<<<<<<<<<<
 *     ValueType.uint8: 'uint8',
 *     ValueType.uint16: 'uint16',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_int64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_int64) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":39
 *     ValueType.int32: 'int32',
 *     ValueType.int64: 'int64',
 *     ValueType.uint8: 'uint8',             # <<<<<<<<<<<<<<
 *     ValueType.uint16: 'uint16',
 *     ValueType.uint32: 'uint32',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_uint8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":40
 *     ValueType.int64: 'int64',
 *     ValueType.uint8: 'uint8',
 *     ValueType.uint16: 'uint16',             # <<<<<<<<<<<<<<
 *     ValueType.uint32: 'uint32',
 *     ValueType.uint64: 'uint64',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_uint16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_uint16) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":41
 *     ValueType.uint8: 'uint8',
 *     ValueType.uint16: 'uint16',
 *     ValueType.uint32: 'uint32',             # <<<<<<<<<<<<<<
 *     ValueType.uint64: 'uint64',
 *     ValueType.float16: 'float16',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_uint32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_uint32) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":42
 *     ValueType.uint16: 'uint16',
 *     ValueType.uint32: 'uint32',
 *     ValueType.uint64: 'uint64',             # <<<<<<<<<<<<<<
 *     ValueType.float16: 'float16',
 *     ValueType.float32: 'float32',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_uint64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_uint64) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":43
 *     ValueType.uint32: 'uint32',
 *     ValueType.uint64: 'uint64',
 *     ValueType.float16: 'float16',             # <<<<<<<<<<<<<<
 *     ValueType.float32: 'float32',
 *     ValueType.float64: 'float64',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_float16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_float16) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":44
 *     ValueType.uint64: 'uint64',
 *     ValueType.float16: 'float16',
 *     ValueType.float32: 'float32',             # <<<<<<<<<<<<<<
 *     ValueType.float64: 'float64',
 *     ValueType.bytes: 'bytes',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_float32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_float32) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":45
 *     ValueType.float16: 'float16',
 *     ValueType.float32: 'float32',
 *     ValueType.float64: 'float64',             # <<<<<<<<<<<<<<
 *     ValueType.bytes: 'bytes',
 *     ValueType.unicode: 'unicode',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_float64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_float64) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":46
 *     ValueType.float32: 'float32',
 *     ValueType.float64: 'float64',
 *     ValueType.bytes: 'bytes',             # <<<<<<<<<<<<<<
 *     ValueType.unicode: 'unicode',
 * }
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_bytes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_bytes) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":47
 *     ValueType.float64: 'float64',
 *     ValueType.bytes: 'bytes',
 *     ValueType.unicode: 'unicode',             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_unicode) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_4mars_9serialize_14jsonserializer_PRIMITIVE_TYPE_TO_NAME);
  __Pyx_DECREF_SET(__pyx_v_4mars_9serialize_14jsonserializer_PRIMITIVE_TYPE_TO_NAME, ((PyObject*)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":51
 * 
 * cdef dict EXTEND_TYPE_TO_NAME = {
 *     ValueType.slice: 'slice',             # <<<<<<<<<<<<<<
 *     ValueType.arr: 'arr',
 *     ValueType.dtype: 'dtype',
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_slice); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_slice) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":52
 * cdef dict EXTEND_TYPE_TO_NAME = {
 *     ValueType.slice: 'slice',
 *     ValueType.arr: 'arr',             # <<<<<<<<<<<<<<
 *     ValueType.dtype: 'dtype',
 *     ValueType.key: 'key',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_arr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_arr) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":53
 *     ValueType.slice: 'slice',
 *     ValueType.arr: 'arr',
 *     ValueType.dtype: 'dtype',             # <<<<<<<<<<<<<<
 *     ValueType.key: 'key',
 *     ValueType.datetime64: 'datetime64',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_dtype) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":54
 *     ValueType.arr: 'arr',
 *     ValueType.dtype: 'dtype',
 *     ValueType.key: 'key',             # <<<<<<<<<<<<<<
 *     ValueType.datetime64: 'datetime64',
 *     ValueType.timedelta64: 'timedelta64',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_key) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":55
 *     ValueType.dtype: 'dtype',
 *     ValueType.key: 'key',
 *     ValueType.datetime64: 'datetime64',             # <<<<<<<<<<<<<<
 *     ValueType.timedelta64: 'timedelta64',
 * }
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_datetime64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_datetime64) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/serialize/jsonserializer.pyx":56
 *     ValueType.key: 'key',
 *     ValueType.datetime64: 'datetime64',
 *     ValueType.timedelta64: 'timedelta64',             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_9serialize_4core_ValueType), __pyx_n_s_timedelta64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_n_s_timedelta64) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_4mars_9serialize_14jsonserializer_EXTEND_TYPE_TO_NAME);
  __Pyx_DECREF_SET(__pyx_v_4mars_9serialize_14jsonserializer_EXTEND_TYPE_TO_NAME, ((PyObject*)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_JsonSerializeProvider(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4mars_9serialize_14jsonserializer_1__pyx_unpickle_JsonSerializeProvider, NULL, __pyx_n_s_mars_serialize_jsonserializer); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_JsonSerializeProv, __pyx_t_1) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/serialize/jsonserializer.pyx":1
 * #!/usr/bin/env python             # <<<<<<<<<<<<<<
 * # -*- coding: utf-8 -*-
 * # Copyright 1999-2018 Alibaba Group Holding Ltd.
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init mars.serialize.jsonserializer", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init mars.serialize.jsonserializer");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* CallNextTpDealloc */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_dealloc != current_tp_dealloc)
        type = type->tp_base;
    while (type && type->tp_dealloc == current_tp_dealloc)
        type = type->tp_base;
    if (type)
        type->tp_dealloc(obj);
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_4mars_9serialize_4core_ProviderType(enum __pyx_t_4mars_9serialize_4core_ProviderType value) {
    const enum __pyx_t_4mars_9serialize_4core_ProviderType neg_one = (enum __pyx_t_4mars_9serialize_4core_ProviderType) ((enum __pyx_t_4mars_9serialize_4core_ProviderType) 0 - (enum __pyx_t_4mars_9serialize_4core_ProviderType) 1), const_zero = (enum __pyx_t_4mars_9serialize_4core_ProviderType) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_4mars_9serialize_4core_ProviderType) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_4mars_9serialize_4core_ProviderType) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_4mars_9serialize_4core_ProviderType) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_4mars_9serialize_4core_ProviderType) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_4mars_9serialize_4core_ProviderType) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_4mars_9serialize_4core_ProviderType),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0, -1);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0, -1);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) ((enum NPY_TYPES) 0 - (enum NPY_TYPES) 1), const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
