/* Generated by Cython 0.29.2 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "C:\\Python36-x64\\lib\\site-packages\\numpy\\core\\include\\numpy\\arrayobject.h",
            "C:\\Python36-x64\\lib\\site-packages\\numpy\\core\\include\\numpy\\ufuncobject.h"
        ],
        "extra_compile_args": [
            "/Ot",
            "/IC:\\projects\\mars\\misc"
        ],
        "include_dirs": [
            "C:\\Python36-x64\\lib\\site-packages\\numpy\\core\\include"
        ],
        "name": "mars.actors.pool.messages",
        "sources": [
            "mars/actors/pool/messages.pyx"
        ]
    },
    "module_name": "mars.actors.pool.messages"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_2"
#define CYTHON_HEX_VERSION 0x001D02F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_USE_DICT_VERSIONS
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
        static PY_UINT64_T __pyx_dict_version = 0;\
        static PyObject *__pyx_dict_cached_value = NULL;\
        if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
            (VAR) = __pyx_dict_cached_value;\
        } else {\
            (VAR) = __pyx_dict_cached_value = (LOOKUP);\
            __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
        }\
    }
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0; // PyThread_create_key reports success always
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif // TSS (Thread Specific Storage) API
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__mars__actors__pool__messages
#define __PYX_HAVE_API__mars__actors__pool__messages
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "mars\\actors\\pool\\messages.pyx",
  "stringsource",
  "__init__.pxd",
  "type.pxd",
  "mars\\actors\\core.pxd",
};

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":776
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":778
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":779
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":784
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":785
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":786
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":791
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":800
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":801
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":804
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":805
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":806
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":808
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":809
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":811
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":812
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "mars/actors/pool/messages.pxd":22
 * 
 * 
 * ctypedef np.int32_t INT32_t             # <<<<<<<<<<<<<<
 * ctypedef np.uint8_t BYTE_t
 * 
 */
typedef __pyx_t_5numpy_int32_t __pyx_t_4mars_6actors_4pool_8messages_INT32_t;

/* "mars/actors/pool/messages.pxd":23
 * 
 * ctypedef np.int32_t INT32_t
 * ctypedef np.uint8_t BYTE_t             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef __pyx_t_5numpy_uint8_t __pyx_t_4mars_6actors_4pool_8messages_BYTE_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_4mars_6actors_4core_ActorRef;
struct __pyx_obj_4mars_6actors_4core_Actor;
struct __pyx_obj_4mars_6actors_4core_FunctionActor;
struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE;
struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE;
struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE;
struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE;
struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE;
struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE;
struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message;
struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr;
struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message;
struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr;
struct __pyx_obj___Pyx_EnumMeta;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":815
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":816
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":817
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":819
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_4mars_6actors_4core_8ActorRef_send;
struct __pyx_opt_args_4mars_6actors_4core_8ActorRef_tell;
struct __pyx_opt_args_4mars_6actors_4core_8ActorRef_destroy;
struct __pyx_opt_args_4mars_6actors_4core_create_actor_pool;
struct __pyx_opt_args_4mars_6actors_4core_new_client;

/* "core.pxd":23
 *     cdef public object _ctx
 * 
 *     cpdef object send(self, object message, bint wait=*, object callback=*)             # <<<<<<<<<<<<<<
 *     cpdef object tell(self, object message, object delay=*, bint wait=*,
 *                       object callback=*)
 */
struct __pyx_opt_args_4mars_6actors_4core_8ActorRef_send {
  int __pyx_n;
  int wait;
  PyObject *callback;
};

/* "core.pxd":24
 * 
 *     cpdef object send(self, object message, bint wait=*, object callback=*)
 *     cpdef object tell(self, object message, object delay=*, bint wait=*,             # <<<<<<<<<<<<<<
 *                       object callback=*)
 *     cpdef object destroy(self, bint wait=*, object callback=*)
 */
struct __pyx_opt_args_4mars_6actors_4core_8ActorRef_tell {
  int __pyx_n;
  PyObject *delay;
  int wait;
  PyObject *callback;
};

/* "core.pxd":26
 *     cpdef object tell(self, object message, object delay=*, bint wait=*,
 *                       object callback=*)
 *     cpdef object destroy(self, bint wait=*, object callback=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_4mars_6actors_4core_8ActorRef_destroy {
  int __pyx_n;
  int wait;
  PyObject *callback;
};

/* "core.pxd":44
 * 
 * 
 * cpdef object create_actor_pool(str address=*, int n_process=*, object distributor=*,             # <<<<<<<<<<<<<<
 *                                object parallel=*, str backend=*)
 * cpdef object new_client(object parallel=*, str backend=*)
 */
struct __pyx_opt_args_4mars_6actors_4core_create_actor_pool {
  int __pyx_n;
  PyObject *address;
  int n_process;
  PyObject *distributor;
  PyObject *parallel;
  PyObject *backend;
};

/* "core.pxd":46
 * cpdef object create_actor_pool(str address=*, int n_process=*, object distributor=*,
 *                                object parallel=*, str backend=*)
 * cpdef object new_client(object parallel=*, str backend=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_4mars_6actors_4core_new_client {
  int __pyx_n;
  PyObject *parallel;
  PyObject *backend;
};
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_send_message;
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_tell_message;
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_create_actor_message;
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_destroy_actor_message;
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_has_actor_message;
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_result_message;
struct __pyx_opt_args_4mars_6actors_4pool_8messages_unpack_result_message;
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_error_message;
struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_message_type;
struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_send_message;
struct __pyx_opt_args_4mars_6actors_4pool_8messages__unpack_send_message;

/* "mars/actors/pool/messages.pxd":26
 * 
 * 
 * cpdef enum MessageType:             # <<<<<<<<<<<<<<
 *     create_actor = 0
 *     destroy_actor = 1
 */
enum __pyx_t_4mars_6actors_4pool_8messages_MessageType {
  __pyx_e_4mars_6actors_4pool_8messages_create_actor = 0,
  __pyx_e_4mars_6actors_4pool_8messages_destroy_actor = 1,
  __pyx_e_4mars_6actors_4pool_8messages_has_actor = 2,
  __pyx_e_4mars_6actors_4pool_8messages_result = 3,
  __pyx_e_4mars_6actors_4pool_8messages_error = 4,
  __pyx_e_4mars_6actors_4pool_8messages_send_all = 5,
  __pyx_e_4mars_6actors_4pool_8messages_tell_all = 6,
  __pyx_e_4mars_6actors_4pool_8messages_send_chunk_start = 7,
  __pyx_e_4mars_6actors_4pool_8messages_send_chunk = 8,
  __pyx_e_4mars_6actors_4pool_8messages_send_chunk_end = 9,
  __pyx_e_4mars_6actors_4pool_8messages_tell_chunk_start = 10,
  __pyx_e_4mars_6actors_4pool_8messages_tell_chunk = 11,
  __pyx_e_4mars_6actors_4pool_8messages_tell_chunk_end = 12
};

/* "mars/actors/pool/messages.pxd":48
 * cpdef void unpack_actor_ref(bytes binary, ActorRef actor_ref)
 * cpdef object unpack_send_message(bytes message)
 * cpdef object pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                object write=*, int protocol=*)
 * cpdef object unpack_send_message(bytes message)
 */
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_send_message {
  int __pyx_n;
  PyObject *write;
  int protocol;
};

/* "mars/actors/pool/messages.pxd":51
 *                                object write=*, int protocol=*)
 * cpdef object unpack_send_message(bytes message)
 * cpdef object pack_tell_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                object write=*, int protocol=*)
 * cpdef object unpack_tell_message(bytes message)
 */
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_tell_message {
  int __pyx_n;
  PyObject *write;
  int protocol;
};

/* "mars/actors/pool/messages.pxd":55
 * cpdef object unpack_tell_message(bytes message)
 * cpdef tuple get_index(bytes binary, object calc_from_uid)
 * cpdef object pack_create_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref, object actor_cls,             # <<<<<<<<<<<<<<
 *                                        tuple args, dict kw, object write=*, int protocol=*)
 * cpdef object unpack_create_actor_message(bytes binary)
 */
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_create_actor_message {
  int __pyx_n;
  PyObject *write;
  int protocol;
};

/* "mars/actors/pool/messages.pxd":58
 *                                        tuple args, dict kw, object write=*, int protocol=*)
 * cpdef object unpack_create_actor_message(bytes binary)
 * cpdef object pack_destroy_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,             # <<<<<<<<<<<<<<
 *                                         object write=*, int protocol=*)
 * cpdef object unpack_destroy_actor_message(bytes binary)
 */
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_destroy_actor_message {
  int __pyx_n;
  PyObject *write;
  int protocol;
};

/* "mars/actors/pool/messages.pxd":61
 *                                         object write=*, int protocol=*)
 * cpdef object unpack_destroy_actor_message(bytes binary)
 * cpdef object pack_has_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref,             # <<<<<<<<<<<<<<
 *                                     object write=*, int protocol=*)
 * cpdef object unpack_has_actor_message(bytes binary)
 */
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_has_actor_message {
  int __pyx_n;
  PyObject *write;
  int protocol;
};

/* "mars/actors/pool/messages.pxd":64
 *                                     object write=*, int protocol=*)
 * cpdef object unpack_has_actor_message(bytes binary)
 * cpdef object pack_result_message(bytes message_id, INT32_t from_index, int to_index, object result,             # <<<<<<<<<<<<<<
 *                                  object write=*, int protocol=*)
 * cpdef object unpack_result_message(bytes binary, object from_index=*, object to_index=*)
 */
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_result_message {
  int __pyx_n;
  PyObject *write;
  int protocol;
};

/* "mars/actors/pool/messages.pxd":66
 * cpdef object pack_result_message(bytes message_id, INT32_t from_index, int to_index, object result,
 *                                  object write=*, int protocol=*)
 * cpdef object unpack_result_message(bytes binary, object from_index=*, object to_index=*)             # <<<<<<<<<<<<<<
 * cpdef object pack_error_message(bytes message_id, INT32_t from_index, INT32_t to_index,
 *                                 object error_type, object error, object tb,
 */
struct __pyx_opt_args_4mars_6actors_4pool_8messages_unpack_result_message {
  int __pyx_n;
  PyObject *from_index;
  PyObject *to_index;
};

/* "mars/actors/pool/messages.pxd":67
 *                                  object write=*, int protocol=*)
 * cpdef object unpack_result_message(bytes binary, object from_index=*, object to_index=*)
 * cpdef object pack_error_message(bytes message_id, INT32_t from_index, INT32_t to_index,             # <<<<<<<<<<<<<<
 *                                 object error_type, object error, object tb,
 *                                 object write=*, int protocol=*)
 */
struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_error_message {
  int __pyx_n;
  PyObject *write;
  int protocol;
};

/* "mars/actors/pool/messages.pyx":50
 * 
 * 
 * cpdef enum MessageSerialType:             # <<<<<<<<<<<<<<
 *     # Internal message serialized type
 *     null = 0
 */
enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType {
  __pyx_e_4mars_6actors_4pool_8messages_null = 0,
  __pyx_e_4mars_6actors_4pool_8messages_raw_bytes = 1,
  __pyx_e_4mars_6actors_4pool_8messages_pickle = 2
};

/* "mars/actors/pool/messages.pyx":202
 * 
 * 
 * cdef inline void _pack_message_type(int message_type, bytearray buf, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef int value
 * 
 */
struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_message_type {
  int __pyx_n;
  int protocol;
};

/* "mars/actors/pool/messages.pyx":468
 * 
 * 
 * cdef inline object _pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                       bint send, object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytes message_id
 */
struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_send_message {
  int __pyx_n;
  PyObject *write;
  int protocol;
};

/* "mars/actors/pool/messages.pyx":494
 * 
 * 
 * cdef inline object _unpack_send_message(bytes binary, send=True):             # <<<<<<<<<<<<<<
 *     cdef object message_type
 *     cdef bytes message_id
 */
struct __pyx_opt_args_4mars_6actors_4pool_8messages__unpack_send_message {
  int __pyx_n;
  PyObject *send;
};

/* "core.pxd":18
 * 
 * 
 * cdef class ActorRef:             # <<<<<<<<<<<<<<
 *     cdef public str address
 *     cdef public object uid
 */
struct __pyx_obj_4mars_6actors_4core_ActorRef {
  PyObject_HEAD
  struct __pyx_vtabstruct_4mars_6actors_4core_ActorRef *__pyx_vtab;
  PyObject *address;
  PyObject *uid;
  PyObject *_ctx;
};


/* "core.pxd":29
 * 
 * 
 * cdef class Actor:             # <<<<<<<<<<<<<<
 *     cdef str _address
 *     cdef object _uid
 */
struct __pyx_obj_4mars_6actors_4core_Actor {
  PyObject_HEAD
  struct __pyx_vtabstruct_4mars_6actors_4core_Actor *__pyx_vtab;
  PyObject *_address;
  PyObject *_uid;
  PyObject *_ctx;
};


/* "core.pxd":40
 * 
 * 
 * cdef class FunctionActor(Actor):             # <<<<<<<<<<<<<<
 *     cpdef on_receive(self, message)
 * 
 */
struct __pyx_obj_4mars_6actors_4core_FunctionActor {
  struct __pyx_obj_4mars_6actors_4core_Actor __pyx_base;
};


/* "mars/actors/pool/messages.pyx":65
 * 
 * 
 * cdef class CREATE_ACTOR_MESSAGE:             # <<<<<<<<<<<<<<
 *     cdef public int message_type
 *     cdef public bytes message_id
 */
struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE {
  PyObject_HEAD
  int message_type;
  PyObject *message_id;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *actor_ref;
  PyObject *actor_cls;
  PyObject *args;
  PyObject *kwargs;
};


/* "mars/actors/pool/messages.pyx":88
 * 
 * 
 * cdef class DESTROY_ACTOR_MESSAGE:             # <<<<<<<<<<<<<<
 *     cdef public int message_type
 *     cdef public bytes message_id
 */
struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE {
  PyObject_HEAD
  int message_type;
  PyObject *message_id;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *actor_ref;
};


/* "mars/actors/pool/messages.pyx":104
 * 
 * 
 * cdef class HAS_ACTOR_MESSAGE:             # <<<<<<<<<<<<<<
 *     cdef public int message_type
 *     cdef public bytes message_id
 */
struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE {
  PyObject_HEAD
  int message_type;
  PyObject *message_id;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *actor_ref;
};


/* "mars/actors/pool/messages.pyx":120
 * 
 * 
 * cdef class RESULT_MESSAGE:             # <<<<<<<<<<<<<<
 *     cdef public int message_type
 *     cdef public bytes message_id
 */
struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE {
  PyObject_HEAD
  int message_type;
  PyObject *message_id;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t to_index;
  PyObject *result;
};


/* "mars/actors/pool/messages.pyx":136
 * 
 * 
 * cdef class ERROR_MESSAGE:             # <<<<<<<<<<<<<<
 *     cdef public int message_type
 *     cdef public bytes message_id
 */
struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE {
  PyObject_HEAD
  int message_type;
  PyObject *message_id;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t to_index;
  PyObject *error_type;
  PyObject *error;
  PyObject *traceback;
};


/* "mars/actors/pool/messages.pyx":157
 * 
 * 
 * cdef class SEND_MESSAGE:             # <<<<<<<<<<<<<<
 *     cdef public object message_type
 *     cdef public bytes message_id
 */
struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE {
  PyObject_HEAD
  PyObject *message_type;
  PyObject *message_id;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *actor_ref;
  PyObject *message;
};


/* "mars/actors/pool/messages.pyx":439
 * 
 * 
 * cdef inline tuple _unpack_tuple_message(bytes buf, size_t size, size_t* pos):             # <<<<<<<<<<<<<<
 *     return tuple(_unpack_sole_message(buf, pos) for _ in range(size))
 * 
 */
struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message {
  PyObject_HEAD
  PyObject *__pyx_v_buf;
  size_t *__pyx_v_pos;
  size_t __pyx_v_size;
};


/* "mars/actors/pool/messages.pyx":440
 * 
 * cdef inline tuple _unpack_tuple_message(bytes buf, size_t size, size_t* pos):
 *     return tuple(_unpack_sole_message(buf, pos) for _ in range(size))             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message *__pyx_outer_scope;
  size_t __pyx_v__;
  size_t __pyx_t_0;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
};


/* "mars/actors/pool/messages.pyx":816
 * 
 * 
 * def write_remote_message(write_func, *binary):             # <<<<<<<<<<<<<<
 *     cdef bytes size
 * 
 */
struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message {
  PyObject_HEAD
  PyObject *__pyx_v_binary;
};


/* "mars/actors/pool/messages.pyx":819
 *     cdef bytes size
 * 
 *     size = _pack_long(sum(len(b) for b in binary))             # <<<<<<<<<<<<<<
 *     write_func(size)
 *     [write_func(b) for b in binary]
 */
struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr {
  PyObject_HEAD
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *__pyx_outer_scope;
  PyObject *__pyx_v_b;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "EnumBase":15
 * 
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):             # <<<<<<<<<<<<<<
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)
 */
struct __pyx_obj___Pyx_EnumMeta {
  PyHeapTypeObject __pyx_base;
};



/* "core.pxd":18
 * 
 * 
 * cdef class ActorRef:             # <<<<<<<<<<<<<<
 *     cdef public str address
 *     cdef public object uid
 */

struct __pyx_vtabstruct_4mars_6actors_4core_ActorRef {
  PyObject *(*send)(struct __pyx_obj_4mars_6actors_4core_ActorRef *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4core_8ActorRef_send *__pyx_optional_args);
  PyObject *(*tell)(struct __pyx_obj_4mars_6actors_4core_ActorRef *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4core_8ActorRef_tell *__pyx_optional_args);
  PyObject *(*destroy)(struct __pyx_obj_4mars_6actors_4core_ActorRef *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4core_8ActorRef_destroy *__pyx_optional_args);
};
static struct __pyx_vtabstruct_4mars_6actors_4core_ActorRef *__pyx_vtabptr_4mars_6actors_4core_ActorRef;


/* "core.pxd":29
 * 
 * 
 * cdef class Actor:             # <<<<<<<<<<<<<<
 *     cdef str _address
 *     cdef object _uid
 */

struct __pyx_vtabstruct_4mars_6actors_4core_Actor {
  struct __pyx_obj_4mars_6actors_4core_ActorRef *(*ref)(struct __pyx_obj_4mars_6actors_4core_Actor *, int __pyx_skip_dispatch);
  PyObject *(*post_create)(struct __pyx_obj_4mars_6actors_4core_Actor *, int __pyx_skip_dispatch);
  PyObject *(*on_receive)(struct __pyx_obj_4mars_6actors_4core_Actor *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*pre_destroy)(struct __pyx_obj_4mars_6actors_4core_Actor *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4mars_6actors_4core_Actor *__pyx_vtabptr_4mars_6actors_4core_Actor;


/* "core.pxd":40
 * 
 * 
 * cdef class FunctionActor(Actor):             # <<<<<<<<<<<<<<
 *     cpdef on_receive(self, message)
 * 
 */

struct __pyx_vtabstruct_4mars_6actors_4core_FunctionActor {
  struct __pyx_vtabstruct_4mars_6actors_4core_Actor __pyx_base;
};
static struct __pyx_vtabstruct_4mars_6actors_4core_FunctionActor *__pyx_vtabptr_4mars_6actors_4core_FunctionActor;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int32(npy_int32 value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint8(npy_uint8 value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(enum __pyx_t_4mars_6actors_4pool_8messages_MessageType value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_int32 __Pyx_PyInt_As_npy_int32(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_uint8 __Pyx_PyInt_As_npy_uint8(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageSerialType(enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType value);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'mars.actors.core' */
static PyTypeObject *__pyx_ptype_4mars_6actors_4core_ActorRef = 0;
static PyTypeObject *__pyx_ptype_4mars_6actors_4core_Actor = 0;
static PyTypeObject *__pyx_ptype_4mars_6actors_4core_FunctionActor = 0;

/* Module declarations from 'cython' */

/* Module declarations from 'cpython.version' */

/* Module declarations from 'mars.actors.pool.messages' */
static PyTypeObject *__pyx_ptype_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE = 0;
static PyTypeObject *__pyx_ptype_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE = 0;
static PyTypeObject *__pyx_ptype_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE = 0;
static PyTypeObject *__pyx_ptype_4mars_6actors_4pool_8messages_RESULT_MESSAGE = 0;
static PyTypeObject *__pyx_ptype_4mars_6actors_4pool_8messages_ERROR_MESSAGE = 0;
static PyTypeObject *__pyx_ptype_4mars_6actors_4pool_8messages_SEND_MESSAGE = 0;
static PyTypeObject *__pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message = 0;
static PyTypeObject *__pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message = 0;
static PyTypeObject *__pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr = 0;
static PyTypeObject *__pyx_ptype___Pyx_EnumMeta = 0;
static __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_4mars_6actors_4pool_8messages_CREATE_ACTOR;
static __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_4mars_6actors_4pool_8messages_DESTROY_ACTOR;
static __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_4mars_6actors_4pool_8messages_HAS_ACTOR;
static __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_4mars_6actors_4pool_8messages_RESULT;
static __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_4mars_6actors_4pool_8messages_ERROR;
static __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_4mars_6actors_4pool_8messages_SEND_ALL;
static __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_4mars_6actors_4pool_8messages_TELL_ALL;
static __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_4mars_6actors_4pool_8messages_NONE;
static __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_4mars_6actors_4pool_8messages_RAW_BYTES;
static __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_4mars_6actors_4pool_8messages_PICKLE;
static int __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;
static PyObject *__pyx_v_4mars_6actors_4pool_8messages__pack_byte = 0;
static PyObject *__pyx_v_4mars_6actors_4pool_8messages__unpack_byte = 0;
static PyObject *__pyx_v_4mars_6actors_4pool_8messages__pack_int = 0;
static PyObject *__pyx_v_4mars_6actors_4pool_8messages__unpack_int = 0;
static PyObject *__pyx_v_4mars_6actors_4pool_8messages__pack_long = 0;
static PyObject *__pyx_v_4mars_6actors_4pool_8messages__unpack_long = 0;
static PyObject *__pyx_v_4mars_6actors_4pool_8messages__pack_short = 0;
static PyObject *__pyx_v_4mars_6actors_4pool_8messages__unpack_short = 0;
static PyObject *__Pyx_OrderedDict = 0;
static PyObject *__Pyx_EnumBase = 0;
static PyObject *__Pyx_globals = 0;
static int __pyx_f_4mars_6actors_4pool_8messages_unpack_message_type_value(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_message_type(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_message_id(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_actor_ref(struct __pyx_obj_4mars_6actors_4core_ActorRef *, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_4mars_6actors_4pool_8messages_unpack_actor_ref(PyObject *, struct __pyx_obj_4mars_6actors_4core_ActorRef *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_send_message(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_send_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, struct __pyx_obj_4mars_6actors_4core_ActorRef *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_send_message *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_tell_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, struct __pyx_obj_4mars_6actors_4core_ActorRef *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_tell_message *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_tell_message(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_get_index(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_create_actor_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t, int, struct __pyx_obj_4mars_6actors_4core_ActorRef *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_create_actor_message *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_create_actor_message(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_destroy_actor_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, struct __pyx_obj_4mars_6actors_4core_ActorRef *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_destroy_actor_message *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_destroy_actor_message(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_has_actor_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t, int, struct __pyx_obj_4mars_6actors_4core_ActorRef *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_has_actor_message *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_has_actor_message(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_result_message(PyObject *, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, int, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_result_message *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_result_message(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_unpack_result_message *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_error_message(PyObject *, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_error_message *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_error_message(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_read_remote_message(PyObject *, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages_new_message_id(void); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__read_bytes(PyObject *, size_t *, size_t); /*proto*/
static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_message_type(int, PyObject *, struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_message_type *__pyx_optional_args); /*proto*/
static CYTHON_INLINE int __pyx_f_4mars_6actors_4pool_8messages__unpack_message_type_value(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_message_type(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__skip_message_type(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_message_id(PyObject *, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_message_id(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__skip_message_id(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_object(PyObject *, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_object(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__skip_object(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_index(int, PyObject *); /*proto*/
static CYTHON_INLINE __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_f_4mars_6actors_4pool_8messages__unpack_index(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__skip_index(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_actor_ref(struct __pyx_obj_4mars_6actors_4core_ActorRef *, PyObject *); /*proto*/
static CYTHON_INLINE struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_f_4mars_6actors_4pool_8messages__unpack_actor_ref(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__skip_address(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_uid(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_message_size(size_t, PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_4mars_6actors_4pool_8messages__unpack_message_size(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__pack_tuple_message(PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__pack_sole_message(PyObject *, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__pack_message(PyObject *, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message(PyObject *, size_t, size_t *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_sole_message(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_message(PyObject *, size_t *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__pack_send_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, struct __pyx_obj_4mars_6actors_4core_ActorRef *, PyObject *, int, struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_send_message *__pyx_optional_args); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_send_message(PyObject *, struct __pyx_opt_args_4mars_6actors_4pool_8messages__unpack_send_message *__pyx_optional_args); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__wrap_read_func(PyObject *, size_t); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *, PyObject *); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *, PyObject *); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *, PyObject *); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_RESULT_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *, PyObject *); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_ERROR_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *, PyObject *); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_SEND_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *, PyObject *); /*proto*/
static PyObject *__pyx_unpickle___Pyx_EnumMeta__set_state(struct __pyx_obj___Pyx_EnumMeta *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "mars.actors.pool.messages"
extern int __pyx_module_is_main_mars__actors__pool__messages;
int __pyx_module_is_main_mars__actors__pool__messages = 0;

/* Implementation of 'mars.actors.pool.messages' */
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_B[] = "<B";
static const char __pyx_k_I[] = "<I";
static const char __pyx_k_Q[] = "<Q";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_h[] = "<h";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_ex[] = "ex";
static const char __pyx_k_kw[] = "kw";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_tb[] = "tb";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_dct[] = "dct";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_s_s[] = "%s.%s";
static const char __pyx_k_str[] = "__str__";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_null[] = "null";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_EPIPE[] = "EPIPE";
static const char __pyx_k_bytes[] = "bytes";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_dumps[] = "dumps";
static const char __pyx_k_errno[] = "errno";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_loads[] = "loads";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_s_s_d[] = "<%s.%s: %d>";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_Struct[] = "Struct";
static const char __pyx_k_binary[] = "binary";
static const char __pyx_k_compat[] = "compat";
static const char __pyx_k_extend[] = "extend";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_socket[] = "socket";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_IntEnum[] = "IntEnum";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_install[] = "install";
static const char __pyx_k_members[] = "__members__";
static const char __pyx_k_message[] = "message";
static const char __pyx_k_parents[] = "parents";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_EnumBase[] = "EnumBase";
static const char __pyx_k_EnumType[] = "EnumType";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_protocol[] = "protocol";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_send_all[] = "send_all";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_tell_all[] = "tell_all";
static const char __pyx_k_to_index[] = "to_index";
static const char __pyx_k_actor_cls[] = "actor_cls";
static const char __pyx_k_actor_ref[] = "actor_ref";
static const char __pyx_k_has_actor[] = "has_actor";
static const char __pyx_k_lib_tblib[] = "lib.tblib";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_raw_bytes[] = "raw_bytes";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_traceback[] = "traceback";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_error_type[] = "error_type";
static const char __pyx_k_from_index[] = "from_index";
static const char __pyx_k_message_id[] = "message_id";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_send_chunk[] = "send_chunk";
static const char __pyx_k_tell_chunk[] = "tell_chunk";
static const char __pyx_k_write_func[] = "write_func";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MessageType[] = "MessageType";
static const char __pyx_k_OrderedDict[] = "OrderedDict";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_Pyx_EnumBase[] = "__Pyx_EnumBase";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_SEND_MESSAGE[] = "SEND_MESSAGE";
static const char __pyx_k_create_actor[] = "create_actor";
static const char __pyx_k_message_type[] = "message_type";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_ERROR_MESSAGE[] = "ERROR_MESSAGE";
static const char __pyx_k_calc_from_uid[] = "calc_from_uid";
static const char __pyx_k_destroy_actor[] = "destroy_actor";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_RESULT_MESSAGE[] = "RESULT_MESSAGE";
static const char __pyx_k_send_chunk_end[] = "send_chunk_end";
static const char __pyx_k_tell_chunk_end[] = "tell_chunk_end";
static const char __pyx_k_BrokenPipeError[] = "BrokenPipeError";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_pickling_support[] = "pickling_support";
static const char __pyx_k_send_chunk_start[] = "send_chunk_start";
static const char __pyx_k_tell_chunk_start[] = "tell_chunk_start";
static const char __pyx_k_HAS_ACTOR_MESSAGE[] = "HAS_ACTOR_MESSAGE";
static const char __pyx_k_MessageSerialType[] = "MessageSerialType";
static const char __pyx_k_Pyx_EnumBase___new[] = "__Pyx_EnumBase.__new__";
static const char __pyx_k_Pyx_EnumBase___str[] = "__Pyx_EnumBase.__str__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_Pyx_EnumBase___repr[] = "__Pyx_EnumBase.__repr__";
static const char __pyx_k_CREATE_ACTOR_MESSAGE[] = "CREATE_ACTOR_MESSAGE";
static const char __pyx_k_ConnectionResetError[] = "ConnectionResetError";
static const char __pyx_k_Unknown_enum_value_s[] = "Unknown enum value: '%s'";
static const char __pyx_k_Unsupported_protocol[] = "Unsupported protocol";
static const char __pyx_k_write_remote_message[] = "_write_remote_message";
static const char __pyx_k_DESTROY_ACTOR_MESSAGE[] = "DESTROY_ACTOR_MESSAGE";
static const char __pyx_k_write_remote_message_2[] = "write_remote_message";
static const char __pyx_k_mars_actors_pool_messages[] = "mars.actors.pool.messages";
static const char __pyx_k_pyx_unpickle_SEND_MESSAGE[] = "__pyx_unpickle_SEND_MESSAGE";
static const char __pyx_k_pyx_unpickle_ERROR_MESSAGE[] = "__pyx_unpickle_ERROR_MESSAGE";
static const char __pyx_k_The_remote_server_is_closed[] = "The remote server is closed";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_pyx_unpickle_RESULT_MESSAGE[] = "__pyx_unpickle_RESULT_MESSAGE";
static const char __pyx_k_pyx_unpickle___Pyx_EnumMeta[] = "__pyx_unpickle___Pyx_EnumMeta";
static const char __pyx_k_mars_actors_pool_messages_pyx[] = "mars\\actors\\pool\\messages.pyx";
static const char __pyx_k_pyx_unpickle_CREATE_ACTOR_MESS[] = "__pyx_unpickle_CREATE_ACTOR_MESSAGE";
static const char __pyx_k_pyx_unpickle_DESTROY_ACTOR_MES[] = "__pyx_unpickle_DESTROY_ACTOR_MESSAGE";
static const char __pyx_k_pyx_unpickle_HAS_ACTOR_MESSAGE[] = "__pyx_unpickle_HAS_ACTOR_MESSAGE";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_unpack_tuple_message_locals_gen[] = "_unpack_tuple_message.<locals>.genexpr";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Incompatible_checksums_s_vs_0x04[] = "Incompatible checksums (%s vs 0x041d35d = (actor_ref, from_index, message, message_id, message_type, to_index))";
static const char __pyx_k_Incompatible_checksums_s_vs_0x2d[] = "Incompatible checksums (%s vs 0x2dc8e2e = (actor_cls, actor_ref, args, from_index, kwargs, message_id, message_type, to_index))";
static const char __pyx_k_Incompatible_checksums_s_vs_0x42[] = "Incompatible checksums (%s vs 0x42c8dac = (error, error_type, from_index, message_id, message_type, to_index, traceback))";
static const char __pyx_k_Incompatible_checksums_s_vs_0x79[] = "Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))";
static const char __pyx_k_Incompatible_checksums_s_vs_0xd4[] = "Incompatible checksums (%s vs 0xd41d8cd = ())";
static const char __pyx_k_Incompatible_checksums_s_vs_0xdd[] = "Incompatible checksums (%s vs 0xddcfada = (from_index, message_id, message_type, result, to_index))";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_write_remote_message_locals_gene[] = "write_remote_message.<locals>.genexpr";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static PyObject *__pyx_kp_s_B;
static PyObject *__pyx_n_s_BrokenPipeError;
static PyObject *__pyx_n_s_CREATE_ACTOR_MESSAGE;
static PyObject *__pyx_n_s_ConnectionResetError;
static PyObject *__pyx_n_s_DESTROY_ACTOR_MESSAGE;
static PyObject *__pyx_n_s_EPIPE;
static PyObject *__pyx_n_s_ERROR_MESSAGE;
static PyObject *__pyx_n_s_EnumBase;
static PyObject *__pyx_n_s_EnumType;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_n_s_HAS_ACTOR_MESSAGE;
static PyObject *__pyx_kp_s_I;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x04;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x2d;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x42;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x79;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xd4;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xdd;
static PyObject *__pyx_n_s_IntEnum;
static PyObject *__pyx_n_s_MessageSerialType;
static PyObject *__pyx_n_s_MessageType;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_OrderedDict;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_Pyx_EnumBase;
static PyObject *__pyx_n_s_Pyx_EnumBase___new;
static PyObject *__pyx_n_s_Pyx_EnumBase___repr;
static PyObject *__pyx_n_s_Pyx_EnumBase___str;
static PyObject *__pyx_kp_s_Q;
static PyObject *__pyx_n_s_RESULT_MESSAGE;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_SEND_MESSAGE;
static PyObject *__pyx_n_s_Struct;
static PyObject *__pyx_kp_s_The_remote_server_is_closed;
static PyObject *__pyx_kp_s_Unknown_enum_value_s;
static PyObject *__pyx_kp_s_Unsupported_protocol;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_actor_cls;
static PyObject *__pyx_n_s_actor_ref;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_n_s_binary;
static PyObject *__pyx_n_s_bytes;
static PyObject *__pyx_n_s_calc_from_uid;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cls;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_compat;
static PyObject *__pyx_n_s_create_actor;
static PyObject *__pyx_n_s_dct;
static PyObject *__pyx_n_s_destroy_actor;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dumps;
static PyObject *__pyx_n_s_enum;
static PyObject *__pyx_n_s_errno;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_error_type;
static PyObject *__pyx_n_s_ex;
static PyObject *__pyx_n_s_extend;
static PyObject *__pyx_n_s_from_index;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_h;
static PyObject *__pyx_n_s_has_actor;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_install;
static PyObject *__pyx_n_s_kw;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_lib_tblib;
static PyObject *__pyx_n_s_loads;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mars_actors_pool_messages;
static PyObject *__pyx_kp_s_mars_actors_pool_messages_pyx;
static PyObject *__pyx_n_s_members;
static PyObject *__pyx_n_s_message;
static PyObject *__pyx_n_s_message_id;
static PyObject *__pyx_n_s_message_type;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_null;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_parents;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pickling_support;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_protocol;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_CREATE_ACTOR_MESS;
static PyObject *__pyx_n_s_pyx_unpickle_DESTROY_ACTOR_MES;
static PyObject *__pyx_n_s_pyx_unpickle_ERROR_MESSAGE;
static PyObject *__pyx_n_s_pyx_unpickle_HAS_ACTOR_MESSAGE;
static PyObject *__pyx_n_s_pyx_unpickle_RESULT_MESSAGE;
static PyObject *__pyx_n_s_pyx_unpickle_SEND_MESSAGE;
static PyObject *__pyx_n_s_pyx_unpickle___Pyx_EnumMeta;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_random;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_raw_bytes;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_kp_s_s_s;
static PyObject *__pyx_kp_s_s_s_d;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_send_all;
static PyObject *__pyx_n_s_send_chunk;
static PyObject *__pyx_n_s_send_chunk_end;
static PyObject *__pyx_n_s_send_chunk_start;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_socket;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_tb;
static PyObject *__pyx_n_s_tell_all;
static PyObject *__pyx_n_s_tell_chunk;
static PyObject *__pyx_n_s_tell_chunk_end;
static PyObject *__pyx_n_s_tell_chunk_start;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_to_index;
static PyObject *__pyx_n_s_traceback;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_unpack_tuple_message_locals_gen;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_s_write_func;
static PyObject *__pyx_n_s_write_remote_message;
static PyObject *__pyx_n_s_write_remote_message_2;
static PyObject *__pyx_n_s_write_remote_message_locals_gene;
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, int __pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_actor_cls, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, int __pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, PyObject *__pyx_v_actor_ref); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, int __pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, int __pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, PyObject *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, int __pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, PyObject *__pyx_v_error_type, PyObject *__pyx_v_error, PyObject *__pyx_v_traceback); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_message); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_unpack_message_type_value(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_2unpack_message_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_4unpack_message_id(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_6pack_actor_ref(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_8unpack_actor_ref(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21_unpack_tuple_message_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_10pack_send_message(CYTHON_UNUSED PyObject *__pyx_self, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_message, PyObject *__pyx_v_write, int __pyx_v_protocol); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12unpack_send_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_message); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14pack_tell_message(CYTHON_UNUSED PyObject *__pyx_self, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_message, PyObject *__pyx_v_write, int __pyx_v_protocol); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_16unpack_tell_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_message); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_18get_index(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary, PyObject *__pyx_v_calc_from_uid); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20pack_create_actor_message(CYTHON_UNUSED PyObject *__pyx_self, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, int __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_actor_cls, PyObject *__pyx_v_args, PyObject *__pyx_v_kw, PyObject *__pyx_v_write, int __pyx_v_protocol); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_22unpack_create_actor_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_24pack_destroy_actor_message(CYTHON_UNUSED PyObject *__pyx_self, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_write, int __pyx_v_protocol); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_26unpack_destroy_actor_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_28pack_has_actor_message(CYTHON_UNUSED PyObject *__pyx_self, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, int __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_write, int __pyx_v_protocol); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_30unpack_has_actor_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_32pack_result_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, int __pyx_v_to_index, PyObject *__pyx_v_result, PyObject *__pyx_v_write, int __pyx_v_protocol); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_34unpack_result_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary, PyObject *__pyx_v_from_index, PyObject *__pyx_v_to_index); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_36pack_error_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, PyObject *__pyx_v_error_type, PyObject *__pyx_v_error, PyObject *__pyx_v_tb, PyObject *__pyx_v_write, int __pyx_v_protocol); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_38unpack_error_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_40read_remote_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read_func); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20write_remote_message_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_42write_remote_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_write_func, PyObject *__pyx_v_binary); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_44write_remote_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_46__pyx_unpickle_CREATE_ACTOR_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_48__pyx_unpickle_DESTROY_ACTOR_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_50__pyx_unpickle_HAS_ACTOR_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_52__pyx_unpickle_RESULT_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_54__pyx_unpickle_ERROR_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_56__pyx_unpickle_SEND_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static int __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name, PyObject *__pyx_v_parents, PyObject *__pyx_v_dct); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_value, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_RESULT_MESSAGE(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_ERROR_MESSAGE(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_SEND_MESSAGE(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___Pyx_EnumMeta(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyByteArray_Type_extend = {0, &__pyx_n_s_extend, 0, 0, 0};
static PyObject *__pyx_int_32;
static PyObject *__pyx_int_4313949;
static PyObject *__pyx_int_48008750;
static PyObject *__pyx_int_70028716;
static PyObject *__pyx_int_127827220;
static PyObject *__pyx_int_222419149;
static PyObject *__pyx_int_232585946;
static PyObject *__pyx_int_neg_1;
static int __pyx_k_;
static int __pyx_k__3;
static int __pyx_k__4;
static int __pyx_k__5;
static int __pyx_k__6;
static int __pyx_k__7;
static int __pyx_k__8;
static int __pyx_k__9;
static int __pyx_k__10;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__46;
/* Late includes */

/* "mars/actors/pool/messages.pyx":75
 *     cdef public dict kwargs
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
 *                  object actor_cls=None, tuple args=None, dict kwargs=None):
 */

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_message_type;
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  PyObject *__pyx_v_actor_cls = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_message_type,&__pyx_n_s_message_id,&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_actor_ref,&__pyx_n_s_actor_cls,&__pyx_n_s_args,&__pyx_n_s_kwargs,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    values[1] = ((PyObject*)Py_None);

    /* "mars/actors/pool/messages.pyx":76
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,             # <<<<<<<<<<<<<<
 *                  object actor_cls=None, tuple args=None, dict kwargs=None):
 *         self.message_type = message_type
 */
    values[4] = (PyObject *)((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None);

    /* "mars/actors/pool/messages.pyx":77
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
 *                  object actor_cls=None, tuple args=None, dict kwargs=None):             # <<<<<<<<<<<<<<
 *         self.message_type = message_type
 *         self.message_id = message_id
 */
    values[5] = ((PyObject *)Py_None);
    values[6] = ((PyObject*)Py_None);
    values[7] = ((PyObject*)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_type);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_id);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_ref);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_cls);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 75, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_message_type = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_message_type == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 75, __pyx_L3_error)
    } else {
      __pyx_v_message_type = ((int)-1);
    }
    __pyx_v_message_id = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[2]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 76, __pyx_L3_error)
    } else {
      __pyx_v_from_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    if (values[3]) {
      __pyx_v_to_index = __Pyx_PyInt_As_npy_int32(values[3]); if (unlikely((__pyx_v_to_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 76, __pyx_L3_error)
    } else {
      __pyx_v_to_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)values[4]);
    __pyx_v_actor_cls = values[5];
    __pyx_v_args = ((PyObject*)values[6]);
    __pyx_v_kwargs = ((PyObject*)values[7]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 75, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message_id), (&PyBytes_Type), 1, "message_id", 1))) __PYX_ERR(0, 75, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_actor_ref), __pyx_ptype_4mars_6actors_4core_ActorRef, 1, "actor_ref", 0))) __PYX_ERR(0, 76, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_args), (&PyTuple_Type), 1, "args", 1))) __PYX_ERR(0, 77, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_kwargs), (&PyDict_Type), 1, "kwargs", 1))) __PYX_ERR(0, 77, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE___init__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self), __pyx_v_message_type, __pyx_v_message_id, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_actor_cls, __pyx_v_args, __pyx_v_kwargs);

  /* "mars/actors/pool/messages.pyx":75
 *     cdef public dict kwargs
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
 *                  object actor_cls=None, tuple args=None, dict kwargs=None):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, int __pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_actor_cls, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "mars/actors/pool/messages.pyx":78
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
 *                  object actor_cls=None, tuple args=None, dict kwargs=None):
 *         self.message_type = message_type             # <<<<<<<<<<<<<<
 *         self.message_id = message_id
 *         self.from_index = from_index
 */
  __pyx_v_self->message_type = __pyx_v_message_type;

  /* "mars/actors/pool/messages.pyx":79
 *                  object actor_cls=None, tuple args=None, dict kwargs=None):
 *         self.message_type = message_type
 *         self.message_id = message_id             # <<<<<<<<<<<<<<
 *         self.from_index = from_index
 *         self.to_index = to_index
 */
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = __pyx_v_message_id;

  /* "mars/actors/pool/messages.pyx":80
 *         self.message_type = message_type
 *         self.message_id = message_id
 *         self.from_index = from_index             # <<<<<<<<<<<<<<
 *         self.to_index = to_index
 *         self.actor_ref = actor_ref
 */
  __pyx_v_self->from_index = __pyx_v_from_index;

  /* "mars/actors/pool/messages.pyx":81
 *         self.message_id = message_id
 *         self.from_index = from_index
 *         self.to_index = to_index             # <<<<<<<<<<<<<<
 *         self.actor_ref = actor_ref
 *         self.actor_cls = actor_cls
 */
  __pyx_v_self->to_index = __pyx_v_to_index;

  /* "mars/actors/pool/messages.pyx":82
 *         self.from_index = from_index
 *         self.to_index = to_index
 *         self.actor_ref = actor_ref             # <<<<<<<<<<<<<<
 *         self.actor_cls = actor_cls
 *         self.args = args
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_actor_ref));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_actor_ref));
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = __pyx_v_actor_ref;

  /* "mars/actors/pool/messages.pyx":83
 *         self.to_index = to_index
 *         self.actor_ref = actor_ref
 *         self.actor_cls = actor_cls             # <<<<<<<<<<<<<<
 *         self.args = args
 *         self.kwargs = kwargs
 */
  __Pyx_INCREF(__pyx_v_actor_cls);
  __Pyx_GIVEREF(__pyx_v_actor_cls);
  __Pyx_GOTREF(__pyx_v_self->actor_cls);
  __Pyx_DECREF(__pyx_v_self->actor_cls);
  __pyx_v_self->actor_cls = __pyx_v_actor_cls;

  /* "mars/actors/pool/messages.pyx":84
 *         self.actor_ref = actor_ref
 *         self.actor_cls = actor_cls
 *         self.args = args             # <<<<<<<<<<<<<<
 *         self.kwargs = kwargs
 * 
 */
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  __Pyx_GOTREF(__pyx_v_self->args);
  __Pyx_DECREF(__pyx_v_self->args);
  __pyx_v_self->args = __pyx_v_args;

  /* "mars/actors/pool/messages.pyx":85
 *         self.actor_cls = actor_cls
 *         self.args = args
 *         self.kwargs = kwargs             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_kwargs);
  __Pyx_GIVEREF(__pyx_v_kwargs);
  __Pyx_GOTREF(__pyx_v_self->kwargs);
  __Pyx_DECREF(__pyx_v_self->kwargs);
  __pyx_v_self->kwargs = __pyx_v_kwargs;

  /* "mars/actors/pool/messages.pyx":75
 *     cdef public dict kwargs
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
 *                  object actor_cls=None, tuple args=None, dict kwargs=None):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":66
 * 
 * cdef class CREATE_ACTOR_MESSAGE:
 *     cdef public int message_type             # <<<<<<<<<<<<<<
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->message_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.message_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L1_error)
  __pyx_v_self->message_type = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.message_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":67
 * cdef class CREATE_ACTOR_MESSAGE:
 *     cdef public int message_type
 *     cdef public bytes message_id             # <<<<<<<<<<<<<<
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __pyx_r = __pyx_v_self->message_id;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyBytes_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.message_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":68
 *     cdef public int message_type
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index             # <<<<<<<<<<<<<<
 *     cdef public INT32_t to_index
 *     cdef public ActorRef actor_ref
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.from_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 68, __pyx_L1_error)
  __pyx_v_self->from_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.from_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":69
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index             # <<<<<<<<<<<<<<
 *     cdef public ActorRef actor_ref
 *     cdef public object actor_cls
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.to_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 69, __pyx_L1_error)
  __pyx_v_self->to_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.to_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":70
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 *     cdef public ActorRef actor_ref             # <<<<<<<<<<<<<<
 *     cdef public object actor_cls
 *     cdef public tuple args
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_r = ((PyObject *)__pyx_v_self->actor_ref);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_4mars_6actors_4core_ActorRef))))) __PYX_ERR(0, 70, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.actor_ref.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":71
 *     cdef public INT32_t to_index
 *     cdef public ActorRef actor_ref
 *     cdef public object actor_cls             # <<<<<<<<<<<<<<
 *     cdef public tuple args
 *     cdef public dict kwargs
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->actor_cls);
  __pyx_r = __pyx_v_self->actor_cls;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->actor_cls);
  __Pyx_DECREF(__pyx_v_self->actor_cls);
  __pyx_v_self->actor_cls = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->actor_cls);
  __Pyx_DECREF(__pyx_v_self->actor_cls);
  __pyx_v_self->actor_cls = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":72
 *     cdef public ActorRef actor_ref
 *     cdef public object actor_cls
 *     cdef public tuple args             # <<<<<<<<<<<<<<
 *     cdef public dict kwargs
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->args);
  __pyx_r = __pyx_v_self->args;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyTuple_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 72, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->args);
  __Pyx_DECREF(__pyx_v_self->args);
  __pyx_v_self->args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.args.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->args);
  __Pyx_DECREF(__pyx_v_self->args);
  __pyx_v_self->args = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":73
 *     cdef public object actor_cls
 *     cdef public tuple args
 *     cdef public dict kwargs             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->kwargs);
  __pyx_r = __pyx_v_self->kwargs;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 73, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->kwargs);
  __Pyx_DECREF(__pyx_v_self->kwargs);
  __pyx_v_self->kwargs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.kwargs.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->kwargs);
  __Pyx_DECREF(__pyx_v_self->kwargs);
  __pyx_v_self->kwargs = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_2__reduce_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.actor_cls, self.actor_ref, self.args, self.from_index, self.kwargs, self.message_id, self.message_type, self.to_index)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->message_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(8); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->actor_cls);
  __Pyx_GIVEREF(__pyx_v_self->actor_cls);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_self->actor_cls);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->actor_ref));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->actor_ref));
  PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_self->actor_ref));
  __Pyx_INCREF(__pyx_v_self->args);
  __Pyx_GIVEREF(__pyx_v_self->args);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_self->args);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->kwargs);
  __Pyx_GIVEREF(__pyx_v_self->kwargs);
  PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_self->kwargs);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __Pyx_GIVEREF(__pyx_v_self->message_id);
  PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_self->message_id);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 7, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.actor_cls, self.actor_ref, self.args, self.from_index, self.kwargs, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.actor_cls, self.actor_ref, self.args, self.from_index, self.kwargs, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.actor_cls is not None or self.actor_ref is not None or self.args is not None or self.kwargs is not None or self.message_id is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.actor_cls, self.actor_ref, self.args, self.from_index, self.kwargs, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.actor_cls is not None or self.actor_ref is not None or self.args is not None or self.kwargs is not None or self.message_id is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CREATE_ACTOR_MESSAGE, (type(self), 0x2dc8e2e, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->actor_cls != Py_None);
    __pyx_t_7 = (__pyx_t_5 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (((PyObject *)__pyx_v_self->actor_ref) != Py_None);
    __pyx_t_5 = (__pyx_t_7 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_6 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->args != ((PyObject*)Py_None));
    __pyx_t_7 = (__pyx_t_5 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_self->kwargs != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_7 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_6 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->message_id != ((PyObject*)Py_None));
    __pyx_t_7 = (__pyx_t_5 != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.actor_cls is not None or self.actor_ref is not None or self.args is not None or self.kwargs is not None or self.message_id is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CREATE_ACTOR_MESSAGE, (type(self), 0x2dc8e2e, None), state
 *     else:
 */
  __pyx_t_6 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":13
 *         use_setstate = self.actor_cls is not None or self.actor_ref is not None or self.args is not None or self.kwargs is not None or self.message_id is not None
 *     if use_setstate:
 *         return __pyx_unpickle_CREATE_ACTOR_MESSAGE, (type(self), 0x2dc8e2e, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CREATE_ACTOR_MESSAGE, (type(self), 0x2dc8e2e, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_CREATE_ACTOR_MESS); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_48008750);
    __Pyx_GIVEREF(__pyx_int_48008750);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_48008750);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.actor_cls is not None or self.actor_ref is not None or self.args is not None or self.kwargs is not None or self.message_id is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CREATE_ACTOR_MESSAGE, (type(self), 0x2dc8e2e, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CREATE_ACTOR_MESSAGE, (type(self), 0x2dc8e2e, None), state
 *     else:
 *         return __pyx_unpickle_CREATE_ACTOR_MESSAGE, (type(self), 0x2dc8e2e, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_CREATE_ACTOR_MESS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_48008750);
    __Pyx_GIVEREF(__pyx_int_48008750);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_48008750);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CREATE_ACTOR_MESSAGE, (type(self), 0x2dc8e2e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4__setstate_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CREATE_ACTOR_MESSAGE, (type(self), 0x2dc8e2e, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CREATE_ACTOR_MESSAGE, (type(self), 0x2dc8e2e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.CREATE_ACTOR_MESSAGE.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":95
 *     cdef public ActorRef actor_ref
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, object actor_ref=None):
 *         self.message_type = message_type
 */

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_message_type;
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  PyObject *__pyx_v_actor_ref = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_message_type,&__pyx_n_s_message_id,&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_actor_ref,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[1] = ((PyObject*)Py_None);

    /* "mars/actors/pool/messages.pyx":96
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, object actor_ref=None):             # <<<<<<<<<<<<<<
 *         self.message_type = message_type
 *         self.message_id = message_id
 */
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_type);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_id);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_ref);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 95, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_message_type = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_message_type == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 95, __pyx_L3_error)
    } else {
      __pyx_v_message_type = ((int)-1);
    }
    __pyx_v_message_id = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[2]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 96, __pyx_L3_error)
    } else {
      __pyx_v_from_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    if (values[3]) {
      __pyx_v_to_index = __Pyx_PyInt_As_npy_int32(values[3]); if (unlikely((__pyx_v_to_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 96, __pyx_L3_error)
    } else {
      __pyx_v_to_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    __pyx_v_actor_ref = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 95, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message_id), (&PyBytes_Type), 1, "message_id", 1))) __PYX_ERR(0, 95, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE___init__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self), __pyx_v_message_type, __pyx_v_message_id, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref);

  /* "mars/actors/pool/messages.pyx":95
 *     cdef public ActorRef actor_ref
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, object actor_ref=None):
 *         self.message_type = message_type
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, int __pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, PyObject *__pyx_v_actor_ref) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "mars/actors/pool/messages.pyx":97
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, object actor_ref=None):
 *         self.message_type = message_type             # <<<<<<<<<<<<<<
 *         self.message_id = message_id
 *         self.from_index = from_index
 */
  __pyx_v_self->message_type = __pyx_v_message_type;

  /* "mars/actors/pool/messages.pyx":98
 *                  INT32_t from_index=0, INT32_t to_index=0, object actor_ref=None):
 *         self.message_type = message_type
 *         self.message_id = message_id             # <<<<<<<<<<<<<<
 *         self.from_index = from_index
 *         self.to_index = to_index
 */
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = __pyx_v_message_id;

  /* "mars/actors/pool/messages.pyx":99
 *         self.message_type = message_type
 *         self.message_id = message_id
 *         self.from_index = from_index             # <<<<<<<<<<<<<<
 *         self.to_index = to_index
 *         self.actor_ref = actor_ref
 */
  __pyx_v_self->from_index = __pyx_v_from_index;

  /* "mars/actors/pool/messages.pyx":100
 *         self.message_id = message_id
 *         self.from_index = from_index
 *         self.to_index = to_index             # <<<<<<<<<<<<<<
 *         self.actor_ref = actor_ref
 * 
 */
  __pyx_v_self->to_index = __pyx_v_to_index;

  /* "mars/actors/pool/messages.pyx":101
 *         self.from_index = from_index
 *         self.to_index = to_index
 *         self.actor_ref = actor_ref             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (!(likely(((__pyx_v_actor_ref) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_actor_ref, __pyx_ptype_4mars_6actors_4core_ActorRef))))) __PYX_ERR(0, 101, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_actor_ref;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":95
 *     cdef public ActorRef actor_ref
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, object actor_ref=None):
 *         self.message_type = message_type
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":89
 * 
 * cdef class DESTROY_ACTOR_MESSAGE:
 *     cdef public int message_type             # <<<<<<<<<<<<<<
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->message_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.message_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 89, __pyx_L1_error)
  __pyx_v_self->message_type = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.message_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":90
 * cdef class DESTROY_ACTOR_MESSAGE:
 *     cdef public int message_type
 *     cdef public bytes message_id             # <<<<<<<<<<<<<<
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __pyx_r = __pyx_v_self->message_id;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyBytes_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 90, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.message_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":91
 *     cdef public int message_type
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index             # <<<<<<<<<<<<<<
 *     cdef public INT32_t to_index
 *     cdef public ActorRef actor_ref
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.from_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 91, __pyx_L1_error)
  __pyx_v_self->from_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.from_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":92
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index             # <<<<<<<<<<<<<<
 *     cdef public ActorRef actor_ref
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.to_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 92, __pyx_L1_error)
  __pyx_v_self->to_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.to_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":93
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 *     cdef public ActorRef actor_ref             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_r = ((PyObject *)__pyx_v_self->actor_ref);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_4mars_6actors_4core_ActorRef))))) __PYX_ERR(0, 93, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.actor_ref.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_2__reduce_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.actor_ref, self.from_index, self.message_id, self.message_type, self.to_index)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->message_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->actor_ref));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->actor_ref));
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_self->actor_ref));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __Pyx_GIVEREF(__pyx_v_self->message_id);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_self->message_id);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.actor_ref, self.from_index, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.actor_ref, self.from_index, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message_id is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.actor_ref, self.from_index, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message_id is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_DESTROY_ACTOR_MESSAGE, (type(self), 0x79e7d14, None), state
 */
  /*else*/ {
    __pyx_t_5 = (((PyObject *)__pyx_v_self->actor_ref) != Py_None);
    __pyx_t_7 = (__pyx_t_5 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_self->message_id != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_7 != 0);
    __pyx_t_6 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message_id is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DESTROY_ACTOR_MESSAGE, (type(self), 0x79e7d14, None), state
 *     else:
 */
  __pyx_t_6 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":13
 *         use_setstate = self.actor_ref is not None or self.message_id is not None
 *     if use_setstate:
 *         return __pyx_unpickle_DESTROY_ACTOR_MESSAGE, (type(self), 0x79e7d14, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_DESTROY_ACTOR_MESSAGE, (type(self), 0x79e7d14, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_DESTROY_ACTOR_MES); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127827220);
    __Pyx_GIVEREF(__pyx_int_127827220);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_127827220);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message_id is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DESTROY_ACTOR_MESSAGE, (type(self), 0x79e7d14, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_DESTROY_ACTOR_MESSAGE, (type(self), 0x79e7d14, None), state
 *     else:
 *         return __pyx_unpickle_DESTROY_ACTOR_MESSAGE, (type(self), 0x79e7d14, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_DESTROY_ACTOR_MES); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127827220);
    __Pyx_GIVEREF(__pyx_int_127827220);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_127827220);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DESTROY_ACTOR_MESSAGE, (type(self), 0x79e7d14, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_4__setstate_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_DESTROY_ACTOR_MESSAGE, (type(self), 0x79e7d14, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DESTROY_ACTOR_MESSAGE, (type(self), 0x79e7d14, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":111
 *     cdef public ActorRef actor_ref
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None):
 *         self.message_type = message_type
 */

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_message_type;
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_message_type,&__pyx_n_s_message_id,&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_actor_ref,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[1] = ((PyObject*)Py_None);

    /* "mars/actors/pool/messages.pyx":112
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None):             # <<<<<<<<<<<<<<
 *         self.message_type = message_type
 *         self.message_id = message_id
 */
    values[4] = (PyObject *)((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_type);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_id);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_ref);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 111, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_message_type = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_message_type == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 111, __pyx_L3_error)
    } else {
      __pyx_v_message_type = ((int)-1);
    }
    __pyx_v_message_id = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[2]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 112, __pyx_L3_error)
    } else {
      __pyx_v_from_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    if (values[3]) {
      __pyx_v_to_index = __Pyx_PyInt_As_npy_int32(values[3]); if (unlikely((__pyx_v_to_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 112, __pyx_L3_error)
    } else {
      __pyx_v_to_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)values[4]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 111, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.HAS_ACTOR_MESSAGE.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message_id), (&PyBytes_Type), 1, "message_id", 1))) __PYX_ERR(0, 111, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_actor_ref), __pyx_ptype_4mars_6actors_4core_ActorRef, 1, "actor_ref", 0))) __PYX_ERR(0, 112, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE___init__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self), __pyx_v_message_type, __pyx_v_message_id, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref);

  /* "mars/actors/pool/messages.pyx":111
 *     cdef public ActorRef actor_ref
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None):
 *         self.message_type = message_type
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, int __pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "mars/actors/pool/messages.pyx":113
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None):
 *         self.message_type = message_type             # <<<<<<<<<<<<<<
 *         self.message_id = message_id
 *         self.from_index = from_index
 */
  __pyx_v_self->message_type = __pyx_v_message_type;

  /* "mars/actors/pool/messages.pyx":114
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None):
 *         self.message_type = message_type
 *         self.message_id = message_id             # <<<<<<<<<<<<<<
 *         self.from_index = from_index
 *         self.to_index = to_index
 */
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = __pyx_v_message_id;

  /* "mars/actors/pool/messages.pyx":115
 *         self.message_type = message_type
 *         self.message_id = message_id
 *         self.from_index = from_index             # <<<<<<<<<<<<<<
 *         self.to_index = to_index
 *         self.actor_ref = actor_ref
 */
  __pyx_v_self->from_index = __pyx_v_from_index;

  /* "mars/actors/pool/messages.pyx":116
 *         self.message_id = message_id
 *         self.from_index = from_index
 *         self.to_index = to_index             # <<<<<<<<<<<<<<
 *         self.actor_ref = actor_ref
 * 
 */
  __pyx_v_self->to_index = __pyx_v_to_index;

  /* "mars/actors/pool/messages.pyx":117
 *         self.from_index = from_index
 *         self.to_index = to_index
 *         self.actor_ref = actor_ref             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_actor_ref));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_actor_ref));
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = __pyx_v_actor_ref;

  /* "mars/actors/pool/messages.pyx":111
 *     cdef public ActorRef actor_ref
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None):
 *         self.message_type = message_type
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":105
 * 
 * cdef class HAS_ACTOR_MESSAGE:
 *     cdef public int message_type             # <<<<<<<<<<<<<<
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->message_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.HAS_ACTOR_MESSAGE.message_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 105, __pyx_L1_error)
  __pyx_v_self->message_type = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.HAS_ACTOR_MESSAGE.message_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":106
 * cdef class HAS_ACTOR_MESSAGE:
 *     cdef public int message_type
 *     cdef public bytes message_id             # <<<<<<<<<<<<<<
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __pyx_r = __pyx_v_self->message_id;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyBytes_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 106, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.HAS_ACTOR_MESSAGE.message_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":107
 *     cdef public int message_type
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index             # <<<<<<<<<<<<<<
 *     cdef public INT32_t to_index
 *     cdef public ActorRef actor_ref
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.HAS_ACTOR_MESSAGE.from_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 107, __pyx_L1_error)
  __pyx_v_self->from_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.HAS_ACTOR_MESSAGE.from_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":108
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index             # <<<<<<<<<<<<<<
 *     cdef public ActorRef actor_ref
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.HAS_ACTOR_MESSAGE.to_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 108, __pyx_L1_error)
  __pyx_v_self->to_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.HAS_ACTOR_MESSAGE.to_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":109
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 *     cdef public ActorRef actor_ref             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_r = ((PyObject *)__pyx_v_self->actor_ref);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_4mars_6actors_4core_ActorRef))))) __PYX_ERR(0, 109, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.HAS_ACTOR_MESSAGE.actor_ref.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_2__reduce_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.actor_ref, self.from_index, self.message_id, self.message_type, self.to_index)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->message_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->actor_ref));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->actor_ref));
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_self->actor_ref));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __Pyx_GIVEREF(__pyx_v_self->message_id);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_self->message_id);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.actor_ref, self.from_index, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.actor_ref, self.from_index, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message_id is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.actor_ref, self.from_index, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message_id is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_HAS_ACTOR_MESSAGE, (type(self), 0x79e7d14, None), state
 */
  /*else*/ {
    __pyx_t_5 = (((PyObject *)__pyx_v_self->actor_ref) != Py_None);
    __pyx_t_7 = (__pyx_t_5 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_self->message_id != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_7 != 0);
    __pyx_t_6 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message_id is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_HAS_ACTOR_MESSAGE, (type(self), 0x79e7d14, None), state
 *     else:
 */
  __pyx_t_6 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":13
 *         use_setstate = self.actor_ref is not None or self.message_id is not None
 *     if use_setstate:
 *         return __pyx_unpickle_HAS_ACTOR_MESSAGE, (type(self), 0x79e7d14, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_HAS_ACTOR_MESSAGE, (type(self), 0x79e7d14, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_HAS_ACTOR_MESSAGE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127827220);
    __Pyx_GIVEREF(__pyx_int_127827220);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_127827220);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message_id is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_HAS_ACTOR_MESSAGE, (type(self), 0x79e7d14, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_HAS_ACTOR_MESSAGE, (type(self), 0x79e7d14, None), state
 *     else:
 *         return __pyx_unpickle_HAS_ACTOR_MESSAGE, (type(self), 0x79e7d14, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_HAS_ACTOR_MESSAGE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127827220);
    __Pyx_GIVEREF(__pyx_int_127827220);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_127827220);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.actors.pool.messages.HAS_ACTOR_MESSAGE.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_HAS_ACTOR_MESSAGE, (type(self), 0x79e7d14, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_4__setstate_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_HAS_ACTOR_MESSAGE, (type(self), 0x79e7d14, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_HAS_ACTOR_MESSAGE, (type(self), 0x79e7d14, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.HAS_ACTOR_MESSAGE.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":127
 *     cdef public object result
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, object result=None):
 *         self.message_type = message_type
 */

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_message_type;
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  PyObject *__pyx_v_result = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_message_type,&__pyx_n_s_message_id,&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_result,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[1] = ((PyObject*)Py_None);

    /* "mars/actors/pool/messages.pyx":128
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, object result=None):             # <<<<<<<<<<<<<<
 *         self.message_type = message_type
 *         self.message_id = message_id
 */
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_type);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_id);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 127, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_message_type = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_message_type == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L3_error)
    } else {
      __pyx_v_message_type = ((int)-1);
    }
    __pyx_v_message_id = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[2]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 128, __pyx_L3_error)
    } else {
      __pyx_v_from_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    if (values[3]) {
      __pyx_v_to_index = __Pyx_PyInt_As_npy_int32(values[3]); if (unlikely((__pyx_v_to_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 128, __pyx_L3_error)
    } else {
      __pyx_v_to_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    __pyx_v_result = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 127, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.RESULT_MESSAGE.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message_id), (&PyBytes_Type), 1, "message_id", 1))) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE___init__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self), __pyx_v_message_type, __pyx_v_message_id, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_result);

  /* "mars/actors/pool/messages.pyx":127
 *     cdef public object result
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, object result=None):
 *         self.message_type = message_type
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, int __pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, PyObject *__pyx_v_result) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "mars/actors/pool/messages.pyx":129
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, object result=None):
 *         self.message_type = message_type             # <<<<<<<<<<<<<<
 *         self.message_id = message_id
 *         self.from_index = from_index
 */
  __pyx_v_self->message_type = __pyx_v_message_type;

  /* "mars/actors/pool/messages.pyx":130
 *                  INT32_t from_index=0, INT32_t to_index=0, object result=None):
 *         self.message_type = message_type
 *         self.message_id = message_id             # <<<<<<<<<<<<<<
 *         self.from_index = from_index
 *         self.to_index = to_index
 */
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = __pyx_v_message_id;

  /* "mars/actors/pool/messages.pyx":131
 *         self.message_type = message_type
 *         self.message_id = message_id
 *         self.from_index = from_index             # <<<<<<<<<<<<<<
 *         self.to_index = to_index
 *         self.result = result
 */
  __pyx_v_self->from_index = __pyx_v_from_index;

  /* "mars/actors/pool/messages.pyx":132
 *         self.message_id = message_id
 *         self.from_index = from_index
 *         self.to_index = to_index             # <<<<<<<<<<<<<<
 *         self.result = result
 * 
 */
  __pyx_v_self->to_index = __pyx_v_to_index;

  /* "mars/actors/pool/messages.pyx":133
 *         self.from_index = from_index
 *         self.to_index = to_index
 *         self.result = result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_result);
  __Pyx_GIVEREF(__pyx_v_result);
  __Pyx_GOTREF(__pyx_v_self->result);
  __Pyx_DECREF(__pyx_v_self->result);
  __pyx_v_self->result = __pyx_v_result;

  /* "mars/actors/pool/messages.pyx":127
 *     cdef public object result
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, object result=None):
 *         self.message_type = message_type
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":121
 * 
 * cdef class RESULT_MESSAGE:
 *     cdef public int message_type             # <<<<<<<<<<<<<<
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->message_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.RESULT_MESSAGE.message_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 121, __pyx_L1_error)
  __pyx_v_self->message_type = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.RESULT_MESSAGE.message_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":122
 * cdef class RESULT_MESSAGE:
 *     cdef public int message_type
 *     cdef public bytes message_id             # <<<<<<<<<<<<<<
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __pyx_r = __pyx_v_self->message_id;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyBytes_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 122, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.RESULT_MESSAGE.message_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":123
 *     cdef public int message_type
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index             # <<<<<<<<<<<<<<
 *     cdef public INT32_t to_index
 *     cdef public object result
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.RESULT_MESSAGE.from_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 123, __pyx_L1_error)
  __pyx_v_self->from_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.RESULT_MESSAGE.from_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":124
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index             # <<<<<<<<<<<<<<
 *     cdef public object result
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.RESULT_MESSAGE.to_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 124, __pyx_L1_error)
  __pyx_v_self->to_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.RESULT_MESSAGE.to_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":125
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 *     cdef public object result             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->result);
  __pyx_r = __pyx_v_self->result;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->result);
  __Pyx_DECREF(__pyx_v_self->result);
  __pyx_v_self->result = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->result);
  __Pyx_DECREF(__pyx_v_self->result);
  __pyx_v_self->result = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_2__reduce_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.from_index, self.message_id, self.message_type, self.result, self.to_index)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->message_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __Pyx_GIVEREF(__pyx_v_self->message_id);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_self->message_id);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->result);
  __Pyx_GIVEREF(__pyx_v_self->result);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_self->result);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.from_index, self.message_id, self.message_type, self.result, self.to_index)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.from_index, self.message_id, self.message_type, self.result, self.to_index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.message_id is not None or self.result is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.from_index, self.message_id, self.message_type, self.result, self.to_index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.message_id is not None or self.result is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_RESULT_MESSAGE, (type(self), 0xddcfada, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->message_id != ((PyObject*)Py_None));
    __pyx_t_7 = (__pyx_t_5 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_self->result != Py_None);
    __pyx_t_5 = (__pyx_t_7 != 0);
    __pyx_t_6 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.message_id is not None or self.result is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RESULT_MESSAGE, (type(self), 0xddcfada, None), state
 *     else:
 */
  __pyx_t_6 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":13
 *         use_setstate = self.message_id is not None or self.result is not None
 *     if use_setstate:
 *         return __pyx_unpickle_RESULT_MESSAGE, (type(self), 0xddcfada, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_RESULT_MESSAGE, (type(self), 0xddcfada, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_RESULT_MESSAGE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_232585946);
    __Pyx_GIVEREF(__pyx_int_232585946);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_232585946);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.message_id is not None or self.result is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RESULT_MESSAGE, (type(self), 0xddcfada, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_RESULT_MESSAGE, (type(self), 0xddcfada, None), state
 *     else:
 *         return __pyx_unpickle_RESULT_MESSAGE, (type(self), 0xddcfada, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RESULT_MESSAGE__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_RESULT_MESSAGE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_232585946);
    __Pyx_GIVEREF(__pyx_int_232585946);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_232585946);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.actors.pool.messages.RESULT_MESSAGE.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RESULT_MESSAGE, (type(self), 0xddcfada, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RESULT_MESSAGE__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_4__setstate_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_RESULT_MESSAGE, (type(self), 0xddcfada, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RESULT_MESSAGE__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_RESULT_MESSAGE__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RESULT_MESSAGE, (type(self), 0xddcfada, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RESULT_MESSAGE__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.RESULT_MESSAGE.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":145
 *     cdef public object traceback
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, object error_type=None,
 *                  object error=None, object traceback=None):
 */

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_message_type;
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  PyObject *__pyx_v_error_type = 0;
  PyObject *__pyx_v_error = 0;
  PyObject *__pyx_v_traceback = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_message_type,&__pyx_n_s_message_id,&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_error_type,&__pyx_n_s_error,&__pyx_n_s_traceback,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    values[1] = ((PyObject*)Py_None);

    /* "mars/actors/pool/messages.pyx":146
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, object error_type=None,             # <<<<<<<<<<<<<<
 *                  object error=None, object traceback=None):
 *         self.message_type = message_type
 */
    values[4] = ((PyObject *)Py_None);

    /* "mars/actors/pool/messages.pyx":147
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, object error_type=None,
 *                  object error=None, object traceback=None):             # <<<<<<<<<<<<<<
 *         self.message_type = message_type
 *         self.message_id = message_id
 */
    values[5] = ((PyObject *)Py_None);
    values[6] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_type);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_id);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_type);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_traceback);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 145, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_message_type = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_message_type == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 145, __pyx_L3_error)
    } else {
      __pyx_v_message_type = ((int)-1);
    }
    __pyx_v_message_id = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[2]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 146, __pyx_L3_error)
    } else {
      __pyx_v_from_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    if (values[3]) {
      __pyx_v_to_index = __Pyx_PyInt_As_npy_int32(values[3]); if (unlikely((__pyx_v_to_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 146, __pyx_L3_error)
    } else {
      __pyx_v_to_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    __pyx_v_error_type = values[4];
    __pyx_v_error = values[5];
    __pyx_v_traceback = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 145, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.ERROR_MESSAGE.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message_id), (&PyBytes_Type), 1, "message_id", 1))) __PYX_ERR(0, 145, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE___init__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self), __pyx_v_message_type, __pyx_v_message_id, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_error_type, __pyx_v_error, __pyx_v_traceback);

  /* "mars/actors/pool/messages.pyx":145
 *     cdef public object traceback
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, object error_type=None,
 *                  object error=None, object traceback=None):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, int __pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, PyObject *__pyx_v_error_type, PyObject *__pyx_v_error, PyObject *__pyx_v_traceback) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "mars/actors/pool/messages.pyx":148
 *                  INT32_t from_index=0, INT32_t to_index=0, object error_type=None,
 *                  object error=None, object traceback=None):
 *         self.message_type = message_type             # <<<<<<<<<<<<<<
 *         self.message_id = message_id
 *         self.from_index = from_index
 */
  __pyx_v_self->message_type = __pyx_v_message_type;

  /* "mars/actors/pool/messages.pyx":149
 *                  object error=None, object traceback=None):
 *         self.message_type = message_type
 *         self.message_id = message_id             # <<<<<<<<<<<<<<
 *         self.from_index = from_index
 *         self.to_index = to_index
 */
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = __pyx_v_message_id;

  /* "mars/actors/pool/messages.pyx":150
 *         self.message_type = message_type
 *         self.message_id = message_id
 *         self.from_index = from_index             # <<<<<<<<<<<<<<
 *         self.to_index = to_index
 *         self.error_type = error_type
 */
  __pyx_v_self->from_index = __pyx_v_from_index;

  /* "mars/actors/pool/messages.pyx":151
 *         self.message_id = message_id
 *         self.from_index = from_index
 *         self.to_index = to_index             # <<<<<<<<<<<<<<
 *         self.error_type = error_type
 *         self.error = error
 */
  __pyx_v_self->to_index = __pyx_v_to_index;

  /* "mars/actors/pool/messages.pyx":152
 *         self.from_index = from_index
 *         self.to_index = to_index
 *         self.error_type = error_type             # <<<<<<<<<<<<<<
 *         self.error = error
 *         self.traceback = traceback
 */
  __Pyx_INCREF(__pyx_v_error_type);
  __Pyx_GIVEREF(__pyx_v_error_type);
  __Pyx_GOTREF(__pyx_v_self->error_type);
  __Pyx_DECREF(__pyx_v_self->error_type);
  __pyx_v_self->error_type = __pyx_v_error_type;

  /* "mars/actors/pool/messages.pyx":153
 *         self.to_index = to_index
 *         self.error_type = error_type
 *         self.error = error             # <<<<<<<<<<<<<<
 *         self.traceback = traceback
 * 
 */
  __Pyx_INCREF(__pyx_v_error);
  __Pyx_GIVEREF(__pyx_v_error);
  __Pyx_GOTREF(__pyx_v_self->error);
  __Pyx_DECREF(__pyx_v_self->error);
  __pyx_v_self->error = __pyx_v_error;

  /* "mars/actors/pool/messages.pyx":154
 *         self.error_type = error_type
 *         self.error = error
 *         self.traceback = traceback             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_traceback);
  __Pyx_GIVEREF(__pyx_v_traceback);
  __Pyx_GOTREF(__pyx_v_self->traceback);
  __Pyx_DECREF(__pyx_v_self->traceback);
  __pyx_v_self->traceback = __pyx_v_traceback;

  /* "mars/actors/pool/messages.pyx":145
 *     cdef public object traceback
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, object error_type=None,
 *                  object error=None, object traceback=None):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":137
 * 
 * cdef class ERROR_MESSAGE:
 *     cdef public int message_type             # <<<<<<<<<<<<<<
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->message_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.ERROR_MESSAGE.message_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 137, __pyx_L1_error)
  __pyx_v_self->message_type = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.ERROR_MESSAGE.message_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":138
 * cdef class ERROR_MESSAGE:
 *     cdef public int message_type
 *     cdef public bytes message_id             # <<<<<<<<<<<<<<
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __pyx_r = __pyx_v_self->message_id;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyBytes_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 138, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.ERROR_MESSAGE.message_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":139
 *     cdef public int message_type
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index             # <<<<<<<<<<<<<<
 *     cdef public INT32_t to_index
 *     cdef public object error_type
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.ERROR_MESSAGE.from_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 139, __pyx_L1_error)
  __pyx_v_self->from_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.ERROR_MESSAGE.from_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":140
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index             # <<<<<<<<<<<<<<
 *     cdef public object error_type
 *     cdef public object error
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.ERROR_MESSAGE.to_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 140, __pyx_L1_error)
  __pyx_v_self->to_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.ERROR_MESSAGE.to_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":141
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 *     cdef public object error_type             # <<<<<<<<<<<<<<
 *     cdef public object error
 *     cdef public object traceback
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->error_type);
  __pyx_r = __pyx_v_self->error_type;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->error_type);
  __Pyx_DECREF(__pyx_v_self->error_type);
  __pyx_v_self->error_type = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->error_type);
  __Pyx_DECREF(__pyx_v_self->error_type);
  __pyx_v_self->error_type = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":142
 *     cdef public INT32_t to_index
 *     cdef public object error_type
 *     cdef public object error             # <<<<<<<<<<<<<<
 *     cdef public object traceback
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->error);
  __pyx_r = __pyx_v_self->error;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->error);
  __Pyx_DECREF(__pyx_v_self->error);
  __pyx_v_self->error = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->error);
  __Pyx_DECREF(__pyx_v_self->error);
  __pyx_v_self->error = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":143
 *     cdef public object error_type
 *     cdef public object error
 *     cdef public object traceback             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int message_type=-1, bytes message_id=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->traceback);
  __pyx_r = __pyx_v_self->traceback;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->traceback);
  __Pyx_DECREF(__pyx_v_self->traceback);
  __pyx_v_self->traceback = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->traceback);
  __Pyx_DECREF(__pyx_v_self->traceback);
  __pyx_v_self->traceback = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_2__reduce_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.error, self.error_type, self.from_index, self.message_id, self.message_type, self.to_index, self.traceback)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->message_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->error);
  __Pyx_GIVEREF(__pyx_v_self->error);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_self->error);
  __Pyx_INCREF(__pyx_v_self->error_type);
  __Pyx_GIVEREF(__pyx_v_self->error_type);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_self->error_type);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __Pyx_GIVEREF(__pyx_v_self->message_id);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_self->message_id);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->traceback);
  __Pyx_GIVEREF(__pyx_v_self->traceback);
  PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_self->traceback);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.error, self.error_type, self.from_index, self.message_id, self.message_type, self.to_index, self.traceback)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.error, self.error_type, self.from_index, self.message_id, self.message_type, self.to_index, self.traceback)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.error is not None or self.error_type is not None or self.message_id is not None or self.traceback is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.error, self.error_type, self.from_index, self.message_id, self.message_type, self.to_index, self.traceback)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.error is not None or self.error_type is not None or self.message_id is not None or self.traceback is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ERROR_MESSAGE, (type(self), 0x42c8dac, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->error != Py_None);
    __pyx_t_7 = (__pyx_t_5 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_self->error_type != Py_None);
    __pyx_t_5 = (__pyx_t_7 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_6 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->message_id != ((PyObject*)Py_None));
    __pyx_t_7 = (__pyx_t_5 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_self->traceback != Py_None);
    __pyx_t_5 = (__pyx_t_7 != 0);
    __pyx_t_6 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.error is not None or self.error_type is not None or self.message_id is not None or self.traceback is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ERROR_MESSAGE, (type(self), 0x42c8dac, None), state
 *     else:
 */
  __pyx_t_6 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":13
 *         use_setstate = self.error is not None or self.error_type is not None or self.message_id is not None or self.traceback is not None
 *     if use_setstate:
 *         return __pyx_unpickle_ERROR_MESSAGE, (type(self), 0x42c8dac, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ERROR_MESSAGE, (type(self), 0x42c8dac, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_ERROR_MESSAGE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_70028716);
    __Pyx_GIVEREF(__pyx_int_70028716);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_70028716);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.error is not None or self.error_type is not None or self.message_id is not None or self.traceback is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ERROR_MESSAGE, (type(self), 0x42c8dac, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ERROR_MESSAGE, (type(self), 0x42c8dac, None), state
 *     else:
 *         return __pyx_unpickle_ERROR_MESSAGE, (type(self), 0x42c8dac, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ERROR_MESSAGE__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_ERROR_MESSAGE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_70028716);
    __Pyx_GIVEREF(__pyx_int_70028716);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_70028716);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.actors.pool.messages.ERROR_MESSAGE.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ERROR_MESSAGE, (type(self), 0x42c8dac, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ERROR_MESSAGE__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_4__setstate_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ERROR_MESSAGE, (type(self), 0x42c8dac, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ERROR_MESSAGE__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_ERROR_MESSAGE__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ERROR_MESSAGE, (type(self), 0x42c8dac, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ERROR_MESSAGE__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.ERROR_MESSAGE.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":165
 *     cdef public object message
 * 
 *     def __init__(self, object message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
 *                  object message=None):
 */

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_message_type = 0;
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  PyObject *__pyx_v_message = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_message_type,&__pyx_n_s_message_id,&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_actor_ref,&__pyx_n_s_message,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[0] = ((PyObject *)__pyx_int_neg_1);
    values[1] = ((PyObject*)Py_None);

    /* "mars/actors/pool/messages.pyx":166
 * 
 *     def __init__(self, object message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,             # <<<<<<<<<<<<<<
 *                  object message=None):
 *         self.message_type = message_type
 */
    values[4] = (PyObject *)((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None);

    /* "mars/actors/pool/messages.pyx":167
 *     def __init__(self, object message_type=-1, bytes message_id=None,
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
 *                  object message=None):             # <<<<<<<<<<<<<<
 *         self.message_type = message_type
 *         self.message_id = message_id
 */
    values[5] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_type);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_id);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_ref);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 165, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_message_type = values[0];
    __pyx_v_message_id = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[2]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L3_error)
    } else {
      __pyx_v_from_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    if (values[3]) {
      __pyx_v_to_index = __Pyx_PyInt_As_npy_int32(values[3]); if (unlikely((__pyx_v_to_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L3_error)
    } else {
      __pyx_v_to_index = ((__pyx_t_4mars_6actors_4pool_8messages_INT32_t)0);
    }
    __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)values[4]);
    __pyx_v_message = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 165, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.SEND_MESSAGE.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message_id), (&PyBytes_Type), 1, "message_id", 1))) __PYX_ERR(0, 165, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_actor_ref), __pyx_ptype_4mars_6actors_4core_ActorRef, 1, "actor_ref", 0))) __PYX_ERR(0, 166, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE___init__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self), __pyx_v_message_type, __pyx_v_message_id, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_message);

  /* "mars/actors/pool/messages.pyx":165
 *     cdef public object message
 * 
 *     def __init__(self, object message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
 *                  object message=None):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE___init__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_message_type, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_message) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "mars/actors/pool/messages.pyx":168
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
 *                  object message=None):
 *         self.message_type = message_type             # <<<<<<<<<<<<<<
 *         self.message_id = message_id
 *         self.from_index = from_index
 */
  __Pyx_INCREF(__pyx_v_message_type);
  __Pyx_GIVEREF(__pyx_v_message_type);
  __Pyx_GOTREF(__pyx_v_self->message_type);
  __Pyx_DECREF(__pyx_v_self->message_type);
  __pyx_v_self->message_type = __pyx_v_message_type;

  /* "mars/actors/pool/messages.pyx":169
 *                  object message=None):
 *         self.message_type = message_type
 *         self.message_id = message_id             # <<<<<<<<<<<<<<
 *         self.from_index = from_index
 *         self.to_index = to_index
 */
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = __pyx_v_message_id;

  /* "mars/actors/pool/messages.pyx":170
 *         self.message_type = message_type
 *         self.message_id = message_id
 *         self.from_index = from_index             # <<<<<<<<<<<<<<
 *         self.to_index = to_index
 *         self.actor_ref = actor_ref
 */
  __pyx_v_self->from_index = __pyx_v_from_index;

  /* "mars/actors/pool/messages.pyx":171
 *         self.message_id = message_id
 *         self.from_index = from_index
 *         self.to_index = to_index             # <<<<<<<<<<<<<<
 *         self.actor_ref = actor_ref
 *         self.message = message
 */
  __pyx_v_self->to_index = __pyx_v_to_index;

  /* "mars/actors/pool/messages.pyx":172
 *         self.from_index = from_index
 *         self.to_index = to_index
 *         self.actor_ref = actor_ref             # <<<<<<<<<<<<<<
 *         self.message = message
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_actor_ref));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_actor_ref));
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = __pyx_v_actor_ref;

  /* "mars/actors/pool/messages.pyx":173
 *         self.to_index = to_index
 *         self.actor_ref = actor_ref
 *         self.message = message             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_message);
  __Pyx_GIVEREF(__pyx_v_message);
  __Pyx_GOTREF(__pyx_v_self->message);
  __Pyx_DECREF(__pyx_v_self->message);
  __pyx_v_self->message = __pyx_v_message;

  /* "mars/actors/pool/messages.pyx":165
 *     cdef public object message
 * 
 *     def __init__(self, object message_type=-1, bytes message_id=None,             # <<<<<<<<<<<<<<
 *                  INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
 *                  object message=None):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":158
 * 
 * cdef class SEND_MESSAGE:
 *     cdef public object message_type             # <<<<<<<<<<<<<<
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->message_type);
  __pyx_r = __pyx_v_self->message_type;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->message_type);
  __Pyx_DECREF(__pyx_v_self->message_type);
  __pyx_v_self->message_type = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->message_type);
  __Pyx_DECREF(__pyx_v_self->message_type);
  __pyx_v_self->message_type = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":159
 * cdef class SEND_MESSAGE:
 *     cdef public object message_type
 *     cdef public bytes message_id             # <<<<<<<<<<<<<<
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __pyx_r = __pyx_v_self->message_id;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyBytes_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 159, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.SEND_MESSAGE.message_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->message_id);
  __Pyx_DECREF(__pyx_v_self->message_id);
  __pyx_v_self->message_id = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":160
 *     cdef public object message_type
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index             # <<<<<<<<<<<<<<
 *     cdef public INT32_t to_index
 *     cdef public ActorRef actor_ref
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.SEND_MESSAGE.from_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 160, __pyx_L1_error)
  __pyx_v_self->from_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.SEND_MESSAGE.from_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":161
 *     cdef public bytes message_id
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index             # <<<<<<<<<<<<<<
 *     cdef public ActorRef actor_ref
 *     cdef public object message
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.SEND_MESSAGE.to_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_v_self->to_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.SEND_MESSAGE.to_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":162
 *     cdef public INT32_t from_index
 *     cdef public INT32_t to_index
 *     cdef public ActorRef actor_ref             # <<<<<<<<<<<<<<
 *     cdef public object message
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_r = ((PyObject *)__pyx_v_self->actor_ref);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_4mars_6actors_4core_ActorRef))))) __PYX_ERR(0, 162, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.SEND_MESSAGE.actor_ref.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->actor_ref));
  __pyx_v_self->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":163
 *     cdef public INT32_t to_index
 *     cdef public ActorRef actor_ref
 *     cdef public object message             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, object message_type=-1, bytes message_id=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message___get__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message___get__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->message);
  __pyx_r = __pyx_v_self->message;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_2__set__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_2__set__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->message);
  __Pyx_DECREF(__pyx_v_self->message);
  __pyx_v_self->message = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_4__del__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_4__del__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->message);
  __Pyx_DECREF(__pyx_v_self->message);
  __pyx_v_self->message = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_2__reduce_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_2__reduce_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.actor_ref, self.from_index, self.message, self.message_id, self.message_type, self.to_index)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->from_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->to_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->actor_ref));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->actor_ref));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->actor_ref));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->message);
  __Pyx_GIVEREF(__pyx_v_self->message);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->message);
  __Pyx_INCREF(__pyx_v_self->message_id);
  __Pyx_GIVEREF(__pyx_v_self->message_id);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->message_id);
  __Pyx_INCREF(__pyx_v_self->message_type);
  __Pyx_GIVEREF(__pyx_v_self->message_type);
  PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_self->message_type);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.actor_ref, self.from_index, self.message, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.actor_ref, self.from_index, self.message, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message is not None or self.message_id is not None or self.message_type is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.actor_ref, self.from_index, self.message, self.message_id, self.message_type, self.to_index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message is not None or self.message_id is not None or self.message_type is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_SEND_MESSAGE, (type(self), 0x041d35d, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->actor_ref) != Py_None);
    __pyx_t_6 = (__pyx_t_4 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->message != Py_None);
    __pyx_t_4 = (__pyx_t_6 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_5 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->message_id != ((PyObject*)Py_None));
    __pyx_t_6 = (__pyx_t_4 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->message_type != Py_None);
    __pyx_t_4 = (__pyx_t_6 != 0);
    __pyx_t_5 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_5;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message is not None or self.message_id is not None or self.message_type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SEND_MESSAGE, (type(self), 0x041d35d, None), state
 *     else:
 */
  __pyx_t_5 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":13
 *         use_setstate = self.actor_ref is not None or self.message is not None or self.message_id is not None or self.message_type is not None
 *     if use_setstate:
 *         return __pyx_unpickle_SEND_MESSAGE, (type(self), 0x041d35d, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_SEND_MESSAGE, (type(self), 0x041d35d, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_SEND_MESSAGE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_4313949);
    __Pyx_GIVEREF(__pyx_int_4313949);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_4313949);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.actor_ref is not None or self.message is not None or self.message_id is not None or self.message_type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SEND_MESSAGE, (type(self), 0x041d35d, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_SEND_MESSAGE, (type(self), 0x041d35d, None), state
 *     else:
 *         return __pyx_unpickle_SEND_MESSAGE, (type(self), 0x041d35d, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SEND_MESSAGE__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_SEND_MESSAGE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_4313949);
    __Pyx_GIVEREF(__pyx_int_4313949);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_4313949);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mars.actors.pool.messages.SEND_MESSAGE.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SEND_MESSAGE, (type(self), 0x041d35d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SEND_MESSAGE__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_4__setstate_cython__(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12SEND_MESSAGE_4__setstate_cython__(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_SEND_MESSAGE, (type(self), 0x041d35d, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SEND_MESSAGE__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_SEND_MESSAGE__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SEND_MESSAGE, (type(self), 0x041d35d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SEND_MESSAGE__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.SEND_MESSAGE.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":186
 * 
 * 
 * cdef inline bytes new_message_id():             # <<<<<<<<<<<<<<
 *     return np.random.bytes(32)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages_new_message_id(void) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("new_message_id", 0);

  /* "mars/actors/pool/messages.pyx":187
 * 
 * cdef inline bytes new_message_id():
 *     return np.random.bytes(32)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_random); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_bytes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_int_32) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_int_32);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 187, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":186
 * 
 * 
 * cdef inline bytes new_message_id():             # <<<<<<<<<<<<<<
 *     return np.random.bytes(32)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mars.actors.pool.messages.new_message_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":193
 * @cython.wraparound(False)
 * @cython.nonecheck(False)
 * cdef inline bytes _read_bytes(bytes binary, size_t* pos, size_t size):             # <<<<<<<<<<<<<<
 *     cdef bytes res
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__read_bytes(PyObject *__pyx_v_binary, size_t *__pyx_v_pos, size_t __pyx_v_size) {
  PyObject *__pyx_v_res = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  __Pyx_RefNannySetupContext("_read_bytes", 0);

  /* "mars/actors/pool/messages.pyx":196
 *     cdef bytes res
 * 
 *     res = binary[pos[0]: pos[0]+size]             # <<<<<<<<<<<<<<
 *     pos[0] += size
 * 
 */
  if (unlikely(__pyx_v_binary == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 196, __pyx_L1_error)
  }
  __pyx_t_1 = PySequence_GetSlice(__pyx_v_binary, (__pyx_v_pos[0]), ((__pyx_v_pos[0]) + __pyx_v_size)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_res = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":197
 * 
 *     res = binary[pos[0]: pos[0]+size]
 *     pos[0] += size             # <<<<<<<<<<<<<<
 * 
 *     return res
 */
  __pyx_t_2 = 0;
  (__pyx_v_pos[__pyx_t_2]) = ((__pyx_v_pos[__pyx_t_2]) + __pyx_v_size);

  /* "mars/actors/pool/messages.pyx":199
 *     pos[0] += size
 * 
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":193
 * @cython.wraparound(False)
 * @cython.nonecheck(False)
 * cdef inline bytes _read_bytes(bytes binary, size_t* pos, size_t size):             # <<<<<<<<<<<<<<
 *     cdef bytes res
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages._read_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":202
 * 
 * 
 * cdef inline void _pack_message_type(int message_type, bytearray buf, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef int value
 * 
 */

static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_message_type(int __pyx_v_message_type, PyObject *__pyx_v_buf, struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_message_type *__pyx_optional_args) {
  int __pyx_v_protocol = __pyx_k_;
  int __pyx_v_value;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_pack_message_type", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_protocol = __pyx_optional_args->protocol;
    }
  }

  /* "mars/actors/pool/messages.pyx":206
 * 
 *     # use 1 byte to represent protocol and message type, from left to right, 0-2, protocol, 3-7 message_type
 *     value = (protocol << 5) | message_type             # <<<<<<<<<<<<<<
 *     buf.extend(_pack_byte(value))
 * 
 */
  __pyx_v_value = ((__pyx_v_protocol << 5) | __pyx_v_message_type);

  /* "mars/actors/pool/messages.pyx":207
 *     # use 1 byte to represent protocol and message type, from left to right, 0-2, protocol, 3-7 message_type
 *     value = (protocol << 5) | message_type
 *     buf.extend(_pack_byte(value))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__pack_byte);
  __pyx_t_3 = __pyx_v_4mars_6actors_4pool_8messages__pack_byte; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_buf, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mars/actors/pool/messages.pyx":202
 * 
 * 
 * cdef inline void _pack_message_type(int message_type, bytearray buf, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef int value
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("mars.actors.pool.messages._pack_message_type", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "mars/actors/pool/messages.pyx":210
 * 
 * 
 * cdef inline int _unpack_message_type_value(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef int value
 *     cdef int protocol
 */

static CYTHON_INLINE int __pyx_f_4mars_6actors_4pool_8messages__unpack_message_type_value(PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  int __pyx_v_value;
  int __pyx_v_protocol;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("_unpack_message_type_value", 0);

  /* "mars/actors/pool/messages.pyx":214
 *     cdef int protocol
 * 
 *     value, = _unpack_byte(_read_bytes(binary, pos, 1))             # <<<<<<<<<<<<<<
 *     protocol = value >> 5
 *     if protocol != 0:
 */
  __pyx_t_2 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_byte);
  __pyx_t_3 = __pyx_v_4mars_6actors_4pool_8messages__unpack_byte; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 214, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_5(__pyx_t_2); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_2), 1) < 0) __PYX_ERR(0, 214, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 214, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_value = __pyx_t_6;

  /* "mars/actors/pool/messages.pyx":215
 * 
 *     value, = _unpack_byte(_read_bytes(binary, pos, 1))
 *     protocol = value >> 5             # <<<<<<<<<<<<<<
 *     if protocol != 0:
 *         raise NotImplementedError('Unsupported protocol')
 */
  __pyx_v_protocol = (__pyx_v_value >> 5);

  /* "mars/actors/pool/messages.pyx":216
 *     value, = _unpack_byte(_read_bytes(binary, pos, 1))
 *     protocol = value >> 5
 *     if protocol != 0:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError('Unsupported protocol')
 * 
 */
  __pyx_t_7 = ((__pyx_v_protocol != 0) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "mars/actors/pool/messages.pyx":217
 *     protocol = value >> 5
 *     if protocol != 0:
 *         raise NotImplementedError('Unsupported protocol')             # <<<<<<<<<<<<<<
 * 
 *     return value & ((1 << 5) - 1)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 217, __pyx_L1_error)

    /* "mars/actors/pool/messages.pyx":216
 *     value, = _unpack_byte(_read_bytes(binary, pos, 1))
 *     protocol = value >> 5
 *     if protocol != 0:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError('Unsupported protocol')
 * 
 */
  }

  /* "mars/actors/pool/messages.pyx":219
 *         raise NotImplementedError('Unsupported protocol')
 * 
 *     return value & ((1 << 5) - 1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_value & 31);
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":210
 * 
 * 
 * cdef inline int _unpack_message_type_value(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef int value
 *     cdef int protocol
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("mars.actors.pool.messages._unpack_message_type_value", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":222
 * 
 * 
 * cdef inline object _unpack_message_type(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     return MessageType(_unpack_message_type_value(binary, pos))
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_message_type(PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_unpack_message_type", 0);

  /* "mars/actors/pool/messages.pyx":223
 * 
 * cdef inline object _unpack_message_type(bytes binary, size_t* pos):
 *     return MessageType(_unpack_message_type_value(binary, pos))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_f_4mars_6actors_4pool_8messages__unpack_message_type_value(__pyx_v_binary, __pyx_v_pos)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":222
 * 
 * 
 * cdef inline object _unpack_message_type(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     return MessageType(_unpack_message_type_value(binary, pos))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.actors.pool.messages._unpack_message_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":226
 * 
 * 
 * cpdef int unpack_message_type_value(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef size_t pos = 0
 *     return _unpack_message_type_value(binary, &pos)
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_1unpack_message_type_value(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static int __pyx_f_4mars_6actors_4pool_8messages_unpack_message_type_value(PyObject *__pyx_v_binary, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_pos;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_message_type_value", 0);

  /* "mars/actors/pool/messages.pyx":227
 * 
 * cpdef int unpack_message_type_value(bytes binary):
 *     cdef size_t pos = 0             # <<<<<<<<<<<<<<
 *     return _unpack_message_type_value(binary, &pos)
 * 
 */
  __pyx_v_pos = 0;

  /* "mars/actors/pool/messages.pyx":228
 * cpdef int unpack_message_type_value(bytes binary):
 *     cdef size_t pos = 0
 *     return _unpack_message_type_value(binary, &pos)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_type_value(__pyx_v_binary, (&__pyx_v_pos));
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":226
 * 
 * 
 * cpdef int unpack_message_type_value(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef size_t pos = 0
 *     return _unpack_message_type_value(binary, &pos)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_1unpack_message_type_value(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_1unpack_message_type_value(PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_message_type_value (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_binary), (&PyBytes_Type), 1, "binary", 1))) __PYX_ERR(0, 226, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_unpack_message_type_value(__pyx_self, ((PyObject*)__pyx_v_binary));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_unpack_message_type_value(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_message_type_value", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4mars_6actors_4pool_8messages_unpack_message_type_value(__pyx_v_binary, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_message_type_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":231
 * 
 * 
 * cpdef object unpack_message_type(bytes binary):             # <<<<<<<<<<<<<<
 *     return MessageType(unpack_message_type_value(binary))
 * 
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_3unpack_message_type(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_message_type(PyObject *__pyx_v_binary, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("unpack_message_type", 0);

  /* "mars/actors/pool/messages.pyx":232
 * 
 * cpdef object unpack_message_type(bytes binary):
 *     return MessageType(unpack_message_type_value(binary))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_f_4mars_6actors_4pool_8messages_unpack_message_type_value(__pyx_v_binary, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":231
 * 
 * 
 * cpdef object unpack_message_type(bytes binary):             # <<<<<<<<<<<<<<
 *     return MessageType(unpack_message_type_value(binary))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_message_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_3unpack_message_type(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_3unpack_message_type(PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_message_type (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_binary), (&PyBytes_Type), 1, "binary", 1))) __PYX_ERR(0, 231, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_2unpack_message_type(__pyx_self, ((PyObject*)__pyx_v_binary));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_2unpack_message_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_message_type", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_unpack_message_type(__pyx_v_binary, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_message_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":238
 * @cython.wraparound(False)
 * @cython.nonecheck(False)
 * cdef inline void _skip_message_type(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     pos[0] += 1
 * 
 */

static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__skip_message_type(CYTHON_UNUSED PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  __Pyx_RefNannySetupContext("_skip_message_type", 0);

  /* "mars/actors/pool/messages.pyx":239
 * @cython.nonecheck(False)
 * cdef inline void _skip_message_type(bytes binary, size_t* pos):
 *     pos[0] += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = 0;
  (__pyx_v_pos[__pyx_t_1]) = ((__pyx_v_pos[__pyx_t_1]) + 1);

  /* "mars/actors/pool/messages.pyx":238
 * @cython.wraparound(False)
 * @cython.nonecheck(False)
 * cdef inline void _skip_message_type(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     pos[0] += 1
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "mars/actors/pool/messages.pyx":242
 * 
 * 
 * cpdef bytes unpack_message_id(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef size_t pos = 1
 * 
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_5unpack_message_id(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_message_id(PyObject *__pyx_v_binary, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_message_id", 0);

  /* "mars/actors/pool/messages.pyx":243
 * 
 * cpdef bytes unpack_message_id(bytes binary):
 *     cdef size_t pos = 1             # <<<<<<<<<<<<<<
 * 
 *     return _unpack_message_id(binary, &pos)
 */
  __pyx_v_pos = 1;

  /* "mars/actors/pool/messages.pyx":245
 *     cdef size_t pos = 1
 * 
 *     return _unpack_message_id(binary, &pos)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_id(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":242
 * 
 * 
 * cpdef bytes unpack_message_id(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef size_t pos = 1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_message_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_5unpack_message_id(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_5unpack_message_id(PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_message_id (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_binary), (&PyBytes_Type), 1, "binary", 1))) __PYX_ERR(0, 242, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_4unpack_message_id(__pyx_self, ((PyObject*)__pyx_v_binary));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_4unpack_message_id(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_message_id", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_unpack_message_id(__pyx_v_binary, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_message_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":248
 * 
 * 
 * cdef inline void _pack_message_id(bytes message_id, bytearray buf):             # <<<<<<<<<<<<<<
 *     _pack_object(message_id, buf)
 * 
 */

static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_message_id(PyObject *__pyx_v_message_id, PyObject *__pyx_v_buf) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_pack_message_id", 0);

  /* "mars/actors/pool/messages.pyx":249
 * 
 * cdef inline void _pack_message_id(bytes message_id, bytearray buf):
 *     _pack_object(message_id, buf)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_object(__pyx_v_message_id, __pyx_v_buf); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 249, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":248
 * 
 * 
 * cdef inline void _pack_message_id(bytes message_id, bytearray buf):             # <<<<<<<<<<<<<<
 *     _pack_object(message_id, buf)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("mars.actors.pool.messages._pack_message_id", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "mars/actors/pool/messages.pyx":252
 * 
 * 
 * cdef inline bytes _unpack_message_id(bytes buf, size_t* pos):             # <<<<<<<<<<<<<<
 *     return _unpack_object(buf, pos)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_message_id(PyObject *__pyx_v_buf, size_t *__pyx_v_pos) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_unpack_message_id", 0);

  /* "mars/actors/pool/messages.pyx":253
 * 
 * cdef inline bytes _unpack_message_id(bytes buf, size_t* pos):
 *     return _unpack_object(buf, pos)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_buf, __pyx_v_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 253, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":252
 * 
 * 
 * cdef inline bytes _unpack_message_id(bytes buf, size_t* pos):             # <<<<<<<<<<<<<<
 *     return _unpack_object(buf, pos)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages._unpack_message_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":256
 * 
 * 
 * cdef inline void _skip_message_id(bytes buf, size_t* pos):             # <<<<<<<<<<<<<<
 *     _skip_object(buf, pos)
 * 
 */

static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__skip_message_id(PyObject *__pyx_v_buf, size_t *__pyx_v_pos) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_skip_message_id", 0);

  /* "mars/actors/pool/messages.pyx":257
 * 
 * cdef inline void _skip_message_id(bytes buf, size_t* pos):
 *     _skip_object(buf, pos)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_f_4mars_6actors_4pool_8messages__skip_object(__pyx_v_buf, __pyx_v_pos);

  /* "mars/actors/pool/messages.pyx":256
 * 
 * 
 * cdef inline void _skip_message_id(bytes buf, size_t* pos):             # <<<<<<<<<<<<<<
 *     _skip_object(buf, pos)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "mars/actors/pool/messages.pyx":260
 * 
 * 
 * cdef inline void _pack_object(object obj, bytearray buf) except *:             # <<<<<<<<<<<<<<
 *     cdef BYTE_t st
 *     cdef bytes m
 */

static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_object(PyObject *__pyx_v_obj, PyObject *__pyx_v_buf) {
  __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_st;
  PyObject *__pyx_v_m = 0;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  __Pyx_RefNannySetupContext("_pack_object", 0);

  /* "mars/actors/pool/messages.pyx":264
 *     cdef bytes m
 * 
 *     if obj is None:             # <<<<<<<<<<<<<<
 *         buf.extend(_pack_byte(NONE))
 *         return
 */
  __pyx_t_1 = (__pyx_v_obj == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "mars/actors/pool/messages.pyx":265
 * 
 *     if obj is None:
 *         buf.extend(_pack_byte(NONE))             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_npy_uint8(__pyx_v_4mars_6actors_4pool_8messages_NONE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__pack_byte);
    __pyx_t_5 = __pyx_v_4mars_6actors_4pool_8messages__pack_byte; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_buf, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "mars/actors/pool/messages.pyx":266
 *     if obj is None:
 *         buf.extend(_pack_byte(NONE))
 *         return             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(obj, bytes):
 */
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":264
 *     cdef bytes m
 * 
 *     if obj is None:             # <<<<<<<<<<<<<<
 *         buf.extend(_pack_byte(NONE))
 *         return
 */
  }

  /* "mars/actors/pool/messages.pyx":268
 *         return
 * 
 *     if isinstance(obj, bytes):             # <<<<<<<<<<<<<<
 *         st = RAW_BYTES
 *         m = obj
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_obj); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "mars/actors/pool/messages.pyx":269
 * 
 *     if isinstance(obj, bytes):
 *         st = RAW_BYTES             # <<<<<<<<<<<<<<
 *         m = obj
 *     else:
 */
    __pyx_v_st = __pyx_v_4mars_6actors_4pool_8messages_RAW_BYTES;

    /* "mars/actors/pool/messages.pyx":270
 *     if isinstance(obj, bytes):
 *         st = RAW_BYTES
 *         m = obj             # <<<<<<<<<<<<<<
 *     else:
 *         st = PICKLE
 */
    if (!(likely(PyBytes_CheckExact(__pyx_v_obj))||((__pyx_v_obj) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_obj)->tp_name), 0))) __PYX_ERR(0, 270, __pyx_L1_error)
    __pyx_t_5 = __pyx_v_obj;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_v_m = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "mars/actors/pool/messages.pyx":268
 *         return
 * 
 *     if isinstance(obj, bytes):             # <<<<<<<<<<<<<<
 *         st = RAW_BYTES
 *         m = obj
 */
    goto __pyx_L4;
  }

  /* "mars/actors/pool/messages.pyx":272
 *         m = obj
 *     else:
 *         st = PICKLE             # <<<<<<<<<<<<<<
 *         m = dumps(obj)
 * 
 */
  /*else*/ {
    __pyx_v_st = __pyx_v_4mars_6actors_4pool_8messages_PICKLE;

    /* "mars/actors/pool/messages.pyx":273
 *     else:
 *         st = PICKLE
 *         m = dumps(obj)             # <<<<<<<<<<<<<<
 * 
 *     buf.extend(_pack_byte(st))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_dumps); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_obj);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 273, __pyx_L1_error)
    __pyx_v_m = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
  }
  __pyx_L4:;

  /* "mars/actors/pool/messages.pyx":275
 *         m = dumps(obj)
 * 
 *     buf.extend(_pack_byte(st))             # <<<<<<<<<<<<<<
 *     buf.extend(_pack_long(len(m)))
 *     buf.extend(m)
 */
  __pyx_t_3 = __Pyx_PyInt_From_npy_uint8(__pyx_v_st); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__pack_byte);
  __pyx_t_4 = __pyx_v_4mars_6actors_4pool_8messages__pack_byte; __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_buf, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "mars/actors/pool/messages.pyx":276
 * 
 *     buf.extend(_pack_byte(st))
 *     buf.extend(_pack_long(len(m)))             # <<<<<<<<<<<<<<
 *     buf.extend(m)
 * 
 */
  if (unlikely(__pyx_v_m == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 276, __pyx_L1_error)
  }
  __pyx_t_7 = PyBytes_GET_SIZE(__pyx_v_m); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 276, __pyx_L1_error)
  __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__pack_long);
  __pyx_t_3 = __pyx_v_4mars_6actors_4pool_8messages__pack_long; __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_buf, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mars/actors/pool/messages.pyx":277
 *     buf.extend(_pack_byte(st))
 *     buf.extend(_pack_long(len(m)))
 *     buf.extend(m)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_buf, __pyx_v_m); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mars/actors/pool/messages.pyx":260
 * 
 * 
 * cdef inline void _pack_object(object obj, bytearray buf) except *:             # <<<<<<<<<<<<<<
 *     cdef BYTE_t st
 *     cdef bytes m
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("mars.actors.pool.messages._pack_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_RefNannyFinishContext();
}

/* "mars/actors/pool/messages.pyx":280
 * 
 * 
 * cdef inline object _unpack_object(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef BYTE_t st
 *     cdef size_t size
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_object(PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_st;
  size_t __pyx_v_size;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_t_6;
  int __pyx_t_7;
  size_t __pyx_t_8;
  __Pyx_RefNannySetupContext("_unpack_object", 0);

  /* "mars/actors/pool/messages.pyx":284
 *     cdef size_t size
 * 
 *     st, = _unpack_byte(_read_bytes(binary, pos, 1))             # <<<<<<<<<<<<<<
 * 
 *     if st == NONE:
 */
  __pyx_t_2 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_byte);
  __pyx_t_3 = __pyx_v_4mars_6actors_4pool_8messages__unpack_byte; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 284, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_5(__pyx_t_2); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_2), 1) < 0) __PYX_ERR(0, 284, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 284, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint8(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint8)-1)) && PyErr_Occurred())) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_st = __pyx_t_6;

  /* "mars/actors/pool/messages.pyx":286
 *     st, = _unpack_byte(_read_bytes(binary, pos, 1))
 * 
 *     if st == NONE:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_7 = ((__pyx_v_st == __pyx_v_4mars_6actors_4pool_8messages_NONE) != 0);
  if (__pyx_t_7) {

    /* "mars/actors/pool/messages.pyx":287
 * 
 *     if st == NONE:
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     size, = _unpack_long(_read_bytes(binary, pos, 8))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":286
 *     st, = _unpack_byte(_read_bytes(binary, pos, 1))
 * 
 *     if st == NONE:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "mars/actors/pool/messages.pyx":289
 *         return None
 * 
 *     size, = _unpack_long(_read_bytes(binary, pos, 8))             # <<<<<<<<<<<<<<
 * 
 *     if st == RAW_BYTES:
 */
  __pyx_t_3 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, 8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_long);
  __pyx_t_2 = __pyx_v_4mars_6actors_4pool_8messages__unpack_long; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 289, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_3), 1) < 0) __PYX_ERR(0, 289, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 289, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __pyx_t_8 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_size = __pyx_t_8;

  /* "mars/actors/pool/messages.pyx":291
 *     size, = _unpack_long(_read_bytes(binary, pos, 8))
 * 
 *     if st == RAW_BYTES:             # <<<<<<<<<<<<<<
 *         return _read_bytes(binary, pos, size)
 *     else:
 */
  __pyx_t_7 = ((__pyx_v_st == __pyx_v_4mars_6actors_4pool_8messages_RAW_BYTES) != 0);
  if (__pyx_t_7) {

    /* "mars/actors/pool/messages.pyx":292
 * 
 *     if st == RAW_BYTES:
 *         return _read_bytes(binary, pos, size)             # <<<<<<<<<<<<<<
 *     else:
 *         return loads(_read_bytes(binary, pos, size))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, __pyx_v_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":291
 *     size, = _unpack_long(_read_bytes(binary, pos, 8))
 * 
 *     if st == RAW_BYTES:             # <<<<<<<<<<<<<<
 *         return _read_bytes(binary, pos, size)
 *     else:
 */
  }

  /* "mars/actors/pool/messages.pyx":294
 *         return _read_bytes(binary, pos, size)
 *     else:
 *         return loads(_read_bytes(binary, pos, size))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_loads); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, __pyx_v_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "mars/actors/pool/messages.pyx":280
 * 
 * 
 * cdef inline object _unpack_object(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef BYTE_t st
 *     cdef size_t size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.actors.pool.messages._unpack_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":297
 * 
 * 
 * cdef inline void _skip_object(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef BYTE_t st
 *     cdef size_t size
 */

static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__skip_object(PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_st;
  size_t __pyx_v_size;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_t_6;
  int __pyx_t_7;
  size_t __pyx_t_8;
  long __pyx_t_9;
  __Pyx_RefNannySetupContext("_skip_object", 0);

  /* "mars/actors/pool/messages.pyx":301
 *     cdef size_t size
 * 
 *     st, = _unpack_byte(_read_bytes(binary, pos, 1))             # <<<<<<<<<<<<<<
 * 
 *     if st == NONE:
 */
  __pyx_t_2 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_byte);
  __pyx_t_3 = __pyx_v_4mars_6actors_4pool_8messages__unpack_byte; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 301, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_5(__pyx_t_2); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_2), 1) < 0) __PYX_ERR(0, 301, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 301, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint8(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint8)-1)) && PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_st = __pyx_t_6;

  /* "mars/actors/pool/messages.pyx":303
 *     st, = _unpack_byte(_read_bytes(binary, pos, 1))
 * 
 *     if st == NONE:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_7 = ((__pyx_v_st == __pyx_v_4mars_6actors_4pool_8messages_NONE) != 0);
  if (__pyx_t_7) {

    /* "mars/actors/pool/messages.pyx":304
 * 
 *     if st == NONE:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     size, = _unpack_long(_read_bytes(binary, pos, 8))
 */
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":303
 *     st, = _unpack_byte(_read_bytes(binary, pos, 1))
 * 
 *     if st == NONE:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  }

  /* "mars/actors/pool/messages.pyx":306
 *         return
 * 
 *     size, = _unpack_long(_read_bytes(binary, pos, 8))             # <<<<<<<<<<<<<<
 *     pos[0] += size
 * 
 */
  __pyx_t_3 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, 8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_long);
  __pyx_t_2 = __pyx_v_4mars_6actors_4pool_8messages__unpack_long; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 306, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_3), 1) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 306, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __pyx_t_8 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_size = __pyx_t_8;

  /* "mars/actors/pool/messages.pyx":307
 * 
 *     size, = _unpack_long(_read_bytes(binary, pos, 8))
 *     pos[0] += size             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_9 = 0;
  (__pyx_v_pos[__pyx_t_9]) = ((__pyx_v_pos[__pyx_t_9]) + __pyx_v_size);

  /* "mars/actors/pool/messages.pyx":297
 * 
 * 
 * cdef inline void _skip_object(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef BYTE_t st
 *     cdef size_t size
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("mars.actors.pool.messages._skip_object", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "mars/actors/pool/messages.pyx":310
 * 
 * 
 * cdef inline void _pack_index(int index, bytearray buf):             # <<<<<<<<<<<<<<
 *     # 2 bytes
 *     buf.extend(_pack_short(index))
 */

static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_index(int __pyx_v_index, PyObject *__pyx_v_buf) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_pack_index", 0);

  /* "mars/actors/pool/messages.pyx":312
 * cdef inline void _pack_index(int index, bytearray buf):
 *     # 2 bytes
 *     buf.extend(_pack_short(index))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__pack_short);
  __pyx_t_3 = __pyx_v_4mars_6actors_4pool_8messages__pack_short; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_buf, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mars/actors/pool/messages.pyx":310
 * 
 * 
 * cdef inline void _pack_index(int index, bytearray buf):             # <<<<<<<<<<<<<<
 *     # 2 bytes
 *     buf.extend(_pack_short(index))
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("mars.actors.pool.messages._pack_index", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "mars/actors/pool/messages.pyx":315
 * 
 * 
 * cdef inline INT32_t _unpack_index(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef int index
 * 
 */

static CYTHON_INLINE __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_f_4mars_6actors_4pool_8messages__unpack_index(PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  int __pyx_v_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("_unpack_index", 0);

  /* "mars/actors/pool/messages.pyx":318
 *     cdef int index
 * 
 *     index, = _unpack_short(_read_bytes(binary, pos, 2))             # <<<<<<<<<<<<<<
 *     return index
 * 
 */
  __pyx_t_2 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_short);
  __pyx_t_3 = __pyx_v_4mars_6actors_4pool_8messages__unpack_short; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 318, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_5(__pyx_t_2); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_2), 1) < 0) __PYX_ERR(0, 318, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 318, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_index = __pyx_t_6;

  /* "mars/actors/pool/messages.pyx":319
 * 
 *     index, = _unpack_short(_read_bytes(binary, pos, 2))
 *     return index             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_index;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":315
 * 
 * 
 * cdef inline INT32_t _unpack_index(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef int index
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("mars.actors.pool.messages._unpack_index", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":325
 * @cython.wraparound(False)
 * @cython.nonecheck(False)
 * cdef inline void _skip_index(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     pos[0] += 2
 * 
 */

static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__skip_index(CYTHON_UNUSED PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  __Pyx_RefNannySetupContext("_skip_index", 0);

  /* "mars/actors/pool/messages.pyx":326
 * @cython.nonecheck(False)
 * cdef inline void _skip_index(bytes binary, size_t* pos):
 *     pos[0] += 2             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = 0;
  (__pyx_v_pos[__pyx_t_1]) = ((__pyx_v_pos[__pyx_t_1]) + 2);

  /* "mars/actors/pool/messages.pyx":325
 * @cython.wraparound(False)
 * @cython.nonecheck(False)
 * cdef inline void _skip_index(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     pos[0] += 2
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "mars/actors/pool/messages.pyx":329
 * 
 * 
 * cdef inline void _pack_actor_ref(ActorRef actor_ref, bytearray buf):             # <<<<<<<<<<<<<<
 *     # this line circumvents a bug in Cython which led to application crash
 *     assert actor_ref is not None
 */

static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_actor_ref(struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_buf) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_pack_actor_ref", 0);

  /* "mars/actors/pool/messages.pyx":331
 * cdef inline void _pack_actor_ref(ActorRef actor_ref, bytearray buf):
 *     # this line circumvents a bug in Cython which led to application crash
 *     assert actor_ref is not None             # <<<<<<<<<<<<<<
 *     _pack_object(actor_ref.address, buf)
 *     _pack_object(actor_ref.uid, buf)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = (((PyObject *)__pyx_v_actor_ref) != Py_None);
    if (unlikely(!(__pyx_t_1 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 331, __pyx_L1_error)
    }
  }
  #endif

  /* "mars/actors/pool/messages.pyx":332
 *     # this line circumvents a bug in Cython which led to application crash
 *     assert actor_ref is not None
 *     _pack_object(actor_ref.address, buf)             # <<<<<<<<<<<<<<
 *     _pack_object(actor_ref.uid, buf)
 * 
 */
  __pyx_t_2 = __pyx_v_actor_ref->address;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_f_4mars_6actors_4pool_8messages__pack_object(__pyx_t_2, __pyx_v_buf); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":333
 *     assert actor_ref is not None
 *     _pack_object(actor_ref.address, buf)
 *     _pack_object(actor_ref.uid, buf)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_actor_ref->uid;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_f_4mars_6actors_4pool_8messages__pack_object(__pyx_t_2, __pyx_v_buf); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":329
 * 
 * 
 * cdef inline void _pack_actor_ref(ActorRef actor_ref, bytearray buf):             # <<<<<<<<<<<<<<
 *     # this line circumvents a bug in Cython which led to application crash
 *     assert actor_ref is not None
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("mars.actors.pool.messages._pack_actor_ref", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "mars/actors/pool/messages.pyx":336
 * 
 * 
 * cdef inline ActorRef _unpack_actor_ref(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef object address
 *     cdef object uid
 */

static CYTHON_INLINE struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_f_4mars_6actors_4pool_8messages__unpack_actor_ref(PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  PyObject *__pyx_v_address = 0;
  PyObject *__pyx_v_uid = 0;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_unpack_actor_ref", 0);

  /* "mars/actors/pool/messages.pyx":340
 *     cdef object uid
 * 
 *     address = _unpack_object(binary, pos)             # <<<<<<<<<<<<<<
 *     uid = _unpack_object(binary, pos)
 *     return ActorRef(address, uid)
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, __pyx_v_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_address = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":341
 * 
 *     address = _unpack_object(binary, pos)
 *     uid = _unpack_object(binary, pos)             # <<<<<<<<<<<<<<
 *     return ActorRef(address, uid)
 * 
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, __pyx_v_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_uid = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":342
 *     address = _unpack_object(binary, pos)
 *     uid = _unpack_object(binary, pos)
 *     return ActorRef(address, uid)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_address);
  __Pyx_GIVEREF(__pyx_v_address);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_address);
  __Pyx_INCREF(__pyx_v_uid);
  __Pyx_GIVEREF(__pyx_v_uid);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_uid);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4mars_6actors_4core_ActorRef), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":336
 * 
 * 
 * cdef inline ActorRef _unpack_actor_ref(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef object address
 *     cdef object uid
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mars.actors.pool.messages._unpack_actor_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_address);
  __Pyx_XDECREF(__pyx_v_uid);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":345
 * 
 * 
 * cpdef bytes pack_actor_ref(ActorRef actor_ref):             # <<<<<<<<<<<<<<
 *     cdef bytearray buf = bytearray()
 * 
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_7pack_actor_ref(PyObject *__pyx_self, PyObject *__pyx_v_actor_ref); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_actor_ref(struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_buf = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pack_actor_ref", 0);

  /* "mars/actors/pool/messages.pyx":346
 * 
 * cpdef bytes pack_actor_ref(ActorRef actor_ref):
 *     cdef bytearray buf = bytearray()             # <<<<<<<<<<<<<<
 * 
 *     _pack_actor_ref(actor_ref, buf)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyByteArray_Type))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buf = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":348
 *     cdef bytearray buf = bytearray()
 * 
 *     _pack_actor_ref(actor_ref, buf)             # <<<<<<<<<<<<<<
 *     return bytes(buf)
 * 
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_actor_ref(__pyx_v_actor_ref, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":349
 * 
 *     _pack_actor_ref(actor_ref, buf)
 *     return bytes(buf)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_v_buf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":345
 * 
 * 
 * cpdef bytes pack_actor_ref(ActorRef actor_ref):             # <<<<<<<<<<<<<<
 *     cdef bytearray buf = bytearray()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_actor_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_7pack_actor_ref(PyObject *__pyx_self, PyObject *__pyx_v_actor_ref); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_7pack_actor_ref(PyObject *__pyx_self, PyObject *__pyx_v_actor_ref) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_actor_ref (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_actor_ref), __pyx_ptype_4mars_6actors_4core_ActorRef, 1, "actor_ref", 0))) __PYX_ERR(0, 345, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_6pack_actor_ref(__pyx_self, ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_v_actor_ref));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_6pack_actor_ref(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pack_actor_ref", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_pack_actor_ref(__pyx_v_actor_ref, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_actor_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":352
 * 
 * 
 * cpdef void unpack_actor_ref(bytes binary, ActorRef actor_ref):             # <<<<<<<<<<<<<<
 *     cdef size_t pos = 0
 *     cdef object address
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_9unpack_actor_ref(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static void __pyx_f_4mars_6actors_4pool_8messages_unpack_actor_ref(PyObject *__pyx_v_binary, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_pos;
  PyObject *__pyx_v_address = 0;
  PyObject *__pyx_v_uid = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_actor_ref", 0);

  /* "mars/actors/pool/messages.pyx":353
 * 
 * cpdef void unpack_actor_ref(bytes binary, ActorRef actor_ref):
 *     cdef size_t pos = 0             # <<<<<<<<<<<<<<
 *     cdef object address
 *     cdef object uid
 */
  __pyx_v_pos = 0;

  /* "mars/actors/pool/messages.pyx":357
 *     cdef object uid
 * 
 *     address = _unpack_object(binary, &pos)             # <<<<<<<<<<<<<<
 *     uid = _unpack_object(binary, &pos)
 * 
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_address = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":358
 * 
 *     address = _unpack_object(binary, &pos)
 *     uid = _unpack_object(binary, &pos)             # <<<<<<<<<<<<<<
 * 
 *     actor_ref.address = address
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_uid = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":360
 *     uid = _unpack_object(binary, &pos)
 * 
 *     actor_ref.address = address             # <<<<<<<<<<<<<<
 *     actor_ref.uid = uid
 * 
 */
  if (!(likely(PyString_CheckExact(__pyx_v_address))||((__pyx_v_address) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_address)->tp_name), 0))) __PYX_ERR(0, 360, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_address;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_actor_ref->address);
  __Pyx_DECREF(__pyx_v_actor_ref->address);
  __pyx_v_actor_ref->address = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":361
 * 
 *     actor_ref.address = address
 *     actor_ref.uid = uid             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_uid);
  __Pyx_GIVEREF(__pyx_v_uid);
  __Pyx_GOTREF(__pyx_v_actor_ref->uid);
  __Pyx_DECREF(__pyx_v_actor_ref->uid);
  __pyx_v_actor_ref->uid = __pyx_v_uid;

  /* "mars/actors/pool/messages.pyx":352
 * 
 * 
 * cpdef void unpack_actor_ref(bytes binary, ActorRef actor_ref):             # <<<<<<<<<<<<<<
 *     cdef size_t pos = 0
 *     cdef object address
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("mars.actors.pool.messages.unpack_actor_ref", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_address);
  __Pyx_XDECREF(__pyx_v_uid);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_9unpack_actor_ref(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_9unpack_actor_ref(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_binary = 0;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_actor_ref (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_binary,&__pyx_n_s_actor_ref,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_binary)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_ref)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("unpack_actor_ref", 1, 2, 2, 1); __PYX_ERR(0, 352, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "unpack_actor_ref") < 0)) __PYX_ERR(0, 352, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_binary = ((PyObject*)values[0]);
    __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("unpack_actor_ref", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 352, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_actor_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_binary), (&PyBytes_Type), 1, "binary", 1))) __PYX_ERR(0, 352, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_actor_ref), __pyx_ptype_4mars_6actors_4core_ActorRef, 1, "actor_ref", 0))) __PYX_ERR(0, 352, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_8unpack_actor_ref(__pyx_self, __pyx_v_binary, __pyx_v_actor_ref);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_8unpack_actor_ref(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_actor_ref", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_4mars_6actors_4pool_8messages_unpack_actor_ref(__pyx_v_binary, __pyx_v_actor_ref, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_actor_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":364
 * 
 * 
 * cdef inline void _skip_address(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     _skip_object(binary, pos)
 * 
 */

static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__skip_address(PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_skip_address", 0);

  /* "mars/actors/pool/messages.pyx":365
 * 
 * cdef inline void _skip_address(bytes binary, size_t* pos):
 *     _skip_object(binary, pos)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_f_4mars_6actors_4pool_8messages__skip_object(__pyx_v_binary, __pyx_v_pos);

  /* "mars/actors/pool/messages.pyx":364
 * 
 * 
 * cdef inline void _skip_address(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     _skip_object(binary, pos)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "mars/actors/pool/messages.pyx":368
 * 
 * 
 * cdef inline object _unpack_uid(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     return _unpack_object(binary, pos)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_uid(PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_unpack_uid", 0);

  /* "mars/actors/pool/messages.pyx":369
 * 
 * cdef inline object _unpack_uid(bytes binary, size_t* pos):
 *     return _unpack_object(binary, pos)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, __pyx_v_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":368
 * 
 * 
 * cdef inline object _unpack_uid(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     return _unpack_object(binary, pos)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages._unpack_uid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":372
 * 
 * 
 * cdef inline void _pack_message_size(size_t size, bytearray buf):             # <<<<<<<<<<<<<<
 *     buf.extend(_pack_long(size))
 * 
 */

static CYTHON_INLINE void __pyx_f_4mars_6actors_4pool_8messages__pack_message_size(size_t __pyx_v_size, PyObject *__pyx_v_buf) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_pack_message_size", 0);

  /* "mars/actors/pool/messages.pyx":373
 * 
 * cdef inline void _pack_message_size(size_t size, bytearray buf):
 *     buf.extend(_pack_long(size))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__pack_long);
  __pyx_t_3 = __pyx_v_4mars_6actors_4pool_8messages__pack_long; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_buf, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mars/actors/pool/messages.pyx":372
 * 
 * 
 * cdef inline void _pack_message_size(size_t size, bytearray buf):             # <<<<<<<<<<<<<<
 *     buf.extend(_pack_long(size))
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("mars.actors.pool.messages._pack_message_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "mars/actors/pool/messages.pyx":376
 * 
 * 
 * cdef inline int _unpack_message_size(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     return _unpack_long(_read_bytes(binary, pos, 8))[0]
 * 
 */

static CYTHON_INLINE int __pyx_f_4mars_6actors_4pool_8messages__unpack_message_size(PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("_unpack_message_size", 0);

  /* "mars/actors/pool/messages.pyx":377
 * 
 * cdef inline int _unpack_message_size(bytes binary, size_t* pos):
 *     return _unpack_long(_read_bytes(binary, pos, 8))[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, 8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_long);
  __pyx_t_3 = __pyx_v_4mars_6actors_4pool_8messages__unpack_long; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":376
 * 
 * 
 * cdef inline int _unpack_message_size(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     return _unpack_long(_read_bytes(binary, pos, 8))[0]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("mars.actors.pool.messages._unpack_message_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":380
 * 
 * 
 * cdef inline list _pack_tuple_message(tuple messages):             # <<<<<<<<<<<<<<
 *     cdef list ret
 *     cdef bytearray bio
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__pack_tuple_message(PyObject *__pyx_v_messages) {
  PyObject *__pyx_v_ret = 0;
  PyObject *__pyx_v_bio = 0;
  __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_st;
  PyObject *__pyx_v_m = 0;
  PyObject *__pyx_v_message = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("_pack_tuple_message", 0);

  /* "mars/actors/pool/messages.pyx":386
 *     cdef bytes m
 * 
 *     ret = []             # <<<<<<<<<<<<<<
 *     for message in messages:
 *         bio = bytearray()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ret = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":387
 * 
 *     ret = []
 *     for message in messages:             # <<<<<<<<<<<<<<
 *         bio = bytearray()
 * 
 */
  if (unlikely(__pyx_v_messages == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 387, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_messages; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 387, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_message, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "mars/actors/pool/messages.pyx":388
 *     ret = []
 *     for message in messages:
 *         bio = bytearray()             # <<<<<<<<<<<<<<
 * 
 *         # use 1 byte to record the serialize type
 */
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyByteArray_Type))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_bio, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "mars/actors/pool/messages.pyx":391
 * 
 *         # use 1 byte to record the serialize type
 *         if isinstance(message, bytes):             # <<<<<<<<<<<<<<
 *             st = RAW_BYTES
 *             m = message
 */
    __pyx_t_4 = PyBytes_Check(__pyx_v_message); 
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "mars/actors/pool/messages.pyx":392
 *         # use 1 byte to record the serialize type
 *         if isinstance(message, bytes):
 *             st = RAW_BYTES             # <<<<<<<<<<<<<<
 *             m = message
 *         else:
 */
      __pyx_v_st = __pyx_v_4mars_6actors_4pool_8messages_RAW_BYTES;

      /* "mars/actors/pool/messages.pyx":393
 *         if isinstance(message, bytes):
 *             st = RAW_BYTES
 *             m = message             # <<<<<<<<<<<<<<
 *         else:
 *             st = PICKLE
 */
      if (!(likely(PyBytes_CheckExact(__pyx_v_message))||((__pyx_v_message) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_message)->tp_name), 0))) __PYX_ERR(0, 393, __pyx_L1_error)
      __pyx_t_3 = __pyx_v_message;
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_m, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "mars/actors/pool/messages.pyx":391
 * 
 *         # use 1 byte to record the serialize type
 *         if isinstance(message, bytes):             # <<<<<<<<<<<<<<
 *             st = RAW_BYTES
 *             m = message
 */
      goto __pyx_L5;
    }

    /* "mars/actors/pool/messages.pyx":395
 *             m = message
 *         else:
 *             st = PICKLE             # <<<<<<<<<<<<<<
 *             m = dumps(message)
 * 
 */
    /*else*/ {
      __pyx_v_st = __pyx_v_4mars_6actors_4pool_8messages_PICKLE;

      /* "mars/actors/pool/messages.pyx":396
 *         else:
 *             st = PICKLE
 *             m = dumps(message)             # <<<<<<<<<<<<<<
 * 
 *         bio.extend(_pack_byte(st))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_dumps); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_v_message) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_message);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 396, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_m, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;
    }
    __pyx_L5:;

    /* "mars/actors/pool/messages.pyx":398
 *             m = dumps(message)
 * 
 *         bio.extend(_pack_byte(st))             # <<<<<<<<<<<<<<
 *         bio.extend(_pack_long(len(m)))
 *         if st == RAW_BYTES:
 */
    __pyx_t_6 = __Pyx_PyInt_From_npy_uint8(__pyx_v_st); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__pack_byte);
    __pyx_t_7 = __pyx_v_4mars_6actors_4pool_8messages__pack_byte; __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_bio, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "mars/actors/pool/messages.pyx":399
 * 
 *         bio.extend(_pack_byte(st))
 *         bio.extend(_pack_long(len(m)))             # <<<<<<<<<<<<<<
 *         if st == RAW_BYTES:
 *             ret.append(<bytes>bio)
 */
    if (unlikely(__pyx_v_m == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 399, __pyx_L1_error)
    }
    __pyx_t_9 = PyBytes_GET_SIZE(__pyx_v_m); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 399, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__pack_long);
    __pyx_t_6 = __pyx_v_4mars_6actors_4pool_8messages__pack_long; __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_bio, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "mars/actors/pool/messages.pyx":400
 *         bio.extend(_pack_byte(st))
 *         bio.extend(_pack_long(len(m)))
 *         if st == RAW_BYTES:             # <<<<<<<<<<<<<<
 *             ret.append(<bytes>bio)
 *             ret.append(m)
 */
    __pyx_t_5 = ((__pyx_v_st == __pyx_v_4mars_6actors_4pool_8messages_RAW_BYTES) != 0);
    if (__pyx_t_5) {

      /* "mars/actors/pool/messages.pyx":401
 *         bio.extend(_pack_long(len(m)))
 *         if st == RAW_BYTES:
 *             ret.append(<bytes>bio)             # <<<<<<<<<<<<<<
 *             ret.append(m)
 *         else:
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_v_bio); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 401, __pyx_L1_error)

      /* "mars/actors/pool/messages.pyx":402
 *         if st == RAW_BYTES:
 *             ret.append(<bytes>bio)
 *             ret.append(m)             # <<<<<<<<<<<<<<
 *         else:
 *             bio.extend(m)
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_v_m); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 402, __pyx_L1_error)

      /* "mars/actors/pool/messages.pyx":400
 *         bio.extend(_pack_byte(st))
 *         bio.extend(_pack_long(len(m)))
 *         if st == RAW_BYTES:             # <<<<<<<<<<<<<<
 *             ret.append(<bytes>bio)
 *             ret.append(m)
 */
      goto __pyx_L6;
    }

    /* "mars/actors/pool/messages.pyx":404
 *             ret.append(m)
 *         else:
 *             bio.extend(m)             # <<<<<<<<<<<<<<
 *             ret.append(<bytes>bio)
 * 
 */
    /*else*/ {
      __pyx_t_6 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_bio, __pyx_v_m); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 404, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "mars/actors/pool/messages.pyx":405
 *         else:
 *             bio.extend(m)
 *             ret.append(<bytes>bio)             # <<<<<<<<<<<<<<
 * 
 *     return ret
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_v_bio); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 405, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "mars/actors/pool/messages.pyx":387
 * 
 *     ret = []
 *     for message in messages:             # <<<<<<<<<<<<<<
 *         bio = bytearray()
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":407
 *             ret.append(<bytes>bio)
 * 
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":380
 * 
 * 
 * cdef inline list _pack_tuple_message(tuple messages):             # <<<<<<<<<<<<<<
 *     cdef list ret
 *     cdef bytearray bio
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("mars.actors.pool.messages._pack_tuple_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_bio);
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_XDECREF(__pyx_v_message);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":410
 * 
 * 
 * cdef inline bytes _pack_sole_message(object message, bytearray buf):             # <<<<<<<<<<<<<<
 *     cdef BYTE_t st
 *     cdef bytes m
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__pack_sole_message(PyObject *__pyx_v_message, PyObject *__pyx_v_buf) {
  __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_st;
  PyObject *__pyx_v_m = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  __Pyx_RefNannySetupContext("_pack_sole_message", 0);

  /* "mars/actors/pool/messages.pyx":415
 * 
 *     # use 1 byte to record the serialize type
 *     if isinstance(message, bytes):             # <<<<<<<<<<<<<<
 *         st = RAW_BYTES
 *         m = message
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_message); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "mars/actors/pool/messages.pyx":416
 *     # use 1 byte to record the serialize type
 *     if isinstance(message, bytes):
 *         st = RAW_BYTES             # <<<<<<<<<<<<<<
 *         m = message
 *     elif message is None:
 */
    __pyx_v_st = __pyx_v_4mars_6actors_4pool_8messages_RAW_BYTES;

    /* "mars/actors/pool/messages.pyx":417
 *     if isinstance(message, bytes):
 *         st = RAW_BYTES
 *         m = message             # <<<<<<<<<<<<<<
 *     elif message is None:
 *         st = NONE
 */
    if (!(likely(PyBytes_CheckExact(__pyx_v_message))||((__pyx_v_message) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_message)->tp_name), 0))) __PYX_ERR(0, 417, __pyx_L1_error)
    __pyx_t_3 = __pyx_v_message;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_m = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "mars/actors/pool/messages.pyx":415
 * 
 *     # use 1 byte to record the serialize type
 *     if isinstance(message, bytes):             # <<<<<<<<<<<<<<
 *         st = RAW_BYTES
 *         m = message
 */
    goto __pyx_L3;
  }

  /* "mars/actors/pool/messages.pyx":418
 *         st = RAW_BYTES
 *         m = message
 *     elif message is None:             # <<<<<<<<<<<<<<
 *         st = NONE
 *         m = bytes()
 */
  __pyx_t_2 = (__pyx_v_message == Py_None);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "mars/actors/pool/messages.pyx":419
 *         m = message
 *     elif message is None:
 *         st = NONE             # <<<<<<<<<<<<<<
 *         m = bytes()
 *     else:
 */
    __pyx_v_st = __pyx_v_4mars_6actors_4pool_8messages_NONE;

    /* "mars/actors/pool/messages.pyx":420
 *     elif message is None:
 *         st = NONE
 *         m = bytes()             # <<<<<<<<<<<<<<
 *     else:
 *         st = PICKLE
 */
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyBytes_Type))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_m = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "mars/actors/pool/messages.pyx":418
 *         st = RAW_BYTES
 *         m = message
 *     elif message is None:             # <<<<<<<<<<<<<<
 *         st = NONE
 *         m = bytes()
 */
    goto __pyx_L3;
  }

  /* "mars/actors/pool/messages.pyx":422
 *         m = bytes()
 *     else:
 *         st = PICKLE             # <<<<<<<<<<<<<<
 *         m = dumps(message)
 * 
 */
  /*else*/ {
    __pyx_v_st = __pyx_v_4mars_6actors_4pool_8messages_PICKLE;

    /* "mars/actors/pool/messages.pyx":423
 *     else:
 *         st = PICKLE
 *         m = dumps(message)             # <<<<<<<<<<<<<<
 * 
 *     buf.extend(_pack_byte(st))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_dumps); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_message) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_message);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 423, __pyx_L1_error)
    __pyx_v_m = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "mars/actors/pool/messages.pyx":425
 *         m = dumps(message)
 * 
 *     buf.extend(_pack_byte(st))             # <<<<<<<<<<<<<<
 *     buf.extend(_pack_long(len(m)))
 *     return m
 */
  __pyx_t_4 = __Pyx_PyInt_From_npy_uint8(__pyx_v_st); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__pack_byte);
  __pyx_t_5 = __pyx_v_4mars_6actors_4pool_8messages__pack_byte; __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_buf, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "mars/actors/pool/messages.pyx":426
 * 
 *     buf.extend(_pack_byte(st))
 *     buf.extend(_pack_long(len(m)))             # <<<<<<<<<<<<<<
 *     return m
 * 
 */
  if (unlikely(__pyx_v_m == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 426, __pyx_L1_error)
  }
  __pyx_t_7 = PyBytes_GET_SIZE(__pyx_v_m); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 426, __pyx_L1_error)
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__pack_long);
  __pyx_t_4 = __pyx_v_4mars_6actors_4pool_8messages__pack_long; __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_buf, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "mars/actors/pool/messages.pyx":427
 *     buf.extend(_pack_byte(st))
 *     buf.extend(_pack_long(len(m)))
 *     return m             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_m);
  __pyx_r = __pyx_v_m;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":410
 * 
 * 
 * cdef inline bytes _pack_sole_message(object message, bytearray buf):             # <<<<<<<<<<<<<<
 *     cdef BYTE_t st
 *     cdef bytes m
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("mars.actors.pool.messages._pack_sole_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":430
 * 
 * 
 * cdef inline list _pack_message(object message, bytearray buf):             # <<<<<<<<<<<<<<
 *     if isinstance(message, tuple):
 *         _pack_message_size(len(message), buf)
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__pack_message(PyObject *__pyx_v_message, PyObject *__pyx_v_buf) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_pack_message", 0);

  /* "mars/actors/pool/messages.pyx":431
 * 
 * cdef inline list _pack_message(object message, bytearray buf):
 *     if isinstance(message, tuple):             # <<<<<<<<<<<<<<
 *         _pack_message_size(len(message), buf)
 *         return _pack_tuple_message(message)
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_message); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "mars/actors/pool/messages.pyx":432
 * cdef inline list _pack_message(object message, bytearray buf):
 *     if isinstance(message, tuple):
 *         _pack_message_size(len(message), buf)             # <<<<<<<<<<<<<<
 *         return _pack_tuple_message(message)
 *     else:
 */
    __pyx_t_3 = PyObject_Length(__pyx_v_message); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 432, __pyx_L1_error)
    __pyx_f_4mars_6actors_4pool_8messages__pack_message_size(__pyx_t_3, __pyx_v_buf);

    /* "mars/actors/pool/messages.pyx":433
 *     if isinstance(message, tuple):
 *         _pack_message_size(len(message), buf)
 *         return _pack_tuple_message(message)             # <<<<<<<<<<<<<<
 *     else:
 *         _pack_message_size(0, buf)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyTuple_CheckExact(__pyx_v_message))||((__pyx_v_message) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_message)->tp_name), 0))) __PYX_ERR(0, 433, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_4mars_6actors_4pool_8messages__pack_tuple_message(((PyObject*)__pyx_v_message)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":431
 * 
 * cdef inline list _pack_message(object message, bytearray buf):
 *     if isinstance(message, tuple):             # <<<<<<<<<<<<<<
 *         _pack_message_size(len(message), buf)
 *         return _pack_tuple_message(message)
 */
  }

  /* "mars/actors/pool/messages.pyx":435
 *         return _pack_tuple_message(message)
 *     else:
 *         _pack_message_size(0, buf)             # <<<<<<<<<<<<<<
 *         return [_pack_sole_message(message, buf)]
 * 
 */
  /*else*/ {
    __pyx_f_4mars_6actors_4pool_8messages__pack_message_size(0, __pyx_v_buf);

    /* "mars/actors/pool/messages.pyx":436
 *     else:
 *         _pack_message_size(0, buf)
 *         return [_pack_sole_message(message, buf)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_4mars_6actors_4pool_8messages__pack_sole_message(__pyx_v_message, __pyx_v_buf); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_r = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "mars/actors/pool/messages.pyx":430
 * 
 * 
 * cdef inline list _pack_message(object message, bytearray buf):             # <<<<<<<<<<<<<<
 *     if isinstance(message, tuple):
 *         _pack_message_size(len(message), buf)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages._pack_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4mars_6actors_4pool_8messages_21_unpack_tuple_message_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "mars/actors/pool/messages.pyx":440
 * 
 * cdef inline tuple _unpack_tuple_message(bytes buf, size_t size, size_t* pos):
 *     return tuple(_unpack_sole_message(buf, pos) for _ in range(size))             # <<<<<<<<<<<<<<
 * 
 * 
 */

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_21_unpack_tuple_message_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr *)__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr(__pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 440, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4mars_6actors_4pool_8messages_21_unpack_tuple_message_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_unpack_tuple_message_locals_gen, __pyx_n_s_mars_actors_pool_messages); if (unlikely(!gen)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages._unpack_tuple_message.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4mars_6actors_4pool_8messages_21_unpack_tuple_message_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 440, __pyx_L1_error)
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_cur_scope->__pyx_v__ = __pyx_t_3;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_buf)) { __Pyx_RaiseClosureNameError("buf"); __PYX_ERR(0, 440, __pyx_L1_error) }
    __pyx_t_4 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_buf;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = __pyx_f_4mars_6actors_4pool_8messages__unpack_sole_message(((PyObject*)__pyx_t_4), __pyx_cur_scope->__pyx_outer_scope->__pyx_v_pos); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 440, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":439
 * 
 * 
 * cdef inline tuple _unpack_tuple_message(bytes buf, size_t size, size_t* pos):             # <<<<<<<<<<<<<<
 *     return tuple(_unpack_sole_message(buf, pos) for _ in range(size))
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message(PyObject *__pyx_v_buf, size_t __pyx_v_size, size_t *__pyx_v_pos) {
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_unpack_tuple_message", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message *)__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message(__pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 439, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_buf = __pyx_v_buf;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_buf);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_buf);
  __pyx_cur_scope->__pyx_v_size = __pyx_v_size;
  __pyx_cur_scope->__pyx_v_pos = __pyx_v_pos;

  /* "mars/actors/pool/messages.pyx":440
 * 
 * cdef inline tuple _unpack_tuple_message(bytes buf, size_t size, size_t* pos):
 *     return tuple(_unpack_sole_message(buf, pos) for _ in range(size))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_4mars_6actors_4pool_8messages_21_unpack_tuple_message_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":439
 * 
 * 
 * cdef inline tuple _unpack_tuple_message(bytes buf, size_t size, size_t* pos):             # <<<<<<<<<<<<<<
 *     return tuple(_unpack_sole_message(buf, pos) for _ in range(size))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mars.actors.pool.messages._unpack_tuple_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":443
 * 
 * 
 * cdef inline object _unpack_sole_message(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef BYTE_t st
 *     cdef size_t size
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_sole_message(PyObject *__pyx_v_binary, size_t *__pyx_v_pos) {
  __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_v_st;
  size_t __pyx_v_size;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_t_6;
  size_t __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("_unpack_sole_message", 0);

  /* "mars/actors/pool/messages.pyx":447
 *     cdef size_t size
 * 
 *     st, = _unpack_byte(_read_bytes(binary, pos, 1))             # <<<<<<<<<<<<<<
 *     size, = _unpack_long(_read_bytes(binary, pos, 8))
 * 
 */
  __pyx_t_2 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_byte);
  __pyx_t_3 = __pyx_v_4mars_6actors_4pool_8messages__unpack_byte; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 447, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_5(__pyx_t_2); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_2), 1) < 0) __PYX_ERR(0, 447, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 447, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint8(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint8)-1)) && PyErr_Occurred())) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_st = __pyx_t_6;

  /* "mars/actors/pool/messages.pyx":448
 * 
 *     st, = _unpack_byte(_read_bytes(binary, pos, 1))
 *     size, = _unpack_long(_read_bytes(binary, pos, 8))             # <<<<<<<<<<<<<<
 * 
 *     if st == RAW_BYTES:
 */
  __pyx_t_3 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, 8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_long);
  __pyx_t_2 = __pyx_v_4mars_6actors_4pool_8messages__unpack_long; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 448, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_3), 1) < 0) __PYX_ERR(0, 448, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 448, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __pyx_t_7 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_7 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_size = __pyx_t_7;

  /* "mars/actors/pool/messages.pyx":450
 *     size, = _unpack_long(_read_bytes(binary, pos, 8))
 * 
 *     if st == RAW_BYTES:             # <<<<<<<<<<<<<<
 *         return _read_bytes(binary, pos, size)
 *     elif st == NONE:
 */
  __pyx_t_8 = ((__pyx_v_st == __pyx_v_4mars_6actors_4pool_8messages_RAW_BYTES) != 0);
  if (__pyx_t_8) {

    /* "mars/actors/pool/messages.pyx":451
 * 
 *     if st == RAW_BYTES:
 *         return _read_bytes(binary, pos, size)             # <<<<<<<<<<<<<<
 *     elif st == NONE:
 *         return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, __pyx_v_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":450
 *     size, = _unpack_long(_read_bytes(binary, pos, 8))
 * 
 *     if st == RAW_BYTES:             # <<<<<<<<<<<<<<
 *         return _read_bytes(binary, pos, size)
 *     elif st == NONE:
 */
  }

  /* "mars/actors/pool/messages.pyx":452
 *     if st == RAW_BYTES:
 *         return _read_bytes(binary, pos, size)
 *     elif st == NONE:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  __pyx_t_8 = ((__pyx_v_st == __pyx_v_4mars_6actors_4pool_8messages_NONE) != 0);
  if (__pyx_t_8) {

    /* "mars/actors/pool/messages.pyx":453
 *         return _read_bytes(binary, pos, size)
 *     elif st == NONE:
 *         return None             # <<<<<<<<<<<<<<
 *     else:
 *         return loads(_read_bytes(binary, pos, size))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":452
 *     if st == RAW_BYTES:
 *         return _read_bytes(binary, pos, size)
 *     elif st == NONE:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  }

  /* "mars/actors/pool/messages.pyx":455
 *         return None
 *     else:
 *         return loads(_read_bytes(binary, pos, size))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_loads); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_f_4mars_6actors_4pool_8messages__read_bytes(__pyx_v_binary, __pyx_v_pos, __pyx_v_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "mars/actors/pool/messages.pyx":443
 * 
 * 
 * cdef inline object _unpack_sole_message(bytes binary, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef BYTE_t st
 *     cdef size_t size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.actors.pool.messages._unpack_sole_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":458
 * 
 * 
 * cdef inline object _unpack_message(bytes buf, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef size_t size
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_message(PyObject *__pyx_v_buf, size_t *__pyx_v_pos) {
  size_t __pyx_v_size;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_unpack_message", 0);

  /* "mars/actors/pool/messages.pyx":461
 *     cdef size_t size
 * 
 *     size = _unpack_message_size(buf, pos)             # <<<<<<<<<<<<<<
 *     if size > 0:
 *         return _unpack_tuple_message(buf, size, pos)
 */
  __pyx_v_size = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_size(__pyx_v_buf, __pyx_v_pos);

  /* "mars/actors/pool/messages.pyx":462
 * 
 *     size = _unpack_message_size(buf, pos)
 *     if size > 0:             # <<<<<<<<<<<<<<
 *         return _unpack_tuple_message(buf, size, pos)
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_size > 0) != 0);
  if (__pyx_t_1) {

    /* "mars/actors/pool/messages.pyx":463
 *     size = _unpack_message_size(buf, pos)
 *     if size > 0:
 *         return _unpack_tuple_message(buf, size, pos)             # <<<<<<<<<<<<<<
 *     else:
 *         return _unpack_sole_message(buf, pos)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message(__pyx_v_buf, __pyx_v_size, __pyx_v_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":462
 * 
 *     size = _unpack_message_size(buf, pos)
 *     if size > 0:             # <<<<<<<<<<<<<<
 *         return _unpack_tuple_message(buf, size, pos)
 *     else:
 */
  }

  /* "mars/actors/pool/messages.pyx":465
 *         return _unpack_tuple_message(buf, size, pos)
 *     else:
 *         return _unpack_sole_message(buf, pos)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_4mars_6actors_4pool_8messages__unpack_sole_message(__pyx_v_buf, __pyx_v_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 465, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "mars/actors/pool/messages.pyx":458
 * 
 * 
 * cdef inline object _unpack_message(bytes buf, size_t* pos):             # <<<<<<<<<<<<<<
 *     cdef size_t size
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mars.actors.pool.messages._unpack_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":468
 * 
 * 
 * cdef inline object _pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                       bint send, object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytes message_id
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__pack_send_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_message, int __pyx_v_send, struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_send_message *__pyx_optional_args) {

  /* "mars/actors/pool/messages.pyx":469
 * 
 * cdef inline object _pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
 *                                       bint send, object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytes message_id
 *     cdef int message_type
 */
  PyObject *__pyx_v_write = ((PyObject *)Py_None);
  int __pyx_v_protocol = __pyx_k__3;
  PyObject *__pyx_v_message_id = 0;
  int __pyx_v_message_type;
  PyObject *__pyx_v_buf = 0;
  PyObject *__pyx_v_m = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_4mars_6actors_4pool_8messages_BYTE_t __pyx_t_2;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_message_type __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("_pack_send_message", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_write = __pyx_optional_args->write;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_protocol = __pyx_optional_args->protocol;
      }
    }
  }

  /* "mars/actors/pool/messages.pyx":476
 * 
 *     # from_index -1 means main process, -2 means remote, other is the subprocess id
 *     message_id = new_message_id()             # <<<<<<<<<<<<<<
 *     message_type = SEND_ALL if send else TELL_ALL
 * 
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_new_message_id(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":477
 *     # from_index -1 means main process, -2 means remote, other is the subprocess id
 *     message_id = new_message_id()
 *     message_type = SEND_ALL if send else TELL_ALL             # <<<<<<<<<<<<<<
 * 
 *     buf = bytearray()
 */
  if ((__pyx_v_send != 0)) {
    __pyx_t_2 = __pyx_v_4mars_6actors_4pool_8messages_SEND_ALL;
  } else {
    __pyx_t_2 = __pyx_v_4mars_6actors_4pool_8messages_TELL_ALL;
  }
  __pyx_v_message_type = __pyx_t_2;

  /* "mars/actors/pool/messages.pyx":479
 *     message_type = SEND_ALL if send else TELL_ALL
 * 
 *     buf = bytearray()             # <<<<<<<<<<<<<<
 *     _pack_message_type(message_type, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyByteArray_Type))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buf = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":480
 * 
 *     buf = bytearray()
 *     _pack_message_type(message_type, buf, protocol=protocol)             # <<<<<<<<<<<<<<
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 */
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.protocol = __pyx_v_protocol;
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_type(__pyx_v_message_type, __pyx_v_buf, &__pyx_t_3); 

  /* "mars/actors/pool/messages.pyx":481
 *     buf = bytearray()
 *     _pack_message_type(message_type, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_id(__pyx_v_message_id, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":482
 *     _pack_message_type(message_type, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(to_index, buf)
 *     _pack_actor_ref(actor_ref, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_from_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":483
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_actor_ref(actor_ref, buf)
 *     m = _pack_message(message, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_to_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":484
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 *     _pack_actor_ref(actor_ref, buf)             # <<<<<<<<<<<<<<
 *     m = _pack_message(message, buf)
 * 
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_actor_ref(__pyx_v_actor_ref, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":485
 *     _pack_index(to_index, buf)
 *     _pack_actor_ref(actor_ref, buf)
 *     m = _pack_message(message, buf)             # <<<<<<<<<<<<<<
 * 
 *     if write:
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__pack_message(__pyx_v_message, __pyx_v_buf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_m = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":487
 *     m = _pack_message(message, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf, *m)
 *         return message_id
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_write); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 487, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "mars/actors/pool/messages.pyx":488
 * 
 *     if write:
 *         write(buf, *m)             # <<<<<<<<<<<<<<
 *         return message_id
 * 
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_buf);
    __Pyx_GIVEREF(__pyx_v_buf);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_buf);
    __pyx_t_5 = __Pyx_PySequence_Tuple(__pyx_v_m); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyNumber_Add(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "mars/actors/pool/messages.pyx":489
 *     if write:
 *         write(buf, *m)
 *         return message_id             # <<<<<<<<<<<<<<
 * 
 *     return [message_id, buf] + m
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_message_id);
    __pyx_r = __pyx_v_message_id;
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":487
 *     m = _pack_message(message, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf, *m)
 *         return message_id
 */
  }

  /* "mars/actors/pool/messages.pyx":491
 *         return message_id
 * 
 *     return [message_id, buf] + m             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyList_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_message_id);
  __Pyx_INCREF(__pyx_v_buf);
  __Pyx_GIVEREF(__pyx_v_buf);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_v_buf);
  __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_v_m); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":468
 * 
 * 
 * cdef inline object _pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                       bint send, object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytes message_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("mars.actors.pool.messages._pack_send_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message_id);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":494
 * 
 * 
 * cdef inline object _unpack_send_message(bytes binary, send=True):             # <<<<<<<<<<<<<<
 *     cdef object message_type
 *     cdef bytes message_id
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__unpack_send_message(PyObject *__pyx_v_binary, struct __pyx_opt_args_4mars_6actors_4pool_8messages__unpack_send_message *__pyx_optional_args) {
  PyObject *__pyx_v_send = ((PyObject *)Py_True);
  PyObject *__pyx_v_message_type = 0;
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  int __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  PyObject *__pyx_v_message = 0;
  size_t __pyx_v_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_unpack_send_message", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_send = __pyx_optional_args->send;
    }
  }

  /* "mars/actors/pool/messages.pyx":501
 *     cdef ActorRef actor_ref
 *     cdef object message
 *     cdef size_t pos = 0             # <<<<<<<<<<<<<<
 * 
 *     _unpack_message_type_value(binary, &pos)
 */
  __pyx_v_pos = 0;

  /* "mars/actors/pool/messages.pyx":503
 *     cdef size_t pos = 0
 * 
 *     _unpack_message_type_value(binary, &pos)             # <<<<<<<<<<<<<<
 *     message_type = MessageType.send_all if send else MessageType.tell_all
 *     message_id = _unpack_message_id(binary, &pos)
 */
  (void)(__pyx_f_4mars_6actors_4pool_8messages__unpack_message_type_value(__pyx_v_binary, (&__pyx_v_pos)));

  /* "mars/actors/pool/messages.pyx":504
 * 
 *     _unpack_message_type_value(binary, &pos)
 *     message_type = MessageType.send_all if send else MessageType.tell_all             # <<<<<<<<<<<<<<
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_send); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 504, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_send_all); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_tell_all); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_v_message_type = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":505
 *     _unpack_message_type_value(binary, &pos)
 *     message_type = MessageType.send_all if send else MessageType.tell_all
 *     message_id = _unpack_message_id(binary, &pos)             # <<<<<<<<<<<<<<
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_id(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":506
 *     message_type = MessageType.send_all if send else MessageType.tell_all
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     to_index = _unpack_index(binary, &pos)
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 */
  __pyx_v_from_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":507
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 *     message = _unpack_message(binary, &pos)
 */
  __pyx_v_to_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":508
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 *     actor_ref = _unpack_actor_ref(binary, &pos)             # <<<<<<<<<<<<<<
 *     message = _unpack_message(binary, &pos)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_4mars_6actors_4pool_8messages__unpack_actor_ref(__pyx_v_binary, (&__pyx_v_pos))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":509
 *     to_index = _unpack_index(binary, &pos)
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 *     message = _unpack_message(binary, &pos)             # <<<<<<<<<<<<<<
 * 
 *     return SEND_MESSAGE(message_type=message_type, message_id=message_id,
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_message(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":511
 *     message = _unpack_message(binary, &pos)
 * 
 *     return SEND_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                         from_index=from_index, to_index=to_index,
 *                         actor_ref=actor_ref, message=message)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_type, __pyx_v_message_type) < 0) __PYX_ERR(0, 511, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_id, __pyx_v_message_id) < 0) __PYX_ERR(0, 511, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":512
 * 
 *     return SEND_MESSAGE(message_type=message_type, message_id=message_id,
 *                         from_index=from_index, to_index=to_index,             # <<<<<<<<<<<<<<
 *                         actor_ref=actor_ref, message=message)
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_from_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_from_index, __pyx_t_3) < 0) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_to_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_to_index, __pyx_t_3) < 0) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mars/actors/pool/messages.pyx":513
 *     return SEND_MESSAGE(message_type=message_type, message_id=message_id,
 *                         from_index=from_index, to_index=to_index,
 *                         actor_ref=actor_ref, message=message)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_actor_ref, ((PyObject *)__pyx_v_actor_ref)) < 0) __PYX_ERR(0, 511, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message, __pyx_v_message) < 0) __PYX_ERR(0, 511, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":511
 *     message = _unpack_message(binary, &pos)
 * 
 *     return SEND_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                         from_index=from_index, to_index=to_index,
 *                         actor_ref=actor_ref, message=message)
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_SEND_MESSAGE), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":494
 * 
 * 
 * cdef inline object _unpack_send_message(bytes binary, send=True):             # <<<<<<<<<<<<<<
 *     cdef object message_type
 *     cdef bytes message_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mars.actors.pool.messages._unpack_send_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message_type);
  __Pyx_XDECREF(__pyx_v_message_id);
  __Pyx_XDECREF((PyObject *)__pyx_v_actor_ref);
  __Pyx_XDECREF(__pyx_v_message);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":516
 * 
 * 
 * cpdef object pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, True,
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_11pack_send_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_send_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_message, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_send_message *__pyx_optional_args) {

  /* "mars/actors/pool/messages.pyx":517
 * 
 * cpdef object pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     return _pack_send_message(from_index, to_index, actor_ref, message, True,
 *                               write=write, protocol=protocol)
 */
  PyObject *__pyx_v_write = ((PyObject *)Py_None);
  int __pyx_v_protocol = __pyx_k__4;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_send_message __pyx_t_2;
  __Pyx_RefNannySetupContext("pack_send_message", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_write = __pyx_optional_args->write;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_protocol = __pyx_optional_args->protocol;
      }
    }
  }

  /* "mars/actors/pool/messages.pyx":518
 * cpdef object pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, True,             # <<<<<<<<<<<<<<
 *                               write=write, protocol=protocol)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "mars/actors/pool/messages.pyx":519
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, True,
 *                               write=write, protocol=protocol)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.write = __pyx_v_write;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__pack_send_message(__pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_message, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":516
 * 
 * 
 * cpdef object pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, True,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_send_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_11pack_send_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_11pack_send_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  PyObject *__pyx_v_message = 0;
  PyObject *__pyx_v_write = 0;
  int __pyx_v_protocol;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_send_message (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_actor_ref,&__pyx_n_s_message,&__pyx_n_s_write,&__pyx_n_s_protocol,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "mars/actors/pool/messages.pyx":517
 * 
 * cpdef object pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     return _pack_send_message(from_index, to_index, actor_ref, message, True,
 *                               write=write, protocol=protocol)
 */
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_send_message", 0, 4, 6, 1); __PYX_ERR(0, 516, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_ref)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_send_message", 0, 4, 6, 2); __PYX_ERR(0, 516, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_send_message", 0, 4, 6, 3); __PYX_ERR(0, 516, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_write);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_protocol);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pack_send_message") < 0)) __PYX_ERR(0, 516, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[0]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L3_error)
    __pyx_v_to_index = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_to_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L3_error)
    __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)values[2]);
    __pyx_v_message = values[3];
    __pyx_v_write = values[4];
    if (values[5]) {
      __pyx_v_protocol = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_protocol == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L3_error)
    } else {
      __pyx_v_protocol = __pyx_k__4;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pack_send_message", 0, 4, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 516, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_send_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_actor_ref), __pyx_ptype_4mars_6actors_4core_ActorRef, 1, "actor_ref", 0))) __PYX_ERR(0, 516, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_10pack_send_message(__pyx_self, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_message, __pyx_v_write, __pyx_v_protocol);

  /* "mars/actors/pool/messages.pyx":516
 * 
 * 
 * cpdef object pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, True,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_10pack_send_message(CYTHON_UNUSED PyObject *__pyx_self, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_message, PyObject *__pyx_v_write, int __pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_send_message __pyx_t_2;
  __Pyx_RefNannySetupContext("pack_send_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.write = __pyx_v_write;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_pack_send_message(__pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_message, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_send_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":522
 * 
 * 
 * cpdef object unpack_send_message(bytes message):             # <<<<<<<<<<<<<<
 *     return _unpack_send_message(message)
 * 
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13unpack_send_message(PyObject *__pyx_self, PyObject *__pyx_v_message); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_send_message(PyObject *__pyx_v_message, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_send_message", 0);

  /* "mars/actors/pool/messages.pyx":523
 * 
 * cpdef object unpack_send_message(bytes message):
 *     return _unpack_send_message(message)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_send_message(__pyx_v_message, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":522
 * 
 * 
 * cpdef object unpack_send_message(bytes message):             # <<<<<<<<<<<<<<
 *     return _unpack_send_message(message)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_send_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13unpack_send_message(PyObject *__pyx_self, PyObject *__pyx_v_message); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_13unpack_send_message(PyObject *__pyx_self, PyObject *__pyx_v_message) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_send_message (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message), (&PyBytes_Type), 1, "message", 1))) __PYX_ERR(0, 522, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_12unpack_send_message(__pyx_self, ((PyObject*)__pyx_v_message));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_12unpack_send_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_message) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_send_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_unpack_send_message(__pyx_v_message, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_send_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":526
 * 
 * 
 * cpdef object pack_tell_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, False,
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_15pack_tell_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_tell_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_message, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_tell_message *__pyx_optional_args) {

  /* "mars/actors/pool/messages.pyx":527
 * 
 * cpdef object pack_tell_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     return _pack_send_message(from_index, to_index, actor_ref, message, False,
 *                               write=write, protocol=protocol)
 */
  PyObject *__pyx_v_write = ((PyObject *)Py_None);
  int __pyx_v_protocol = __pyx_k__5;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_send_message __pyx_t_2;
  __Pyx_RefNannySetupContext("pack_tell_message", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_write = __pyx_optional_args->write;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_protocol = __pyx_optional_args->protocol;
      }
    }
  }

  /* "mars/actors/pool/messages.pyx":528
 * cpdef object pack_tell_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, False,             # <<<<<<<<<<<<<<
 *                               write=write, protocol=protocol)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "mars/actors/pool/messages.pyx":529
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, False,
 *                               write=write, protocol=protocol)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.write = __pyx_v_write;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__pack_send_message(__pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_message, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":526
 * 
 * 
 * cpdef object pack_tell_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, False,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_tell_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_15pack_tell_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_15pack_tell_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  PyObject *__pyx_v_message = 0;
  PyObject *__pyx_v_write = 0;
  int __pyx_v_protocol;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_tell_message (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_actor_ref,&__pyx_n_s_message,&__pyx_n_s_write,&__pyx_n_s_protocol,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "mars/actors/pool/messages.pyx":527
 * 
 * cpdef object pack_tell_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     return _pack_send_message(from_index, to_index, actor_ref, message, False,
 *                               write=write, protocol=protocol)
 */
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_tell_message", 0, 4, 6, 1); __PYX_ERR(0, 526, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_ref)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_tell_message", 0, 4, 6, 2); __PYX_ERR(0, 526, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_tell_message", 0, 4, 6, 3); __PYX_ERR(0, 526, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_write);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_protocol);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pack_tell_message") < 0)) __PYX_ERR(0, 526, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[0]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 526, __pyx_L3_error)
    __pyx_v_to_index = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_to_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 526, __pyx_L3_error)
    __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)values[2]);
    __pyx_v_message = values[3];
    __pyx_v_write = values[4];
    if (values[5]) {
      __pyx_v_protocol = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_protocol == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 527, __pyx_L3_error)
    } else {
      __pyx_v_protocol = __pyx_k__5;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pack_tell_message", 0, 4, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 526, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_tell_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_actor_ref), __pyx_ptype_4mars_6actors_4core_ActorRef, 1, "actor_ref", 0))) __PYX_ERR(0, 526, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_14pack_tell_message(__pyx_self, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_message, __pyx_v_write, __pyx_v_protocol);

  /* "mars/actors/pool/messages.pyx":526
 * 
 * 
 * cpdef object pack_tell_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, False,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_14pack_tell_message(CYTHON_UNUSED PyObject *__pyx_self, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_message, PyObject *__pyx_v_write, int __pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_tell_message __pyx_t_2;
  __Pyx_RefNannySetupContext("pack_tell_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.write = __pyx_v_write;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_pack_tell_message(__pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_message, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_tell_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":532
 * 
 * 
 * cpdef object unpack_tell_message(bytes message):             # <<<<<<<<<<<<<<
 *     return _unpack_send_message(message, send=False)
 * 
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17unpack_tell_message(PyObject *__pyx_self, PyObject *__pyx_v_message); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_tell_message(PyObject *__pyx_v_message, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages__unpack_send_message __pyx_t_2;
  __Pyx_RefNannySetupContext("unpack_tell_message", 0);

  /* "mars/actors/pool/messages.pyx":533
 * 
 * cpdef object unpack_tell_message(bytes message):
 *     return _unpack_send_message(message, send=False)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.send = Py_False;
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_send_message(__pyx_v_message, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":532
 * 
 * 
 * cpdef object unpack_tell_message(bytes message):             # <<<<<<<<<<<<<<
 *     return _unpack_send_message(message, send=False)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_tell_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17unpack_tell_message(PyObject *__pyx_self, PyObject *__pyx_v_message); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_17unpack_tell_message(PyObject *__pyx_self, PyObject *__pyx_v_message) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_tell_message (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message), (&PyBytes_Type), 1, "message", 1))) __PYX_ERR(0, 532, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_16unpack_tell_message(__pyx_self, ((PyObject*)__pyx_v_message));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_16unpack_tell_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_message) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_tell_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_unpack_tell_message(__pyx_v_message, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_tell_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":536
 * 
 * 
 * cpdef tuple get_index(bytes binary, object calc_from_uid):             # <<<<<<<<<<<<<<
 *     cdef object buf
 *     cdef INT32_t from_index
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_19get_index(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_get_index(PyObject *__pyx_v_binary, PyObject *__pyx_v_calc_from_uid, CYTHON_UNUSED int __pyx_skip_dispatch) {
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  PyObject *__pyx_v_uid = 0;
  size_t __pyx_v_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("get_index", 0);

  /* "mars/actors/pool/messages.pyx":541
 *     cdef INT32_t to_index
 *     cdef object uid
 *     cdef size_t pos = 0             # <<<<<<<<<<<<<<
 * 
 *     _skip_message_type(binary, &pos)
 */
  __pyx_v_pos = 0;

  /* "mars/actors/pool/messages.pyx":543
 *     cdef size_t pos = 0
 * 
 *     _skip_message_type(binary, &pos)             # <<<<<<<<<<<<<<
 *     _skip_message_id(binary, &pos)
 * 
 */
  __pyx_f_4mars_6actors_4pool_8messages__skip_message_type(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":544
 * 
 *     _skip_message_type(binary, &pos)
 *     _skip_message_id(binary, &pos)             # <<<<<<<<<<<<<<
 * 
 *     from_index = _unpack_index(binary, &pos)
 */
  __pyx_f_4mars_6actors_4pool_8messages__skip_message_id(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":546
 *     _skip_message_id(binary, &pos)
 * 
 *     from_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     to_index = _unpack_index(binary, &pos)
 *     if to_index != -1 or calc_from_uid is None:
 */
  __pyx_v_from_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":547
 * 
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     if to_index != -1 or calc_from_uid is None:
 *         return from_index, to_index
 */
  __pyx_v_to_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":548
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 *     if to_index != -1 or calc_from_uid is None:             # <<<<<<<<<<<<<<
 *         return from_index, to_index
 * 
 */
  __pyx_t_2 = ((__pyx_v_to_index != -1L) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_calc_from_uid == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "mars/actors/pool/messages.pyx":549
 *     to_index = _unpack_index(binary, &pos)
 *     if to_index != -1 or calc_from_uid is None:
 *         return from_index, to_index             # <<<<<<<<<<<<<<
 * 
 *     _skip_address(binary, &pos)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_from_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_From_npy_int32(__pyx_v_to_index); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":548
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 *     if to_index != -1 or calc_from_uid is None:             # <<<<<<<<<<<<<<
 *         return from_index, to_index
 * 
 */
  }

  /* "mars/actors/pool/messages.pyx":551
 *         return from_index, to_index
 * 
 *     _skip_address(binary, &pos)             # <<<<<<<<<<<<<<
 * 
 *     uid = _unpack_uid(binary, &pos)
 */
  __pyx_f_4mars_6actors_4pool_8messages__skip_address(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":553
 *     _skip_address(binary, &pos)
 * 
 *     uid = _unpack_uid(binary, &pos)             # <<<<<<<<<<<<<<
 *     return from_index, calc_from_uid(uid)
 * 
 */
  __pyx_t_6 = __pyx_f_4mars_6actors_4pool_8messages__unpack_uid(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_uid = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "mars/actors/pool/messages.pyx":554
 * 
 *     uid = _unpack_uid(binary, &pos)
 *     return from_index, calc_from_uid(uid)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = __Pyx_PyInt_From_npy_int32(__pyx_v_from_index); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_calc_from_uid);
  __pyx_t_4 = __pyx_v_calc_from_uid; __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_v_uid) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_uid);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":536
 * 
 * 
 * cpdef tuple get_index(bytes binary, object calc_from_uid):             # <<<<<<<<<<<<<<
 *     cdef object buf
 *     cdef INT32_t from_index
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("mars.actors.pool.messages.get_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_uid);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_19get_index(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_19get_index(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_binary = 0;
  PyObject *__pyx_v_calc_from_uid = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_index (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_binary,&__pyx_n_s_calc_from_uid,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_binary)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_calc_from_uid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_index", 1, 2, 2, 1); __PYX_ERR(0, 536, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_index") < 0)) __PYX_ERR(0, 536, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_binary = ((PyObject*)values[0]);
    __pyx_v_calc_from_uid = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_index", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 536, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.get_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_binary), (&PyBytes_Type), 1, "binary", 1))) __PYX_ERR(0, 536, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_18get_index(__pyx_self, __pyx_v_binary, __pyx_v_calc_from_uid);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_18get_index(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary, PyObject *__pyx_v_calc_from_uid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_index", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_get_index(__pyx_v_binary, __pyx_v_calc_from_uid, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.get_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":557
 * 
 * 
 * cpdef object pack_create_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref, object actor_cls,             # <<<<<<<<<<<<<<
 *                                        tuple args, dict kw, object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytes message_id
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21pack_create_actor_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_create_actor_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, int __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_actor_cls, PyObject *__pyx_v_args, PyObject *__pyx_v_kw, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_create_actor_message *__pyx_optional_args) {

  /* "mars/actors/pool/messages.pyx":558
 * 
 * cpdef object pack_create_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref, object actor_cls,
 *                                        tuple args, dict kw, object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytes message_id
 *     cdef bytearray buf
 */
  PyObject *__pyx_v_write = ((PyObject *)Py_None);
  int __pyx_v_protocol = __pyx_k__6;
  PyObject *__pyx_v_message_id = 0;
  PyObject *__pyx_v_buf = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_message_type __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("pack_create_actor_message", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_write = __pyx_optional_args->write;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_protocol = __pyx_optional_args->protocol;
      }
    }
  }

  /* "mars/actors/pool/messages.pyx":563
 * 
 *     # from_index -1 means main process, -2 means remote, other is the subprocess id
 *     message_id = new_message_id()             # <<<<<<<<<<<<<<
 * 
 *     buf = bytearray()
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_new_message_id(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":565
 *     message_id = new_message_id()
 * 
 *     buf = bytearray()             # <<<<<<<<<<<<<<
 *     _pack_message_type(CREATE_ACTOR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyByteArray_Type))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buf = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":566
 * 
 *     buf = bytearray()
 *     _pack_message_type(CREATE_ACTOR, buf, protocol=protocol)             # <<<<<<<<<<<<<<
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_type(__pyx_v_4mars_6actors_4pool_8messages_CREATE_ACTOR, __pyx_v_buf, &__pyx_t_2); 

  /* "mars/actors/pool/messages.pyx":567
 *     buf = bytearray()
 *     _pack_message_type(CREATE_ACTOR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_id(__pyx_v_message_id, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":568
 *     _pack_message_type(CREATE_ACTOR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(to_index, buf)
 *     _pack_actor_ref(actor_ref, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_from_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":569
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_actor_ref(actor_ref, buf)
 *     _pack_object(actor_cls, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_to_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":570
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 *     _pack_actor_ref(actor_ref, buf)             # <<<<<<<<<<<<<<
 *     _pack_object(actor_cls, buf)
 *     _pack_object(args if args else None, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_actor_ref(__pyx_v_actor_ref, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":571
 *     _pack_index(to_index, buf)
 *     _pack_actor_ref(actor_ref, buf)
 *     _pack_object(actor_cls, buf)             # <<<<<<<<<<<<<<
 *     _pack_object(args if args else None, buf)
 *     _pack_object(kw if kw else None, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_object(__pyx_v_actor_cls, __pyx_v_buf); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 571, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":572
 *     _pack_actor_ref(actor_ref, buf)
 *     _pack_object(actor_cls, buf)
 *     _pack_object(args if args else None, buf)             # <<<<<<<<<<<<<<
 *     _pack_object(kw if kw else None, buf)
 * 
 */
  __pyx_t_3 = (__pyx_v_args != Py_None)&&(PyTuple_GET_SIZE(__pyx_v_args) != 0);
  if (__pyx_t_3) {
    __Pyx_INCREF(__pyx_v_args);
    __pyx_t_1 = __pyx_v_args;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_1 = Py_None;
  }
  __pyx_f_4mars_6actors_4pool_8messages__pack_object(__pyx_t_1, __pyx_v_buf); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":573
 *     _pack_object(actor_cls, buf)
 *     _pack_object(args if args else None, buf)
 *     _pack_object(kw if kw else None, buf)             # <<<<<<<<<<<<<<
 * 
 *     if write:
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_kw); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 573, __pyx_L1_error)
  if (__pyx_t_3) {
    __Pyx_INCREF(__pyx_v_kw);
    __pyx_t_1 = __pyx_v_kw;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_1 = Py_None;
  }
  __pyx_f_4mars_6actors_4pool_8messages__pack_object(__pyx_t_1, __pyx_v_buf); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":575
 *     _pack_object(kw if kw else None, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf)
 *         return message_id
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_write); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 575, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "mars/actors/pool/messages.pyx":576
 * 
 *     if write:
 *         write(buf)             # <<<<<<<<<<<<<<
 *         return message_id
 * 
 */
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_4 = __pyx_v_write; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_buf) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_buf);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mars/actors/pool/messages.pyx":577
 *     if write:
 *         write(buf)
 *         return message_id             # <<<<<<<<<<<<<<
 * 
 *     return message_id, buf
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_message_id);
    __pyx_r = __pyx_v_message_id;
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":575
 *     _pack_object(kw if kw else None, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf)
 *         return message_id
 */
  }

  /* "mars/actors/pool/messages.pyx":579
 *         return message_id
 * 
 *     return message_id, buf             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_message_id);
  __Pyx_INCREF(__pyx_v_buf);
  __Pyx_GIVEREF(__pyx_v_buf);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_buf);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":557
 * 
 * 
 * cpdef object pack_create_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref, object actor_cls,             # <<<<<<<<<<<<<<
 *                                        tuple args, dict kw, object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytes message_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_create_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message_id);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21pack_create_actor_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_21pack_create_actor_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  int __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  PyObject *__pyx_v_actor_cls = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kw = 0;
  PyObject *__pyx_v_write = 0;
  int __pyx_v_protocol;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_create_actor_message (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_actor_ref,&__pyx_n_s_actor_cls,&__pyx_n_s_args,&__pyx_n_s_kw,&__pyx_n_s_write,&__pyx_n_s_protocol,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};

    /* "mars/actors/pool/messages.pyx":558
 * 
 * cpdef object pack_create_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref, object actor_cls,
 *                                        tuple args, dict kw, object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytes message_id
 *     cdef bytearray buf
 */
    values[6] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_create_actor_message", 0, 6, 8, 1); __PYX_ERR(0, 557, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_ref)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_create_actor_message", 0, 6, 8, 2); __PYX_ERR(0, 557, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_cls)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_create_actor_message", 0, 6, 8, 3); __PYX_ERR(0, 557, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_create_actor_message", 0, 6, 8, 4); __PYX_ERR(0, 557, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kw)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_create_actor_message", 0, 6, 8, 5); __PYX_ERR(0, 557, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_write);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_protocol);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pack_create_actor_message") < 0)) __PYX_ERR(0, 557, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[0]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 557, __pyx_L3_error)
    __pyx_v_to_index = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_to_index == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 557, __pyx_L3_error)
    __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)values[2]);
    __pyx_v_actor_cls = values[3];
    __pyx_v_args = ((PyObject*)values[4]);
    __pyx_v_kw = ((PyObject*)values[5]);
    __pyx_v_write = values[6];
    if (values[7]) {
      __pyx_v_protocol = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_protocol == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 558, __pyx_L3_error)
    } else {
      __pyx_v_protocol = __pyx_k__6;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pack_create_actor_message", 0, 6, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 557, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_create_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_actor_ref), __pyx_ptype_4mars_6actors_4core_ActorRef, 1, "actor_ref", 0))) __PYX_ERR(0, 557, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_args), (&PyTuple_Type), 1, "args", 1))) __PYX_ERR(0, 558, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_kw), (&PyDict_Type), 1, "kw", 1))) __PYX_ERR(0, 558, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_20pack_create_actor_message(__pyx_self, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_actor_cls, __pyx_v_args, __pyx_v_kw, __pyx_v_write, __pyx_v_protocol);

  /* "mars/actors/pool/messages.pyx":557
 * 
 * 
 * cpdef object pack_create_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref, object actor_cls,             # <<<<<<<<<<<<<<
 *                                        tuple args, dict kw, object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytes message_id
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20pack_create_actor_message(CYTHON_UNUSED PyObject *__pyx_self, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, int __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_actor_cls, PyObject *__pyx_v_args, PyObject *__pyx_v_kw, PyObject *__pyx_v_write, int __pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_create_actor_message __pyx_t_2;
  __Pyx_RefNannySetupContext("pack_create_actor_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.write = __pyx_v_write;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_pack_create_actor_message(__pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_actor_cls, __pyx_v_args, __pyx_v_kw, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_create_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":582
 * 
 * 
 * cpdef object unpack_create_actor_message(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef object buf
 *     cdef object message_type
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_23unpack_create_actor_message(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_create_actor_message(PyObject *__pyx_v_binary, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_message_type = 0;
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  PyObject *__pyx_v_actor_cls = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kw = 0;
  size_t __pyx_v_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("unpack_create_actor_message", 0);

  /* "mars/actors/pool/messages.pyx":592
 *     cdef tuple args
 *     cdef dict kw
 *     cdef size_t pos = 0             # <<<<<<<<<<<<<<
 * 
 *     _unpack_message_type_value(binary, &pos)
 */
  __pyx_v_pos = 0;

  /* "mars/actors/pool/messages.pyx":594
 *     cdef size_t pos = 0
 * 
 *     _unpack_message_type_value(binary, &pos)             # <<<<<<<<<<<<<<
 *     message_type = MessageType.create_actor
 *     message_id = _unpack_message_id(binary, &pos)
 */
  (void)(__pyx_f_4mars_6actors_4pool_8messages__unpack_message_type_value(__pyx_v_binary, (&__pyx_v_pos)));

  /* "mars/actors/pool/messages.pyx":595
 * 
 *     _unpack_message_type_value(binary, &pos)
 *     message_type = MessageType.create_actor             # <<<<<<<<<<<<<<
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_create_actor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_type = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":596
 *     _unpack_message_type_value(binary, &pos)
 *     message_type = MessageType.create_actor
 *     message_id = _unpack_message_id(binary, &pos)             # <<<<<<<<<<<<<<
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_id(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":597
 *     message_type = MessageType.create_actor
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     to_index = _unpack_index(binary, &pos)
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 */
  __pyx_v_from_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":598
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 *     actor_cls = _unpack_object(binary, &pos)
 */
  __pyx_v_to_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":599
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 *     actor_ref = _unpack_actor_ref(binary, &pos)             # <<<<<<<<<<<<<<
 *     actor_cls = _unpack_object(binary, &pos)
 *     args = _unpack_object(binary, &pos) or tuple()
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_4mars_6actors_4pool_8messages__unpack_actor_ref(__pyx_v_binary, (&__pyx_v_pos))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":600
 *     to_index = _unpack_index(binary, &pos)
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 *     actor_cls = _unpack_object(binary, &pos)             # <<<<<<<<<<<<<<
 *     args = _unpack_object(binary, &pos) or tuple()
 *     kw = _unpack_object(binary, &pos) or dict()
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_actor_cls = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":601
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 *     actor_cls = _unpack_object(binary, &pos)
 *     args = _unpack_object(binary, &pos) or tuple()             # <<<<<<<<<<<<<<
 *     kw = _unpack_object(binary, &pos) or dict()
 * 
 */
  __pyx_t_2 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 601, __pyx_L1_error)
  if (!__pyx_t_3) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    if (!(likely(PyTuple_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyTuple_Type))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_v_args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":602
 *     actor_cls = _unpack_object(binary, &pos)
 *     args = _unpack_object(binary, &pos) or tuple()
 *     kw = _unpack_object(binary, &pos) or dict()             # <<<<<<<<<<<<<<
 * 
 *     return CREATE_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
 */
  __pyx_t_2 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 602, __pyx_L1_error)
  if (!__pyx_t_3) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 602, __pyx_L1_error)
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L5_bool_binop_done:;
  __pyx_v_kw = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":604
 *     kw = _unpack_object(binary, &pos) or dict()
 * 
 *     return CREATE_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                                 from_index=from_index, to_index=to_index,
 *                                 actor_ref=actor_ref, actor_cls=actor_cls,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_type, __pyx_v_message_type) < 0) __PYX_ERR(0, 604, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_id, __pyx_v_message_id) < 0) __PYX_ERR(0, 604, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":605
 * 
 *     return CREATE_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
 *                                 from_index=from_index, to_index=to_index,             # <<<<<<<<<<<<<<
 *                                 actor_ref=actor_ref, actor_cls=actor_cls,
 *                                 args=args, kwargs=kw)
 */
  __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_from_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_from_index, __pyx_t_2) < 0) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_to_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_to_index, __pyx_t_2) < 0) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":606
 *     return CREATE_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
 *                                 from_index=from_index, to_index=to_index,
 *                                 actor_ref=actor_ref, actor_cls=actor_cls,             # <<<<<<<<<<<<<<
 *                                 args=args, kwargs=kw)
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_actor_ref, ((PyObject *)__pyx_v_actor_ref)) < 0) __PYX_ERR(0, 604, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_actor_cls, __pyx_v_actor_cls) < 0) __PYX_ERR(0, 604, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":607
 *                                 from_index=from_index, to_index=to_index,
 *                                 actor_ref=actor_ref, actor_cls=actor_cls,
 *                                 args=args, kwargs=kw)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_args, __pyx_v_args) < 0) __PYX_ERR(0, 604, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_kwargs, __pyx_v_kw) < 0) __PYX_ERR(0, 604, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":604
 *     kw = _unpack_object(binary, &pos) or dict()
 * 
 *     return CREATE_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                                 from_index=from_index, to_index=to_index,
 *                                 actor_ref=actor_ref, actor_cls=actor_cls,
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":582
 * 
 * 
 * cpdef object unpack_create_actor_message(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef object buf
 *     cdef object message_type
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_create_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message_type);
  __Pyx_XDECREF(__pyx_v_message_id);
  __Pyx_XDECREF((PyObject *)__pyx_v_actor_ref);
  __Pyx_XDECREF(__pyx_v_actor_cls);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kw);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_23unpack_create_actor_message(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_23unpack_create_actor_message(PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_create_actor_message (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_binary), (&PyBytes_Type), 1, "binary", 1))) __PYX_ERR(0, 582, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_22unpack_create_actor_message(__pyx_self, ((PyObject*)__pyx_v_binary));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_22unpack_create_actor_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_create_actor_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_unpack_create_actor_message(__pyx_v_binary, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_create_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":610
 * 
 * 
 * cpdef object pack_destroy_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,             # <<<<<<<<<<<<<<
 *                                         object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytes message_id
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_25pack_destroy_actor_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_destroy_actor_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_destroy_actor_message *__pyx_optional_args) {

  /* "mars/actors/pool/messages.pyx":611
 * 
 * cpdef object pack_destroy_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,
 *                                         object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytes message_id
 *     cdef bytearray buf
 */
  PyObject *__pyx_v_write = ((PyObject *)Py_None);
  int __pyx_v_protocol = __pyx_k__7;
  PyObject *__pyx_v_message_id = 0;
  PyObject *__pyx_v_buf = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_message_type __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("pack_destroy_actor_message", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_write = __pyx_optional_args->write;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_protocol = __pyx_optional_args->protocol;
      }
    }
  }

  /* "mars/actors/pool/messages.pyx":616
 * 
 *     # from_index -1 means main process, -2 means remote, other is the subprocess id
 *     message_id = new_message_id()             # <<<<<<<<<<<<<<
 * 
 *     buf = bytearray()
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_new_message_id(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":618
 *     message_id = new_message_id()
 * 
 *     buf = bytearray()             # <<<<<<<<<<<<<<
 *     _pack_message_type(DESTROY_ACTOR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyByteArray_Type))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buf = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":619
 * 
 *     buf = bytearray()
 *     _pack_message_type(DESTROY_ACTOR, buf, protocol=protocol)             # <<<<<<<<<<<<<<
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_type(__pyx_v_4mars_6actors_4pool_8messages_DESTROY_ACTOR, __pyx_v_buf, &__pyx_t_2); 

  /* "mars/actors/pool/messages.pyx":620
 *     buf = bytearray()
 *     _pack_message_type(DESTROY_ACTOR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_id(__pyx_v_message_id, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":621
 *     _pack_message_type(DESTROY_ACTOR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(to_index, buf)
 *     _pack_actor_ref(actor_ref, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_from_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":622
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_actor_ref(actor_ref, buf)
 * 
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_to_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":623
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 *     _pack_actor_ref(actor_ref, buf)             # <<<<<<<<<<<<<<
 * 
 *     if write:
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_actor_ref(__pyx_v_actor_ref, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":625
 *     _pack_actor_ref(actor_ref, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf)
 *         return message_id
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_write); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 625, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "mars/actors/pool/messages.pyx":626
 * 
 *     if write:
 *         write(buf)             # <<<<<<<<<<<<<<
 *         return message_id
 * 
 */
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_4 = __pyx_v_write; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_buf) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_buf);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mars/actors/pool/messages.pyx":627
 *     if write:
 *         write(buf)
 *         return message_id             # <<<<<<<<<<<<<<
 * 
 *     return message_id, buf
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_message_id);
    __pyx_r = __pyx_v_message_id;
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":625
 *     _pack_actor_ref(actor_ref, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf)
 *         return message_id
 */
  }

  /* "mars/actors/pool/messages.pyx":629
 *         return message_id
 * 
 *     return message_id, buf             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_message_id);
  __Pyx_INCREF(__pyx_v_buf);
  __Pyx_GIVEREF(__pyx_v_buf);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_buf);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":610
 * 
 * 
 * cpdef object pack_destroy_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,             # <<<<<<<<<<<<<<
 *                                         object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytes message_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_destroy_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message_id);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_25pack_destroy_actor_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_25pack_destroy_actor_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  PyObject *__pyx_v_write = 0;
  int __pyx_v_protocol;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_destroy_actor_message (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_actor_ref,&__pyx_n_s_write,&__pyx_n_s_protocol,0};
    PyObject* values[5] = {0,0,0,0,0};

    /* "mars/actors/pool/messages.pyx":611
 * 
 * cpdef object pack_destroy_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,
 *                                         object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytes message_id
 *     cdef bytearray buf
 */
    values[3] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_destroy_actor_message", 0, 3, 5, 1); __PYX_ERR(0, 610, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_ref)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_destroy_actor_message", 0, 3, 5, 2); __PYX_ERR(0, 610, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_write);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_protocol);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pack_destroy_actor_message") < 0)) __PYX_ERR(0, 610, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[0]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 610, __pyx_L3_error)
    __pyx_v_to_index = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_to_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 610, __pyx_L3_error)
    __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)values[2]);
    __pyx_v_write = values[3];
    if (values[4]) {
      __pyx_v_protocol = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_protocol == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 611, __pyx_L3_error)
    } else {
      __pyx_v_protocol = __pyx_k__7;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pack_destroy_actor_message", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 610, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_destroy_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_actor_ref), __pyx_ptype_4mars_6actors_4core_ActorRef, 1, "actor_ref", 0))) __PYX_ERR(0, 610, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_24pack_destroy_actor_message(__pyx_self, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_write, __pyx_v_protocol);

  /* "mars/actors/pool/messages.pyx":610
 * 
 * 
 * cpdef object pack_destroy_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,             # <<<<<<<<<<<<<<
 *                                         object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytes message_id
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_24pack_destroy_actor_message(CYTHON_UNUSED PyObject *__pyx_self, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_write, int __pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_destroy_actor_message __pyx_t_2;
  __Pyx_RefNannySetupContext("pack_destroy_actor_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.write = __pyx_v_write;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_pack_destroy_actor_message(__pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_destroy_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":632
 * 
 * 
 * cpdef object unpack_destroy_actor_message(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef object buf
 *     cdef object message_type
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_27unpack_destroy_actor_message(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_destroy_actor_message(PyObject *__pyx_v_binary, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_message_type = 0;
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  int __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  size_t __pyx_v_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("unpack_destroy_actor_message", 0);

  /* "mars/actors/pool/messages.pyx":639
 *     cdef int to_index
 *     cdef ActorRef actor_ref
 *     cdef size_t pos = 0             # <<<<<<<<<<<<<<
 * 
 *     message_type = _unpack_message_type(binary, &pos)
 */
  __pyx_v_pos = 0;

  /* "mars/actors/pool/messages.pyx":641
 *     cdef size_t pos = 0
 * 
 *     message_type = _unpack_message_type(binary, &pos)             # <<<<<<<<<<<<<<
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_type(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_type = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":642
 * 
 *     message_type = _unpack_message_type(binary, &pos)
 *     message_id = _unpack_message_id(binary, &pos)             # <<<<<<<<<<<<<<
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_id(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":643
 *     message_type = _unpack_message_type(binary, &pos)
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     to_index = _unpack_index(binary, &pos)
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 */
  __pyx_v_from_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":644
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 * 
 */
  __pyx_v_to_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":645
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 *     actor_ref = _unpack_actor_ref(binary, &pos)             # <<<<<<<<<<<<<<
 * 
 *     return DESTROY_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_4mars_6actors_4pool_8messages__unpack_actor_ref(__pyx_v_binary, (&__pyx_v_pos))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":647
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 * 
 *     return DESTROY_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                                  from_index=from_index, to_index=to_index,
 *                                  actor_ref=actor_ref)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_type, __pyx_v_message_type) < 0) __PYX_ERR(0, 647, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_id, __pyx_v_message_id) < 0) __PYX_ERR(0, 647, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":648
 * 
 *     return DESTROY_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
 *                                  from_index=from_index, to_index=to_index,             # <<<<<<<<<<<<<<
 *                                  actor_ref=actor_ref)
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_from_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_from_index, __pyx_t_2) < 0) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_to_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_to_index, __pyx_t_2) < 0) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":649
 *     return DESTROY_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
 *                                  from_index=from_index, to_index=to_index,
 *                                  actor_ref=actor_ref)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_actor_ref, ((PyObject *)__pyx_v_actor_ref)) < 0) __PYX_ERR(0, 647, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":647
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 * 
 *     return DESTROY_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                                  from_index=from_index, to_index=to_index,
 *                                  actor_ref=actor_ref)
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":632
 * 
 * 
 * cpdef object unpack_destroy_actor_message(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef object buf
 *     cdef object message_type
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_destroy_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message_type);
  __Pyx_XDECREF(__pyx_v_message_id);
  __Pyx_XDECREF((PyObject *)__pyx_v_actor_ref);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_27unpack_destroy_actor_message(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_27unpack_destroy_actor_message(PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_destroy_actor_message (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_binary), (&PyBytes_Type), 1, "binary", 1))) __PYX_ERR(0, 632, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_26unpack_destroy_actor_message(__pyx_self, ((PyObject*)__pyx_v_binary));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_26unpack_destroy_actor_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_destroy_actor_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_unpack_destroy_actor_message(__pyx_v_binary, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_destroy_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":652
 * 
 * 
 * cpdef object pack_has_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref,             # <<<<<<<<<<<<<<
 *                                     object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytearray buf
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_29pack_has_actor_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_has_actor_message(__pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, int __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_has_actor_message *__pyx_optional_args) {

  /* "mars/actors/pool/messages.pyx":653
 * 
 * cpdef object pack_has_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref,
 *                                     object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytearray buf
 * 
 */
  PyObject *__pyx_v_write = ((PyObject *)Py_None);
  int __pyx_v_protocol = __pyx_k__8;
  PyObject *__pyx_v_buf = 0;
  PyObject *__pyx_v_message_id = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_message_type __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("pack_has_actor_message", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_write = __pyx_optional_args->write;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_protocol = __pyx_optional_args->protocol;
      }
    }
  }

  /* "mars/actors/pool/messages.pyx":657
 * 
 *     # from_index -1 means main process, -2 means remote, other is the subprocess id
 *     message_id = new_message_id()             # <<<<<<<<<<<<<<
 * 
 *     buf = bytearray()
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_new_message_id(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":659
 *     message_id = new_message_id()
 * 
 *     buf = bytearray()             # <<<<<<<<<<<<<<
 *     _pack_message_type(HAS_ACTOR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyByteArray_Type))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buf = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":660
 * 
 *     buf = bytearray()
 *     _pack_message_type(HAS_ACTOR, buf, protocol=protocol)             # <<<<<<<<<<<<<<
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_type(__pyx_v_4mars_6actors_4pool_8messages_HAS_ACTOR, __pyx_v_buf, &__pyx_t_2); 

  /* "mars/actors/pool/messages.pyx":661
 *     buf = bytearray()
 *     _pack_message_type(HAS_ACTOR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_id(__pyx_v_message_id, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":662
 *     _pack_message_type(HAS_ACTOR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(to_index, buf)
 *     _pack_actor_ref(actor_ref, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_from_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":663
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_actor_ref(actor_ref, buf)
 * 
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_to_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":664
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 *     _pack_actor_ref(actor_ref, buf)             # <<<<<<<<<<<<<<
 * 
 *     if write:
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_actor_ref(__pyx_v_actor_ref, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":666
 *     _pack_actor_ref(actor_ref, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf)
 *         return message_id
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_write); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 666, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "mars/actors/pool/messages.pyx":667
 * 
 *     if write:
 *         write(buf)             # <<<<<<<<<<<<<<
 *         return message_id
 * 
 */
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_4 = __pyx_v_write; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_buf) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_buf);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mars/actors/pool/messages.pyx":668
 *     if write:
 *         write(buf)
 *         return message_id             # <<<<<<<<<<<<<<
 * 
 *     return message_id, buf
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_message_id);
    __pyx_r = __pyx_v_message_id;
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":666
 *     _pack_actor_ref(actor_ref, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf)
 *         return message_id
 */
  }

  /* "mars/actors/pool/messages.pyx":670
 *         return message_id
 * 
 *     return message_id, buf             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_message_id);
  __Pyx_INCREF(__pyx_v_buf);
  __Pyx_GIVEREF(__pyx_v_buf);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_buf);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":652
 * 
 * 
 * cpdef object pack_has_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref,             # <<<<<<<<<<<<<<
 *                                     object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytearray buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_has_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XDECREF(__pyx_v_message_id);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_29pack_has_actor_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_29pack_has_actor_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  int __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  PyObject *__pyx_v_write = 0;
  int __pyx_v_protocol;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_has_actor_message (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_actor_ref,&__pyx_n_s_write,&__pyx_n_s_protocol,0};
    PyObject* values[5] = {0,0,0,0,0};

    /* "mars/actors/pool/messages.pyx":653
 * 
 * cpdef object pack_has_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref,
 *                                     object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytearray buf
 * 
 */
    values[3] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_has_actor_message", 0, 3, 5, 1); __PYX_ERR(0, 652, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_actor_ref)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_has_actor_message", 0, 3, 5, 2); __PYX_ERR(0, 652, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_write);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_protocol);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pack_has_actor_message") < 0)) __PYX_ERR(0, 652, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[0]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 652, __pyx_L3_error)
    __pyx_v_to_index = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_to_index == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 652, __pyx_L3_error)
    __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)values[2]);
    __pyx_v_write = values[3];
    if (values[4]) {
      __pyx_v_protocol = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_protocol == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 653, __pyx_L3_error)
    } else {
      __pyx_v_protocol = __pyx_k__8;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pack_has_actor_message", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 652, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_has_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_actor_ref), __pyx_ptype_4mars_6actors_4core_ActorRef, 1, "actor_ref", 0))) __PYX_ERR(0, 652, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_28pack_has_actor_message(__pyx_self, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, __pyx_v_write, __pyx_v_protocol);

  /* "mars/actors/pool/messages.pyx":652
 * 
 * 
 * cpdef object pack_has_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref,             # <<<<<<<<<<<<<<
 *                                     object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytearray buf
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_28pack_has_actor_message(CYTHON_UNUSED PyObject *__pyx_self, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, int __pyx_v_to_index, struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref, PyObject *__pyx_v_write, int __pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_has_actor_message __pyx_t_2;
  __Pyx_RefNannySetupContext("pack_has_actor_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.write = __pyx_v_write;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_pack_has_actor_message(__pyx_v_from_index, __pyx_v_to_index, __pyx_v_actor_ref, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_has_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":673
 * 
 * 
 * cpdef object unpack_has_actor_message(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef object message_type
 *     cdef bytes message_id
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_31unpack_has_actor_message(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_has_actor_message(PyObject *__pyx_v_binary, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_message_type = 0;
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  struct __pyx_obj_4mars_6actors_4core_ActorRef *__pyx_v_actor_ref = 0;
  size_t __pyx_v_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("unpack_has_actor_message", 0);

  /* "mars/actors/pool/messages.pyx":679
 *     cdef INT32_t to_index
 *     cdef ActorRef actor_ref
 *     cdef size_t pos = 0             # <<<<<<<<<<<<<<
 * 
 *     message_type = _unpack_message_type(binary, &pos)
 */
  __pyx_v_pos = 0;

  /* "mars/actors/pool/messages.pyx":681
 *     cdef size_t pos = 0
 * 
 *     message_type = _unpack_message_type(binary, &pos)             # <<<<<<<<<<<<<<
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_type(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_type = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":682
 * 
 *     message_type = _unpack_message_type(binary, &pos)
 *     message_id = _unpack_message_id(binary, &pos)             # <<<<<<<<<<<<<<
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_id(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":683
 *     message_type = _unpack_message_type(binary, &pos)
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     to_index = _unpack_index(binary, &pos)
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 */
  __pyx_v_from_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":684
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 * 
 */
  __pyx_v_to_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":685
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 *     actor_ref = _unpack_actor_ref(binary, &pos)             # <<<<<<<<<<<<<<
 * 
 *     return HAS_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_4mars_6actors_4pool_8messages__unpack_actor_ref(__pyx_v_binary, (&__pyx_v_pos))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":687
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 * 
 *     return HAS_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                              from_index=from_index, to_index=to_index,
 *                              actor_ref=actor_ref)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_type, __pyx_v_message_type) < 0) __PYX_ERR(0, 687, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_id, __pyx_v_message_id) < 0) __PYX_ERR(0, 687, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":688
 * 
 *     return HAS_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
 *                              from_index=from_index, to_index=to_index,             # <<<<<<<<<<<<<<
 *                              actor_ref=actor_ref)
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_from_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_from_index, __pyx_t_2) < 0) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_to_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_to_index, __pyx_t_2) < 0) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":689
 *     return HAS_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
 *                              from_index=from_index, to_index=to_index,
 *                              actor_ref=actor_ref)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_actor_ref, ((PyObject *)__pyx_v_actor_ref)) < 0) __PYX_ERR(0, 687, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":687
 *     actor_ref = _unpack_actor_ref(binary, &pos)
 * 
 *     return HAS_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                              from_index=from_index, to_index=to_index,
 *                              actor_ref=actor_ref)
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":673
 * 
 * 
 * cpdef object unpack_has_actor_message(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef object message_type
 *     cdef bytes message_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_has_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message_type);
  __Pyx_XDECREF(__pyx_v_message_id);
  __Pyx_XDECREF((PyObject *)__pyx_v_actor_ref);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_31unpack_has_actor_message(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_31unpack_has_actor_message(PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_has_actor_message (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_binary), (&PyBytes_Type), 1, "binary", 1))) __PYX_ERR(0, 673, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_30unpack_has_actor_message(__pyx_self, ((PyObject*)__pyx_v_binary));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_30unpack_has_actor_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_has_actor_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_unpack_has_actor_message(__pyx_v_binary, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_has_actor_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":692
 * 
 * 
 * cpdef object pack_result_message(bytes message_id, INT32_t from_index, int to_index, object result,             # <<<<<<<<<<<<<<
 *                                  object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytearray buf
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_33pack_result_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_result_message(PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, int __pyx_v_to_index, PyObject *__pyx_v_result, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_result_message *__pyx_optional_args) {

  /* "mars/actors/pool/messages.pyx":693
 * 
 * cpdef object pack_result_message(bytes message_id, INT32_t from_index, int to_index, object result,
 *                                  object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytearray buf
 * 
 */
  PyObject *__pyx_v_write = ((PyObject *)Py_None);
  int __pyx_v_protocol = __pyx_k__9;
  PyObject *__pyx_v_buf = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_message_type __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("pack_result_message", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_write = __pyx_optional_args->write;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_protocol = __pyx_optional_args->protocol;
      }
    }
  }

  /* "mars/actors/pool/messages.pyx":696
 *     cdef bytearray buf
 * 
 *     buf = bytearray()             # <<<<<<<<<<<<<<
 *     _pack_message_type(RESULT, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyByteArray_Type))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buf = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":697
 * 
 *     buf = bytearray()
 *     _pack_message_type(RESULT, buf, protocol=protocol)             # <<<<<<<<<<<<<<
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_type(__pyx_v_4mars_6actors_4pool_8messages_RESULT, __pyx_v_buf, &__pyx_t_2); 

  /* "mars/actors/pool/messages.pyx":698
 *     buf = bytearray()
 *     _pack_message_type(RESULT, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_id(__pyx_v_message_id, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":699
 *     _pack_message_type(RESULT, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(to_index, buf)
 *     _pack_object(result, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_from_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":700
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_object(result, buf)
 * 
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_to_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":701
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 *     _pack_object(result, buf)             # <<<<<<<<<<<<<<
 * 
 *     if write:
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_object(__pyx_v_result, __pyx_v_buf); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 701, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":703
 *     _pack_object(result, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf)
 *         return message_id
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_write); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 703, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "mars/actors/pool/messages.pyx":704
 * 
 *     if write:
 *         write(buf)             # <<<<<<<<<<<<<<
 *         return message_id
 * 
 */
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_4 = __pyx_v_write; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_buf) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_buf);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mars/actors/pool/messages.pyx":705
 *     if write:
 *         write(buf)
 *         return message_id             # <<<<<<<<<<<<<<
 * 
 *     return message_id, buf
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_message_id);
    __pyx_r = __pyx_v_message_id;
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":703
 *     _pack_object(result, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf)
 *         return message_id
 */
  }

  /* "mars/actors/pool/messages.pyx":707
 *         return message_id
 * 
 *     return message_id, buf             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_message_id);
  __Pyx_INCREF(__pyx_v_buf);
  __Pyx_GIVEREF(__pyx_v_buf);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_buf);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":692
 * 
 * 
 * cpdef object pack_result_message(bytes message_id, INT32_t from_index, int to_index, object result,             # <<<<<<<<<<<<<<
 *                                  object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytearray buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_result_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_33pack_result_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_33pack_result_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  int __pyx_v_to_index;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_v_write = 0;
  int __pyx_v_protocol;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_result_message (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_message_id,&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_result,&__pyx_n_s_write,&__pyx_n_s_protocol,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "mars/actors/pool/messages.pyx":693
 * 
 * cpdef object pack_result_message(bytes message_id, INT32_t from_index, int to_index, object result,
 *                                  object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytearray buf
 * 
 */
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_id)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_result_message", 0, 4, 6, 1); __PYX_ERR(0, 692, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_result_message", 0, 4, 6, 2); __PYX_ERR(0, 692, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_result_message", 0, 4, 6, 3); __PYX_ERR(0, 692, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_write);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_protocol);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pack_result_message") < 0)) __PYX_ERR(0, 692, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_message_id = ((PyObject*)values[0]);
    __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 692, __pyx_L3_error)
    __pyx_v_to_index = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_to_index == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 692, __pyx_L3_error)
    __pyx_v_result = values[3];
    __pyx_v_write = values[4];
    if (values[5]) {
      __pyx_v_protocol = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_protocol == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 693, __pyx_L3_error)
    } else {
      __pyx_v_protocol = __pyx_k__9;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pack_result_message", 0, 4, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 692, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_result_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message_id), (&PyBytes_Type), 1, "message_id", 1))) __PYX_ERR(0, 692, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_32pack_result_message(__pyx_self, __pyx_v_message_id, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_result, __pyx_v_write, __pyx_v_protocol);

  /* "mars/actors/pool/messages.pyx":692
 * 
 * 
 * cpdef object pack_result_message(bytes message_id, INT32_t from_index, int to_index, object result,             # <<<<<<<<<<<<<<
 *                                  object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytearray buf
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_32pack_result_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, int __pyx_v_to_index, PyObject *__pyx_v_result, PyObject *__pyx_v_write, int __pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_result_message __pyx_t_2;
  __Pyx_RefNannySetupContext("pack_result_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.write = __pyx_v_write;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_pack_result_message(__pyx_v_message_id, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_result, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_result_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":710
 * 
 * 
 * cpdef object unpack_result_message(bytes binary, object from_index=None, object to_index=None):             # <<<<<<<<<<<<<<
 *     cdef object message_type
 *     cdef bytes message_id
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_35unpack_result_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_result_message(PyObject *__pyx_v_binary, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_unpack_result_message *__pyx_optional_args) {
  PyObject *__pyx_v_from_index = ((PyObject *)Py_None);
  PyObject *__pyx_v_to_index = ((PyObject *)Py_None);
  PyObject *__pyx_v_message_type = 0;
  PyObject *__pyx_v_message_id = 0;
  PyObject *__pyx_v_result = 0;
  size_t __pyx_v_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("unpack_result_message", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_from_index = __pyx_optional_args->from_index;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_to_index = __pyx_optional_args->to_index;
      }
    }
  }
  __Pyx_INCREF(__pyx_v_from_index);
  __Pyx_INCREF(__pyx_v_to_index);

  /* "mars/actors/pool/messages.pyx":714
 *     cdef bytes message_id
 *     cdef object result
 *     cdef size_t pos = 0             # <<<<<<<<<<<<<<
 * 
 *     _unpack_message_type_value(binary, &pos)
 */
  __pyx_v_pos = 0;

  /* "mars/actors/pool/messages.pyx":716
 *     cdef size_t pos = 0
 * 
 *     _unpack_message_type_value(binary, &pos)             # <<<<<<<<<<<<<<
 *     message_type = MessageType.result
 *     message_id = _unpack_message_id(binary, &pos)
 */
  (void)(__pyx_f_4mars_6actors_4pool_8messages__unpack_message_type_value(__pyx_v_binary, (&__pyx_v_pos)));

  /* "mars/actors/pool/messages.pyx":717
 * 
 *     _unpack_message_type_value(binary, &pos)
 *     message_type = MessageType.result             # <<<<<<<<<<<<<<
 *     message_id = _unpack_message_id(binary, &pos)
 *     if from_index is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_type = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":718
 *     _unpack_message_type_value(binary, &pos)
 *     message_type = MessageType.result
 *     message_id = _unpack_message_id(binary, &pos)             # <<<<<<<<<<<<<<
 *     if from_index is not None:
 *         _skip_index(binary, &pos)
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_id(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":719
 *     message_type = MessageType.result
 *     message_id = _unpack_message_id(binary, &pos)
 *     if from_index is not None:             # <<<<<<<<<<<<<<
 *         _skip_index(binary, &pos)
 *     else:
 */
  __pyx_t_2 = (__pyx_v_from_index != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "mars/actors/pool/messages.pyx":720
 *     message_id = _unpack_message_id(binary, &pos)
 *     if from_index is not None:
 *         _skip_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     else:
 *         from_index = _unpack_index(binary, &pos)
 */
    __pyx_f_4mars_6actors_4pool_8messages__skip_index(__pyx_v_binary, (&__pyx_v_pos));

    /* "mars/actors/pool/messages.pyx":719
 *     message_type = MessageType.result
 *     message_id = _unpack_message_id(binary, &pos)
 *     if from_index is not None:             # <<<<<<<<<<<<<<
 *         _skip_index(binary, &pos)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "mars/actors/pool/messages.pyx":722
 *         _skip_index(binary, &pos)
 *     else:
 *         from_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     if to_index is not None:
 *         _skip_index(binary, &pos)
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_from_index, __pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "mars/actors/pool/messages.pyx":723
 *     else:
 *         from_index = _unpack_index(binary, &pos)
 *     if to_index is not None:             # <<<<<<<<<<<<<<
 *         _skip_index(binary, &pos)
 *     else:
 */
  __pyx_t_3 = (__pyx_v_to_index != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "mars/actors/pool/messages.pyx":724
 *         from_index = _unpack_index(binary, &pos)
 *     if to_index is not None:
 *         _skip_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     else:
 *         to_index = _unpack_index(binary, &pos)
 */
    __pyx_f_4mars_6actors_4pool_8messages__skip_index(__pyx_v_binary, (&__pyx_v_pos));

    /* "mars/actors/pool/messages.pyx":723
 *     else:
 *         from_index = _unpack_index(binary, &pos)
 *     if to_index is not None:             # <<<<<<<<<<<<<<
 *         _skip_index(binary, &pos)
 *     else:
 */
    goto __pyx_L4;
  }

  /* "mars/actors/pool/messages.pyx":726
 *         _skip_index(binary, &pos)
 *     else:
 *         to_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     result = _unpack_object(binary, &pos)
 * 
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_to_index, __pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L4:;

  /* "mars/actors/pool/messages.pyx":727
 *     else:
 *         to_index = _unpack_index(binary, &pos)
 *     result = _unpack_object(binary, &pos)             # <<<<<<<<<<<<<<
 * 
 *     return RESULT_MESSAGE(message_type=message_type, message_id=message_id,
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":729
 *     result = _unpack_object(binary, &pos)
 * 
 *     return RESULT_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                           from_index=from_index, to_index=to_index,
 *                           result=result)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_type, __pyx_v_message_type) < 0) __PYX_ERR(0, 729, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_id, __pyx_v_message_id) < 0) __PYX_ERR(0, 729, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":730
 * 
 *     return RESULT_MESSAGE(message_type=message_type, message_id=message_id,
 *                           from_index=from_index, to_index=to_index,             # <<<<<<<<<<<<<<
 *                           result=result)
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_from_index, __pyx_v_from_index) < 0) __PYX_ERR(0, 729, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_to_index, __pyx_v_to_index) < 0) __PYX_ERR(0, 729, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":731
 *     return RESULT_MESSAGE(message_type=message_type, message_id=message_id,
 *                           from_index=from_index, to_index=to_index,
 *                           result=result)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_result, __pyx_v_result) < 0) __PYX_ERR(0, 729, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":729
 *     result = _unpack_object(binary, &pos)
 * 
 *     return RESULT_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                           from_index=from_index, to_index=to_index,
 *                           result=result)
 */
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_RESULT_MESSAGE), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":710
 * 
 * 
 * cpdef object unpack_result_message(bytes binary, object from_index=None, object to_index=None):             # <<<<<<<<<<<<<<
 *     cdef object message_type
 *     cdef bytes message_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_result_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message_type);
  __Pyx_XDECREF(__pyx_v_message_id);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_from_index);
  __Pyx_XDECREF(__pyx_v_to_index);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_35unpack_result_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_35unpack_result_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_binary = 0;
  PyObject *__pyx_v_from_index = 0;
  PyObject *__pyx_v_to_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_result_message (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_binary,&__pyx_n_s_from_index,&__pyx_n_s_to_index,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)Py_None);
    values[2] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_binary)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "unpack_result_message") < 0)) __PYX_ERR(0, 710, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_binary = ((PyObject*)values[0]);
    __pyx_v_from_index = values[1];
    __pyx_v_to_index = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("unpack_result_message", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 710, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_result_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_binary), (&PyBytes_Type), 1, "binary", 1))) __PYX_ERR(0, 710, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_34unpack_result_message(__pyx_self, __pyx_v_binary, __pyx_v_from_index, __pyx_v_to_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_34unpack_result_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary, PyObject *__pyx_v_from_index, PyObject *__pyx_v_to_index) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages_unpack_result_message __pyx_t_2;
  __Pyx_RefNannySetupContext("unpack_result_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.from_index = __pyx_v_from_index;
  __pyx_t_2.to_index = __pyx_v_to_index;
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_unpack_result_message(__pyx_v_binary, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_result_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":734
 * 
 * 
 * cpdef object pack_error_message(bytes message_id, INT32_t from_index, INT32_t to_index,             # <<<<<<<<<<<<<<
 *                                 object error_type, object error, object tb,
 *                                 object write=None, int protocol=DEFAULT_PROTOCOL):
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_37pack_error_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_pack_error_message(PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, PyObject *__pyx_v_error_type, PyObject *__pyx_v_error, PyObject *__pyx_v_tb, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_error_message *__pyx_optional_args) {

  /* "mars/actors/pool/messages.pyx":736
 * cpdef object pack_error_message(bytes message_id, INT32_t from_index, INT32_t to_index,
 *                                 object error_type, object error, object tb,
 *                                 object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytearray buf
 * 
 */
  PyObject *__pyx_v_write = ((PyObject *)Py_None);
  int __pyx_v_protocol = __pyx_k__10;
  PyObject *__pyx_v_buf = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages__pack_message_type __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("pack_error_message", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_write = __pyx_optional_args->write;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_protocol = __pyx_optional_args->protocol;
      }
    }
  }

  /* "mars/actors/pool/messages.pyx":739
 *     cdef bytearray buf
 * 
 *     buf = bytearray()             # <<<<<<<<<<<<<<
 *     _pack_message_type(ERROR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyByteArray_Type))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buf = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":740
 * 
 *     buf = bytearray()
 *     _pack_message_type(ERROR, buf, protocol=protocol)             # <<<<<<<<<<<<<<
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_type(__pyx_v_4mars_6actors_4pool_8messages_ERROR, __pyx_v_buf, &__pyx_t_2); 

  /* "mars/actors/pool/messages.pyx":741
 *     buf = bytearray()
 *     _pack_message_type(ERROR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_message_id(__pyx_v_message_id, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":742
 *     _pack_message_type(ERROR, buf, protocol=protocol)
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_index(to_index, buf)
 *     _pack_object(error_type, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_from_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":743
 *     _pack_message_id(message_id, buf)
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)             # <<<<<<<<<<<<<<
 *     _pack_object(error_type, buf)
 *     _pack_object(error, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_index(__pyx_v_to_index, __pyx_v_buf);

  /* "mars/actors/pool/messages.pyx":744
 *     _pack_index(from_index, buf)
 *     _pack_index(to_index, buf)
 *     _pack_object(error_type, buf)             # <<<<<<<<<<<<<<
 *     _pack_object(error, buf)
 *     _pack_object(tb, buf)
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_object(__pyx_v_error_type, __pyx_v_buf); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":745
 *     _pack_index(to_index, buf)
 *     _pack_object(error_type, buf)
 *     _pack_object(error, buf)             # <<<<<<<<<<<<<<
 *     _pack_object(tb, buf)
 * 
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_object(__pyx_v_error, __pyx_v_buf); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 745, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":746
 *     _pack_object(error_type, buf)
 *     _pack_object(error, buf)
 *     _pack_object(tb, buf)             # <<<<<<<<<<<<<<
 * 
 *     if write:
 */
  __pyx_f_4mars_6actors_4pool_8messages__pack_object(__pyx_v_tb, __pyx_v_buf); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 746, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":748
 *     _pack_object(tb, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf)
 *         return message_id
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_write); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 748, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "mars/actors/pool/messages.pyx":749
 * 
 *     if write:
 *         write(buf)             # <<<<<<<<<<<<<<
 *         return message_id
 * 
 */
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_4 = __pyx_v_write; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_buf) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_buf);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 749, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mars/actors/pool/messages.pyx":750
 *     if write:
 *         write(buf)
 *         return message_id             # <<<<<<<<<<<<<<
 * 
 *     return message_id, buf
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_message_id);
    __pyx_r = __pyx_v_message_id;
    goto __pyx_L0;

    /* "mars/actors/pool/messages.pyx":748
 *     _pack_object(tb, buf)
 * 
 *     if write:             # <<<<<<<<<<<<<<
 *         write(buf)
 *         return message_id
 */
  }

  /* "mars/actors/pool/messages.pyx":752
 *         return message_id
 * 
 *     return message_id, buf             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_message_id);
  __Pyx_GIVEREF(__pyx_v_message_id);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_message_id);
  __Pyx_INCREF(__pyx_v_buf);
  __Pyx_GIVEREF(__pyx_v_buf);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_buf);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":734
 * 
 * 
 * cpdef object pack_error_message(bytes message_id, INT32_t from_index, INT32_t to_index,             # <<<<<<<<<<<<<<
 *                                 object error_type, object error, object tb,
 *                                 object write=None, int protocol=DEFAULT_PROTOCOL):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_error_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_37pack_error_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_37pack_error_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  PyObject *__pyx_v_error_type = 0;
  PyObject *__pyx_v_error = 0;
  PyObject *__pyx_v_tb = 0;
  PyObject *__pyx_v_write = 0;
  int __pyx_v_protocol;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_error_message (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_message_id,&__pyx_n_s_from_index,&__pyx_n_s_to_index,&__pyx_n_s_error_type,&__pyx_n_s_error,&__pyx_n_s_tb,&__pyx_n_s_write,&__pyx_n_s_protocol,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};

    /* "mars/actors/pool/messages.pyx":736
 * cpdef object pack_error_message(bytes message_id, INT32_t from_index, INT32_t to_index,
 *                                 object error_type, object error, object tb,
 *                                 object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytearray buf
 * 
 */
    values[6] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message_id)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_from_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_error_message", 0, 6, 8, 1); __PYX_ERR(0, 734, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_error_message", 0, 6, 8, 2); __PYX_ERR(0, 734, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_error_message", 0, 6, 8, 3); __PYX_ERR(0, 734, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_error_message", 0, 6, 8, 4); __PYX_ERR(0, 734, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_error_message", 0, 6, 8, 5); __PYX_ERR(0, 734, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_write);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_protocol);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pack_error_message") < 0)) __PYX_ERR(0, 734, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_message_id = ((PyObject*)values[0]);
    __pyx_v_from_index = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_from_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 734, __pyx_L3_error)
    __pyx_v_to_index = __Pyx_PyInt_As_npy_int32(values[2]); if (unlikely((__pyx_v_to_index == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 734, __pyx_L3_error)
    __pyx_v_error_type = values[3];
    __pyx_v_error = values[4];
    __pyx_v_tb = values[5];
    __pyx_v_write = values[6];
    if (values[7]) {
      __pyx_v_protocol = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_protocol == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 736, __pyx_L3_error)
    } else {
      __pyx_v_protocol = __pyx_k__10;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pack_error_message", 0, 6, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 734, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_error_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message_id), (&PyBytes_Type), 1, "message_id", 1))) __PYX_ERR(0, 734, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_36pack_error_message(__pyx_self, __pyx_v_message_id, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_error_type, __pyx_v_error, __pyx_v_tb, __pyx_v_write, __pyx_v_protocol);

  /* "mars/actors/pool/messages.pyx":734
 * 
 * 
 * cpdef object pack_error_message(bytes message_id, INT32_t from_index, INT32_t to_index,             # <<<<<<<<<<<<<<
 *                                 object error_type, object error, object tb,
 *                                 object write=None, int protocol=DEFAULT_PROTOCOL):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_36pack_error_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_message_id, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index, __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index, PyObject *__pyx_v_error_type, PyObject *__pyx_v_error, PyObject *__pyx_v_tb, PyObject *__pyx_v_write, int __pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_error_message __pyx_t_2;
  __Pyx_RefNannySetupContext("pack_error_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.write = __pyx_v_write;
  __pyx_t_2.protocol = __pyx_v_protocol;
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_pack_error_message(__pyx_v_message_id, __pyx_v_from_index, __pyx_v_to_index, __pyx_v_error_type, __pyx_v_error, __pyx_v_tb, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.pack_error_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":755
 * 
 * 
 * cpdef object unpack_error_message(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef object message_type
 *     cdef bytes message_id
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_39unpack_error_message(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_unpack_error_message(PyObject *__pyx_v_binary, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_message_type = 0;
  PyObject *__pyx_v_message_id = 0;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_from_index;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_v_to_index;
  PyObject *__pyx_v_error_type = 0;
  PyObject *__pyx_v_error = 0;
  PyObject *__pyx_v_tb = 0;
  size_t __pyx_v_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("unpack_error_message", 0);

  /* "mars/actors/pool/messages.pyx":763
 *     cdef object error
 *     cdef object tb
 *     cdef size_t pos = 0             # <<<<<<<<<<<<<<
 * 
 *     message_type = _unpack_message_type(binary, &pos)
 */
  __pyx_v_pos = 0;

  /* "mars/actors/pool/messages.pyx":765
 *     cdef size_t pos = 0
 * 
 *     message_type = _unpack_message_type(binary, &pos)             # <<<<<<<<<<<<<<
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_type(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_type = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":766
 * 
 *     message_type = _unpack_message_type(binary, &pos)
 *     message_id = _unpack_message_id(binary, &pos)             # <<<<<<<<<<<<<<
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_message_id(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":767
 *     message_type = _unpack_message_type(binary, &pos)
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     to_index = _unpack_index(binary, &pos)
 *     error_type = _unpack_object(binary, &pos)
 */
  __pyx_v_from_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":768
 *     message_id = _unpack_message_id(binary, &pos)
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)             # <<<<<<<<<<<<<<
 *     error_type = _unpack_object(binary, &pos)
 *     error = _unpack_object(binary, &pos)
 */
  __pyx_v_to_index = __pyx_f_4mars_6actors_4pool_8messages__unpack_index(__pyx_v_binary, (&__pyx_v_pos));

  /* "mars/actors/pool/messages.pyx":769
 *     from_index = _unpack_index(binary, &pos)
 *     to_index = _unpack_index(binary, &pos)
 *     error_type = _unpack_object(binary, &pos)             # <<<<<<<<<<<<<<
 *     error = _unpack_object(binary, &pos)
 *     tb = _unpack_object(binary, &pos)
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_error_type = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":770
 *     to_index = _unpack_index(binary, &pos)
 *     error_type = _unpack_object(binary, &pos)
 *     error = _unpack_object(binary, &pos)             # <<<<<<<<<<<<<<
 *     tb = _unpack_object(binary, &pos)
 * 
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_error = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":771
 *     error_type = _unpack_object(binary, &pos)
 *     error = _unpack_object(binary, &pos)
 *     tb = _unpack_object(binary, &pos)             # <<<<<<<<<<<<<<
 * 
 *     return ERROR_MESSAGE(message_type=message_type, message_id=message_id,
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__unpack_object(__pyx_v_binary, (&__pyx_v_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tb = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":773
 *     tb = _unpack_object(binary, &pos)
 * 
 *     return ERROR_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                          from_index=from_index, to_index=to_index,
 *                          error_type=error_type, error=error, traceback=tb)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_type, __pyx_v_message_type) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_message_id, __pyx_v_message_id) < 0) __PYX_ERR(0, 773, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":774
 * 
 *     return ERROR_MESSAGE(message_type=message_type, message_id=message_id,
 *                          from_index=from_index, to_index=to_index,             # <<<<<<<<<<<<<<
 *                          error_type=error_type, error=error, traceback=tb)
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_from_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_from_index, __pyx_t_2) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_to_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_to_index, __pyx_t_2) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":775
 *     return ERROR_MESSAGE(message_type=message_type, message_id=message_id,
 *                          from_index=from_index, to_index=to_index,
 *                          error_type=error_type, error=error, traceback=tb)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_error_type, __pyx_v_error_type) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_error, __pyx_v_error) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_traceback, __pyx_v_tb) < 0) __PYX_ERR(0, 773, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":773
 *     tb = _unpack_object(binary, &pos)
 * 
 *     return ERROR_MESSAGE(message_type=message_type, message_id=message_id,             # <<<<<<<<<<<<<<
 *                          from_index=from_index, to_index=to_index,
 *                          error_type=error_type, error=error, traceback=tb)
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_ERROR_MESSAGE), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":755
 * 
 * 
 * cpdef object unpack_error_message(bytes binary):             # <<<<<<<<<<<<<<
 *     cdef object message_type
 *     cdef bytes message_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_error_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message_type);
  __Pyx_XDECREF(__pyx_v_message_id);
  __Pyx_XDECREF(__pyx_v_error_type);
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XDECREF(__pyx_v_tb);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_39unpack_error_message(PyObject *__pyx_self, PyObject *__pyx_v_binary); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_39unpack_error_message(PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unpack_error_message (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_binary), (&PyBytes_Type), 1, "binary", 1))) __PYX_ERR(0, 755, __pyx_L1_error)
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_38unpack_error_message(__pyx_self, ((PyObject*)__pyx_v_binary));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_38unpack_error_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_binary) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("unpack_error_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_unpack_error_message(__pyx_v_binary, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.unpack_error_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":778
 * 
 * 
 * cdef inline bytes _wrap_read_func(object read_func, size_t size):             # <<<<<<<<<<<<<<
 *     cdef bytes read_bytes
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_4mars_6actors_4pool_8messages__wrap_read_func(PyObject *__pyx_v_read_func, size_t __pyx_v_size) {
  PyObject *__pyx_v_read_bytes = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  __Pyx_RefNannySetupContext("_wrap_read_func", 0);

  /* "mars/actors/pool/messages.pyx":781
 *     cdef bytes read_bytes
 * 
 *     try:             # <<<<<<<<<<<<<<
 *          read_bytes = read_func(size)
 *     except ConnectionResetError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "mars/actors/pool/messages.pyx":782
 * 
 *     try:
 *          read_bytes = read_func(size)             # <<<<<<<<<<<<<<
 *     except ConnectionResetError:
 *         raise BrokenPipeError('The remote server is closed')
 */
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 782, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_read_func);
      __pyx_t_6 = __pyx_v_read_func; __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 782, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 782, __pyx_L3_error)
      __pyx_v_read_bytes = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "mars/actors/pool/messages.pyx":781
 *     cdef bytes read_bytes
 * 
 *     try:             # <<<<<<<<<<<<<<
 *          read_bytes = read_func(size)
 *     except ConnectionResetError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "mars/actors/pool/messages.pyx":783
 *     try:
 *          read_bytes = read_func(size)
 *     except ConnectionResetError:             # <<<<<<<<<<<<<<
 *         raise BrokenPipeError('The remote server is closed')
 * 
 */
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ConnectionResetError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 783, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_6, __pyx_t_5);
    __pyx_t_4 = 0; __pyx_t_6 = 0; __pyx_t_5 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("mars.actors.pool.messages._wrap_read_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 783, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);

      /* "mars/actors/pool/messages.pyx":784
 *          read_bytes = read_func(size)
 *     except ConnectionResetError:
 *         raise BrokenPipeError('The remote server is closed')             # <<<<<<<<<<<<<<
 * 
 *     if len(read_bytes) == 0:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_BrokenPipeError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 784, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_7 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_kp_s_The_remote_server_is_closed) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_s_The_remote_server_is_closed);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 784, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 784, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "mars/actors/pool/messages.pyx":781
 *     cdef bytes read_bytes
 * 
 *     try:             # <<<<<<<<<<<<<<
 *          read_bytes = read_func(size)
 *     except ConnectionResetError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "mars/actors/pool/messages.pyx":786
 *         raise BrokenPipeError('The remote server is closed')
 * 
 *     if len(read_bytes) == 0:             # <<<<<<<<<<<<<<
 *         raise BrokenPipeError('The remote server is closed')
 * 
 */
  if (unlikely(__pyx_v_read_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 786, __pyx_L1_error)
  }
  __pyx_t_11 = PyBytes_GET_SIZE(__pyx_v_read_bytes); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 786, __pyx_L1_error)
  __pyx_t_12 = ((__pyx_t_11 == 0) != 0);
  if (unlikely(__pyx_t_12)) {

    /* "mars/actors/pool/messages.pyx":787
 * 
 *     if len(read_bytes) == 0:
 *         raise BrokenPipeError('The remote server is closed')             # <<<<<<<<<<<<<<
 * 
 *     return read_bytes
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_BrokenPipeError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_kp_s_The_remote_server_is_closed) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_s_The_remote_server_is_closed);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 787, __pyx_L1_error)

    /* "mars/actors/pool/messages.pyx":786
 *         raise BrokenPipeError('The remote server is closed')
 * 
 *     if len(read_bytes) == 0:             # <<<<<<<<<<<<<<
 *         raise BrokenPipeError('The remote server is closed')
 * 
 */
  }

  /* "mars/actors/pool/messages.pyx":789
 *         raise BrokenPipeError('The remote server is closed')
 * 
 *     return read_bytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_read_bytes);
  __pyx_r = __pyx_v_read_bytes;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":778
 * 
 * 
 * cdef inline bytes _wrap_read_func(object read_func, size_t size):             # <<<<<<<<<<<<<<
 *     cdef bytes read_bytes
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("mars.actors.pool.messages._wrap_read_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_read_bytes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":792
 * 
 * 
 * cpdef bytes read_remote_message(object read_func):             # <<<<<<<<<<<<<<
 *     cdef size_t size
 *     cdef bytearray buf
 */

static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_41read_remote_message(PyObject *__pyx_self, PyObject *__pyx_v_read_func); /*proto*/
static PyObject *__pyx_f_4mars_6actors_4pool_8messages_read_remote_message(PyObject *__pyx_v_read_func, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_size;
  PyObject *__pyx_v_buf = 0;
  int __pyx_v_received_size;
  int __pyx_v_left;
  PyObject *__pyx_v_read_bytes = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("read_remote_message", 0);

  /* "mars/actors/pool/messages.pyx":799
 *     cdef bytes read_bytes
 * 
 *     read_bytes = _wrap_read_func(read_func, 8)             # <<<<<<<<<<<<<<
 *     size = _unpack_long(read_bytes)[0]
 *     buf = bytearray()
 */
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages__wrap_read_func(__pyx_v_read_func, 8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_read_bytes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":800
 * 
 *     read_bytes = _wrap_read_func(read_func, 8)
 *     size = _unpack_long(read_bytes)[0]             # <<<<<<<<<<<<<<
 *     buf = bytearray()
 *     received_size = 0
 */
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_long);
  __pyx_t_2 = __pyx_v_4mars_6actors_4pool_8messages__unpack_long; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_read_bytes) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_read_bytes);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_4 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_size = __pyx_t_4;

  /* "mars/actors/pool/messages.pyx":801
 *     read_bytes = _wrap_read_func(read_func, 8)
 *     size = _unpack_long(read_bytes)[0]
 *     buf = bytearray()             # <<<<<<<<<<<<<<
 *     received_size = 0
 *     left = size
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyByteArray_Type))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_buf = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":802
 *     size = _unpack_long(read_bytes)[0]
 *     buf = bytearray()
 *     received_size = 0             # <<<<<<<<<<<<<<
 *     left = size
 * 
 */
  __pyx_v_received_size = 0;

  /* "mars/actors/pool/messages.pyx":803
 *     buf = bytearray()
 *     received_size = 0
 *     left = size             # <<<<<<<<<<<<<<
 * 
 *     while True:
 */
  __pyx_v_left = __pyx_v_size;

  /* "mars/actors/pool/messages.pyx":805
 *     left = size
 * 
 *     while True:             # <<<<<<<<<<<<<<
 *         read_bytes = _wrap_read_func(read_func, left)
 *         buf.extend(read_bytes)
 */
  while (1) {

    /* "mars/actors/pool/messages.pyx":806
 * 
 *     while True:
 *         read_bytes = _wrap_read_func(read_func, left)             # <<<<<<<<<<<<<<
 *         buf.extend(read_bytes)
 *         received_size += len(read_bytes)
 */
    __pyx_t_2 = __pyx_f_4mars_6actors_4pool_8messages__wrap_read_func(__pyx_v_read_func, __pyx_v_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_read_bytes, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "mars/actors/pool/messages.pyx":807
 *     while True:
 *         read_bytes = _wrap_read_func(read_func, left)
 *         buf.extend(read_bytes)             # <<<<<<<<<<<<<<
 *         received_size += len(read_bytes)
 *         if received_size >= size:
 */
    __pyx_t_2 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_buf, __pyx_v_read_bytes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "mars/actors/pool/messages.pyx":808
 *         read_bytes = _wrap_read_func(read_func, left)
 *         buf.extend(read_bytes)
 *         received_size += len(read_bytes)             # <<<<<<<<<<<<<<
 *         if received_size >= size:
 *             break
 */
    if (unlikely(__pyx_v_read_bytes == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 808, __pyx_L1_error)
    }
    __pyx_t_5 = PyBytes_GET_SIZE(__pyx_v_read_bytes); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 808, __pyx_L1_error)
    __pyx_v_received_size = (__pyx_v_received_size + __pyx_t_5);

    /* "mars/actors/pool/messages.pyx":809
 *         buf.extend(read_bytes)
 *         received_size += len(read_bytes)
 *         if received_size >= size:             # <<<<<<<<<<<<<<
 *             break
 *         left = size - received_size
 */
    __pyx_t_6 = ((__pyx_v_received_size >= __pyx_v_size) != 0);
    if (__pyx_t_6) {

      /* "mars/actors/pool/messages.pyx":810
 *         received_size += len(read_bytes)
 *         if received_size >= size:
 *             break             # <<<<<<<<<<<<<<
 *         left = size - received_size
 * 
 */
      goto __pyx_L4_break;

      /* "mars/actors/pool/messages.pyx":809
 *         buf.extend(read_bytes)
 *         received_size += len(read_bytes)
 *         if received_size >= size:             # <<<<<<<<<<<<<<
 *             break
 *         left = size - received_size
 */
    }

    /* "mars/actors/pool/messages.pyx":811
 *         if received_size >= size:
 *             break
 *         left = size - received_size             # <<<<<<<<<<<<<<
 * 
 *     return bytes(buf)
 */
    __pyx_v_left = (__pyx_v_size - __pyx_v_received_size);
  }
  __pyx_L4_break:;

  /* "mars/actors/pool/messages.pyx":813
 *         left = size - received_size
 * 
 *     return bytes(buf)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_v_buf); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "mars/actors/pool/messages.pyx":792
 * 
 * 
 * cpdef bytes read_remote_message(object read_func):             # <<<<<<<<<<<<<<
 *     cdef size_t size
 *     cdef bytearray buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mars.actors.pool.messages.read_remote_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XDECREF(__pyx_v_read_bytes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_41read_remote_message(PyObject *__pyx_self, PyObject *__pyx_v_read_func); /*proto*/
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_41read_remote_message(PyObject *__pyx_self, PyObject *__pyx_v_read_func) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_remote_message (wrapper)", 0);
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_40read_remote_message(__pyx_self, ((PyObject *)__pyx_v_read_func));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_40read_remote_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("read_remote_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4mars_6actors_4pool_8messages_read_remote_message(__pyx_v_read_func, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mars.actors.pool.messages.read_remote_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":816
 * 
 * 
 * def write_remote_message(write_func, *binary):             # <<<<<<<<<<<<<<
 *     cdef bytes size
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_43write_remote_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4mars_6actors_4pool_8messages_43write_remote_message = {"write_remote_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_43write_remote_message, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_43write_remote_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_write_func = 0;
  PyObject *__pyx_v_binary = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_remote_message (wrapper)", 0);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_binary = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_binary)) {
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_binary);
  } else {
    __pyx_v_binary = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_write_func,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_write_func)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, used_pos_args, "write_remote_message") < 0)) __PYX_ERR(0, 816, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_write_func = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("write_remote_message", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 816, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_CLEAR(__pyx_v_binary);
  __Pyx_AddTraceback("mars.actors.pool.messages.write_remote_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_42write_remote_message(__pyx_self, __pyx_v_write_func, __pyx_v_binary);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_binary);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4mars_6actors_4pool_8messages_20write_remote_message_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "mars/actors/pool/messages.pyx":819
 *     cdef bytes size
 * 
 *     size = _pack_long(sum(len(b) for b in binary))             # <<<<<<<<<<<<<<
 *     write_func(size)
 *     [write_func(b) for b in binary]
 */

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_20write_remote_message_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr *)__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr(__pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 819, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4mars_6actors_4pool_8messages_20write_remote_message_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_write_remote_message_locals_gene, __pyx_n_s_mars_actors_pool_messages); if (unlikely(!gen)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.write_remote_message.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4mars_6actors_4pool_8messages_20write_remote_message_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr *__pyx_cur_scope = ((struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 819, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_binary)) { __Pyx_RaiseClosureNameError("binary"); __PYX_ERR(0, 819, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_binary == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 819, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_binary; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 819, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_b);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_b, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = PyObject_Length(__pyx_cur_scope->__pyx_v_b); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 819, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 819, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":816
 * 
 * 
 * def write_remote_message(write_func, *binary):             # <<<<<<<<<<<<<<
 *     cdef bytes size
 * 
 */

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_42write_remote_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_write_func, PyObject *__pyx_v_binary) {
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *__pyx_cur_scope;
  PyObject *__pyx_v_size = 0;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("write_remote_message", 0);
  __pyx_cur_scope = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *)__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message(__pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 816, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_binary = __pyx_v_binary;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_binary);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_binary);

  /* "mars/actors/pool/messages.pyx":819
 *     cdef bytes size
 * 
 *     size = _pack_long(sum(len(b) for b in binary))             # <<<<<<<<<<<<<<
 *     write_func(size)
 *     [write_func(b) for b in binary]
 */
  __pyx_t_2 = __pyx_pf_4mars_6actors_4pool_8messages_20write_remote_message_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_INCREF(__pyx_v_4mars_6actors_4pool_8messages__pack_long);
  __pyx_t_2 = __pyx_v_4mars_6actors_4pool_8messages__pack_long; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 819, __pyx_L1_error)
  __pyx_v_size = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":820
 * 
 *     size = _pack_long(sum(len(b) for b in binary))
 *     write_func(size)             # <<<<<<<<<<<<<<
 *     [write_func(b) for b in binary]
 * 
 */
  __Pyx_INCREF(__pyx_v_write_func);
  __pyx_t_2 = __pyx_v_write_func; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_size) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_size);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":821
 *     size = _pack_long(sum(len(b) for b in binary))
 *     write_func(size)
 *     [write_func(b) for b in binary]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_binary; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 821, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 821, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_b, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_v_write_func);
    __pyx_t_4 = __pyx_v_write_func; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_b) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_b);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 821, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 821, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":816
 * 
 * 
 * def write_remote_message(write_func, *binary):             # <<<<<<<<<<<<<<
 *     cdef bytes size
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("mars.actors.pool.messages.write_remote_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mars/actors/pool/messages.pyx":827
 *     _write_remote_message = write_remote_message
 * 
 *     def write_remote_message(*args, **kwargs):             # <<<<<<<<<<<<<<
 *         try:
 *             return _write_remote_message(*args, **kwargs)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_45write_remote_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4mars_6actors_4pool_8messages_45write_remote_message = {"write_remote_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_45write_remote_message, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_45write_remote_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_remote_message (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "write_remote_message", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = PyDict_Copy(__pyx_kwds); if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = NULL;
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_44write_remote_message(__pyx_self, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_44write_remote_message(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("write_remote_message", 0);

  /* "mars/actors/pool/messages.pyx":828
 * 
 *     def write_remote_message(*args, **kwargs):
 *         try:             # <<<<<<<<<<<<<<
 *             return _write_remote_message(*args, **kwargs)
 *         except socket.error as ex:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "mars/actors/pool/messages.pyx":829
 *     def write_remote_message(*args, **kwargs):
 *         try:
 *             return _write_remote_message(*args, **kwargs)             # <<<<<<<<<<<<<<
 *         except socket.error as ex:
 *             if ex.errno == errno.EPIPE:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_write_remote_message); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 829, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_v_args, __pyx_v_kwargs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 829, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L7_try_return;

      /* "mars/actors/pool/messages.pyx":828
 * 
 *     def write_remote_message(*args, **kwargs):
 *         try:             # <<<<<<<<<<<<<<
 *             return _write_remote_message(*args, **kwargs)
 *         except socket.error as ex:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "mars/actors/pool/messages.pyx":830
 *         try:
 *             return _write_remote_message(*args, **kwargs)
 *         except socket.error as ex:             # <<<<<<<<<<<<<<
 *             if ex.errno == errno.EPIPE:
 *                 raise BrokenPipeError
 */
    __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_4, &__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_socket); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 830, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_error); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 830, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_5, __pyx_t_8);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_ErrRestore(__pyx_t_5, __pyx_t_4, __pyx_t_6);
    __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_6 = 0;
    if (__pyx_t_9) {
      __Pyx_AddTraceback("mars.actors.pool.messages.write_remote_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 830, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_ex = __pyx_t_4;

      /* "mars/actors/pool/messages.pyx":831
 *             return _write_remote_message(*args, **kwargs)
 *         except socket.error as ex:
 *             if ex.errno == errno.EPIPE:             # <<<<<<<<<<<<<<
 *                 raise BrokenPipeError
 *             else:
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_ex, __pyx_n_s_errno); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 831, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_errno); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 831, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_EPIPE); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 831, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyObject_RichCompare(__pyx_t_8, __pyx_t_10, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 831, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 831, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(__pyx_t_11)) {

        /* "mars/actors/pool/messages.pyx":832
 *         except socket.error as ex:
 *             if ex.errno == errno.EPIPE:
 *                 raise BrokenPipeError             # <<<<<<<<<<<<<<
 *             else:
 *                 raise
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_BrokenPipeError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 832, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 832, __pyx_L5_except_error)

        /* "mars/actors/pool/messages.pyx":831
 *             return _write_remote_message(*args, **kwargs)
 *         except socket.error as ex:
 *             if ex.errno == errno.EPIPE:             # <<<<<<<<<<<<<<
 *                 raise BrokenPipeError
 *             else:
 */
      }

      /* "mars/actors/pool/messages.pyx":834
 *                 raise BrokenPipeError
 *             else:
 *                 raise             # <<<<<<<<<<<<<<
 */
      /*else*/ {
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_4, __pyx_t_5);
        __pyx_t_6 = 0; __pyx_t_4 = 0; __pyx_t_5 = 0; 
        __PYX_ERR(0, 834, __pyx_L5_except_error)
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "mars/actors/pool/messages.pyx":828
 * 
 *     def write_remote_message(*args, **kwargs):
 *         try:             # <<<<<<<<<<<<<<
 *             return _write_remote_message(*args, **kwargs)
 *         except socket.error as ex:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "mars/actors/pool/messages.pyx":827
 *     _write_remote_message = write_remote_message
 * 
 *     def write_remote_message(*args, **kwargs):             # <<<<<<<<<<<<<<
 *         try:
 *             return _write_remote_message(*args, **kwargs)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("mars.actors.pool.messages.write_remote_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CREATE_ACTOR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_47__pyx_unpickle_CREATE_ACTOR_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4mars_6actors_4pool_8messages_47__pyx_unpickle_CREATE_ACTOR_MESSAGE = {"__pyx_unpickle_CREATE_ACTOR_MESSAGE", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_47__pyx_unpickle_CREATE_ACTOR_MESSAGE, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_47__pyx_unpickle_CREATE_ACTOR_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CREATE_ACTOR_MESSAGE (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CREATE_ACTOR_MESSAGE", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CREATE_ACTOR_MESSAGE", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_CREATE_ACTOR_MESSAGE") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CREATE_ACTOR_MESSAGE", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_CREATE_ACTOR_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_46__pyx_unpickle_CREATE_ACTOR_MESSAGE(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_46__pyx_unpickle_CREATE_ACTOR_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CREATE_ACTOR_MESSAGE", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x2dc8e2e:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x2dc8e2e = (actor_cls, actor_ref, args, from_index, kwargs, message_id, message_type, to_index))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x2dc8e2e) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x2dc8e2e:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x2dc8e2e = (actor_cls, actor_ref, args, from_index, kwargs, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = CREATE_ACTOR_MESSAGE.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x2dc8e2e:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x2dc8e2e = (actor_cls, actor_ref, args, from_index, kwargs, message_id, message_type, to_index))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = CREATE_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x2d, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x2dc8e2e:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x2dc8e2e = (actor_cls, actor_ref, args, from_index, kwargs, message_id, message_type, to_index))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x2dc8e2e = (actor_cls, actor_ref, args, from_index, kwargs, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = CREATE_ACTOR_MESSAGE.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(<CREATE_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x2dc8e2e = (actor_cls, actor_ref, args, from_index, kwargs, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = CREATE_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(<CREATE_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = CREATE_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(<CREATE_ACTOR_MESSAGE> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(CREATE_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x2dc8e2e = (actor_cls, actor_ref, args, from_index, kwargs, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = CREATE_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(<CREATE_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(<CREATE_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(CREATE_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_cls = __pyx_state[0]; __pyx_result.actor_ref = __pyx_state[1]; __pyx_result.args = __pyx_state[2]; __pyx_result.from_index = __pyx_state[3]; __pyx_result.kwargs = __pyx_state[4]; __pyx_result.message_id = __pyx_state[5]; __pyx_result.message_type = __pyx_state[6]; __pyx_result.to_index = __pyx_state[7]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CREATE_ACTOR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_CREATE_ACTOR_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(<CREATE_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(CREATE_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.actor_cls = __pyx_state[0]; __pyx_result.actor_ref = __pyx_state[1]; __pyx_result.args = __pyx_state[2]; __pyx_result.from_index = __pyx_state[3]; __pyx_result.kwargs = __pyx_state[4]; __pyx_result.message_id = __pyx_state[5]; __pyx_result.message_type = __pyx_state[6]; __pyx_result.to_index = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(CREATE_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_cls = __pyx_state[0]; __pyx_result.actor_ref = __pyx_state[1]; __pyx_result.args = __pyx_state[2]; __pyx_result.from_index = __pyx_state[3]; __pyx_result.kwargs = __pyx_state[4]; __pyx_result.message_id = __pyx_state[5]; __pyx_result.message_type = __pyx_state[6]; __pyx_result.to_index = __pyx_state[7]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->actor_cls);
  __Pyx_DECREF(__pyx_v___pyx_result->actor_cls);
  __pyx_v___pyx_result->actor_cls = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_4mars_6actors_4core_ActorRef))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->actor_ref));
  __pyx_v___pyx_result->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->args);
  __Pyx_DECREF(__pyx_v___pyx_result->args);
  __pyx_v___pyx_result->args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->from_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->kwargs);
  __Pyx_DECREF(__pyx_v___pyx_result->kwargs);
  __pyx_v___pyx_result->kwargs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->message_id);
  __Pyx_DECREF(__pyx_v___pyx_result->message_id);
  __pyx_v___pyx_result->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->message_type = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->to_index = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(CREATE_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_cls = __pyx_state[0]; __pyx_result.actor_ref = __pyx_state[1]; __pyx_result.args = __pyx_state[2]; __pyx_result.from_index = __pyx_state[3]; __pyx_result.kwargs = __pyx_state[4]; __pyx_result.message_id = __pyx_state[5]; __pyx_result.message_type = __pyx_state[6]; __pyx_result.to_index = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 8) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.actor_cls = __pyx_state[0]; __pyx_result.actor_ref = __pyx_state[1]; __pyx_result.args = __pyx_state[2]; __pyx_result.from_index = __pyx_state[3]; __pyx_result.kwargs = __pyx_state[4]; __pyx_result.message_id = __pyx_state[5]; __pyx_result.message_type = __pyx_state[6]; __pyx_result.to_index = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(CREATE_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_cls = __pyx_state[0]; __pyx_result.actor_ref = __pyx_state[1]; __pyx_result.args = __pyx_state[2]; __pyx_result.from_index = __pyx_state[3]; __pyx_result.kwargs = __pyx_state[4]; __pyx_result.message_id = __pyx_state[5]; __pyx_result.message_type = __pyx_state[6]; __pyx_result.to_index = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(<CREATE_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(CREATE_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.actor_cls = __pyx_state[0]; __pyx_result.actor_ref = __pyx_state[1]; __pyx_result.args = __pyx_state[2]; __pyx_result.from_index = __pyx_state[3]; __pyx_result.kwargs = __pyx_state[4]; __pyx_result.message_id = __pyx_state[5]; __pyx_result.message_type = __pyx_state[6]; __pyx_result.to_index = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_DESTROY_ACTOR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_49__pyx_unpickle_DESTROY_ACTOR_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4mars_6actors_4pool_8messages_49__pyx_unpickle_DESTROY_ACTOR_MESSAGE = {"__pyx_unpickle_DESTROY_ACTOR_MESSAGE", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_49__pyx_unpickle_DESTROY_ACTOR_MESSAGE, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_49__pyx_unpickle_DESTROY_ACTOR_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_DESTROY_ACTOR_MESSAGE (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DESTROY_ACTOR_MESSAGE", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DESTROY_ACTOR_MESSAGE", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_DESTROY_ACTOR_MESSAGE") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DESTROY_ACTOR_MESSAGE", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_DESTROY_ACTOR_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_48__pyx_unpickle_DESTROY_ACTOR_MESSAGE(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_48__pyx_unpickle_DESTROY_ACTOR_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DESTROY_ACTOR_MESSAGE", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x79e7d14:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x79e7d14) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x79e7d14:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = DESTROY_ACTOR_MESSAGE.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x79e7d14:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = DESTROY_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x79, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x79e7d14:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = DESTROY_ACTOR_MESSAGE.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(<DESTROY_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = DESTROY_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(<DESTROY_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = DESTROY_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(<DESTROY_ACTOR_MESSAGE> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(DESTROY_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = DESTROY_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(<DESTROY_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(<DESTROY_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(DESTROY_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_DESTROY_ACTOR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_DESTROY_ACTOR_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(<DESTROY_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(DESTROY_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(DESTROY_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_4mars_6actors_4core_ActorRef))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->actor_ref));
  __pyx_v___pyx_result->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->from_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->message_id);
  __Pyx_DECREF(__pyx_v___pyx_result->message_id);
  __pyx_v___pyx_result->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->message_type = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->to_index = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(DESTROY_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 5) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(DESTROY_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(<DESTROY_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state(DESTROY_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_DESTROY_ACTOR_MESSAGE__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_HAS_ACTOR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_51__pyx_unpickle_HAS_ACTOR_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4mars_6actors_4pool_8messages_51__pyx_unpickle_HAS_ACTOR_MESSAGE = {"__pyx_unpickle_HAS_ACTOR_MESSAGE", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_51__pyx_unpickle_HAS_ACTOR_MESSAGE, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_51__pyx_unpickle_HAS_ACTOR_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_HAS_ACTOR_MESSAGE (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_HAS_ACTOR_MESSAGE", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_HAS_ACTOR_MESSAGE", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_HAS_ACTOR_MESSAGE") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_HAS_ACTOR_MESSAGE", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_HAS_ACTOR_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_50__pyx_unpickle_HAS_ACTOR_MESSAGE(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_50__pyx_unpickle_HAS_ACTOR_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_HAS_ACTOR_MESSAGE", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x79e7d14:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x79e7d14) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x79e7d14:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = HAS_ACTOR_MESSAGE.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x79e7d14:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = HAS_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x79, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x79e7d14:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = HAS_ACTOR_MESSAGE.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(<HAS_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = HAS_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(<HAS_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = HAS_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(<HAS_ACTOR_MESSAGE> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(HAS_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x79e7d14 = (actor_ref, from_index, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = HAS_ACTOR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(<HAS_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(<HAS_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(HAS_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_HAS_ACTOR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_HAS_ACTOR_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(<HAS_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(HAS_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_HAS_ACTOR_MESSAGE__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(HAS_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_4mars_6actors_4core_ActorRef))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->actor_ref));
  __pyx_v___pyx_result->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->from_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->message_id);
  __Pyx_DECREF(__pyx_v___pyx_result->message_id);
  __pyx_v___pyx_result->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->message_type = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->to_index = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(HAS_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 5) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(HAS_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(<HAS_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(HAS_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_HAS_ACTOR_MESSAGE__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_RESULT_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_53__pyx_unpickle_RESULT_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4mars_6actors_4pool_8messages_53__pyx_unpickle_RESULT_MESSAGE = {"__pyx_unpickle_RESULT_MESSAGE", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_53__pyx_unpickle_RESULT_MESSAGE, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_53__pyx_unpickle_RESULT_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_RESULT_MESSAGE (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RESULT_MESSAGE", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RESULT_MESSAGE", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_RESULT_MESSAGE") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RESULT_MESSAGE", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_RESULT_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_52__pyx_unpickle_RESULT_MESSAGE(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_52__pyx_unpickle_RESULT_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_RESULT_MESSAGE", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xddcfada:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xddcfada = (from_index, message_id, message_type, result, to_index))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xddcfada) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xddcfada:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xddcfada = (from_index, message_id, message_type, result, to_index))" % __pyx_checksum)
 *     __pyx_result = RESULT_MESSAGE.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xddcfada:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xddcfada = (from_index, message_id, message_type, result, to_index))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = RESULT_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xdd, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xddcfada:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xddcfada = (from_index, message_id, message_type, result, to_index))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xddcfada = (from_index, message_id, message_type, result, to_index))" % __pyx_checksum)
 *     __pyx_result = RESULT_MESSAGE.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RESULT_MESSAGE__set_state(<RESULT_MESSAGE> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_RESULT_MESSAGE), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xddcfada = (from_index, message_id, message_type, result, to_index))" % __pyx_checksum)
 *     __pyx_result = RESULT_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RESULT_MESSAGE__set_state(<RESULT_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = RESULT_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RESULT_MESSAGE__set_state(<RESULT_MESSAGE> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_RESULT_MESSAGE__set_state(RESULT_MESSAGE __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_RESULT_MESSAGE__set_state(((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xddcfada = (from_index, message_id, message_type, result, to_index))" % __pyx_checksum)
 *     __pyx_result = RESULT_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RESULT_MESSAGE__set_state(<RESULT_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RESULT_MESSAGE__set_state(<RESULT_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_RESULT_MESSAGE__set_state(RESULT_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.from_index = __pyx_state[0]; __pyx_result.message_id = __pyx_state[1]; __pyx_result.message_type = __pyx_state[2]; __pyx_result.result = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_RESULT_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_RESULT_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_RESULT_MESSAGE__set_state(<RESULT_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RESULT_MESSAGE__set_state(RESULT_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.from_index = __pyx_state[0]; __pyx_result.message_id = __pyx_state[1]; __pyx_result.message_type = __pyx_state[2]; __pyx_result.result = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_RESULT_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_RESULT_MESSAGE__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_RESULT_MESSAGE__set_state(RESULT_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.from_index = __pyx_state[0]; __pyx_result.message_id = __pyx_state[1]; __pyx_result.message_type = __pyx_state[2]; __pyx_result.result = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->from_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->message_id);
  __Pyx_DECREF(__pyx_v___pyx_result->message_id);
  __pyx_v___pyx_result->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->message_type = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->result);
  __Pyx_DECREF(__pyx_v___pyx_result->result);
  __pyx_v___pyx_result->result = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->to_index = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_RESULT_MESSAGE__set_state(RESULT_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.from_index = __pyx_state[0]; __pyx_result.message_id = __pyx_state[1]; __pyx_result.message_type = __pyx_state[2]; __pyx_result.result = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 5) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.from_index = __pyx_state[0]; __pyx_result.message_id = __pyx_state[1]; __pyx_result.message_type = __pyx_state[2]; __pyx_result.result = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_RESULT_MESSAGE__set_state(RESULT_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.from_index = __pyx_state[0]; __pyx_result.message_id = __pyx_state[1]; __pyx_result.message_type = __pyx_state[2]; __pyx_result.result = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_RESULT_MESSAGE__set_state(<RESULT_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RESULT_MESSAGE__set_state(RESULT_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.from_index = __pyx_state[0]; __pyx_result.message_id = __pyx_state[1]; __pyx_result.message_type = __pyx_state[2]; __pyx_result.result = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_RESULT_MESSAGE__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ERROR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_55__pyx_unpickle_ERROR_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4mars_6actors_4pool_8messages_55__pyx_unpickle_ERROR_MESSAGE = {"__pyx_unpickle_ERROR_MESSAGE", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_55__pyx_unpickle_ERROR_MESSAGE, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_55__pyx_unpickle_ERROR_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ERROR_MESSAGE (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ERROR_MESSAGE", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ERROR_MESSAGE", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_ERROR_MESSAGE") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ERROR_MESSAGE", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_ERROR_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_54__pyx_unpickle_ERROR_MESSAGE(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_54__pyx_unpickle_ERROR_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ERROR_MESSAGE", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x42c8dac:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x42c8dac = (error, error_type, from_index, message_id, message_type, to_index, traceback))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x42c8dac) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x42c8dac:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x42c8dac = (error, error_type, from_index, message_id, message_type, to_index, traceback))" % __pyx_checksum)
 *     __pyx_result = ERROR_MESSAGE.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x42c8dac:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x42c8dac = (error, error_type, from_index, message_id, message_type, to_index, traceback))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = ERROR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x42, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x42c8dac:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x42c8dac = (error, error_type, from_index, message_id, message_type, to_index, traceback))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x42c8dac = (error, error_type, from_index, message_id, message_type, to_index, traceback))" % __pyx_checksum)
 *     __pyx_result = ERROR_MESSAGE.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ERROR_MESSAGE__set_state(<ERROR_MESSAGE> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_ERROR_MESSAGE), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x42c8dac = (error, error_type, from_index, message_id, message_type, to_index, traceback))" % __pyx_checksum)
 *     __pyx_result = ERROR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ERROR_MESSAGE__set_state(<ERROR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = ERROR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ERROR_MESSAGE__set_state(<ERROR_MESSAGE> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ERROR_MESSAGE__set_state(ERROR_MESSAGE __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_ERROR_MESSAGE__set_state(((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x42c8dac = (error, error_type, from_index, message_id, message_type, to_index, traceback))" % __pyx_checksum)
 *     __pyx_result = ERROR_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ERROR_MESSAGE__set_state(<ERROR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ERROR_MESSAGE__set_state(<ERROR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ERROR_MESSAGE__set_state(ERROR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.error = __pyx_state[0]; __pyx_result.error_type = __pyx_state[1]; __pyx_result.from_index = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]; __pyx_result.traceback = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ERROR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_ERROR_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ERROR_MESSAGE__set_state(<ERROR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ERROR_MESSAGE__set_state(ERROR_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.error = __pyx_state[0]; __pyx_result.error_type = __pyx_state[1]; __pyx_result.from_index = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]; __pyx_result.traceback = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_ERROR_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ERROR_MESSAGE__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ERROR_MESSAGE__set_state(ERROR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.error = __pyx_state[0]; __pyx_result.error_type = __pyx_state[1]; __pyx_result.from_index = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]; __pyx_result.traceback = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->error);
  __Pyx_DECREF(__pyx_v___pyx_result->error);
  __pyx_v___pyx_result->error = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->error_type);
  __Pyx_DECREF(__pyx_v___pyx_result->error_type);
  __pyx_v___pyx_result->error_type = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->from_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->message_id);
  __Pyx_DECREF(__pyx_v___pyx_result->message_id);
  __pyx_v___pyx_result->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->message_type = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->to_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->traceback);
  __Pyx_DECREF(__pyx_v___pyx_result->traceback);
  __pyx_v___pyx_result->traceback = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_ERROR_MESSAGE__set_state(ERROR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.error = __pyx_state[0]; __pyx_result.error_type = __pyx_state[1]; __pyx_result.from_index = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]; __pyx_result.traceback = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 7) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.error = __pyx_state[0]; __pyx_result.error_type = __pyx_state[1]; __pyx_result.from_index = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]; __pyx_result.traceback = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ERROR_MESSAGE__set_state(ERROR_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.error = __pyx_state[0]; __pyx_result.error_type = __pyx_state[1]; __pyx_result.from_index = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]; __pyx_result.traceback = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ERROR_MESSAGE__set_state(<ERROR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ERROR_MESSAGE__set_state(ERROR_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.error = __pyx_state[0]; __pyx_result.error_type = __pyx_state[1]; __pyx_result.from_index = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]; __pyx_result.traceback = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_ERROR_MESSAGE__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_SEND_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_57__pyx_unpickle_SEND_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4mars_6actors_4pool_8messages_57__pyx_unpickle_SEND_MESSAGE = {"__pyx_unpickle_SEND_MESSAGE", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_57__pyx_unpickle_SEND_MESSAGE, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4mars_6actors_4pool_8messages_57__pyx_unpickle_SEND_MESSAGE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_SEND_MESSAGE (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SEND_MESSAGE", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SEND_MESSAGE", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_SEND_MESSAGE") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SEND_MESSAGE", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_SEND_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4mars_6actors_4pool_8messages_56__pyx_unpickle_SEND_MESSAGE(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4mars_6actors_4pool_8messages_56__pyx_unpickle_SEND_MESSAGE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_SEND_MESSAGE", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x041d35d:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x041d35d = (actor_ref, from_index, message, message_id, message_type, to_index))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x041d35d) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x041d35d:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x041d35d = (actor_ref, from_index, message, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = SEND_MESSAGE.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x041d35d:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x041d35d = (actor_ref, from_index, message, message_id, message_type, to_index))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = SEND_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x04, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x041d35d:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x041d35d = (actor_ref, from_index, message, message_id, message_type, to_index))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x041d35d = (actor_ref, from_index, message, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = SEND_MESSAGE.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SEND_MESSAGE__set_state(<SEND_MESSAGE> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_4mars_6actors_4pool_8messages_SEND_MESSAGE), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x041d35d = (actor_ref, from_index, message, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = SEND_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SEND_MESSAGE__set_state(<SEND_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = SEND_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SEND_MESSAGE__set_state(<SEND_MESSAGE> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_SEND_MESSAGE__set_state(SEND_MESSAGE __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_SEND_MESSAGE__set_state(((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x041d35d = (actor_ref, from_index, message, message_id, message_type, to_index))" % __pyx_checksum)
 *     __pyx_result = SEND_MESSAGE.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SEND_MESSAGE__set_state(<SEND_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SEND_MESSAGE__set_state(<SEND_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_SEND_MESSAGE__set_state(SEND_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_SEND_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_SEND_MESSAGE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_SEND_MESSAGE__set_state(<SEND_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SEND_MESSAGE__set_state(SEND_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_4mars_6actors_4pool_8messages___pyx_unpickle_SEND_MESSAGE__set_state(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_4mars_6actors_4pool_8messages_INT32_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_SEND_MESSAGE__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_SEND_MESSAGE__set_state(SEND_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_4mars_6actors_4core_ActorRef))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->actor_ref);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->actor_ref));
  __pyx_v___pyx_result->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->from_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->message);
  __Pyx_DECREF(__pyx_v___pyx_result->message);
  __pyx_v___pyx_result->message = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->message_id);
  __Pyx_DECREF(__pyx_v___pyx_result->message_id);
  __pyx_v___pyx_result->message_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->message_type);
  __Pyx_DECREF(__pyx_v___pyx_result->message_type);
  __pyx_v___pyx_result->message_type = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->to_index = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_SEND_MESSAGE__set_state(SEND_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 6) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_SEND_MESSAGE__set_state(SEND_MESSAGE __pyx_result, tuple __pyx_state):
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_SEND_MESSAGE__set_state(<SEND_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SEND_MESSAGE__set_state(SEND_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("mars.actors.pool.messages.__pyx_unpickle_SEND_MESSAGE__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyArray_Descr *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":265
 * 
 *             cdef int i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *             cdef int i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":271
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 272, __pyx_L1_error)

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":275
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 276, __pyx_L1_error)

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":279
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * 2) * ((size_t)__pyx_v_ndim))));

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":286
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":287
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L9;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":289
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L9:;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":291
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":296
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":297
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_7 = PyArray_DESCR(__pyx_v_self);
  __pyx_t_3 = ((PyObject *)__pyx_t_7);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":300
 *             cdef int offset
 * 
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not PyDataType_HASFIELDS(descr):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(PyDataType_HASFIELDS(__pyx_v_descr) != 0)) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":303
 * 
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L15_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_L15_next_or:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":305
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_1)) {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 306, __pyx_L1_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":307
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;
      case NPY_UBYTE:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":308
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      __pyx_v_f = ((char *)"B");
      break;
      case NPY_SHORT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":309
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      __pyx_v_f = ((char *)"h");
      break;
      case NPY_USHORT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":310
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      __pyx_v_f = ((char *)"H");
      break;
      case NPY_INT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":311
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      __pyx_v_f = ((char *)"i");
      break;
      case NPY_UINT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":312
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      __pyx_v_f = ((char *)"I");
      break;
      case NPY_LONG:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":313
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      __pyx_v_f = ((char *)"l");
      break;
      case NPY_ULONG:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":314
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      __pyx_v_f = ((char *)"L");
      break;
      case NPY_LONGLONG:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":315
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      __pyx_v_f = ((char *)"q");
      break;
      case NPY_ULONGLONG:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":316
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      __pyx_v_f = ((char *)"Q");
      break;
      case NPY_FLOAT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":317
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      __pyx_v_f = ((char *)"f");
      break;
      case NPY_DOUBLE:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":318
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      __pyx_v_f = ((char *)"d");
      break;
      case NPY_LONGDOUBLE:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":319
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      __pyx_v_f = ((char *)"g");
      break;
      case NPY_CFLOAT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":320
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      __pyx_v_f = ((char *)"Zf");
      break;
      case NPY_CDOUBLE:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":321
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      __pyx_v_f = ((char *)"Zd");
      break;
      case NPY_CLONGDOUBLE:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":322
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      __pyx_v_f = ((char *)"Zg");
      break;
      case NPY_OBJECT:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":323
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":325
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 325, __pyx_L1_error)
      break;
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":326
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":327
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":329
 *                 return
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)PyObject_Malloc(0xFF));

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":330
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":331
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":332
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(2, 332, __pyx_L1_error)
    __pyx_v_f = __pyx_t_9;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":335
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":339
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)
 */
    PyObject_Free(__pyx_v_info->format);

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":341
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    PyObject_Free(__pyx_v_info->strides);

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":822
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":825
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":828
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":831
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":834
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":838
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":847
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":848
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(2, 851, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 851, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":852
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 852, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(2, 852, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":853
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(2, 853, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 853, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(2, 853, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 856, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 856, __pyx_L1_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":859
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_6)) {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":860
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 860, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 860, __pyx_L1_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":870
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":871
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":872
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":873
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":875
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":878
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (unlikely(__pyx_t_6)) {

        /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 880, __pyx_L1_error)

        /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":883
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":884
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":885
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":886
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":887
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":888
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":889
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":890
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":891
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":892
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":893
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":894
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":895
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":896
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":897
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":898
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":899
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(__pyx_t_6)) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":901
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 901, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":902
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":906
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(2, 906, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":907
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1023
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1024
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1029
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1030
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1036
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1036, __pyx_L3_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1037, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1038, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1038, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1042
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1042, __pyx_L3_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1044, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1044, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1048
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1048, __pyx_L3_error)

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1049
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1050
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1050, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1050, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":16
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):             # <<<<<<<<<<<<<<
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 */

/* Python wrapper */
static int __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_parents = 0;
  PyObject *__pyx_v_dct = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_parents,&__pyx_n_s_dct,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parents)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(1, 16, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dct)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(1, 16, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_name = values[0];
    __pyx_v_parents = values[1];
    __pyx_v_dct = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name, PyObject *__pyx_v_parents, PyObject *__pyx_v_dct) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "EnumBase":17
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)             # <<<<<<<<<<<<<<
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyType_Type)), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, ((PyObject *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, ((PyObject *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 17, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_cls));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_cls));
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, ((PyObject *)__pyx_v_cls));
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_parents);
    __Pyx_GIVEREF(__pyx_v_parents);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_parents);
    __Pyx_INCREF(__pyx_v_dct);
    __Pyx_GIVEREF(__pyx_v_dct);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_4, __pyx_v_dct);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":18
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()             # <<<<<<<<<<<<<<
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 */
  __Pyx_INCREF(__Pyx_OrderedDict);
  __pyx_t_2 = __Pyx_OrderedDict; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members, __pyx_t_1) < 0) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":16
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):             # <<<<<<<<<<<<<<
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":19
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):             # <<<<<<<<<<<<<<
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__(PyObject *__pyx_v_cls); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__(PyObject *__pyx_v_cls) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "EnumBase":20
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())             # <<<<<<<<<<<<<<
 *     def __getitem__(cls, name):
 *         return cls.__members__[name]
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "EnumBase":19
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):             # <<<<<<<<<<<<<<
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":21
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):             # <<<<<<<<<<<<<<
 *         return cls.__members__[name]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__(PyObject *__pyx_v_cls, PyObject *__pyx_v_name); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__(PyObject *__pyx_v_cls, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls), ((PyObject *)__pyx_v_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "EnumBase":22
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 *         return cls.__members__[name]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "EnumBase":21
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):             # <<<<<<<<<<<<<<
 *         return cls.__members__[name]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle___Pyx_EnumMeta__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_1__new__ = {"__new__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__new__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_value,&__pyx_n_s_name,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__new__", 0, 2, 3, 1); __PYX_ERR(1, 28, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__new__") < 0)) __PYX_ERR(1, 28, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cls = values[0];
    __pyx_v_value = values[1];
    __pyx_v_name = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__new__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 28, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(__pyx_self, __pyx_v_cls, __pyx_v_value, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_value, PyObject *__pyx_v_name) {
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("__new__", 0);

  /* "EnumBase":29
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 *         for v in cls:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return v
 */
  if (likely(PyList_CheckExact(__pyx_v_cls)) || PyTuple_CheckExact(__pyx_v_cls)) {
    __pyx_t_1 = __pyx_v_cls; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_cls); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 29, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 29, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 29, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 29, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "EnumBase":30
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return v
 *         if name is None:
 */
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_v, __pyx_v_value, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 30, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "EnumBase":31
 *         for v in cls:
 *             if v == value:
 *                 return v             # <<<<<<<<<<<<<<
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_v);
      __pyx_r = __pyx_v_v;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "EnumBase":30
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return v
 *         if name is None:
 */
    }

    /* "EnumBase":29
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 *         for v in cls:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":32
 *             if v == value:
 *                 return v
 *         if name is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 */
  __pyx_t_5 = (__pyx_v_name == Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (unlikely(__pyx_t_6)) {

    /* "EnumBase":33
 *                 return v
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)             # <<<<<<<<<<<<<<
 *         res = int.__new__(cls, value)
 *         res.name = name
 */
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Unknown_enum_value_s, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 33, __pyx_L1_error)

    /* "EnumBase":32
 *             if v == value:
 *                 return v
 *         if name is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 */
  }

  /* "EnumBase":34
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)             # <<<<<<<<<<<<<<
 *         res.name = name
 *         setattr(cls, name, res)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyInt_Type)), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_cls, __pyx_v_value};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_cls, __pyx_v_value};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_cls);
    __Pyx_GIVEREF(__pyx_v_cls);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_cls);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_value);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_res = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "EnumBase":35
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 *         res.name = name             # <<<<<<<<<<<<<<
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_res, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(1, 35, __pyx_L1_error)

  /* "EnumBase":36
 *         res = int.__new__(cls, value)
 *         res.name = name
 *         setattr(cls, name, res)             # <<<<<<<<<<<<<<
 *         cls.__members__[name] = res
 *         return res
 */
  __pyx_t_10 = PyObject_SetAttr(__pyx_v_cls, __pyx_v_name, __pyx_v_res); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(1, 36, __pyx_L1_error)

  /* "EnumBase":37
 *         res.name = name
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res             # <<<<<<<<<<<<<<
 *         return res
 *     def __repr__(self):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_members); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(PyObject_SetItem(__pyx_t_4, __pyx_v_name, __pyx_v_res) < 0)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumBase":38
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res
 *         return res             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_3__repr__ = {"__repr__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__, METH_O, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "EnumBase":40
 *         return res
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)             # <<<<<<<<<<<<<<
 *     def __str__(self):
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s_d, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_5__str__ = {"__str__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__, METH_O, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "EnumBase":42
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 *         return "%s.%s" % (self.__class__.__name__, self.name)             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta = {"__pyx_unpickle___Pyx_EnumMeta", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle___Pyx_EnumMeta") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype___Pyx_EnumMeta), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_unpickle___Pyx_EnumMeta__set_state(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_unpickle___Pyx_EnumMeta__set_state(struct __pyx_obj___Pyx_EnumMeta *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)o);
  p->message_id = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None); Py_INCREF(Py_None);
  p->actor_cls = Py_None; Py_INCREF(Py_None);
  p->args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->kwargs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE(PyObject *o) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->message_id);
  Py_CLEAR(p->actor_ref);
  Py_CLEAR(p->actor_cls);
  Py_CLEAR(p->args);
  Py_CLEAR(p->kwargs);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)o;
  if (p->actor_ref) {
    e = (*v)(((PyObject *)p->actor_ref), a); if (e) return e;
  }
  if (p->actor_cls) {
    e = (*v)(p->actor_cls, a); if (e) return e;
  }
  if (p->args) {
    e = (*v)(p->args, a); if (e) return e;
  }
  if (p->kwargs) {
    e = (*v)(p->kwargs, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE *)o;
  tmp = ((PyObject*)p->actor_ref);
  p->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->actor_cls);
  p->actor_cls = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->args);
  p->args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->kwargs);
  p->kwargs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_message_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_message_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_12message_type_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_message_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_message_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10message_id_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_from_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_from_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_10from_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_to_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_to_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_8to_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_actor_ref(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_actor_ref(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_ref_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_actor_cls(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_actor_cls(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_9actor_cls_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_args(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_args(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_4args_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_kwargs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_kwargs(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_6kwargs_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE[] = {
  {(char *)"message_type", __pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_message_type, __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_message_type, (char *)0, 0},
  {(char *)"message_id", __pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_message_id, __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_message_id, (char *)0, 0},
  {(char *)"from_index", __pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_from_index, __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_from_index, (char *)0, 0},
  {(char *)"to_index", __pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_to_index, __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_to_index, (char *)0, 0},
  {(char *)"actor_ref", __pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_actor_ref, __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_actor_ref, (char *)0, 0},
  {(char *)"actor_cls", __pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_actor_cls, __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_actor_cls, (char *)0, 0},
  {(char *)"args", __pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_args, __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_args, (char *)0, 0},
  {(char *)"kwargs", __pyx_getprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_kwargs, __pyx_setprop_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_kwargs, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.actors.pool.messages.CREATE_ACTOR_MESSAGE", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE, /*tp_traverse*/
  __pyx_tp_clear_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_4mars_6actors_4pool_8messages_20CREATE_ACTOR_MESSAGE_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)o);
  p->message_id = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE(PyObject *o) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->message_id);
  Py_CLEAR(p->actor_ref);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)o;
  if (p->actor_ref) {
    e = (*v)(((PyObject *)p->actor_ref), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE *)o;
  tmp = ((PyObject*)p->actor_ref);
  p->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_message_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_message_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_12message_type_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_message_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_message_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10message_id_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_from_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_from_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_10from_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_to_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_to_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_8to_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_actor_ref(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_actor_ref(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_9actor_ref_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE[] = {
  {(char *)"message_type", __pyx_getprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_message_type, __pyx_setprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_message_type, (char *)0, 0},
  {(char *)"message_id", __pyx_getprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_message_id, __pyx_setprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_message_id, (char *)0, 0},
  {(char *)"from_index", __pyx_getprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_from_index, __pyx_setprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_from_index, (char *)0, 0},
  {(char *)"to_index", __pyx_getprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_to_index, __pyx_setprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_to_index, (char *)0, 0},
  {(char *)"actor_ref", __pyx_getprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_actor_ref, __pyx_setprop_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_actor_ref, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.actors.pool.messages.DESTROY_ACTOR_MESSAGE", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE, /*tp_traverse*/
  __pyx_tp_clear_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_4mars_6actors_4pool_8messages_21DESTROY_ACTOR_MESSAGE_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)o);
  p->message_id = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE(PyObject *o) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->message_id);
  Py_CLEAR(p->actor_ref);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)o;
  if (p->actor_ref) {
    e = (*v)(((PyObject *)p->actor_ref), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE *)o;
  tmp = ((PyObject*)p->actor_ref);
  p->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_message_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_message_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_12message_type_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_message_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_message_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10message_id_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_from_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_from_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_10from_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_to_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_to_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_8to_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_actor_ref(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_actor_ref(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_9actor_ref_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE[] = {
  {(char *)"message_type", __pyx_getprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_message_type, __pyx_setprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_message_type, (char *)0, 0},
  {(char *)"message_id", __pyx_getprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_message_id, __pyx_setprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_message_id, (char *)0, 0},
  {(char *)"from_index", __pyx_getprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_from_index, __pyx_setprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_from_index, (char *)0, 0},
  {(char *)"to_index", __pyx_getprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_to_index, __pyx_setprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_to_index, (char *)0, 0},
  {(char *)"actor_ref", __pyx_getprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_actor_ref, __pyx_setprop_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_actor_ref, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.actors.pool.messages.HAS_ACTOR_MESSAGE", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE, /*tp_traverse*/
  __pyx_tp_clear_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_4mars_6actors_4pool_8messages_17HAS_ACTOR_MESSAGE_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_RESULT_MESSAGE(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)o);
  p->message_id = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->result = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4mars_6actors_4pool_8messages_RESULT_MESSAGE(PyObject *o) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->message_id);
  Py_CLEAR(p->result);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4mars_6actors_4pool_8messages_RESULT_MESSAGE(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)o;
  if (p->result) {
    e = (*v)(p->result, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_6actors_4pool_8messages_RESULT_MESSAGE(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE *)o;
  tmp = ((PyObject*)p->result);
  p->result = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_message_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_message_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_12message_type_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_message_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_message_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10message_id_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_from_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_from_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_10from_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_to_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_to_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_8to_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_result(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_result(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_6result_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_4mars_6actors_4pool_8messages_RESULT_MESSAGE[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4mars_6actors_4pool_8messages_RESULT_MESSAGE[] = {
  {(char *)"message_type", __pyx_getprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_message_type, __pyx_setprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_message_type, (char *)0, 0},
  {(char *)"message_id", __pyx_getprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_message_id, __pyx_setprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_message_id, (char *)0, 0},
  {(char *)"from_index", __pyx_getprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_from_index, __pyx_setprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_from_index, (char *)0, 0},
  {(char *)"to_index", __pyx_getprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_to_index, __pyx_setprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_to_index, (char *)0, 0},
  {(char *)"result", __pyx_getprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_result, __pyx_setprop_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_result, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4mars_6actors_4pool_8messages_RESULT_MESSAGE = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.actors.pool.messages.RESULT_MESSAGE", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages_RESULT_MESSAGE), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_6actors_4pool_8messages_RESULT_MESSAGE, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_6actors_4pool_8messages_RESULT_MESSAGE, /*tp_traverse*/
  __pyx_tp_clear_4mars_6actors_4pool_8messages_RESULT_MESSAGE, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4mars_6actors_4pool_8messages_RESULT_MESSAGE, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4mars_6actors_4pool_8messages_RESULT_MESSAGE, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_4mars_6actors_4pool_8messages_14RESULT_MESSAGE_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_6actors_4pool_8messages_RESULT_MESSAGE, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_ERROR_MESSAGE(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)o);
  p->message_id = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->error_type = Py_None; Py_INCREF(Py_None);
  p->error = Py_None; Py_INCREF(Py_None);
  p->traceback = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4mars_6actors_4pool_8messages_ERROR_MESSAGE(PyObject *o) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->message_id);
  Py_CLEAR(p->error_type);
  Py_CLEAR(p->error);
  Py_CLEAR(p->traceback);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4mars_6actors_4pool_8messages_ERROR_MESSAGE(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)o;
  if (p->error_type) {
    e = (*v)(p->error_type, a); if (e) return e;
  }
  if (p->error) {
    e = (*v)(p->error, a); if (e) return e;
  }
  if (p->traceback) {
    e = (*v)(p->traceback, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_6actors_4pool_8messages_ERROR_MESSAGE(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE *)o;
  tmp = ((PyObject*)p->error_type);
  p->error_type = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->error);
  p->error = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->traceback);
  p->traceback = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_message_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_message_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_12message_type_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_message_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_message_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10message_id_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_from_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_from_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10from_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_to_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_to_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_8to_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_error_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_error_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_10error_type_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_error(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_error(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5error_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_traceback(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_traceback(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_9traceback_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_4mars_6actors_4pool_8messages_ERROR_MESSAGE[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4mars_6actors_4pool_8messages_ERROR_MESSAGE[] = {
  {(char *)"message_type", __pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_message_type, __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_message_type, (char *)0, 0},
  {(char *)"message_id", __pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_message_id, __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_message_id, (char *)0, 0},
  {(char *)"from_index", __pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_from_index, __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_from_index, (char *)0, 0},
  {(char *)"to_index", __pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_to_index, __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_to_index, (char *)0, 0},
  {(char *)"error_type", __pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_error_type, __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_error_type, (char *)0, 0},
  {(char *)"error", __pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_error, __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_error, (char *)0, 0},
  {(char *)"traceback", __pyx_getprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_traceback, __pyx_setprop_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_traceback, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4mars_6actors_4pool_8messages_ERROR_MESSAGE = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.actors.pool.messages.ERROR_MESSAGE", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages_ERROR_MESSAGE), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_6actors_4pool_8messages_ERROR_MESSAGE, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_6actors_4pool_8messages_ERROR_MESSAGE, /*tp_traverse*/
  __pyx_tp_clear_4mars_6actors_4pool_8messages_ERROR_MESSAGE, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4mars_6actors_4pool_8messages_ERROR_MESSAGE, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4mars_6actors_4pool_8messages_ERROR_MESSAGE, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_4mars_6actors_4pool_8messages_13ERROR_MESSAGE_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_6actors_4pool_8messages_ERROR_MESSAGE, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages_SEND_MESSAGE(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)o);
  p->message_type = Py_None; Py_INCREF(Py_None);
  p->message_id = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None); Py_INCREF(Py_None);
  p->message = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_4mars_6actors_4pool_8messages_SEND_MESSAGE(PyObject *o) {
  struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->message_type);
  Py_CLEAR(p->message_id);
  Py_CLEAR(p->actor_ref);
  Py_CLEAR(p->message);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4mars_6actors_4pool_8messages_SEND_MESSAGE(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)o;
  if (p->message_type) {
    e = (*v)(p->message_type, a); if (e) return e;
  }
  if (p->actor_ref) {
    e = (*v)(((PyObject *)p->actor_ref), a); if (e) return e;
  }
  if (p->message) {
    e = (*v)(p->message, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_6actors_4pool_8messages_SEND_MESSAGE(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *p = (struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE *)o;
  tmp = ((PyObject*)p->message_type);
  p->message_type = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->actor_ref);
  p->actor_ref = ((struct __pyx_obj_4mars_6actors_4core_ActorRef *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->message);
  p->message = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_12message_type_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10message_id_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_from_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_from_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_10from_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_to_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_to_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_8to_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_actor_ref(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_actor_ref(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_9actor_ref_5__del__(o);
  }
}

static PyObject *__pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_1__get__(o);
}

static int __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_3__set__(o, v);
  }
  else {
    return __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_7message_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_4mars_6actors_4pool_8messages_SEND_MESSAGE[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4mars_6actors_4pool_8messages_SEND_MESSAGE[] = {
  {(char *)"message_type", __pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message_type, __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message_type, (char *)0, 0},
  {(char *)"message_id", __pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message_id, __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message_id, (char *)0, 0},
  {(char *)"from_index", __pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_from_index, __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_from_index, (char *)0, 0},
  {(char *)"to_index", __pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_to_index, __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_to_index, (char *)0, 0},
  {(char *)"actor_ref", __pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_actor_ref, __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_actor_ref, (char *)0, 0},
  {(char *)"message", __pyx_getprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message, __pyx_setprop_4mars_6actors_4pool_8messages_12SEND_MESSAGE_message, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4mars_6actors_4pool_8messages_SEND_MESSAGE = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.actors.pool.messages.SEND_MESSAGE", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages_SEND_MESSAGE), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_6actors_4pool_8messages_SEND_MESSAGE, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_6actors_4pool_8messages_SEND_MESSAGE, /*tp_traverse*/
  __pyx_tp_clear_4mars_6actors_4pool_8messages_SEND_MESSAGE, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4mars_6actors_4pool_8messages_SEND_MESSAGE, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4mars_6actors_4pool_8messages_SEND_MESSAGE, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_4mars_6actors_4pool_8messages_12SEND_MESSAGE_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_6actors_4pool_8messages_SEND_MESSAGE, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message *__pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message[8];
static int __pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message = 0;

static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message)))) {
    o = (PyObject*)__pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message[--__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message(PyObject *o) {
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message *p = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message *)o;
  Py_CLEAR(p->__pyx_v_buf);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message)))) {
    __pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message[__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message++] = ((struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.actors.pool.messages.__pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr *__pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr[--__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr[__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.actors.pool.messages.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *__pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message[8];
static int __pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message = 0;

static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message)))) {
    o = (PyObject*)__pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message[--__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message(PyObject *o) {
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *p = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_binary);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message)))) {
    __pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message[__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message++] = ((struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *p = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *)o;
  if (p->__pyx_v_binary) {
    e = (*v)(p->__pyx_v_binary, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *p = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message *)o;
  tmp = ((PyObject*)p->__pyx_v_binary);
  p->__pyx_v_binary = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.actors.pool.messages.__pyx_scope_struct_2_write_remote_message", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message, /*tp_traverse*/
  __pyx_tp_clear_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr *__pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr[8];
static int __pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr = 0;

static PyObject *__pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr)))) {
    o = (PyObject*)__pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr[--__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr(PyObject *o) {
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_b);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr)))) {
    __pyx_freelist_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr[__pyx_freecount_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr++] = ((struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_b) {
    e = (*v)(p->__pyx_v_b, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.actors.pool.messages.__pyx_scope_struct_3_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new___Pyx_EnumMeta(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = (&PyType_Type)->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc___Pyx_EnumMeta(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  PyObject_GC_Track(o);
  (&PyType_Type)->tp_dealloc(o);
}

static int __pyx_tp_traverse___Pyx_EnumMeta(PyObject *o, visitproc v, void *a) {
  int e;
  if (!(&PyType_Type)->tp_traverse); else { e = (&PyType_Type)->tp_traverse(o,v,a); if (e) return e; }
  return 0;
}

static int __pyx_tp_clear___Pyx_EnumMeta(PyObject *o) {
  if (!(&PyType_Type)->tp_clear); else (&PyType_Type)->tp_clear(o);
  return 0;
}
static PyObject *__pyx_sq_item___Pyx_EnumMeta(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods___Pyx_EnumMeta[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence___Pyx_EnumMeta = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item___Pyx_EnumMeta, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping___Pyx_EnumMeta = {
  0, /*mp_length*/
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __Pyx_EnumMeta = {
  PyVarObject_HEAD_INIT(0, 0)
  "mars.actors.pool.messages.__Pyx_EnumMeta", /*tp_name*/
  sizeof(struct __pyx_obj___Pyx_EnumMeta), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___Pyx_EnumMeta, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence___Pyx_EnumMeta, /*tp_as_sequence*/
  &__pyx_tp_as_mapping___Pyx_EnumMeta, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse___Pyx_EnumMeta, /*tp_traverse*/
  __pyx_tp_clear___Pyx_EnumMeta, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods___Pyx_EnumMeta, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___Pyx_EnumMeta, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"unpack_message_type_value", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_1unpack_message_type_value, METH_O, 0},
  {"unpack_message_type", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_3unpack_message_type, METH_O, 0},
  {"unpack_message_id", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_5unpack_message_id, METH_O, 0},
  {"pack_actor_ref", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_7pack_actor_ref, METH_O, 0},
  {"unpack_actor_ref", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_9unpack_actor_ref, METH_VARARGS|METH_KEYWORDS, 0},
  {"pack_send_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_11pack_send_message, METH_VARARGS|METH_KEYWORDS, 0},
  {"unpack_send_message", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_13unpack_send_message, METH_O, 0},
  {"pack_tell_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_15pack_tell_message, METH_VARARGS|METH_KEYWORDS, 0},
  {"unpack_tell_message", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_17unpack_tell_message, METH_O, 0},
  {"get_index", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_19get_index, METH_VARARGS|METH_KEYWORDS, 0},
  {"pack_create_actor_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_21pack_create_actor_message, METH_VARARGS|METH_KEYWORDS, 0},
  {"unpack_create_actor_message", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_23unpack_create_actor_message, METH_O, 0},
  {"pack_destroy_actor_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_25pack_destroy_actor_message, METH_VARARGS|METH_KEYWORDS, 0},
  {"unpack_destroy_actor_message", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_27unpack_destroy_actor_message, METH_O, 0},
  {"pack_has_actor_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_29pack_has_actor_message, METH_VARARGS|METH_KEYWORDS, 0},
  {"unpack_has_actor_message", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_31unpack_has_actor_message, METH_O, 0},
  {"pack_result_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_33pack_result_message, METH_VARARGS|METH_KEYWORDS, 0},
  {"unpack_result_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_35unpack_result_message, METH_VARARGS|METH_KEYWORDS, 0},
  {"pack_error_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4mars_6actors_4pool_8messages_37pack_error_message, METH_VARARGS|METH_KEYWORDS, 0},
  {"unpack_error_message", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_39unpack_error_message, METH_O, 0},
  {"read_remote_message", (PyCFunction)__pyx_pw_4mars_6actors_4pool_8messages_41read_remote_message, METH_O, 0},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_messages(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_messages},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "messages",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_B, __pyx_k_B, sizeof(__pyx_k_B), 0, 0, 1, 0},
  {&__pyx_n_s_BrokenPipeError, __pyx_k_BrokenPipeError, sizeof(__pyx_k_BrokenPipeError), 0, 0, 1, 1},
  {&__pyx_n_s_CREATE_ACTOR_MESSAGE, __pyx_k_CREATE_ACTOR_MESSAGE, sizeof(__pyx_k_CREATE_ACTOR_MESSAGE), 0, 0, 1, 1},
  {&__pyx_n_s_ConnectionResetError, __pyx_k_ConnectionResetError, sizeof(__pyx_k_ConnectionResetError), 0, 0, 1, 1},
  {&__pyx_n_s_DESTROY_ACTOR_MESSAGE, __pyx_k_DESTROY_ACTOR_MESSAGE, sizeof(__pyx_k_DESTROY_ACTOR_MESSAGE), 0, 0, 1, 1},
  {&__pyx_n_s_EPIPE, __pyx_k_EPIPE, sizeof(__pyx_k_EPIPE), 0, 0, 1, 1},
  {&__pyx_n_s_ERROR_MESSAGE, __pyx_k_ERROR_MESSAGE, sizeof(__pyx_k_ERROR_MESSAGE), 0, 0, 1, 1},
  {&__pyx_n_s_EnumBase, __pyx_k_EnumBase, sizeof(__pyx_k_EnumBase), 0, 0, 1, 1},
  {&__pyx_n_s_EnumType, __pyx_k_EnumType, sizeof(__pyx_k_EnumType), 0, 0, 1, 1},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_n_s_HAS_ACTOR_MESSAGE, __pyx_k_HAS_ACTOR_MESSAGE, sizeof(__pyx_k_HAS_ACTOR_MESSAGE), 0, 0, 1, 1},
  {&__pyx_kp_s_I, __pyx_k_I, sizeof(__pyx_k_I), 0, 0, 1, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x04, __pyx_k_Incompatible_checksums_s_vs_0x04, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x04), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x2d, __pyx_k_Incompatible_checksums_s_vs_0x2d, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x2d), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x42, __pyx_k_Incompatible_checksums_s_vs_0x42, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x42), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x79, __pyx_k_Incompatible_checksums_s_vs_0x79, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x79), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_k_Incompatible_checksums_s_vs_0xd4, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xd4), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xdd, __pyx_k_Incompatible_checksums_s_vs_0xdd, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xdd), 0, 0, 1, 0},
  {&__pyx_n_s_IntEnum, __pyx_k_IntEnum, sizeof(__pyx_k_IntEnum), 0, 0, 1, 1},
  {&__pyx_n_s_MessageSerialType, __pyx_k_MessageSerialType, sizeof(__pyx_k_MessageSerialType), 0, 0, 1, 1},
  {&__pyx_n_s_MessageType, __pyx_k_MessageType, sizeof(__pyx_k_MessageType), 0, 0, 1, 1},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_OrderedDict, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase, __pyx_k_Pyx_EnumBase, sizeof(__pyx_k_Pyx_EnumBase), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___new, __pyx_k_Pyx_EnumBase___new, sizeof(__pyx_k_Pyx_EnumBase___new), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___repr, __pyx_k_Pyx_EnumBase___repr, sizeof(__pyx_k_Pyx_EnumBase___repr), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___str, __pyx_k_Pyx_EnumBase___str, sizeof(__pyx_k_Pyx_EnumBase___str), 0, 0, 1, 1},
  {&__pyx_kp_s_Q, __pyx_k_Q, sizeof(__pyx_k_Q), 0, 0, 1, 0},
  {&__pyx_n_s_RESULT_MESSAGE, __pyx_k_RESULT_MESSAGE, sizeof(__pyx_k_RESULT_MESSAGE), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_SEND_MESSAGE, __pyx_k_SEND_MESSAGE, sizeof(__pyx_k_SEND_MESSAGE), 0, 0, 1, 1},
  {&__pyx_n_s_Struct, __pyx_k_Struct, sizeof(__pyx_k_Struct), 0, 0, 1, 1},
  {&__pyx_kp_s_The_remote_server_is_closed, __pyx_k_The_remote_server_is_closed, sizeof(__pyx_k_The_remote_server_is_closed), 0, 0, 1, 0},
  {&__pyx_kp_s_Unknown_enum_value_s, __pyx_k_Unknown_enum_value_s, sizeof(__pyx_k_Unknown_enum_value_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Unsupported_protocol, __pyx_k_Unsupported_protocol, sizeof(__pyx_k_Unsupported_protocol), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_actor_cls, __pyx_k_actor_cls, sizeof(__pyx_k_actor_cls), 0, 0, 1, 1},
  {&__pyx_n_s_actor_ref, __pyx_k_actor_ref, sizeof(__pyx_k_actor_ref), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_n_s_binary, __pyx_k_binary, sizeof(__pyx_k_binary), 0, 0, 1, 1},
  {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_calc_from_uid, __pyx_k_calc_from_uid, sizeof(__pyx_k_calc_from_uid), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_compat, __pyx_k_compat, sizeof(__pyx_k_compat), 0, 0, 1, 1},
  {&__pyx_n_s_create_actor, __pyx_k_create_actor, sizeof(__pyx_k_create_actor), 0, 0, 1, 1},
  {&__pyx_n_s_dct, __pyx_k_dct, sizeof(__pyx_k_dct), 0, 0, 1, 1},
  {&__pyx_n_s_destroy_actor, __pyx_k_destroy_actor, sizeof(__pyx_k_destroy_actor), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dumps, __pyx_k_dumps, sizeof(__pyx_k_dumps), 0, 0, 1, 1},
  {&__pyx_n_s_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
  {&__pyx_n_s_errno, __pyx_k_errno, sizeof(__pyx_k_errno), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_error_type, __pyx_k_error_type, sizeof(__pyx_k_error_type), 0, 0, 1, 1},
  {&__pyx_n_s_ex, __pyx_k_ex, sizeof(__pyx_k_ex), 0, 0, 1, 1},
  {&__pyx_n_s_extend, __pyx_k_extend, sizeof(__pyx_k_extend), 0, 0, 1, 1},
  {&__pyx_n_s_from_index, __pyx_k_from_index, sizeof(__pyx_k_from_index), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 0, 1, 0},
  {&__pyx_n_s_has_actor, __pyx_k_has_actor, sizeof(__pyx_k_has_actor), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_install, __pyx_k_install, sizeof(__pyx_k_install), 0, 0, 1, 1},
  {&__pyx_n_s_kw, __pyx_k_kw, sizeof(__pyx_k_kw), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_lib_tblib, __pyx_k_lib_tblib, sizeof(__pyx_k_lib_tblib), 0, 0, 1, 1},
  {&__pyx_n_s_loads, __pyx_k_loads, sizeof(__pyx_k_loads), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mars_actors_pool_messages, __pyx_k_mars_actors_pool_messages, sizeof(__pyx_k_mars_actors_pool_messages), 0, 0, 1, 1},
  {&__pyx_kp_s_mars_actors_pool_messages_pyx, __pyx_k_mars_actors_pool_messages_pyx, sizeof(__pyx_k_mars_actors_pool_messages_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_members, __pyx_k_members, sizeof(__pyx_k_members), 0, 0, 1, 1},
  {&__pyx_n_s_message, __pyx_k_message, sizeof(__pyx_k_message), 0, 0, 1, 1},
  {&__pyx_n_s_message_id, __pyx_k_message_id, sizeof(__pyx_k_message_id), 0, 0, 1, 1},
  {&__pyx_n_s_message_type, __pyx_k_message_type, sizeof(__pyx_k_message_type), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_null, __pyx_k_null, sizeof(__pyx_k_null), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_parents, __pyx_k_parents, sizeof(__pyx_k_parents), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pickling_support, __pyx_k_pickling_support, sizeof(__pyx_k_pickling_support), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_protocol, __pyx_k_protocol, sizeof(__pyx_k_protocol), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_CREATE_ACTOR_MESS, __pyx_k_pyx_unpickle_CREATE_ACTOR_MESS, sizeof(__pyx_k_pyx_unpickle_CREATE_ACTOR_MESS), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_DESTROY_ACTOR_MES, __pyx_k_pyx_unpickle_DESTROY_ACTOR_MES, sizeof(__pyx_k_pyx_unpickle_DESTROY_ACTOR_MES), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_ERROR_MESSAGE, __pyx_k_pyx_unpickle_ERROR_MESSAGE, sizeof(__pyx_k_pyx_unpickle_ERROR_MESSAGE), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_HAS_ACTOR_MESSAGE, __pyx_k_pyx_unpickle_HAS_ACTOR_MESSAGE, sizeof(__pyx_k_pyx_unpickle_HAS_ACTOR_MESSAGE), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_RESULT_MESSAGE, __pyx_k_pyx_unpickle_RESULT_MESSAGE, sizeof(__pyx_k_pyx_unpickle_RESULT_MESSAGE), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_SEND_MESSAGE, __pyx_k_pyx_unpickle_SEND_MESSAGE, sizeof(__pyx_k_pyx_unpickle_SEND_MESSAGE), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle___Pyx_EnumMeta, __pyx_k_pyx_unpickle___Pyx_EnumMeta, sizeof(__pyx_k_pyx_unpickle___Pyx_EnumMeta), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_raw_bytes, __pyx_k_raw_bytes, sizeof(__pyx_k_raw_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_kp_s_s_s, __pyx_k_s_s, sizeof(__pyx_k_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_d, __pyx_k_s_s_d, sizeof(__pyx_k_s_s_d), 0, 0, 1, 0},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_send_all, __pyx_k_send_all, sizeof(__pyx_k_send_all), 0, 0, 1, 1},
  {&__pyx_n_s_send_chunk, __pyx_k_send_chunk, sizeof(__pyx_k_send_chunk), 0, 0, 1, 1},
  {&__pyx_n_s_send_chunk_end, __pyx_k_send_chunk_end, sizeof(__pyx_k_send_chunk_end), 0, 0, 1, 1},
  {&__pyx_n_s_send_chunk_start, __pyx_k_send_chunk_start, sizeof(__pyx_k_send_chunk_start), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_socket, __pyx_k_socket, sizeof(__pyx_k_socket), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_tb, __pyx_k_tb, sizeof(__pyx_k_tb), 0, 0, 1, 1},
  {&__pyx_n_s_tell_all, __pyx_k_tell_all, sizeof(__pyx_k_tell_all), 0, 0, 1, 1},
  {&__pyx_n_s_tell_chunk, __pyx_k_tell_chunk, sizeof(__pyx_k_tell_chunk), 0, 0, 1, 1},
  {&__pyx_n_s_tell_chunk_end, __pyx_k_tell_chunk_end, sizeof(__pyx_k_tell_chunk_end), 0, 0, 1, 1},
  {&__pyx_n_s_tell_chunk_start, __pyx_k_tell_chunk_start, sizeof(__pyx_k_tell_chunk_start), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_to_index, __pyx_k_to_index, sizeof(__pyx_k_to_index), 0, 0, 1, 1},
  {&__pyx_n_s_traceback, __pyx_k_traceback, sizeof(__pyx_k_traceback), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_unpack_tuple_message_locals_gen, __pyx_k_unpack_tuple_message_locals_gen, sizeof(__pyx_k_unpack_tuple_message_locals_gen), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_s_write_func, __pyx_k_write_func, sizeof(__pyx_k_write_func), 0, 0, 1, 1},
  {&__pyx_n_s_write_remote_message, __pyx_k_write_remote_message, sizeof(__pyx_k_write_remote_message), 0, 0, 1, 1},
  {&__pyx_n_s_write_remote_message_2, __pyx_k_write_remote_message_2, sizeof(__pyx_k_write_remote_message_2), 0, 0, 1, 1},
  {&__pyx_n_s_write_remote_message_locals_gene, __pyx_k_write_remote_message_locals_gene, sizeof(__pyx_k_write_remote_message_locals_gene), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 217, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 440, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 819, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(2, 272, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(2, 856, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 1038, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "mars/actors/pool/messages.pyx":217
 *     protocol = value >> 5
 *     if protocol != 0:
 *         raise NotImplementedError('Unsupported protocol')             # <<<<<<<<<<<<<<
 * 
 *     return value & ((1 << 5) - 1)
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_Unsupported_protocol); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(2, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(2, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(2, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(2, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(2, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(2, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "../../Python36-x64/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(2, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "mars/actors/pool/messages.pyx":176
 * 
 * 
 * cdef object _pack_byte = struct.Struct('<B').pack             # <<<<<<<<<<<<<<
 * cdef object _unpack_byte = struct.Struct('<B').unpack
 * cdef object _pack_int = struct.Struct('<I').pack
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_B); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "mars/actors/pool/messages.pyx":178
 * cdef object _pack_byte = struct.Struct('<B').pack
 * cdef object _unpack_byte = struct.Struct('<B').unpack
 * cdef object _pack_int = struct.Struct('<I').pack             # <<<<<<<<<<<<<<
 * cdef object _unpack_int = struct.Struct('<I').unpack
 * cdef object _pack_long = struct.Struct('<Q').pack
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_I); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "mars/actors/pool/messages.pyx":180
 * cdef object _pack_int = struct.Struct('<I').pack
 * cdef object _unpack_int = struct.Struct('<I').unpack
 * cdef object _pack_long = struct.Struct('<Q').pack             # <<<<<<<<<<<<<<
 * cdef object _unpack_long = struct.Struct('<Q').unpack
 * cdef object _pack_short = struct.Struct('<h').pack
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_Q); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "mars/actors/pool/messages.pyx":182
 * cdef object _pack_long = struct.Struct('<Q').pack
 * cdef object _unpack_long = struct.Struct('<Q').unpack
 * cdef object _pack_short = struct.Struct('<h').pack             # <<<<<<<<<<<<<<
 * cdef object _unpack_short = struct.Struct('<h').unpack
 * 
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_h); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "mars/actors/pool/messages.pyx":816
 * 
 * 
 * def write_remote_message(write_func, *binary):             # <<<<<<<<<<<<<<
 *     cdef bytes size
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(6, __pyx_n_s_write_func, __pyx_n_s_binary, __pyx_n_s_size, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_b); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_mars_actors_pool_messages_pyx, __pyx_n_s_write_remote_message_2, 816, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 816, __pyx_L1_error)

  /* "mars/actors/pool/messages.pyx":827
 *     _write_remote_message = write_remote_message
 * 
 *     def write_remote_message(*args, **kwargs):             # <<<<<<<<<<<<<<
 *         try:
 *             return _write_remote_message(*args, **kwargs)
 */
  __pyx_tuple__24 = PyTuple_Pack(3, __pyx_n_s_args, __pyx_n_s_kwargs, __pyx_n_s_ex); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_mars_actors_pool_messages_pyx, __pyx_n_s_write_remote_message_2, 827, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 827, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_CREATE_ACTOR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__26 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CREATE_ACTOR_MESS, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__28 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_DESTROY_ACTOR_MES, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__30 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_HAS_ACTOR_MESSAGE, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__32 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_RESULT_MESSAGE, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__34 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ERROR_MESSAGE, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__36 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_SEND_MESSAGE, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */
  __pyx_tuple__38 = PyTuple_Pack(5, __pyx_n_s_cls, __pyx_n_s_value, __pyx_n_s_name, __pyx_n_s_v, __pyx_n_s_res); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_new, 28, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(1, 28, __pyx_L1_error)
  __pyx_tuple__40 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_repr, 39, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(1, 39, __pyx_L1_error)

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_str, 41, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(1, 41, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__45 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyByteArray_Type_extend.type = (PyObject*)&PyByteArray_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_32 = PyInt_FromLong(32); if (unlikely(!__pyx_int_32)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4313949 = PyInt_FromLong(4313949L); if (unlikely(!__pyx_int_4313949)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_48008750 = PyInt_FromLong(48008750L); if (unlikely(!__pyx_int_48008750)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_70028716 = PyInt_FromLong(70028716L); if (unlikely(!__pyx_int_70028716)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_127827220 = PyInt_FromLong(127827220L); if (unlikely(!__pyx_int_127827220)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_232585946 = PyInt_FromLong(232585946L); if (unlikely(!__pyx_int_232585946)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_4mars_6actors_4pool_8messages__pack_byte = Py_None; Py_INCREF(Py_None);
  __pyx_v_4mars_6actors_4pool_8messages__unpack_byte = Py_None; Py_INCREF(Py_None);
  __pyx_v_4mars_6actors_4pool_8messages__pack_int = Py_None; Py_INCREF(Py_None);
  __pyx_v_4mars_6actors_4pool_8messages__unpack_int = Py_None; Py_INCREF(Py_None);
  __pyx_v_4mars_6actors_4pool_8messages__pack_long = Py_None; Py_INCREF(Py_None);
  __pyx_v_4mars_6actors_4pool_8messages__unpack_long = Py_None; Py_INCREF(Py_None);
  __pyx_v_4mars_6actors_4pool_8messages__pack_short = Py_None; Py_INCREF(Py_None);
  __pyx_v_4mars_6actors_4pool_8messages__unpack_short = Py_None; Py_INCREF(Py_None);
  __Pyx_OrderedDict = Py_None; Py_INCREF(Py_None);
  __Pyx_EnumBase = Py_None; Py_INCREF(Py_None);
  __Pyx_globals = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("unpack_message_type_value", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_unpack_message_type_value, "int (PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("unpack_message_type", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_unpack_message_type, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("unpack_message_id", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_unpack_message_id, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("pack_actor_ref", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_pack_actor_ref, "PyObject *(struct __pyx_obj_4mars_6actors_4core_ActorRef *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("unpack_actor_ref", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_unpack_actor_ref, "void (PyObject *, struct __pyx_obj_4mars_6actors_4core_ActorRef *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("unpack_send_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_unpack_send_message, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("pack_send_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_pack_send_message, "PyObject *(__pyx_t_4mars_6actors_4pool_8messages_INT32_t, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, struct __pyx_obj_4mars_6actors_4core_ActorRef *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_send_message *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("pack_tell_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_pack_tell_message, "PyObject *(__pyx_t_4mars_6actors_4pool_8messages_INT32_t, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, struct __pyx_obj_4mars_6actors_4core_ActorRef *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_tell_message *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("unpack_tell_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_unpack_tell_message, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("get_index", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_get_index, "PyObject *(PyObject *, PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("pack_create_actor_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_pack_create_actor_message, "PyObject *(__pyx_t_4mars_6actors_4pool_8messages_INT32_t, int, struct __pyx_obj_4mars_6actors_4core_ActorRef *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_create_actor_message *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("unpack_create_actor_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_unpack_create_actor_message, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("pack_destroy_actor_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_pack_destroy_actor_message, "PyObject *(__pyx_t_4mars_6actors_4pool_8messages_INT32_t, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, struct __pyx_obj_4mars_6actors_4core_ActorRef *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_destroy_actor_message *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("unpack_destroy_actor_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_unpack_destroy_actor_message, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("pack_has_actor_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_pack_has_actor_message, "PyObject *(__pyx_t_4mars_6actors_4pool_8messages_INT32_t, int, struct __pyx_obj_4mars_6actors_4core_ActorRef *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_has_actor_message *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("unpack_has_actor_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_unpack_has_actor_message, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("pack_result_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_pack_result_message, "PyObject *(PyObject *, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, int, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_result_message *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("unpack_result_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_unpack_result_message, "PyObject *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_unpack_result_message *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("pack_error_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_pack_error_message, "PyObject *(PyObject *, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, __pyx_t_4mars_6actors_4pool_8messages_INT32_t, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4mars_6actors_4pool_8messages_pack_error_message *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("unpack_error_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_unpack_error_message, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("read_remote_message", (void (*)(void))__pyx_f_4mars_6actors_4pool_8messages_read_remote_message, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  __pyx_type_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE.tp_dictoffset && __pyx_type_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CREATE_ACTOR_MESSAGE, (PyObject *)&__pyx_type_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  __pyx_ptype_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE = &__pyx_type_4mars_6actors_4pool_8messages_CREATE_ACTOR_MESSAGE;
  if (PyType_Ready(&__pyx_type_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  __pyx_type_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE.tp_dictoffset && __pyx_type_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DESTROY_ACTOR_MESSAGE, (PyObject *)&__pyx_type_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  __pyx_ptype_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE = &__pyx_type_4mars_6actors_4pool_8messages_DESTROY_ACTOR_MESSAGE;
  if (PyType_Ready(&__pyx_type_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE) < 0) __PYX_ERR(0, 104, __pyx_L1_error)
  __pyx_type_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE.tp_dictoffset && __pyx_type_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_HAS_ACTOR_MESSAGE, (PyObject *)&__pyx_type_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE) < 0) __PYX_ERR(0, 104, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE) < 0) __PYX_ERR(0, 104, __pyx_L1_error)
  __pyx_ptype_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE = &__pyx_type_4mars_6actors_4pool_8messages_HAS_ACTOR_MESSAGE;
  if (PyType_Ready(&__pyx_type_4mars_6actors_4pool_8messages_RESULT_MESSAGE) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_type_4mars_6actors_4pool_8messages_RESULT_MESSAGE.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_6actors_4pool_8messages_RESULT_MESSAGE.tp_dictoffset && __pyx_type_4mars_6actors_4pool_8messages_RESULT_MESSAGE.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_6actors_4pool_8messages_RESULT_MESSAGE.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_RESULT_MESSAGE, (PyObject *)&__pyx_type_4mars_6actors_4pool_8messages_RESULT_MESSAGE) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_4mars_6actors_4pool_8messages_RESULT_MESSAGE) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_ptype_4mars_6actors_4pool_8messages_RESULT_MESSAGE = &__pyx_type_4mars_6actors_4pool_8messages_RESULT_MESSAGE;
  if (PyType_Ready(&__pyx_type_4mars_6actors_4pool_8messages_ERROR_MESSAGE) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  __pyx_type_4mars_6actors_4pool_8messages_ERROR_MESSAGE.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_6actors_4pool_8messages_ERROR_MESSAGE.tp_dictoffset && __pyx_type_4mars_6actors_4pool_8messages_ERROR_MESSAGE.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_6actors_4pool_8messages_ERROR_MESSAGE.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ERROR_MESSAGE, (PyObject *)&__pyx_type_4mars_6actors_4pool_8messages_ERROR_MESSAGE) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_4mars_6actors_4pool_8messages_ERROR_MESSAGE) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  __pyx_ptype_4mars_6actors_4pool_8messages_ERROR_MESSAGE = &__pyx_type_4mars_6actors_4pool_8messages_ERROR_MESSAGE;
  if (PyType_Ready(&__pyx_type_4mars_6actors_4pool_8messages_SEND_MESSAGE) < 0) __PYX_ERR(0, 157, __pyx_L1_error)
  __pyx_type_4mars_6actors_4pool_8messages_SEND_MESSAGE.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_6actors_4pool_8messages_SEND_MESSAGE.tp_dictoffset && __pyx_type_4mars_6actors_4pool_8messages_SEND_MESSAGE.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_6actors_4pool_8messages_SEND_MESSAGE.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SEND_MESSAGE, (PyObject *)&__pyx_type_4mars_6actors_4pool_8messages_SEND_MESSAGE) < 0) __PYX_ERR(0, 157, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_4mars_6actors_4pool_8messages_SEND_MESSAGE) < 0) __PYX_ERR(0, 157, __pyx_L1_error)
  __pyx_ptype_4mars_6actors_4pool_8messages_SEND_MESSAGE = &__pyx_type_4mars_6actors_4pool_8messages_SEND_MESSAGE;
  if (PyType_Ready(&__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message) < 0) __PYX_ERR(0, 439, __pyx_L1_error)
  __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message.tp_dictoffset && __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message = &__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct____pyx_f_4mars_6actors_4pool_8messages__unpack_tuple_message;
  if (PyType_Ready(&__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 440, __pyx_L1_error)
  __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr = &__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message) < 0) __PYX_ERR(0, 816, __pyx_L1_error)
  __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message.tp_dictoffset && __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message = &__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_2_write_remote_message;
  if (PyType_Ready(&__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 819, __pyx_L1_error)
  __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr.tp_dictoffset && __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr = &__pyx_type_4mars_6actors_4pool_8messages___pyx_scope_struct_3_genexpr;
  __Pyx_EnumMeta.tp_base = (&PyType_Type);
  if (PyType_Ready(&__Pyx_EnumMeta) < 0) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_EnumMeta.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__Pyx_EnumMeta.tp_dictoffset && __Pyx_EnumMeta.tp_getattro == PyObject_GenericGetAttr)) {
    __Pyx_EnumMeta.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__Pyx_EnumMeta) < 0) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_ptype___Pyx_EnumMeta = &__Pyx_EnumMeta;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 206, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 233, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 242, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("mars.actors.core"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4mars_6actors_4core_ActorRef = __Pyx_ImportType(__pyx_t_1, "mars.actors.core", "ActorRef", sizeof(struct __pyx_obj_4mars_6actors_4core_ActorRef), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_6actors_4core_ActorRef) __PYX_ERR(4, 18, __pyx_L1_error)
  __pyx_vtabptr_4mars_6actors_4core_ActorRef = (struct __pyx_vtabstruct_4mars_6actors_4core_ActorRef*)__Pyx_GetVtable(__pyx_ptype_4mars_6actors_4core_ActorRef->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_6actors_4core_ActorRef)) __PYX_ERR(4, 18, __pyx_L1_error)
  __pyx_ptype_4mars_6actors_4core_Actor = __Pyx_ImportType(__pyx_t_1, "mars.actors.core", "Actor", sizeof(struct __pyx_obj_4mars_6actors_4core_Actor), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_6actors_4core_Actor) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_vtabptr_4mars_6actors_4core_Actor = (struct __pyx_vtabstruct_4mars_6actors_4core_Actor*)__Pyx_GetVtable(__pyx_ptype_4mars_6actors_4core_Actor->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_6actors_4core_Actor)) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_ptype_4mars_6actors_4core_FunctionActor = __Pyx_ImportType(__pyx_t_1, "mars.actors.core", "FunctionActor", sizeof(struct __pyx_obj_4mars_6actors_4core_FunctionActor), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4mars_6actors_4core_FunctionActor) __PYX_ERR(4, 40, __pyx_L1_error)
  __pyx_vtabptr_4mars_6actors_4core_FunctionActor = (struct __pyx_vtabstruct_4mars_6actors_4core_FunctionActor*)__Pyx_GetVtable(__pyx_ptype_4mars_6actors_4core_FunctionActor->tp_dict); if (unlikely(!__pyx_vtabptr_4mars_6actors_4core_FunctionActor)) __PYX_ERR(4, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initmessages(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initmessages(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_messages(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_messages(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_messages(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'messages' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_messages(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("messages", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_mars__actors__pool__messages) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "mars.actors.pool.messages")) {
      if (unlikely(PyDict_SetItemString(modules, "mars.actors.pool.messages", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "mars/actors/pool/messages.pyx":17
 * # limitations under the License.
 * 
 * import errno             # <<<<<<<<<<<<<<
 * import socket
 * import struct
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_errno, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_errno, __pyx_t_1) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":18
 * 
 * import errno
 * import socket             # <<<<<<<<<<<<<<
 * import struct
 * from pickle import dumps, loads
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_socket, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_socket, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":19
 * import errno
 * import socket
 * import struct             # <<<<<<<<<<<<<<
 * from pickle import dumps, loads
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_struct, __pyx_t_1) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":20
 * import socket
 * import struct
 * from pickle import dumps, loads             # <<<<<<<<<<<<<<
 * 
 * import numpy as np
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_dumps);
  __Pyx_GIVEREF(__pyx_n_s_dumps);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_dumps);
  __Pyx_INCREF(__pyx_n_s_loads);
  __Pyx_GIVEREF(__pyx_n_s_loads);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_loads);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_dumps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dumps, __pyx_t_1) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_loads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_loads, __pyx_t_1) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":22
 * from pickle import dumps, loads
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * cimport cython
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_2) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":28
 * 
 * from ..core cimport ActorRef
 * from ...compat import BrokenPipeError, ConnectionResetError             # <<<<<<<<<<<<<<
 * from ...lib.tblib import pickling_support
 * 
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_BrokenPipeError);
  __Pyx_GIVEREF(__pyx_n_s_BrokenPipeError);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_BrokenPipeError);
  __Pyx_INCREF(__pyx_n_s_ConnectionResetError);
  __Pyx_GIVEREF(__pyx_n_s_ConnectionResetError);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_ConnectionResetError);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_compat, __pyx_t_2, 3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_BrokenPipeError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BrokenPipeError, __pyx_t_2) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ConnectionResetError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ConnectionResetError, __pyx_t_2) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":29
 * from ..core cimport ActorRef
 * from ...compat import BrokenPipeError, ConnectionResetError
 * from ...lib.tblib import pickling_support             # <<<<<<<<<<<<<<
 * 
 * pickling_support.install()
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_pickling_support);
  __Pyx_GIVEREF(__pyx_n_s_pickling_support);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_pickling_support);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_lib_tblib, __pyx_t_1, 3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_pickling_support); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pickling_support, __pyx_t_1) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":31
 * from ...lib.tblib import pickling_support
 * 
 * pickling_support.install()             # <<<<<<<<<<<<<<
 * 
 * # Internal message types includes:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pickling_support); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_install); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":41
 * # 6) send all, the send actor message
 * # 7) tell all, the tell actor message
 * cpdef BYTE_t CREATE_ACTOR = MessageType.create_actor             # <<<<<<<<<<<<<<
 * cpdef BYTE_t DESTROY_ACTOR = MessageType.destroy_actor
 * cpdef BYTE_t HAS_ACTOR = MessageType.has_actor
 */
  __pyx_v_4mars_6actors_4pool_8messages_CREATE_ACTOR = __pyx_e_4mars_6actors_4pool_8messages_create_actor;

  /* "mars/actors/pool/messages.pyx":42
 * # 7) tell all, the tell actor message
 * cpdef BYTE_t CREATE_ACTOR = MessageType.create_actor
 * cpdef BYTE_t DESTROY_ACTOR = MessageType.destroy_actor             # <<<<<<<<<<<<<<
 * cpdef BYTE_t HAS_ACTOR = MessageType.has_actor
 * cpdef BYTE_t RESULT = MessageType.result
 */
  __pyx_v_4mars_6actors_4pool_8messages_DESTROY_ACTOR = __pyx_e_4mars_6actors_4pool_8messages_destroy_actor;

  /* "mars/actors/pool/messages.pyx":43
 * cpdef BYTE_t CREATE_ACTOR = MessageType.create_actor
 * cpdef BYTE_t DESTROY_ACTOR = MessageType.destroy_actor
 * cpdef BYTE_t HAS_ACTOR = MessageType.has_actor             # <<<<<<<<<<<<<<
 * cpdef BYTE_t RESULT = MessageType.result
 * cpdef BYTE_t ERROR = MessageType.error
 */
  __pyx_v_4mars_6actors_4pool_8messages_HAS_ACTOR = __pyx_e_4mars_6actors_4pool_8messages_has_actor;

  /* "mars/actors/pool/messages.pyx":44
 * cpdef BYTE_t DESTROY_ACTOR = MessageType.destroy_actor
 * cpdef BYTE_t HAS_ACTOR = MessageType.has_actor
 * cpdef BYTE_t RESULT = MessageType.result             # <<<<<<<<<<<<<<
 * cpdef BYTE_t ERROR = MessageType.error
 * cpdef BYTE_t SEND_ALL = MessageType.send_all
 */
  __pyx_v_4mars_6actors_4pool_8messages_RESULT = __pyx_e_4mars_6actors_4pool_8messages_result;

  /* "mars/actors/pool/messages.pyx":45
 * cpdef BYTE_t HAS_ACTOR = MessageType.has_actor
 * cpdef BYTE_t RESULT = MessageType.result
 * cpdef BYTE_t ERROR = MessageType.error             # <<<<<<<<<<<<<<
 * cpdef BYTE_t SEND_ALL = MessageType.send_all
 * cpdef BYTE_t TELL_ALL = MessageType.tell_all
 */
  __pyx_v_4mars_6actors_4pool_8messages_ERROR = __pyx_e_4mars_6actors_4pool_8messages_error;

  /* "mars/actors/pool/messages.pyx":46
 * cpdef BYTE_t RESULT = MessageType.result
 * cpdef BYTE_t ERROR = MessageType.error
 * cpdef BYTE_t SEND_ALL = MessageType.send_all             # <<<<<<<<<<<<<<
 * cpdef BYTE_t TELL_ALL = MessageType.tell_all
 * 
 */
  __pyx_v_4mars_6actors_4pool_8messages_SEND_ALL = __pyx_e_4mars_6actors_4pool_8messages_send_all;

  /* "mars/actors/pool/messages.pyx":47
 * cpdef BYTE_t ERROR = MessageType.error
 * cpdef BYTE_t SEND_ALL = MessageType.send_all
 * cpdef BYTE_t TELL_ALL = MessageType.tell_all             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_4mars_6actors_4pool_8messages_TELL_ALL = __pyx_e_4mars_6actors_4pool_8messages_tell_all;

  /* "mars/actors/pool/messages.pyx":57
 * 
 * 
 * cdef BYTE_t NONE = MessageSerialType.null             # <<<<<<<<<<<<<<
 * cdef BYTE_t RAW_BYTES = MessageSerialType.raw_bytes
 * cdef BYTE_t PICKLE = MessageSerialType.pickle
 */
  __pyx_v_4mars_6actors_4pool_8messages_NONE = __pyx_e_4mars_6actors_4pool_8messages_null;

  /* "mars/actors/pool/messages.pyx":58
 * 
 * cdef BYTE_t NONE = MessageSerialType.null
 * cdef BYTE_t RAW_BYTES = MessageSerialType.raw_bytes             # <<<<<<<<<<<<<<
 * cdef BYTE_t PICKLE = MessageSerialType.pickle
 * 
 */
  __pyx_v_4mars_6actors_4pool_8messages_RAW_BYTES = __pyx_e_4mars_6actors_4pool_8messages_raw_bytes;

  /* "mars/actors/pool/messages.pyx":59
 * cdef BYTE_t NONE = MessageSerialType.null
 * cdef BYTE_t RAW_BYTES = MessageSerialType.raw_bytes
 * cdef BYTE_t PICKLE = MessageSerialType.pickle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_4mars_6actors_4pool_8messages_PICKLE = __pyx_e_4mars_6actors_4pool_8messages_pickle;

  /* "mars/actors/pool/messages.pyx":62
 * 
 * 
 * cdef int DEFAULT_PROTOCOL = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL = 0;

  /* "mars/actors/pool/messages.pyx":176
 * 
 * 
 * cdef object _pack_byte = struct.Struct('<B').pack             # <<<<<<<<<<<<<<
 * cdef object _unpack_byte = struct.Struct('<B').unpack
 * cdef object _pack_int = struct.Struct('<I').pack
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_4mars_6actors_4pool_8messages__pack_byte);
  __Pyx_DECREF_SET(__pyx_v_4mars_6actors_4pool_8messages__pack_byte, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":177
 * 
 * cdef object _pack_byte = struct.Struct('<B').pack
 * cdef object _unpack_byte = struct.Struct('<B').unpack             # <<<<<<<<<<<<<<
 * cdef object _pack_int = struct.Struct('<I').pack
 * cdef object _unpack_int = struct.Struct('<I').unpack
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_unpack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_byte);
  __Pyx_DECREF_SET(__pyx_v_4mars_6actors_4pool_8messages__unpack_byte, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":178
 * cdef object _pack_byte = struct.Struct('<B').pack
 * cdef object _unpack_byte = struct.Struct('<B').unpack
 * cdef object _pack_int = struct.Struct('<I').pack             # <<<<<<<<<<<<<<
 * cdef object _unpack_int = struct.Struct('<I').unpack
 * cdef object _pack_long = struct.Struct('<Q').pack
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_4mars_6actors_4pool_8messages__pack_int);
  __Pyx_DECREF_SET(__pyx_v_4mars_6actors_4pool_8messages__pack_int, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":179
 * cdef object _unpack_byte = struct.Struct('<B').unpack
 * cdef object _pack_int = struct.Struct('<I').pack
 * cdef object _unpack_int = struct.Struct('<I').unpack             # <<<<<<<<<<<<<<
 * cdef object _pack_long = struct.Struct('<Q').pack
 * cdef object _unpack_long = struct.Struct('<Q').unpack
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_unpack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_int);
  __Pyx_DECREF_SET(__pyx_v_4mars_6actors_4pool_8messages__unpack_int, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":180
 * cdef object _pack_int = struct.Struct('<I').pack
 * cdef object _unpack_int = struct.Struct('<I').unpack
 * cdef object _pack_long = struct.Struct('<Q').pack             # <<<<<<<<<<<<<<
 * cdef object _unpack_long = struct.Struct('<Q').unpack
 * cdef object _pack_short = struct.Struct('<h').pack
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_4mars_6actors_4pool_8messages__pack_long);
  __Pyx_DECREF_SET(__pyx_v_4mars_6actors_4pool_8messages__pack_long, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":181
 * cdef object _unpack_int = struct.Struct('<I').unpack
 * cdef object _pack_long = struct.Struct('<Q').pack
 * cdef object _unpack_long = struct.Struct('<Q').unpack             # <<<<<<<<<<<<<<
 * cdef object _pack_short = struct.Struct('<h').pack
 * cdef object _unpack_short = struct.Struct('<h').unpack
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_unpack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_long);
  __Pyx_DECREF_SET(__pyx_v_4mars_6actors_4pool_8messages__unpack_long, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":182
 * cdef object _pack_long = struct.Struct('<Q').pack
 * cdef object _unpack_long = struct.Struct('<Q').unpack
 * cdef object _pack_short = struct.Struct('<h').pack             # <<<<<<<<<<<<<<
 * cdef object _unpack_short = struct.Struct('<h').unpack
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_4mars_6actors_4pool_8messages__pack_short);
  __Pyx_DECREF_SET(__pyx_v_4mars_6actors_4pool_8messages__pack_short, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mars/actors/pool/messages.pyx":183
 * cdef object _unpack_long = struct.Struct('<Q').unpack
 * cdef object _pack_short = struct.Struct('<h').pack
 * cdef object _unpack_short = struct.Struct('<h').unpack             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_unpack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_v_4mars_6actors_4pool_8messages__unpack_short);
  __Pyx_DECREF_SET(__pyx_v_4mars_6actors_4pool_8messages__unpack_short, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":202
 * 
 * 
 * cdef inline void _pack_message_type(int message_type, bytearray buf, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef int value
 * 
 */
  __pyx_k_ = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":469
 * 
 * cdef inline object _pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
 *                                       bint send, object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytes message_id
 *     cdef int message_type
 */
  __pyx_k__3 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":517
 * 
 * cpdef object pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     return _pack_send_message(from_index, to_index, actor_ref, message, True,
 *                               write=write, protocol=protocol)
 */
  __pyx_k__4 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":516
 * 
 * 
 * cpdef object pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, True,
 */
  __pyx_k__4 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":527
 * 
 * cpdef object pack_tell_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     return _pack_send_message(from_index, to_index, actor_ref, message, False,
 *                               write=write, protocol=protocol)
 */
  __pyx_k__5 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":526
 * 
 * 
 * cpdef object pack_tell_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,             # <<<<<<<<<<<<<<
 *                                object write=None, int protocol=DEFAULT_PROTOCOL):
 *     return _pack_send_message(from_index, to_index, actor_ref, message, False,
 */
  __pyx_k__5 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":558
 * 
 * cpdef object pack_create_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref, object actor_cls,
 *                                        tuple args, dict kw, object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytes message_id
 *     cdef bytearray buf
 */
  __pyx_k__6 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":557
 * 
 * 
 * cpdef object pack_create_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref, object actor_cls,             # <<<<<<<<<<<<<<
 *                                        tuple args, dict kw, object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytes message_id
 */
  __pyx_k__6 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":611
 * 
 * cpdef object pack_destroy_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,
 *                                         object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytes message_id
 *     cdef bytearray buf
 */
  __pyx_k__7 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":610
 * 
 * 
 * cpdef object pack_destroy_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,             # <<<<<<<<<<<<<<
 *                                         object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytes message_id
 */
  __pyx_k__7 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":653
 * 
 * cpdef object pack_has_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref,
 *                                     object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytearray buf
 * 
 */
  __pyx_k__8 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":652
 * 
 * 
 * cpdef object pack_has_actor_message(INT32_t from_index, int to_index, ActorRef actor_ref,             # <<<<<<<<<<<<<<
 *                                     object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytearray buf
 */
  __pyx_k__8 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":693
 * 
 * cpdef object pack_result_message(bytes message_id, INT32_t from_index, int to_index, object result,
 *                                  object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytearray buf
 * 
 */
  __pyx_k__9 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":692
 * 
 * 
 * cpdef object pack_result_message(bytes message_id, INT32_t from_index, int to_index, object result,             # <<<<<<<<<<<<<<
 *                                  object write=None, int protocol=DEFAULT_PROTOCOL):
 *     cdef bytearray buf
 */
  __pyx_k__9 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":736
 * cpdef object pack_error_message(bytes message_id, INT32_t from_index, INT32_t to_index,
 *                                 object error_type, object error, object tb,
 *                                 object write=None, int protocol=DEFAULT_PROTOCOL):             # <<<<<<<<<<<<<<
 *     cdef bytearray buf
 * 
 */
  __pyx_k__10 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":734
 * 
 * 
 * cpdef object pack_error_message(bytes message_id, INT32_t from_index, INT32_t to_index,             # <<<<<<<<<<<<<<
 *                                 object error_type, object error, object tb,
 *                                 object write=None, int protocol=DEFAULT_PROTOCOL):
 */
  __pyx_k__10 = __pyx_v_4mars_6actors_4pool_8messages_DEFAULT_PROTOCOL;

  /* "mars/actors/pool/messages.pyx":816
 * 
 * 
 * def write_remote_message(write_func, *binary):             # <<<<<<<<<<<<<<
 *     cdef bytes size
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4mars_6actors_4pool_8messages_43write_remote_message, NULL, __pyx_n_s_mars_actors_pool_messages); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_write_remote_message_2, __pyx_t_2) < 0) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":824
 * 
 * 
 * if PY_MAJOR_VERSION < 3:             # <<<<<<<<<<<<<<
 *     _write_remote_message = write_remote_message
 * 
 */
  __pyx_t_3 = ((PY_MAJOR_VERSION < 3) != 0);
  if (__pyx_t_3) {

    /* "mars/actors/pool/messages.pyx":825
 * 
 * if PY_MAJOR_VERSION < 3:
 *     _write_remote_message = write_remote_message             # <<<<<<<<<<<<<<
 * 
 *     def write_remote_message(*args, **kwargs):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_write_remote_message_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 825, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_write_remote_message, __pyx_t_2) < 0) __PYX_ERR(0, 825, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "mars/actors/pool/messages.pyx":827
 *     _write_remote_message = write_remote_message
 * 
 *     def write_remote_message(*args, **kwargs):             # <<<<<<<<<<<<<<
 *         try:
 *             return _write_remote_message(*args, **kwargs)
 */
    __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4mars_6actors_4pool_8messages_45write_remote_message, NULL, __pyx_n_s_mars_actors_pool_messages); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_write_remote_message_2, __pyx_t_2) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "mars/actors/pool/messages.pyx":824
 * 
 * 
 * if PY_MAJOR_VERSION < 3:             # <<<<<<<<<<<<<<
 *     _write_remote_message = write_remote_message
 * 
 */
  }

  /* "(tree fragment)":1
 * def __pyx_unpickle_CREATE_ACTOR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4mars_6actors_4pool_8messages_47__pyx_unpickle_CREATE_ACTOR_MESSAGE, NULL, __pyx_n_s_mars_actors_pool_messages); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CREATE_ACTOR_MESS, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(<CREATE_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CREATE_ACTOR_MESSAGE__set_state(CREATE_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.actor_cls = __pyx_state[0]; __pyx_result.actor_ref = __pyx_state[1]; __pyx_result.args = __pyx_state[2]; __pyx_result.from_index = __pyx_state[3]; __pyx_result.kwargs = __pyx_state[4]; __pyx_result.message_id = __pyx_state[5]; __pyx_result.message_type = __pyx_state[6]; __pyx_result.to_index = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4mars_6actors_4pool_8messages_49__pyx_unpickle_DESTROY_ACTOR_MESSAGE, NULL, __pyx_n_s_mars_actors_pool_messages); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_DESTROY_ACTOR_MES, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_HAS_ACTOR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4mars_6actors_4pool_8messages_51__pyx_unpickle_HAS_ACTOR_MESSAGE, NULL, __pyx_n_s_mars_actors_pool_messages); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_HAS_ACTOR_MESSAGE, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(<HAS_ACTOR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_HAS_ACTOR_MESSAGE__set_state(HAS_ACTOR_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.actor_ref = __pyx_state[0]; __pyx_result.from_index = __pyx_state[1]; __pyx_result.message_id = __pyx_state[2]; __pyx_result.message_type = __pyx_state[3]; __pyx_result.to_index = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4mars_6actors_4pool_8messages_53__pyx_unpickle_RESULT_MESSAGE, NULL, __pyx_n_s_mars_actors_pool_messages); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_RESULT_MESSAGE, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ERROR_MESSAGE(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4mars_6actors_4pool_8messages_55__pyx_unpickle_ERROR_MESSAGE, NULL, __pyx_n_s_mars_actors_pool_messages); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ERROR_MESSAGE, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_ERROR_MESSAGE__set_state(<ERROR_MESSAGE> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ERROR_MESSAGE__set_state(ERROR_MESSAGE __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.error = __pyx_state[0]; __pyx_result.error_type = __pyx_state[1]; __pyx_result.from_index = __pyx_state[2]; __pyx_result.message_id = __pyx_state[3]; __pyx_result.message_type = __pyx_state[4]; __pyx_result.to_index = __pyx_state[5]; __pyx_result.traceback = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4mars_6actors_4pool_8messages_57__pyx_unpickle_SEND_MESSAGE, NULL, __pyx_n_s_mars_actors_pool_messages); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_SEND_MESSAGE, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mars/actors/pool/messages.pyx":1
 * #!/usr/bin/env python             # <<<<<<<<<<<<<<
 * # -*- coding: utf-8 -*-
 * # Copyright 1999-2018 Alibaba Group Holding Ltd.
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":9
 * 
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:             # <<<<<<<<<<<<<<
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 */
  __pyx_t_3 = ((PY_VERSION_HEX >= 0x02070000) != 0);
  if (__pyx_t_3) {

    /* "EnumBase":10
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:
 *     from collections import OrderedDict as __Pyx_OrderedDict             # <<<<<<<<<<<<<<
 * else:
 *     __Pyx_OrderedDict = dict
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_OrderedDict);
    __Pyx_GIVEREF(__pyx_n_s_OrderedDict);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_OrderedDict);
    __pyx_t_1 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XGOTREF(__Pyx_OrderedDict);
    __Pyx_DECREF_SET(__Pyx_OrderedDict, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "EnumBase":9
 * 
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:             # <<<<<<<<<<<<<<
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 */
    goto __pyx_L3;
  }

  /* "EnumBase":12
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 *     __Pyx_OrderedDict = dict             # <<<<<<<<<<<<<<
 * 
 * @cython.internal
 */
  /*else*/ {
    __Pyx_INCREF(((PyObject *)(&PyDict_Type)));
    __Pyx_XGOTREF(__Pyx_OrderedDict);
    __Pyx_DECREF_SET(__Pyx_OrderedDict, ((PyObject *)(&PyDict_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyDict_Type)));
  }
  __pyx_L3:;

  /* "EnumBase":26
 * 
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):             # <<<<<<<<<<<<<<
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&PyInt_Type)));
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_Pyx_EnumBase, __pyx_n_s_Pyx_EnumBase, (PyObject *) NULL, __pyx_n_s_EnumBase, (PyObject *) NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "EnumBase":27
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta             # <<<<<<<<<<<<<<
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 */
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_metaclass, ((PyObject *)__pyx_ptype___Pyx_EnumMeta)) < 0) __PYX_ERR(1, 27, __pyx_L1_error)

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_1__new__, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_Pyx_EnumBase___new, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__40);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_new, __pyx_t_5) < 0) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_3__repr__, 0, __pyx_n_s_Pyx_EnumBase___repr, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_repr, __pyx_t_5) < 0) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_5__str__, 0, __pyx_n_s_Pyx_EnumBase___str, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_str, __pyx_t_5) < 0) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumBase":26
 * 
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):             # <<<<<<<<<<<<<<
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_Pyx_EnumBase, __pyx_t_1, __pyx_t_4, NULL, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XGOTREF(__Pyx_EnumBase);
  __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":44
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 *     from enum import IntEnum as __Pyx_EnumBase
 * 
 */
  __pyx_t_3 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_3) {

    /* "EnumBase":45
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 *     from enum import IntEnum as __Pyx_EnumBase             # <<<<<<<<<<<<<<
 * 
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_IntEnum);
    __Pyx_GIVEREF(__pyx_n_s_IntEnum);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_IntEnum);
    __pyx_t_2 = __Pyx_Import(__pyx_n_s_enum, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XGOTREF(__Pyx_EnumBase);
    __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumBase":44
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 *     from enum import IntEnum as __Pyx_EnumBase
 * 
 */
  }

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta, NULL, __pyx_n_s_EnumBase); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_2 = __Pyx_Globals(); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     MessageSerialType = __Pyx_EnumBase('MessageSerialType', __Pyx_OrderedDict([
 */
  __pyx_t_3 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_3) {

    /* "EnumType":54
 * 
 *     MessageSerialType = __Pyx_EnumBase('MessageSerialType', __Pyx_OrderedDict([
 *         ('null', null),             # <<<<<<<<<<<<<<
 *         ('raw_bytes', raw_bytes),
 *         ('pickle', pickle),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageSerialType(__pyx_e_4mars_6actors_4pool_8messages_null); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_null);
    __Pyx_GIVEREF(__pyx_n_s_null);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_null);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":55
 *     MessageSerialType = __Pyx_EnumBase('MessageSerialType', __Pyx_OrderedDict([
 *         ('null', null),
 *         ('raw_bytes', raw_bytes),             # <<<<<<<<<<<<<<
 *         ('pickle', pickle),
 *     ]))
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageSerialType(__pyx_e_4mars_6actors_4pool_8messages_raw_bytes); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_n_s_raw_bytes);
    __Pyx_GIVEREF(__pyx_n_s_raw_bytes);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_raw_bytes);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":56
 *         ('null', null),
 *         ('raw_bytes', raw_bytes),
 *         ('pickle', pickle),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['null'] = MessageSerialType.null
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageSerialType(__pyx_e_4mars_6actors_4pool_8messages_pickle); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_pickle);
    __Pyx_GIVEREF(__pyx_n_s_pickle);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_pickle);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     MessageSerialType = __Pyx_EnumBase('MessageSerialType', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('null', null),
 *         ('raw_bytes', raw_bytes),
 */
    __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_MessageSerialType);
    __Pyx_GIVEREF(__pyx_n_s_MessageSerialType);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_MessageSerialType);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_MessageSerialType, __pyx_t_5) < 0) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "EnumType":58
 *         ('pickle', pickle),
 *     ]))
 *     __Pyx_globals['null'] = MessageSerialType.null             # <<<<<<<<<<<<<<
 *     __Pyx_globals['raw_bytes'] = MessageSerialType.raw_bytes
 *     __Pyx_globals['pickle'] = MessageSerialType.pickle
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_MessageSerialType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_null); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 58, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_null, __pyx_t_2) < 0)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":59
 *     ]))
 *     __Pyx_globals['null'] = MessageSerialType.null
 *     __Pyx_globals['raw_bytes'] = MessageSerialType.raw_bytes             # <<<<<<<<<<<<<<
 *     __Pyx_globals['pickle'] = MessageSerialType.pickle
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageSerialType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_raw_bytes); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 59, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_raw_bytes, __pyx_t_5) < 0)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "EnumType":60
 *     __Pyx_globals['null'] = MessageSerialType.null
 *     __Pyx_globals['raw_bytes'] = MessageSerialType.raw_bytes
 *     __Pyx_globals['pickle'] = MessageSerialType.pickle             # <<<<<<<<<<<<<<
 * else:
 *     class MessageSerialType(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_MessageSerialType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_pickle); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 60, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_pickle, __pyx_t_2) < 0)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     MessageSerialType = __Pyx_EnumBase('MessageSerialType', __Pyx_OrderedDict([
 */
    goto __pyx_L5;
  }

  /* "EnumType":62
 *     __Pyx_globals['pickle'] = MessageSerialType.pickle
 * else:
 *     class MessageSerialType(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['null'] = MessageSerialType(null, 'null')
 */
  /*else*/ {
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __Pyx_EnumBase);
    __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_2, __pyx_n_s_MessageSerialType, __pyx_n_s_MessageSerialType, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_MessageSerialType, __pyx_t_2, __pyx_t_4, NULL, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_MessageSerialType, __pyx_t_1) < 0) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":64
 *     class MessageSerialType(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['null'] = MessageSerialType(null, 'null')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['raw_bytes'] = MessageSerialType(raw_bytes, 'raw_bytes')
 *     __Pyx_globals['pickle'] = MessageSerialType(pickle, 'pickle')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageSerialType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageSerialType(__pyx_e_4mars_6actors_4pool_8messages_null); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_null);
    __Pyx_GIVEREF(__pyx_n_s_null);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_n_s_null);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 64, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_null, __pyx_t_5) < 0)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "EnumType":65
 *         pass
 *     __Pyx_globals['null'] = MessageSerialType(null, 'null')
 *     __Pyx_globals['raw_bytes'] = MessageSerialType(raw_bytes, 'raw_bytes')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['pickle'] = MessageSerialType(pickle, 'pickle')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_MessageSerialType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageSerialType(__pyx_e_4mars_6actors_4pool_8messages_raw_bytes); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
    __Pyx_INCREF(__pyx_n_s_raw_bytes);
    __Pyx_GIVEREF(__pyx_n_s_raw_bytes);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_raw_bytes);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 65, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_raw_bytes, __pyx_t_4) < 0)) __PYX_ERR(1, 65, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":66
 *     __Pyx_globals['null'] = MessageSerialType(null, 'null')
 *     __Pyx_globals['raw_bytes'] = MessageSerialType(raw_bytes, 'raw_bytes')
 *     __Pyx_globals['pickle'] = MessageSerialType(pickle, 'pickle')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_MessageSerialType); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageSerialType(__pyx_e_4mars_6actors_4pool_8messages_pickle); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_pickle);
    __Pyx_GIVEREF(__pyx_n_s_pickle);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_n_s_pickle);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 66, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_pickle, __pyx_t_2) < 0)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L5:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_2 = __Pyx_Globals(); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     MessageType = __Pyx_EnumBase('MessageType', __Pyx_OrderedDict([
 */
  __pyx_t_3 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_3) {

    /* "EnumType":54
 * 
 *     MessageType = __Pyx_EnumBase('MessageType', __Pyx_OrderedDict([
 *         ('create_actor', create_actor),             # <<<<<<<<<<<<<<
 *         ('destroy_actor', destroy_actor),
 *         ('has_actor', has_actor),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_create_actor); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_create_actor);
    __Pyx_GIVEREF(__pyx_n_s_create_actor);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_create_actor);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":55
 *     MessageType = __Pyx_EnumBase('MessageType', __Pyx_OrderedDict([
 *         ('create_actor', create_actor),
 *         ('destroy_actor', destroy_actor),             # <<<<<<<<<<<<<<
 *         ('has_actor', has_actor),
 *         ('result', result),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_destroy_actor); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_n_s_destroy_actor);
    __Pyx_GIVEREF(__pyx_n_s_destroy_actor);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_destroy_actor);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":56
 *         ('create_actor', create_actor),
 *         ('destroy_actor', destroy_actor),
 *         ('has_actor', has_actor),             # <<<<<<<<<<<<<<
 *         ('result', result),
 *         ('error', error),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_has_actor); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_has_actor);
    __Pyx_GIVEREF(__pyx_n_s_has_actor);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_has_actor);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":57
 *         ('destroy_actor', destroy_actor),
 *         ('has_actor', has_actor),
 *         ('result', result),             # <<<<<<<<<<<<<<
 *         ('error', error),
 *         ('send_all', send_all),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_result); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_n_s_result);
    __Pyx_GIVEREF(__pyx_n_s_result);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_result);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":58
 *         ('has_actor', has_actor),
 *         ('result', result),
 *         ('error', error),             # <<<<<<<<<<<<<<
 *         ('send_all', send_all),
 *         ('tell_all', tell_all),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_error); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_n_s_error);
    __Pyx_GIVEREF(__pyx_n_s_error);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_error);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":59
 *         ('result', result),
 *         ('error', error),
 *         ('send_all', send_all),             # <<<<<<<<<<<<<<
 *         ('tell_all', tell_all),
 *         ('send_chunk_start', send_chunk_start),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_send_all); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_send_all);
    __Pyx_GIVEREF(__pyx_n_s_send_all);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_send_all);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":60
 *         ('error', error),
 *         ('send_all', send_all),
 *         ('tell_all', tell_all),             # <<<<<<<<<<<<<<
 *         ('send_chunk_start', send_chunk_start),
 *         ('send_chunk', send_chunk),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_tell_all); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_tell_all);
    __Pyx_GIVEREF(__pyx_n_s_tell_all);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_tell_all);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":61
 *         ('send_all', send_all),
 *         ('tell_all', tell_all),
 *         ('send_chunk_start', send_chunk_start),             # <<<<<<<<<<<<<<
 *         ('send_chunk', send_chunk),
 *         ('send_chunk_end', send_chunk_end),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_send_chunk_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_n_s_send_chunk_start);
    __Pyx_GIVEREF(__pyx_n_s_send_chunk_start);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_send_chunk_start);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":62
 *         ('tell_all', tell_all),
 *         ('send_chunk_start', send_chunk_start),
 *         ('send_chunk', send_chunk),             # <<<<<<<<<<<<<<
 *         ('send_chunk_end', send_chunk_end),
 *         ('tell_chunk_start', tell_chunk_start),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_send_chunk); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_n_s_send_chunk);
    __Pyx_GIVEREF(__pyx_n_s_send_chunk);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_send_chunk);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":63
 *         ('send_chunk_start', send_chunk_start),
 *         ('send_chunk', send_chunk),
 *         ('send_chunk_end', send_chunk_end),             # <<<<<<<<<<<<<<
 *         ('tell_chunk_start', tell_chunk_start),
 *         ('tell_chunk', tell_chunk),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_send_chunk_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_n_s_send_chunk_end);
    __Pyx_GIVEREF(__pyx_n_s_send_chunk_end);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_s_send_chunk_end);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":64
 *         ('send_chunk', send_chunk),
 *         ('send_chunk_end', send_chunk_end),
 *         ('tell_chunk_start', tell_chunk_start),             # <<<<<<<<<<<<<<
 *         ('tell_chunk', tell_chunk),
 *         ('tell_chunk_end', tell_chunk_end),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_tell_chunk_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_n_s_tell_chunk_start);
    __Pyx_GIVEREF(__pyx_n_s_tell_chunk_start);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_s_tell_chunk_start);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":65
 *         ('send_chunk_end', send_chunk_end),
 *         ('tell_chunk_start', tell_chunk_start),
 *         ('tell_chunk', tell_chunk),             # <<<<<<<<<<<<<<
 *         ('tell_chunk_end', tell_chunk_end),
 *     ]))
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_tell_chunk); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_tell_chunk);
    __Pyx_GIVEREF(__pyx_n_s_tell_chunk);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_n_s_tell_chunk);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":66
 *         ('tell_chunk_start', tell_chunk_start),
 *         ('tell_chunk', tell_chunk),
 *         ('tell_chunk_end', tell_chunk_end),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['create_actor'] = MessageType.create_actor
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_tell_chunk_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_tell_chunk_end);
    __Pyx_GIVEREF(__pyx_n_s_tell_chunk_end);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_n_s_tell_chunk_end);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     MessageType = __Pyx_EnumBase('MessageType', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('create_actor', create_actor),
 *         ('destroy_actor', destroy_actor),
 */
    __pyx_t_2 = PyList_New(13); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyList_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    PyList_SET_ITEM(__pyx_t_2, 3, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyList_SET_ITEM(__pyx_t_2, 4, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyList_SET_ITEM(__pyx_t_2, 5, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    PyList_SET_ITEM(__pyx_t_2, 6, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_10);
    PyList_SET_ITEM(__pyx_t_2, 7, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_11);
    PyList_SET_ITEM(__pyx_t_2, 8, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_12);
    PyList_SET_ITEM(__pyx_t_2, 9, __pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyList_SET_ITEM(__pyx_t_2, 10, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyList_SET_ITEM(__pyx_t_2, 11, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyList_SET_ITEM(__pyx_t_2, 12, __pyx_t_15);
    __pyx_t_5 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_t_10 = 0;
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_t_13 = 0;
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_MessageType);
    __Pyx_GIVEREF(__pyx_n_s_MessageType);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_MessageType);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_15);
    __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_2, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_MessageType, __pyx_t_15) < 0) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":68
 *         ('tell_chunk_end', tell_chunk_end),
 *     ]))
 *     __Pyx_globals['create_actor'] = MessageType.create_actor             # <<<<<<<<<<<<<<
 *     __Pyx_globals['destroy_actor'] = MessageType.destroy_actor
 *     __Pyx_globals['has_actor'] = MessageType.has_actor
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_create_actor); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 68, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_create_actor, __pyx_t_2) < 0)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":69
 *     ]))
 *     __Pyx_globals['create_actor'] = MessageType.create_actor
 *     __Pyx_globals['destroy_actor'] = MessageType.destroy_actor             # <<<<<<<<<<<<<<
 *     __Pyx_globals['has_actor'] = MessageType.has_actor
 *     __Pyx_globals['result'] = MessageType.result
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_destroy_actor); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 69, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_destroy_actor, __pyx_t_15) < 0)) __PYX_ERR(1, 69, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":70
 *     __Pyx_globals['create_actor'] = MessageType.create_actor
 *     __Pyx_globals['destroy_actor'] = MessageType.destroy_actor
 *     __Pyx_globals['has_actor'] = MessageType.has_actor             # <<<<<<<<<<<<<<
 *     __Pyx_globals['result'] = MessageType.result
 *     __Pyx_globals['error'] = MessageType.error
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_has_actor); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 70, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_has_actor, __pyx_t_2) < 0)) __PYX_ERR(1, 70, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":71
 *     __Pyx_globals['destroy_actor'] = MessageType.destroy_actor
 *     __Pyx_globals['has_actor'] = MessageType.has_actor
 *     __Pyx_globals['result'] = MessageType.result             # <<<<<<<<<<<<<<
 *     __Pyx_globals['error'] = MessageType.error
 *     __Pyx_globals['send_all'] = MessageType.send_all
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_result); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 71, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_result, __pyx_t_15) < 0)) __PYX_ERR(1, 71, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":72
 *     __Pyx_globals['has_actor'] = MessageType.has_actor
 *     __Pyx_globals['result'] = MessageType.result
 *     __Pyx_globals['error'] = MessageType.error             # <<<<<<<<<<<<<<
 *     __Pyx_globals['send_all'] = MessageType.send_all
 *     __Pyx_globals['tell_all'] = MessageType.tell_all
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_error); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 72, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_error, __pyx_t_2) < 0)) __PYX_ERR(1, 72, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":73
 *     __Pyx_globals['result'] = MessageType.result
 *     __Pyx_globals['error'] = MessageType.error
 *     __Pyx_globals['send_all'] = MessageType.send_all             # <<<<<<<<<<<<<<
 *     __Pyx_globals['tell_all'] = MessageType.tell_all
 *     __Pyx_globals['send_chunk_start'] = MessageType.send_chunk_start
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_send_all); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 73, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_send_all, __pyx_t_15) < 0)) __PYX_ERR(1, 73, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":74
 *     __Pyx_globals['error'] = MessageType.error
 *     __Pyx_globals['send_all'] = MessageType.send_all
 *     __Pyx_globals['tell_all'] = MessageType.tell_all             # <<<<<<<<<<<<<<
 *     __Pyx_globals['send_chunk_start'] = MessageType.send_chunk_start
 *     __Pyx_globals['send_chunk'] = MessageType.send_chunk
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_tell_all); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 74, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_tell_all, __pyx_t_2) < 0)) __PYX_ERR(1, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":75
 *     __Pyx_globals['send_all'] = MessageType.send_all
 *     __Pyx_globals['tell_all'] = MessageType.tell_all
 *     __Pyx_globals['send_chunk_start'] = MessageType.send_chunk_start             # <<<<<<<<<<<<<<
 *     __Pyx_globals['send_chunk'] = MessageType.send_chunk
 *     __Pyx_globals['send_chunk_end'] = MessageType.send_chunk_end
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_send_chunk_start); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 75, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_send_chunk_start, __pyx_t_15) < 0)) __PYX_ERR(1, 75, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":76
 *     __Pyx_globals['tell_all'] = MessageType.tell_all
 *     __Pyx_globals['send_chunk_start'] = MessageType.send_chunk_start
 *     __Pyx_globals['send_chunk'] = MessageType.send_chunk             # <<<<<<<<<<<<<<
 *     __Pyx_globals['send_chunk_end'] = MessageType.send_chunk_end
 *     __Pyx_globals['tell_chunk_start'] = MessageType.tell_chunk_start
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_send_chunk); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 76, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_send_chunk, __pyx_t_2) < 0)) __PYX_ERR(1, 76, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":77
 *     __Pyx_globals['send_chunk_start'] = MessageType.send_chunk_start
 *     __Pyx_globals['send_chunk'] = MessageType.send_chunk
 *     __Pyx_globals['send_chunk_end'] = MessageType.send_chunk_end             # <<<<<<<<<<<<<<
 *     __Pyx_globals['tell_chunk_start'] = MessageType.tell_chunk_start
 *     __Pyx_globals['tell_chunk'] = MessageType.tell_chunk
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_send_chunk_end); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 77, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_send_chunk_end, __pyx_t_15) < 0)) __PYX_ERR(1, 77, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":78
 *     __Pyx_globals['send_chunk'] = MessageType.send_chunk
 *     __Pyx_globals['send_chunk_end'] = MessageType.send_chunk_end
 *     __Pyx_globals['tell_chunk_start'] = MessageType.tell_chunk_start             # <<<<<<<<<<<<<<
 *     __Pyx_globals['tell_chunk'] = MessageType.tell_chunk
 *     __Pyx_globals['tell_chunk_end'] = MessageType.tell_chunk_end
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_tell_chunk_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 78, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_tell_chunk_start, __pyx_t_2) < 0)) __PYX_ERR(1, 78, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":79
 *     __Pyx_globals['send_chunk_end'] = MessageType.send_chunk_end
 *     __Pyx_globals['tell_chunk_start'] = MessageType.tell_chunk_start
 *     __Pyx_globals['tell_chunk'] = MessageType.tell_chunk             # <<<<<<<<<<<<<<
 *     __Pyx_globals['tell_chunk_end'] = MessageType.tell_chunk_end
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_tell_chunk); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 79, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_tell_chunk, __pyx_t_15) < 0)) __PYX_ERR(1, 79, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":80
 *     __Pyx_globals['tell_chunk_start'] = MessageType.tell_chunk_start
 *     __Pyx_globals['tell_chunk'] = MessageType.tell_chunk
 *     __Pyx_globals['tell_chunk_end'] = MessageType.tell_chunk_end             # <<<<<<<<<<<<<<
 * else:
 *     class MessageType(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_tell_chunk_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 80, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_tell_chunk_end, __pyx_t_2) < 0)) __PYX_ERR(1, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     MessageType = __Pyx_EnumBase('MessageType', __Pyx_OrderedDict([
 */
    goto __pyx_L6;
  }

  /* "EnumType":82
 *     __Pyx_globals['tell_chunk_end'] = MessageType.tell_chunk_end
 * else:
 *     class MessageType(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['create_actor'] = MessageType(create_actor, 'create_actor')
 */
  /*else*/ {
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __Pyx_EnumBase);
    __pyx_t_15 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = __Pyx_Py3MetaclassPrepare(__pyx_t_15, __pyx_t_2, __pyx_n_s_MessageType, __pyx_n_s_MessageType, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_15, __pyx_n_s_MessageType, __pyx_t_2, __pyx_t_14, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_MessageType, __pyx_t_13) < 0) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":84
 *     class MessageType(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['create_actor'] = MessageType(create_actor, 'create_actor')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['destroy_actor'] = MessageType(destroy_actor, 'destroy_actor')
 *     __Pyx_globals['has_actor'] = MessageType(has_actor, 'has_actor')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_create_actor); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_create_actor);
    __Pyx_GIVEREF(__pyx_n_s_create_actor);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_n_s_create_actor);
    __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_14, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 84, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_create_actor, __pyx_t_15) < 0)) __PYX_ERR(1, 84, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":85
 *         pass
 *     __Pyx_globals['create_actor'] = MessageType(create_actor, 'create_actor')
 *     __Pyx_globals['destroy_actor'] = MessageType(destroy_actor, 'destroy_actor')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['has_actor'] = MessageType(has_actor, 'has_actor')
 *     __Pyx_globals['result'] = MessageType(result, 'result')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_destroy_actor); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_destroy_actor);
    __Pyx_GIVEREF(__pyx_n_s_destroy_actor);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_destroy_actor);
    __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_2, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 85, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_destroy_actor, __pyx_t_14) < 0)) __PYX_ERR(1, 85, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "EnumType":86
 *     __Pyx_globals['create_actor'] = MessageType(create_actor, 'create_actor')
 *     __Pyx_globals['destroy_actor'] = MessageType(destroy_actor, 'destroy_actor')
 *     __Pyx_globals['has_actor'] = MessageType(has_actor, 'has_actor')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['result'] = MessageType(result, 'result')
 *     __Pyx_globals['error'] = MessageType(error, 'error')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_has_actor); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_has_actor);
    __Pyx_GIVEREF(__pyx_n_s_has_actor);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_n_s_has_actor);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 86, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_has_actor, __pyx_t_2) < 0)) __PYX_ERR(1, 86, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":87
 *     __Pyx_globals['destroy_actor'] = MessageType(destroy_actor, 'destroy_actor')
 *     __Pyx_globals['has_actor'] = MessageType(has_actor, 'has_actor')
 *     __Pyx_globals['result'] = MessageType(result, 'result')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['error'] = MessageType(error, 'error')
 *     __Pyx_globals['send_all'] = MessageType(send_all, 'send_all')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_result); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_result);
    __Pyx_GIVEREF(__pyx_n_s_result);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_n_s_result);
    __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_14, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 87, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_result, __pyx_t_15) < 0)) __PYX_ERR(1, 87, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":88
 *     __Pyx_globals['has_actor'] = MessageType(has_actor, 'has_actor')
 *     __Pyx_globals['result'] = MessageType(result, 'result')
 *     __Pyx_globals['error'] = MessageType(error, 'error')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['send_all'] = MessageType(send_all, 'send_all')
 *     __Pyx_globals['tell_all'] = MessageType(tell_all, 'tell_all')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_error); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_error);
    __Pyx_GIVEREF(__pyx_n_s_error);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_error);
    __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_2, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 88, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_error, __pyx_t_14) < 0)) __PYX_ERR(1, 88, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "EnumType":89
 *     __Pyx_globals['result'] = MessageType(result, 'result')
 *     __Pyx_globals['error'] = MessageType(error, 'error')
 *     __Pyx_globals['send_all'] = MessageType(send_all, 'send_all')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['tell_all'] = MessageType(tell_all, 'tell_all')
 *     __Pyx_globals['send_chunk_start'] = MessageType(send_chunk_start, 'send_chunk_start')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_send_all); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_send_all);
    __Pyx_GIVEREF(__pyx_n_s_send_all);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_n_s_send_all);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 89, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_send_all, __pyx_t_2) < 0)) __PYX_ERR(1, 89, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":90
 *     __Pyx_globals['error'] = MessageType(error, 'error')
 *     __Pyx_globals['send_all'] = MessageType(send_all, 'send_all')
 *     __Pyx_globals['tell_all'] = MessageType(tell_all, 'tell_all')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['send_chunk_start'] = MessageType(send_chunk_start, 'send_chunk_start')
 *     __Pyx_globals['send_chunk'] = MessageType(send_chunk, 'send_chunk')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_tell_all); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_tell_all);
    __Pyx_GIVEREF(__pyx_n_s_tell_all);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_n_s_tell_all);
    __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_14, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 90, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_tell_all, __pyx_t_15) < 0)) __PYX_ERR(1, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":91
 *     __Pyx_globals['send_all'] = MessageType(send_all, 'send_all')
 *     __Pyx_globals['tell_all'] = MessageType(tell_all, 'tell_all')
 *     __Pyx_globals['send_chunk_start'] = MessageType(send_chunk_start, 'send_chunk_start')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['send_chunk'] = MessageType(send_chunk, 'send_chunk')
 *     __Pyx_globals['send_chunk_end'] = MessageType(send_chunk_end, 'send_chunk_end')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_send_chunk_start); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_send_chunk_start);
    __Pyx_GIVEREF(__pyx_n_s_send_chunk_start);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_send_chunk_start);
    __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_2, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 91, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_send_chunk_start, __pyx_t_14) < 0)) __PYX_ERR(1, 91, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "EnumType":92
 *     __Pyx_globals['tell_all'] = MessageType(tell_all, 'tell_all')
 *     __Pyx_globals['send_chunk_start'] = MessageType(send_chunk_start, 'send_chunk_start')
 *     __Pyx_globals['send_chunk'] = MessageType(send_chunk, 'send_chunk')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['send_chunk_end'] = MessageType(send_chunk_end, 'send_chunk_end')
 *     __Pyx_globals['tell_chunk_start'] = MessageType(tell_chunk_start, 'tell_chunk_start')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_send_chunk); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_send_chunk);
    __Pyx_GIVEREF(__pyx_n_s_send_chunk);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_n_s_send_chunk);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 92, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_send_chunk, __pyx_t_2) < 0)) __PYX_ERR(1, 92, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":93
 *     __Pyx_globals['send_chunk_start'] = MessageType(send_chunk_start, 'send_chunk_start')
 *     __Pyx_globals['send_chunk'] = MessageType(send_chunk, 'send_chunk')
 *     __Pyx_globals['send_chunk_end'] = MessageType(send_chunk_end, 'send_chunk_end')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['tell_chunk_start'] = MessageType(tell_chunk_start, 'tell_chunk_start')
 *     __Pyx_globals['tell_chunk'] = MessageType(tell_chunk, 'tell_chunk')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_send_chunk_end); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_send_chunk_end);
    __Pyx_GIVEREF(__pyx_n_s_send_chunk_end);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_n_s_send_chunk_end);
    __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_14, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 93, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_send_chunk_end, __pyx_t_15) < 0)) __PYX_ERR(1, 93, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":94
 *     __Pyx_globals['send_chunk'] = MessageType(send_chunk, 'send_chunk')
 *     __Pyx_globals['send_chunk_end'] = MessageType(send_chunk_end, 'send_chunk_end')
 *     __Pyx_globals['tell_chunk_start'] = MessageType(tell_chunk_start, 'tell_chunk_start')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['tell_chunk'] = MessageType(tell_chunk, 'tell_chunk')
 *     __Pyx_globals['tell_chunk_end'] = MessageType(tell_chunk_end, 'tell_chunk_end')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_tell_chunk_start); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_tell_chunk_start);
    __Pyx_GIVEREF(__pyx_n_s_tell_chunk_start);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_tell_chunk_start);
    __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_2, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 94, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_tell_chunk_start, __pyx_t_14) < 0)) __PYX_ERR(1, 94, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "EnumType":95
 *     __Pyx_globals['send_chunk_end'] = MessageType(send_chunk_end, 'send_chunk_end')
 *     __Pyx_globals['tell_chunk_start'] = MessageType(tell_chunk_start, 'tell_chunk_start')
 *     __Pyx_globals['tell_chunk'] = MessageType(tell_chunk, 'tell_chunk')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['tell_chunk_end'] = MessageType(tell_chunk_end, 'tell_chunk_end')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_tell_chunk); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_tell_chunk);
    __Pyx_GIVEREF(__pyx_n_s_tell_chunk);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_n_s_tell_chunk);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 95, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_tell_chunk, __pyx_t_2) < 0)) __PYX_ERR(1, 95, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":96
 *     __Pyx_globals['tell_chunk_start'] = MessageType(tell_chunk_start, 'tell_chunk_start')
 *     __Pyx_globals['tell_chunk'] = MessageType(tell_chunk, 'tell_chunk')
 *     __Pyx_globals['tell_chunk_end'] = MessageType(tell_chunk_end, 'tell_chunk_end')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MessageType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(__pyx_e_4mars_6actors_4pool_8messages_tell_chunk_end); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_tell_chunk_end);
    __Pyx_GIVEREF(__pyx_n_s_tell_chunk_end);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_n_s_tell_chunk_end);
    __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_14, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 96, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_tell_chunk_end, __pyx_t_15) < 0)) __PYX_ERR(1, 96, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  }
  __pyx_L6:;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init mars.actors.pool.messages", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init mars.actors.pool.messages");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* Globals */
static PyObject* __Pyx_Globals(void) {
    Py_ssize_t i;
    PyObject *names;
    PyObject *globals = __pyx_d;
    Py_INCREF(globals);
    names = PyObject_Dir(__pyx_m);
    if (!names)
        goto bad;
    for (i = PyList_GET_SIZE(names)-1; i >= 0; i--) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject* name = PySequence_ITEM(names, i);
        if (!name)
            goto bad;
#else
        PyObject* name = PyList_GET_ITEM(names, i);
#endif
        if (!PyDict_Contains(globals, name)) {
            PyObject* value = __Pyx_GetAttr(__pyx_m, name);
            if (!value) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                goto bad;
            }
            if (PyDict_SetItem(globals, name, value) < 0) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                Py_DECREF(value);
                goto bad;
            }
        }
#if CYTHON_COMPILING_IN_PYPY
        Py_DECREF(name);
#endif
    }
    Py_DECREF(names);
    return globals;
bad:
    Py_XDECREF(names);
    Py_XDECREF(globals);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int32(npy_int32 value) {
    const npy_int32 neg_one = (npy_int32) ((npy_int32) 0 - (npy_int32) 1), const_zero = (npy_int32) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_int32) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_int32) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int32) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_int32) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int32) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int32),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint8(npy_uint8 value) {
    const npy_uint8 neg_one = (npy_uint8) ((npy_uint8) 0 - (npy_uint8) 1), const_zero = (npy_uint8) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_uint8) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_uint8) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint8) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_uint8) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint8) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_uint8),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageType(enum __pyx_t_4mars_6actors_4pool_8messages_MessageType value) {
    const enum __pyx_t_4mars_6actors_4pool_8messages_MessageType neg_one = (enum __pyx_t_4mars_6actors_4pool_8messages_MessageType) ((enum __pyx_t_4mars_6actors_4pool_8messages_MessageType) 0 - (enum __pyx_t_4mars_6actors_4pool_8messages_MessageType) 1), const_zero = (enum __pyx_t_4mars_6actors_4pool_8messages_MessageType) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageType) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageType) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageType) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageType) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageType) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageType),
                                     little, !is_unsigned);
    }
}

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0, -1);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0, -1);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) ((enum NPY_TYPES) 0 - (enum NPY_TYPES) 1), const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE npy_int32 __Pyx_PyInt_As_npy_int32(PyObject *x) {
    const npy_int32 neg_one = (npy_int32) ((npy_int32) 0 - (npy_int32) 1), const_zero = (npy_int32) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_int32) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_int32, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_int32) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int32) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_int32, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_int32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 2 * PyLong_SHIFT) {
                            return (npy_int32) (((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 3 * PyLong_SHIFT) {
                            return (npy_int32) (((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 4 * PyLong_SHIFT) {
                            return (npy_int32) (((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_int32) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_int32) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int32) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int32) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_int32, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_int32,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_int32) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_int32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int32) ((((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int32) ((((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int32) ((((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_int32) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int32) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_int32 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_int32) -1;
        }
    } else {
        npy_int32 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_int32) -1;
        val = __Pyx_PyInt_As_npy_int32(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_int32");
    return (npy_int32) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_int32");
    return (npy_int32) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE npy_uint8 __Pyx_PyInt_As_npy_uint8(PyObject *x) {
    const npy_uint8 neg_one = (npy_uint8) ((npy_uint8) 0 - (npy_uint8) 1), const_zero = (npy_uint8) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_uint8) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_uint8, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_uint8) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint8) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint8, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_uint8) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) >= 2 * PyLong_SHIFT) {
                            return (npy_uint8) (((((npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint8) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) >= 3 * PyLong_SHIFT) {
                            return (npy_uint8) (((((((npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint8) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) >= 4 * PyLong_SHIFT) {
                            return (npy_uint8) (((((((((npy_uint8)digits[3]) << PyLong_SHIFT) | (npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_uint8) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_uint8) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint8) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint8) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_uint8, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint8,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_uint8) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint8) (((npy_uint8)-1)*(((((npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_uint8) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint8) ((((((npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_uint8) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint8) (((npy_uint8)-1)*(((((((npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint8) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint8) ((((((((npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_uint8) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint8) (((npy_uint8)-1)*(((((((((npy_uint8)digits[3]) << PyLong_SHIFT) | (npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint8) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint8) ((((((((((npy_uint8)digits[3]) << PyLong_SHIFT) | (npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_uint8) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint8) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_uint8 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_uint8) -1;
        }
    } else {
        npy_uint8 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_uint8) -1;
        val = __Pyx_PyInt_As_npy_uint8(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_uint8");
    return (npy_uint8) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_uint8");
    return (npy_uint8) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_4mars_6actors_4pool_8messages_MessageSerialType(enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType value) {
    const enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType neg_one = (enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType) ((enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType) 0 - (enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType) 1), const_zero = (enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_4mars_6actors_4pool_8messages_MessageSerialType),
                                     little, !is_unsigned);
    }
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
